package de.lmu.ifi.pms.cep.compilation;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.ElementList;

import de.lmu.ifi.pms.java.util.collection.array.Arrays;



public class StackCompilerFactory<SourceProgramType, DestinationProgramType> extends CompilerFactory<SourceProgramType, DestinationProgramType>{
	
	public final static String defaultIdentifierPrefix = "de.lmu.ifi.pms.cep.compiler.stack";
	
	protected final CompilerFactory<?, ?>[] compilerFactoryStack;
	
	@ElementList(inline=true, entry="compiler", required=false, empty=false, type=CompilerFactory.class)
	java.util.ArrayList<CompilerFactory<?, ?>> getCompilerStack(){
		return new java.util.ArrayList<CompilerFactory<?,?>>(java.util.Arrays.asList(this.compilerFactoryStack));
	}
	
	@Attribute(name="source-type")
	protected String getSourceProgramType(){
		return this.sourceProgramType.getName();
	}
	
	@Attribute(name="destination-type")
	public String getDestinationProgramType(){
		return this.destinationProgramType.getName();
	}
	
	@SuppressWarnings("unchecked")
	StackCompilerFactory(@Attribute(name="iri", required=true) final String identifier, @Attribute(name="source-type", required=false) final String sourceProgramType, @Attribute(name="destination-type", required=false) final String destinationProgramType, @ElementList(inline=true, entry="compiler", required=false, empty=false, type=CompilerFactory.class) final java.util.ArrayList<CompilerFactory<?, ?>> compilerFactoryStack) throws ClassNotFoundException{
		this(identifier, (Class<SourceProgramType>) Class.forName(sourceProgramType), (Class<DestinationProgramType>)Class.forName(destinationProgramType), compilerFactoryStack.toArray(new CompilerFactory[0]));
	}
	
	public static String defaultIdentifier(final CompilerFactory<?, ?>... compilerStack){
		StringBuilder sb = new StringBuilder();
		sb.append(defaultIdentifierPrefix);
		sb.append(":");
		for(CompilerFactory<?, ?> compiler: compilerStack){
			sb.append(compiler.identifier);
			sb.append(":");
		}
		sb.setLength(sb.length() - ":".length());
		return sb.toString();
	}
	
	public StackCompilerFactory(final Class<SourceProgramType> sourceProgramType, final Class<DestinationProgramType> destinationProgramType, final CompilerFactory<?, ?>... compilerStack) {
		this(defaultIdentifier(compilerStack), sourceProgramType, destinationProgramType, compilerStack);
	}

	public StackCompilerFactory(final String identifier, final Class<SourceProgramType> sourceProgramType, final Class<DestinationProgramType> destinationProgramType, final CompilerFactory<?, ?>... compilerStack) {
		super(sourceProgramType, destinationProgramType, identifier);
		this.compilerFactoryStack = Arrays.clone(compilerStack);
		if(compilerStack.length <= 0){
			if(!destinationProgramType.isAssignableFrom(sourceProgramType)){
				throw new IllegalArgumentException("Empty compiler stack, but source type does not correspond to destination type.");
			}
		}
		else{
			if(this.compilerFactoryStack[0].sourceProgramType.isAssignableFrom(sourceProgramType)){
				if(this.destinationProgramType.isAssignableFrom(this.compilerFactoryStack[this.compilerFactoryStack.length - 1].destinationProgramType)){
					for(int compiler = 1; compiler < this.compilerFactoryStack.length; ++compiler){
						if(!this.compilerFactoryStack[compiler].sourceProgramType.isAssignableFrom(this.compilerFactoryStack[compiler - 1].destinationProgramType)){
							throw new IllegalArgumentException("Destination type of compiler " + (compiler - 1) + "does not correspond to source type of compiler " + compiler + ".");
						}
					}
				}
				else{
					throw new IllegalArgumentException("Last compiler has invalid destination type.");
				}
			}
			else{
				throw new IllegalArgumentException("First compiler has invalid source type.");
			}
		}
	}

	@Override
	public Compiler<SourceProgramType, DestinationProgramType> make() {
		final Compiler<?, ?>[] compilerStack = new Compiler<?, ?>[this.compilerFactoryStack.length];
		for(int compiler = 0; compiler < compilerStack.length; ++compiler){
			compilerStack[compiler] = this.compilerFactoryStack[compiler].make();
		}
		return new StackCompiler<SourceProgramType, DestinationProgramType>(this.sourceProgramType, this.destinationProgramType, compilerStack);
	}
	
}
