tree grammar StatefulObjectQueryTransformation;

options { output=AST; ASTLabelType=CommonTree; tokenVocab=DuraC; }

@header {
package de.lmu.ifi.pms.cep.dura.compiler.antlr;

import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.*;
}

@members {
public static final TreeAdaptor treeAdaptor = new CommonTreeAdaptor();
}

createdQuery:
  ^(STATE_QUERY ^(IDENTIFIER ID) ^(QUERY ^(PATH query+=.*)) ^(SUPPLEMENT supplement+=.*))
  
  {
  	String id = String.format("\%s\%s\%s", $ID.text, identifierSeparator, soCreatedSuffix); 
  	String eventType = String.format("\%s\%s\%s", $PATH.text, labelSeparator, soCreatedSuffix);
  }
    
  -> ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[id]) ^(QUERY ^(PATH[eventType] ^(TERM[payloadLabel] $query*))) ^(SUPPLEMENT $supplement*))
  ;

terminatedQuery:
  ^(STATE_QUERY ^(IDENTIFIER ID) ^(QUERY ^(PATH .*)) ^(SUPPLEMENT .*))
  
  {
  	String id = String.format("\%s\%s\%s", $ID.text, identifierSeparator, soTerminatedSuffix);
  	String eventType = String.format("\%s\%s\%s", $PATH.text, labelSeparator, soTerminatedSuffix);
  }
  
  -> ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[id]) ^(QUERY PATH[eventType]) SUPPLEMENT)
  ;

createdRule:
  ^(SCHEMA ^(TERM .*) ^(SUPPLEMENT .*))
  
  {
    String eId = "e";
    
    String objectCreatedLabel = String.format("\%s\%s\%s", $TERM.text, labelSeparator, soCreatedSuffix);
    String objectCreateInitiatedLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soCreateSuffix, labelSeparator, actionInitiatedSuffix);
    
    String valuesPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel, valuesLabel);
    String actionIdPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel, identifierAttribute);
  }
  
  -> ^(DETECT 
        ^(HEAD ^(TERM[objectCreatedLabel] 
        		^(TERM[payloadLabel] 
							^(TERM[identifierAttribute] CLOCK_FUNCTION[identifierClockName])
							PATH[valuesPath])
        		^(TERM[provenanceLabel] ^(TERM[identifierAttribute] PATH[actionIdPath])))
        )
        ^(BODY ^(CONJUNCTION ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(QUERY ^(PATH[objectCreateInitiatedLabel])) SUPPLEMENT) SUPPLEMENT))
      )
  ;

createSucceededRule:
  ^(SCHEMA ^(TERM .*) ^(SUPPLEMENT .*))
  
  {
  	String eId = "e";
  	
  	String objectCreateInitiatedLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soCreateSuffix, labelSeparator, actionInitiatedSuffix);
  	String objectCreateSucceededLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soCreateSuffix, labelSeparator, actionSucceededSuffix);
  	
  	String payloadPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel);
  }
  
  -> ^(DETECT
  			^(HEAD ^(TERM[objectCreateSucceededLabel]
  				^(TERM[payloadLabel] PATH[payloadPath]))
  			)
  			^(BODY ^(CONJUNCTION ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(QUERY ^(PATH[objectCreateInitiatedLabel])) SUPPLEMENT) SUPPLEMENT))
  		)
  ;

terminatedRule:
	^(SCHEMA ^(TERM .*) ^(SUPPLEMENT .*))
  
  {
    String eId = "e";
    String sId = "s";
    
    String objectLabel = $TERM.text;
    String objectTerminatedLabel = String.format("\%s\%s\%s", $TERM.text, labelSeparator, soTerminatedSuffix);
    String objectTerminateInitiatedLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soTerminateSuffix, labelSeparator, actionInitiatedSuffix);
        
    String actionIdPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel, identifierAttribute);
    
    String stateIdPath = TSA_SyntaxPatterns.attributeName(sId, identifierAttribute);
    String queryStateIdPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel, queryLabel, identifierAttribute);
  }
  
  -> ^(DETECT 
        ^(HEAD ^(TERM[objectTerminatedLabel] 
        		^(TERM[payloadLabel] PATH[sId])
        		^(TERM[provenanceLabel] ^(TERM[identifierAttribute] PATH[actionIdPath])))
        )
        ^(BODY ^(CONJUNCTION
            ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(QUERY ^(PATH[objectTerminateInitiatedLabel])) SUPPLEMENT)
            ^(STATE_QUERY ^(IDENTIFIER[stateIdentifierPrefix] ID[sId]) ^(QUERY ^(PATH[objectLabel])) SUPPLEMENT)
            ^(SUPPLEMENT ^(CONDITION ^(CONJUNCTION 
                ^(ARITHMETIC_RELATION["="] PATH[queryStateIdPath] PATH[stateIdPath])
                ^(TIME_RELATION["valid-at"] ^(IDENTIFIER[stateIdentifierPrefix] ID[sId]) ^(INTERVAL_OP["end"] ^(IDENTIFIER[eventIdentifierPrefix] ID[eId])))
             )))
         ))
      )
  ;
  
terminateSucceededRule:
	^(SCHEMA ^(TERM .*) ^(SUPPLEMENT .*))
  
  {
    String eId = "e";
    String fId = "f";
    
    String objectTerminatedLabel = String.format("\%s\%s\%s", $TERM.text, labelSeparator, soTerminatedSuffix);
    String objectTerminateInitiatedLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soTerminateSuffix, labelSeparator, actionInitiatedSuffix);
    String objectTerminateSucceededLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soTerminateSuffix, labelSeparator, actionSucceededSuffix);
        
    String payloadPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel);
    String actionIdPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel, identifierAttribute);
    String provenanceIdPath = TSA_SyntaxPatterns.attributeName(fId, provenanceLabel, identifierAttribute);
  }
  
  -> ^(DETECT 
        ^(HEAD ^(TERM[objectTerminateSucceededLabel] 
        		^(TERM[payloadLabel] PATH[payloadPath]))
        )
        ^(BODY ^(CONJUNCTION
            ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(QUERY ^(PATH[objectTerminateInitiatedLabel])) SUPPLEMENT)
            ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[fId]) ^(QUERY ^(PATH[objectTerminatedLabel])) SUPPLEMENT)
            ^(SUPPLEMENT ^(CONDITION ^(CONJUNCTION 
                ^(ARITHMETIC_RELATION["="] PATH[actionIdPath] PATH[provenanceIdPath])
                ^(TIME_RELATION["equals"] ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(IDENTIFIER[eventIdentifierPrefix] ID[fId]))
             )))
         ))
      )
  ;
  
terminateFailedRule:
	^(SCHEMA ^(TERM .*) ^(SUPPLEMENT .*))
  
  {
    String eId = "e";
    String fId = "f";
    
    String objectTerminatedLabel = String.format("\%s\%s\%s", $TERM.text, labelSeparator, soTerminatedSuffix);
    String objectTerminateInitiatedLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soTerminateSuffix, labelSeparator, actionInitiatedSuffix);
    String objectTerminateFailedLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soTerminateSuffix, labelSeparator, actionFailedSuffix);
        
    String payloadPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel);
    String actionIdPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel, identifierAttribute);
    String provenanceIdPath = TSA_SyntaxPatterns.attributeName(fId, provenanceLabel, identifierAttribute);
  }
  
  -> ^(DETECT 
        ^(HEAD ^(TERM[objectTerminateFailedLabel] 
        		^(TERM[payloadLabel] PATH[payloadPath]))
        )
        ^(BODY ^(CONJUNCTION
            ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(QUERY ^(PATH[objectTerminateInitiatedLabel])) SUPPLEMENT)
            ^(NEGATION ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[fId]) ^(QUERY ^(PATH[objectTerminatedLabel])) SUPPLEMENT) SUPPLEMENT)
            ^(SUPPLEMENT ^(CONDITION ^(CONJUNCTION 
                ^(ARITHMETIC_RELATION["="] PATH[actionIdPath] PATH[provenanceIdPath])
                ^(TIME_RELATION["equals"] ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(IDENTIFIER[eventIdentifierPrefix] ID[fId]))
             )))
         ))
      )
  ;
  
  
updateTerminatedRule:
	^(SCHEMA ^(TERM .*) ^(SUPPLEMENT .*))
  
  {
    String eId = "e";
    String sId = "s";
    
    String objectLabel = $TERM.text;
    String objectTerminatedLabel = String.format("\%s\%s\%s", $TERM.text, labelSeparator, soTerminatedSuffix);
    String objectUpdateInitiatedLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soUpdateSuffix, labelSeparator, actionInitiatedSuffix);
        
    String actionIdPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel, identifierAttribute);
    String stateIdPath = TSA_SyntaxPatterns.attributeName(sId, identifierAttribute);
    String queryStateIdPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel, queryLabel, identifierAttribute);
  }
  
  -> ^(DETECT 
        ^(HEAD ^(TERM[objectTerminatedLabel] 
        		^(TERM[payloadLabel] PATH[sId])
        		^(TERM[provenanceLabel] ^(TERM[identifierAttribute] PATH[actionIdPath])))
        )
        ^(BODY ^(CONJUNCTION
            ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(QUERY ^(PATH[objectUpdateInitiatedLabel])) SUPPLEMENT)
            ^(STATE_QUERY ^(IDENTIFIER[stateIdentifierPrefix] ID[sId]) ^(QUERY ^(PATH[objectLabel])) SUPPLEMENT)
            ^(SUPPLEMENT ^(CONDITION ^(CONJUNCTION 
                ^(ARITHMETIC_RELATION["="] PATH[queryStateIdPath] PATH[stateIdPath])
								^(TIME_RELATION["valid-at"] ^(IDENTIFIER[stateIdentifierPrefix] ID[sId]) ^(INTERVAL_OP["end"] ^(IDENTIFIER[eventIdentifierPrefix] ID[eId])))
             )))
         ))
      )
  ;
  
updateCreatedRule:
  ^(SCHEMA ^(TERM .*) ^(SUPPLEMENT .*))
  
  {
    String eId = "e";
    String fId = "f";

    String objectCreatedLabel = String.format("\%s\%s\%s", $TERM.text, labelSeparator, soCreatedSuffix);    
    String objectTerminatedLabel = String.format("\%s\%s\%s", $TERM.text, labelSeparator, soTerminatedSuffix);
    String objectUpdateInitiatedLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soUpdateSuffix, labelSeparator, actionInitiatedSuffix);
    
    String valuesPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel, valuesLabel);
    String actionIdPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel, identifierAttribute);
    String provenanceIdPath = TSA_SyntaxPatterns.attributeName(fId, provenanceLabel, identifierAttribute);
  }
  
  -> ^(DETECT 
        ^(HEAD ^(TERM[objectCreatedLabel] 
        		^(TERM[payloadLabel] ^(TERM[identifierAttribute] CLOCK_FUNCTION[identifierClockName])	PATH[valuesPath])
        		^(TERM[provenanceLabel] ^(TERM[identifierAttribute] PATH[actionIdPath]))))
        ^(BODY ^(CONJUNCTION
            ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(QUERY ^(PATH[objectUpdateInitiatedLabel])) SUPPLEMENT)
            ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[fId]) ^(QUERY ^(PATH[objectTerminatedLabel])) SUPPLEMENT)
            ^(SUPPLEMENT ^(CONDITION ^(CONJUNCTION 
                ^(ARITHMETIC_RELATION["="] PATH[actionIdPath] PATH[provenanceIdPath])
                ^(TIME_RELATION["equals"] ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(IDENTIFIER[eventIdentifierPrefix] ID[fId]))
             )))
         ))
       )
  ;
  
updatedRule:
	^(SCHEMA ^(TERM .*) ^(SUPPLEMENT .*))
  
  {
    String eId = "e";
    String fId = "f";
    String gId = "g";
    
    String objectUpdatedLabel = String.format("\%s\%s\%s", $TERM.text, labelSeparator, soUpdatedSuffix);
    String objectUpdateInitiatedLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soUpdateSuffix, labelSeparator, actionInitiatedSuffix);
    String objectCreatedLabel = String.format("\%s\%s\%s", $TERM.text, labelSeparator, soCreatedSuffix);    
    String objectTerminatedLabel = String.format("\%s\%s\%s", $TERM.text, labelSeparator, soTerminatedSuffix);
    
    String fPayload = TSA_SyntaxPatterns.attributeName(fId, payloadLabel);
    String gPayload = TSA_SyntaxPatterns.attributeName(gId, payloadLabel);
    
    String actionIdPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel, identifierAttribute);
    String fProvenanceIdPath = TSA_SyntaxPatterns.attributeName(fId, provenanceLabel, identifierAttribute);
    String gProvenanceIdPath = TSA_SyntaxPatterns.attributeName(gId, provenanceLabel, identifierAttribute);
  }
  
  -> ^(DETECT 
        ^(HEAD ^(TERM[objectUpdatedLabel] 
        		^(TERM[payloadLabel] PATH[gPayload])
        		^(TERM[obsoletePayloadLabel] PATH[fPayload])
        		^(TERM[provenanceLabel] ^(TERM[identifierAttribute] PATH[actionIdPath])))
        )
        ^(BODY ^(CONJUNCTION
            ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(QUERY ^(PATH[objectUpdateInitiatedLabel])) SUPPLEMENT)
            ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[fId]) ^(QUERY ^(PATH[objectTerminatedLabel])) SUPPLEMENT)
            ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[gId]) ^(QUERY ^(PATH[objectCreatedLabel])) SUPPLEMENT)
            ^(SUPPLEMENT ^(CONDITION ^(CONJUNCTION 
                ^(ARITHMETIC_RELATION["="] PATH[actionIdPath] PATH[fProvenanceIdPath])
                ^(ARITHMETIC_RELATION["="] PATH[actionIdPath] PATH[gProvenanceIdPath])
                ^(TIME_RELATION["equals"] ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(IDENTIFIER[eventIdentifierPrefix] ID[fId]))
                ^(TIME_RELATION["equals"] ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(IDENTIFIER[eventIdentifierPrefix] ID[gId]))
             )))
         ))
      )
  ;

updateSucceededRule:
	^(SCHEMA ^(TERM .*) ^(SUPPLEMENT .*))
  
  {
    String eId = "e";
    String fId = "f";
    
    String objectUpdatedLabel = String.format("\%s\%s\%s", $TERM.text, labelSeparator, soUpdatedSuffix);
    String objectUpdateInitiatedLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soUpdateSuffix, labelSeparator, actionInitiatedSuffix);
    String objectUpdateSucceededLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soUpdateSuffix, labelSeparator, actionSucceededSuffix);
        
    String payloadPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel);
    String actionIdPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel, identifierAttribute);
    String provenanceIdPath = TSA_SyntaxPatterns.attributeName(fId, provenanceLabel, identifierAttribute);
  }
  
  -> ^(DETECT 
        ^(HEAD ^(TERM[objectUpdateSucceededLabel] 
        		^(TERM[payloadLabel] PATH[payloadPath]))
        )
        ^(BODY ^(CONJUNCTION
            ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(QUERY ^(PATH[objectUpdateInitiatedLabel])) SUPPLEMENT)
            ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[fId]) ^(QUERY ^(PATH[objectUpdatedLabel])) SUPPLEMENT)
            ^(SUPPLEMENT ^(CONDITION ^(CONJUNCTION 
                ^(ARITHMETIC_RELATION["="] PATH[actionIdPath] PATH[provenanceIdPath])
                ^(TIME_RELATION["equals"] ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(IDENTIFIER[eventIdentifierPrefix] ID[fId]))
             )))
         ))
      )
  ;
  
updateFailedRule:
	^(SCHEMA ^(TERM .*) ^(SUPPLEMENT .*))
  
  {
    String eId = "e";
    String fId = "f";
    
    String objectUpdatedLabel = String.format("\%s\%s\%s", $TERM.text, labelSeparator, soUpdatedSuffix);
    String objectUpdateInitiatedLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soUpdateSuffix, labelSeparator, actionInitiatedSuffix);
    String objectUpdateFailedLabel = String.format("\%s\%s\%s\%s\%s", $TERM.text, labelSeparator, soUpdateSuffix, labelSeparator, actionFailedSuffix);
    
    String payloadPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel);
    String actionIdPath = TSA_SyntaxPatterns.attributeName(eId, payloadLabel, identifierAttribute);
    String provenanceIdPath = TSA_SyntaxPatterns.attributeName(fId, provenanceLabel, identifierAttribute);
  }
  
  -> ^(DETECT 
        ^(HEAD ^(TERM[objectUpdateFailedLabel] 
        		^(TERM[payloadLabel] PATH[payloadPath]))
        )
        ^(BODY ^(CONJUNCTION
            ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(QUERY ^(PATH[objectUpdateInitiatedLabel])) SUPPLEMENT)
            ^(NEGATION ^(EVENT_QUERY ^(IDENTIFIER[eventIdentifierPrefix] ID[fId]) ^(QUERY ^(PATH[objectUpdatedLabel])) SUPPLEMENT) SUPPLEMENT)
            ^(SUPPLEMENT ^(CONDITION ^(CONJUNCTION 
                ^(ARITHMETIC_RELATION["="] PATH[actionIdPath] PATH[provenanceIdPath])
                ^(TIME_RELATION["equals"] ^(IDENTIFIER[eventIdentifierPrefix] ID[eId]) ^(IDENTIFIER[eventIdentifierPrefix] ID[fId]))
             )))
         ))
      )
  ;
  