package de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import org.antlr.runtime.CommonToken;
import org.antlr.runtime.tree.CommonTree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;


public class AnonymousActionCompositions extends Transformation {

	@Override
	public void apply(CommonTree tree) {
		for (CommonTree actionSpecification : collectActionSpecificationNodes(tree)) {	
			checkForAndResolveAnonCompositions(actionSpecification);
		}
		
		for (CommonTree reactiveRule : collectReactiveRuleNodes(tree)) {
			CommonTree body = (CommonTree) reactiveRule.getChild(0);
			
			
			// Only relevant if we need to offer access to the query part / body name space from "succeeds on" or
			// "fails on" blocks... (see below)	
			// This is not entirely uncommented but "null" because some methods require it.
			// HashSet<CommonTreeSetWrapper> allBindingsInBody = collectAllBindingsInBody(body);
			HashSet<CommonTreeSetWrapper> allBindingsInBody = null;
			
			
			CommonTree head = (CommonTree) reactiveRule.getChild(1);
			CommonTree concurrent = (CommonTree) head.getChild(0);
		
			int childCount = concurrent.getChildCount();
			if (concurrent.getChild(childCount - 3).getChildCount() > 0 ||
				concurrent.getChild(childCount - 2).getChildCount() > 0 ||
				concurrent.getChild(childCount - 1).getChildCount() > 0) {
				
				CommonTree newConcurrent = new CommonTree(new CommonToken(DuraParser.CONCURRENT, "CONCURRENT"));
				head.deleteChild(0);
				head.addChild(newConcurrent);
				
				
				CommonTree actionExec = new CommonTree(new CommonToken(DuraParser.ACTION_EXECUTION, "ACTION_EXECUTION"));
				newConcurrent.addChild(actionExec);
				
				
				actionExec.addChild(concurrent);
				
				CommonTree actionDefinition = resolveAnonComposition(reactiveRule, concurrent, body, allBindingsInBody);
				
				if (EMIT_ANNOTATIONS) {
					CommonTree annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
					actionDefinition.addChild(annotation);
					annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
					annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Translation of anonymous action composition " + concurrent.getParent().toStringTree() + " within reactive rule " + reactiveRule.toStringTree() + ".\"")));
				}
			} else {
				for (int i = 0; i < childCount - 3; i++) {				
					CommonTree childConcurrent = (CommonTree) concurrent.getChild(i).getChild(1);
					
					if (childConcurrent.getType() == DuraParser.CONCURRENT) {
						CommonTree actionDefinition = resolveAnonComposition(reactiveRule, childConcurrent, body, allBindingsInBody);
						
						if (EMIT_ANNOTATIONS) {
							CommonTree annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
							actionDefinition.addChild(annotation);
							annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
							annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Translation of anonymous action composition " + childConcurrent.getParent().toStringTree() + " within reactive rule " + reactiveRule.toStringTree() + ".\"")));
						}
					}
				}
				
				concurrent.deleteChild(childCount - 1);
				concurrent.deleteChild(childCount - 2);
				concurrent.deleteChild(childCount - 3);
				
				for (int i = 0; i < concurrent.getChildCount(); i++) {
					// DuraC does not allow an identifier.
					concurrent.getChild(i).deleteChild(0);
				}
			}
		}
	}
	
	private HashSet<CommonTreeSetWrapper> collectAllBindingsInBody(CommonTree body) {
		HashSet<CommonTreeSetWrapper> set = new HashSet<CommonTreeSetWrapper>();
		
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(body);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.UNIFICATION) {
				set.add(new CommonTreeSetWrapper(getVariableFromUnification(node)));
			} else if (node.getType() == DuraParser.VARIABLE || (node.getType() == DuraParser.IDENTIFIER && node.getChildCount() > 0)) {
				set.add(new CommonTreeSetWrapper(node));
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					
					if (node.getType() != DuraParser.CONDITION) {
						nodeQueue.add(child);
					}
				}
			}
		}
			
		return set;
	}

	private void checkForAndResolveAnonCompositions(CommonTree actionSpecification) {
		CommonTree head = (CommonTree) actionSpecification.getChild(1);
		CommonTree concurrent = (CommonTree) head.getChild(0);
		
		for (int i = 0; i < concurrent.getChildCount() - 3; i++) {
			CommonTree child = (CommonTree) concurrent.getChild(i);
			
			CommonTree childConcurrent = (CommonTree) child.getChild(1);
			if (childConcurrent.getType() == DuraParser.CONCURRENT) {
				CommonTree annotation = null;
				if (EMIT_ANNOTATIONS) {
					annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
					annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
					annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Translation of anonymous action composition " + child.toStringTree() + " within action specification for action " + actionSpecification.getChild(0).getChild(1).getText() + ".\"")));
				}
				
				CommonTree actionDefinition = resolveAnonComposition((CommonTree) actionSpecification.getParent(), childConcurrent, null, null);
				
				if (EMIT_ANNOTATIONS) {
					actionDefinition.addChild(annotation);
				}
			}
		}
	}

	private CommonTree resolveAnonComposition(CommonTree def, CommonTree childConcurrent, CommonTree ruleBody, HashSet<CommonTreeSetWrapper> allBindingsInBody) {
		
		HashMap<String, CommonTree> varNameToExpressionMap = new HashMap<String, CommonTree>();
		HashMap<String, List<CommonTree>> varNameToTypeNodesMap = new HashMap<String, List<CommonTree>>();
		
		int nextId = getNextNameId();
		resetNextNameId();
		
		replaceExpressionsInConcurrent(childConcurrent, varNameToExpressionMap, varNameToTypeNodesMap);
		
		// Only relevant if we need to offer access to the query part / body name space from "succeeds on" or
		// "fails on" blocks... (see below)		
		// if (ruleBody != null) {
		//   replaceReferencesInSucceedsOrFailsOn(childConcurrent, varNameToExpressionMap, varNameToTypeNodesMap, ruleBody, allBindingsInBody);
		// }
		
		setNextNameId(nextId);
		
		
		
		CommonTree parentTerm = createParentTerm(childConcurrent, varNameToExpressionMap);
		
		CommonTree pseudoSchemaTerm = copyTree(parentTerm);
		
		for (int i = 0; i < pseudoSchemaTerm.getChildCount(); i++) {
			CommonTree subTerm = (CommonTree) pseudoSchemaTerm.getChild(i);
			
			CommonTree var = (CommonTree) subTerm.getChild(0);
			
			String varName = var.getChild(0).getText();
			
			
			subTerm.deleteChild(0);
			
			for (CommonTree t : varNameToTypeNodesMap.get(varName)) {
				subTerm.addChild(copyTree(t));
			}
		}
		
		CommonTree defs = (CommonTree) def.getParent();
		int addIndex = def.getChildIndex() + 1;

		CommonTree actionDefinition = new CommonTree(new CommonToken(DuraParser.ACTION, "ACTION"));
		defs.getChildren().add(addIndex, actionDefinition);
		
		
		CommonTree modifiers = new CommonTree(new CommonToken(DuraParser.MODIFIERS, "MODIFIERS"));
		actionDefinition.addChild(modifiers);
		
		
		CommonTree schema = new CommonTree(new CommonToken(DuraParser.SCHEMA, "SCHEMA"));
		actionDefinition.addChild(schema);
		
		
		schema.addChild(pseudoSchemaTerm);
		
		
		
		CommonTree schemaTermCopy = copyTree(pseudoSchemaTerm);
		actionSchemataMap.put(pseudoSchemaTerm.getText(), schemaTermCopy);	
		explicitActionSchemataMap.put(pseudoSchemaTerm.getText(), pseudoSchemaTerm);
		
		addImplicitActionEntailedEventSchemata(pseudoSchemaTerm.getText(), schemaTermCopy);
		
		
			
		schema.addChild(new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT")));
		
		CommonTree actionSpec = new CommonTree(new CommonToken(DuraParser.FOR, "FOR"));
		actionDefinition.addChild(actionSpec);
		
		CommonTree body = new CommonTree(new CommonToken(DuraParser.BODY, "BODY"));
		actionSpec.addChild(body);
		
		CommonTree identifier;
		CommonTree temp = (CommonTree) childConcurrent.getParent().getChild(0);
		if (childConcurrent.getParent().getType() == DuraParser.ACTION_EXECUTION && temp.getType() == DuraParser.IDENTIFIER) {
			identifier = copyTree(temp);
		} else {
			identifier = new CommonTree(new CommonToken(DuraParser.IDENTIFIER, "action"));
		}
		body.addChild(identifier);
		
		
		body.addChild(parentTerm);
		
		
		CommonTree head = new CommonTree(new CommonToken(DuraParser.HEAD, "HEAD"));
		actionSpec.addChild(head);
		
		
		CommonTree headTerm = copyTree(parentTerm);
		for (int i = 0; i < headTerm.getChildCount(); i++) {
			CommonTree subTerm = (CommonTree) headTerm.getChild(i);
			
			CommonTree var = (CommonTree) subTerm.getChild(0);
			
			String varName = var.getChild(0).getText();
			
			
			subTerm.deleteChild(0);
			
			subTerm.addChild(varNameToExpressionMap.get(varName));
		}		
		
		CommonTree p = (CommonTree) childConcurrent.getParent();
		p.deleteChild(childConcurrent.getChildIndex());
		p.addChild(headTerm);
		
		
		head.addChild(childConcurrent);
		
		defs.freshenParentAndChildIndexes();
		
		// Necessary if there is deeper nesting...
		checkForAndResolveAnonCompositions(actionSpec);
		
		return actionDefinition;
	}

	
	// Only relevant if we need to offer access to the query part / body name space from "succeeds on" or
	// "fails on" blocks...
	//
	// NOTE: This is experimental / not finished at all!!!
	//
	//
	// Keep in mind that ActionSpecification needs to be adjusted to that as well!!
	// 
	//
	//
	// Though it is possible, in my opinion this is not the best idea. 
	//
	// Why?
	//
	// Consider the following situation:
	//
	// FOR
	//   action y: x{ p{ var X } }
	// DO
	//   concurrent {
	//     ...
	//   } succeeds on {
	//     and {
	//       event: or {
	//         event: u{},
	//         event: l{ z{ var X } }
    //	     },
	//       event: k{ p{ var X } }
	//     }
	//   }
	// END
	//
	// In a normal event query this would be illegal because the disjunction does not expose
	// the var X binding, but still uses the same name!!!
	//
	// While we can somehow fix this during transformation by standardization apart and matching the right
	// x.initiated{ ... } contents ... this still goes counter to our normal naming rules (equal names match;
	// if they can't match due to hiding => error).
	// 
	// Additionally, unlike the special construct for instance matching (event: action y$succeeded{ ... }, these variables
	// are not marked as special constructs.
	
	/*
	private void replaceReferencesInSucceedsOrFailsOn(
			CommonTree concurrent,
			HashMap<String, CommonTree> varNameToExpressionMap,
			HashMap<String, List<CommonTree>> varNameToTypeNodesMap,
			CommonTree body,
			HashSet<CommonTreeSetWrapper> allBindingsInBody) {

		Stack<CommonTree> nodeStack = new Stack<CommonTree>();
		nodeStack.add(concurrent);
		
		CommonTree bodyComposition = (CommonTree) body.getChild(0);
		
		
		HashMap<String, String> eventIdentNameToVarNameMap = new HashMap<String, String>();
		HashMap<String, String> stateIdentNameToVarNameMap = new HashMap<String, String>();
		HashMap<String, String> actionIdentNameToVarNameMap = new HashMap<String, String>();
		
		
		while(!nodeStack.isEmpty()) {
			CommonTree node = (CommonTree) nodeStack.pop();			
			
			
			if (node.getParent().getType() == DuraParser.ACTION_EXECUTION && node.getType() == DuraParser.TERM) {
				continue;
			}
			
			
			CommonTree var;
			String varName;
			List<CommonTree> typeNodes;
			
			if (node.getType() == DuraParser.IDENTIFIER && allBindingsInBody.contains(new CommonTreeSetWrapper(node))) {
				String identName = node.getChild(0).getText();
				varName = null;
				
				if (node.getText().equals("event")) {
					varName = eventIdentNameToVarNameMap.get(identName);
				} else if (node.getText().equals("state")) {
					varName = stateIdentNameToVarNameMap.get(identName);
				} else if (node.getText().equals("action")) {
					varName = actionIdentNameToVarNameMap.get(identName);
				}
				
				
				if (varName == null) {
					var = createVariableNode();
					varName = var.getChild(0).getText();
					
					varNameToExpressionMap.put(varName, node);
					
					typeNodes = determineAndGetTypeNodes(bodyComposition, node);
					
					if (typeNodes.size() > 1 || typeNodes.get(0).getType() == DuraParser.TERM) {
						throw new RuntimeException("Placeholder or unification variable for expression " + node.toStringTree() + " is of complex type!" );
					}
									
					varNameToTypeNodesMap.put(varName, typeNodes);
					
					
					if (node.getText().equals("event")) {
						eventIdentNameToVarNameMap.put(identName, varName);
					} else if (node.getText().equals("state")) {
						stateIdentNameToVarNameMap.put(identName, varName);
					} else if (node.getText().equals("action")) {
						actionIdentNameToVarNameMap.put(identName, varName);
					}
				} else {
					var = new CommonTree(new CommonToken(DuraParser.VARIABLE, "VARIABLE"));
					var.addChild(new CommonTree(new CommonToken(DuraParser.ID, varName)));
				}
				
				int index = node.getChildIndex();
				node.getParent().replaceChildren(index, index, var);
				
				
				//  Not ready yet... works if identifier used like variables (in a term), but not otherwise
				//  
				//  The following needs to be done:
				//
				//
				//
				//  event i: t{} ->
				//  
				//  succeeds on {
				//    and {
				//      event i: t{ id{ var U } }
				//		event k: t{ id{ var U } }
				//    }
				//  }
				//
				//  Identifier event i becomes variable _X...
				//
				//  var X ->
				//   
				//  succeeds on {
				//    and {
				//      event _x: t{ id{ var X } }
				//      
				//      event  i: t{ id{ var U } }
				//		event  k: t{ id{ var U } }
				//    } where { event _x = event i }
				//  end
			}
			
			
			if (node.getType() == DuraParser.VARIABLE && allBindingsInBody.contains(new CommonTreeSetWrapper(node))) {
				var = node;
				varName = var.getChild(0).getText();
				
				if (varNameToTypeNodesMap.get(varName) == null) {
					
					varNameToExpressionMap.put(varName, copyTree(var));
					
					typeNodes = determineAndGetTypeNodes(bodyComposition, node);
					
					varNameToTypeNodesMap.put(varName, typeNodes);
				}
				
				continue;
			}
			
			
			if (node.getChildCount() > 0) {
				for (int i = node.getChildCount() - 1; i >= 0; i--) {
					CommonTree child = (CommonTree) node.getChild(i);	
					
					if (child.getType() != DuraParser.EXECUTION_CONSTRAINTS) {
						nodeStack.push(child);
					}
				}
			}
		}
	}
	*/

	private CommonTree createParentTerm(CommonTree childConcurrent, HashMap<String, CommonTree> varNameToExpressionMap) {
		String termName = createName();
		
		CommonTree term = new CommonTree(new CommonToken(DuraParser.TERM, termName));
		
		int localNameId = 0; 
		
		for (String varName : varNameToExpressionMap.keySet()) {		
			CommonTree subTerm = new CommonTree(new CommonToken(DuraParser.TERM, termName + "_" + localNameId));
			localNameId++;
			term.addChild(subTerm);
			
			CommonTree varNode = new CommonTree(new CommonToken(DuraParser.VARIABLE, "VARIABLE"));
			subTerm.addChild(varNode);
			
			varNode.addChild(new CommonTree(new CommonToken(DuraParser.ID, varName)));
		}
		
		return term;
	}
}
