package de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring;

import java.util.HashMap;
import java.util.List;

import org.antlr.runtime.CommonToken;
import org.antlr.runtime.tree.CommonTree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;


public class ReactiveRuleSimplification extends Transformation {

	@Override
	public void apply(CommonTree tree) {
		for (CommonTree reactiveRule : collectReactiveRuleNodes(tree)) {
			CommonTree body = (CommonTree) reactiveRule.getChild(0);
			
			CommonTree head = (CommonTree) reactiveRule.getChild(1);
			CommonTree concurrent = (CommonTree) head.getChild(0);
			
			HashMap<String, CommonTree> varNameToExpressionMap = new HashMap<String, CommonTree>();
			HashMap<String, List<CommonTree>> varNameToTypeNodesMap = new HashMap<String, List<CommonTree>>();
			
			
			int nextId = getNextNameId();
			resetNextNameId();
			
			replaceExpressionsInConcurrent(concurrent, varNameToExpressionMap, varNameToTypeNodesMap);
			
			setNextNameId(nextId);
			
			
			CommonTree queryTerm = createQueryTerm(varNameToExpressionMap);
			
			
			createEventFromReactiveRule(queryTerm, reactiveRule, varNameToExpressionMap, varNameToTypeNodesMap);
			
			CommonTree conjunction = new CommonTree(new CommonToken(DuraParser.CONJUNCTION, "CONJUNCTION"));
			body.addChild(conjunction);
			
			CommonTree eventQuery = new CommonTree(new CommonToken(DuraParser.EVENT_QUERY, "EVENT_QUERY"));
			conjunction.addChild(eventQuery);
			
			// ^(EVENT_QUERY ^(IDENTIFIER[$i] ID?) ^(QUERY complexSubquery) ^(SUPPLEMENT))
			
			CommonTree identifier = new CommonTree(new CommonToken(DuraParser.IDENTIFIER, "event"));
			eventQuery.addChild(identifier);
			
			CommonTree query = new CommonTree(new CommonToken(DuraParser.QUERY, "QUERY"));
			eventQuery.addChild(query);
			
			
			
			query.addChild(queryTerm);
			
			
			
			CommonTree querySupp = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
			eventQuery.addChild(querySupp);
			
			
			CommonTree supp = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
			conjunction.addChild(supp);
		}
	}

	
	private CommonTree createQueryTerm(HashMap<String, CommonTree> varNameToExpressionMap) {
		String termName = createName();
		
		CommonTree term = new CommonTree(new CommonToken(DuraParser.TERM, termName));
		
		int localNameId = 0; 
		
		for (String varName : varNameToExpressionMap.keySet()) {		
			CommonTree subTerm = new CommonTree(new CommonToken(DuraParser.TERM, termName + "_" + localNameId));
			localNameId++;
			term.addChild(subTerm);
			
			CommonTree varNode = new CommonTree(new CommonToken(DuraParser.VARIABLE, "VARIABLE"));
			subTerm.addChild(varNode);
			
			varNode.addChild(new CommonTree(new CommonToken(DuraParser.ID, varName)));
		}
		
		return term;
	}


	private void createEventFromReactiveRule(CommonTree queryTerm, CommonTree reactiveRule, HashMap<String, CommonTree> varNameToExpressionMap,	HashMap<String, List<CommonTree>> varNameToTypeNodesMap) {
		String eventName = queryTerm.getText();
		
		
		CommonTree schemaTerm = copyTree(queryTerm);
		for (int i = 0; i < schemaTerm.getChildCount(); i++) {
			CommonTree subTerm = (CommonTree) schemaTerm.getChild(i);
			
			CommonTree var = (CommonTree) subTerm.getChild(0);
			
			String varName = var.getChild(0).getText();
			
			
			subTerm.deleteChild(0);
			
			for (CommonTree t : varNameToTypeNodesMap.get(varName)) {
				subTerm.addChild(copyTree(t));
			}
		}
		
		
		
		CommonTree schemaTermCopy = copyTree(schemaTerm);
		
		joinTermCopiesWithTermChildren(implicitEventSchemaTerms, schemaTermCopy);
		eventSchemataMap.put(eventName, schemaTermCopy);
		
		explicitEventSchemataMap.put(eventName, schemaTerm);
		
		
		
		CommonTree headTerm = copyTree(queryTerm);
		for (int i = 0; i < headTerm.getChildCount(); i++) {
			CommonTree subTerm = (CommonTree) headTerm.getChild(i);
			
			CommonTree var = (CommonTree) subTerm.getChild(0);
			
			String varName = var.getChild(0).getText();
			
			
			subTerm.deleteChild(0);
			
			subTerm.addChild(varNameToExpressionMap.get(varName));
		}
		
		
		
		CommonTree defs = (CommonTree) reactiveRule.getParent();
		int addIndex = reactiveRule.getChildIndex() + 1;
		
		CommonTree eventDef = new CommonTree(new CommonToken(DuraParser.EVENT, "EVENT"));
		defs.getChildren().add(addIndex, eventDef);
		defs.freshenParentAndChildIndexes();
		
		
		eventDef.addChild(new CommonTree(new CommonToken(DuraParser.MODIFIERS, "MODIFIERS")));
		
		CommonTree schema = new CommonTree(new CommonToken(DuraParser.SCHEMA, "SCHEMA"));
		eventDef.addChild(schema);
		
		schema.addChild(schemaTerm);
		schema.addChild(new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT")));
		
		
		
		
		CommonTree ruleBody = (CommonTree) reactiveRule.getChild(0);
		CommonTree ruleBodyComposition = (CommonTree) ruleBody.getChild(0);
		
		
		
		if (EMIT_ANNOTATIONS) {
			CommonTree annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
			reactiveRule.addChild(annotation);
			annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
			annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Extraction of query " + ruleBodyComposition.toStringTree() + ".\"")));
		}
		
		String tempTreeString = null;
		if (EMIT_ANNOTATIONS) {
			tempTreeString = reactiveRule.toStringTree();
		}
		
		
		ruleBody.deleteChild(0);
		
		
		
		CommonTree eventSpecification = new CommonTree(new CommonToken(DuraParser.DETECT, "DETECT"));
		eventDef.addChild(eventSpecification);
		
		CommonTree head = new CommonTree(new CommonToken(DuraParser.HEAD, "HEAD"));
		head.addChild(headTerm);
		eventSpecification.addChild(head);
		
		CommonTree body = new CommonTree(new CommonToken(DuraParser.BODY, "BODY"));
		body.addChild(ruleBodyComposition);
		eventSpecification.addChild(body);
		
		
		
		if (EMIT_ANNOTATIONS) {
			CommonTree annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
			eventDef.addChild(annotation);
			annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
			annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Extracted query from reactive rule " + tempTreeString + ".\"")));
		}
	}
}
