package de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring;

import java.util.Arrays;
import java.util.HashMap;

import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;


/*
 * TODO: Adjustment to optional where { } supplement...
 *       Implicit actions themselves, e.g. for creating stateful objects etc. 
 */
public class SchemaAnalysis extends Transformation {
	// TODO: Whole implicit schemata for system generated events & actions.
	
	public SchemaAnalysis() {
		super();
		
		Arrays.sort(implicitEventSchemaTermStrings);
		for (String termStr : implicitEventSchemaTermStrings) {
			ANTLRStringStream stream = new ANTLRStringStream(termStr);
			DuraLexer lexer = new DuraLexer(stream);
			CommonTokenStream tokens = new CommonTokenStream(lexer);
			DuraParser parser = new DuraParser(tokens);
			
			try {
				CommonTree term = (CommonTree) parser.schemaDefinition().getTree();
				normalizeTerm(term);
				implicitEventSchemaTerms.add(term);
			} catch (RecognitionException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		Arrays.sort(implicitStateSchemaTermString);
		for (String termStr : implicitStateSchemaTermString) {
			ANTLRStringStream stream = new ANTLRStringStream(termStr);
			DuraLexer lexer = new DuraLexer(stream);
			CommonTokenStream tokens = new CommonTokenStream(lexer);
			DuraParser parser = new DuraParser(tokens);
			
			try {
				CommonTree term = (CommonTree) parser.schemaDefinition().getTree();
				normalizeTerm(term);
				implicitStateSchemaTerms.add(term);
			} catch (RecognitionException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		Arrays.sort(implicitActionEntailedEventSchemaTermStrings);
		for (String termStr : implicitActionEntailedEventSchemaTermStrings) {
			ANTLRStringStream stream = new ANTLRStringStream(termStr);
			DuraLexer lexer = new DuraLexer(stream);
			CommonTokenStream tokens = new CommonTokenStream(lexer);
			DuraParser parser = new DuraParser(tokens);
			
			try {
				CommonTree term = (CommonTree) parser.schemaDefinition().getTree();
				normalizeTerm(term);
				implicitActionEntailedEventSchemaTerms.add(term);
			} catch (RecognitionException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}
	
	@Override
	public void apply(CommonTree tree) {
		explicitEventSchemataMap = new HashMap<String, CommonTree>();
		eventSchemataMap = new HashMap<String, CommonTree>();
		explicitStateSchemataMap = new HashMap<String, CommonTree>();
		stateSchemataMap = new HashMap<String, CommonTree>();
		explicitActionSchemataMap = new HashMap<String, CommonTree>();
		actionSchemataMap = new HashMap<String, CommonTree>();
		
		for (CommonTree schema : collectSchemata(tree)) {
			// We do not want to change (by adding implicit terms) the actual schemata in the syntax tree.
			// Instead we make copies, amend them and add them to the maps.
			
			CommonTree schemaTerm = (CommonTree) schema.getChild(0);
			CommonTree schemaTermCopy = copyTree(schemaTerm);
			String schemaTermCopyText = schemaTermCopy.getText();
			
			switch(schema.getParent().getType()) {
			case DuraParser.EVENT:
				if (eventSchemataMap.get(schemaTermCopyText) != null) {
					throw new RuntimeException("Event schema with name \"" + schemaTermCopyText + "\" has already been defined.");
				}
				
				explicitEventSchemataMap.put(schemaTermCopyText, schemaTerm);
				
				joinTermCopiesWithTermChildren(implicitEventSchemaTerms, schemaTermCopy);
				eventSchemataMap.put(schemaTermCopyText, schemaTermCopy);
				break;
				
			case DuraParser.STATEFUL_OBJECT:
				if (stateSchemataMap.get(schemaTermCopyText) != null) {
					throw new RuntimeException("State schema with name \"" + schemaTermCopyText + "\" has already been defined.");
				}
				
				explicitStateSchemataMap.put(schemaTermCopyText, schemaTerm);
				
				joinTermCopiesWithTermChildren(implicitStateSchemaTerms, schemaTermCopy);
				stateSchemataMap.put(schemaTermCopyText, schemaTermCopy);
				break;
				
			case DuraParser.ACTION:
				if (actionSchemataMap.get(schemaTermCopyText) != null) {
					throw new RuntimeException("Action schema with name \"" + schemaTermCopyText + "\" has already been defined.");
				}
				
				explicitActionSchemataMap.put(schemaTermCopyText, schemaTerm);
				
				actionSchemataMap.put(schemaTermCopyText, schemaTermCopy);
				
				addImplicitActionEntailedEventSchemata(schemaTermCopyText, schemaTermCopy);
			}
		}
		
		/*
		// Debug
		for (CommonTree c : actionSchemataMap.values()) {
			System.out.println(c.toStringTree());
		}
		
		// Debug
		System.out.println();
		
		// Debug
		for (CommonTree c : eventSchemataMap.values()) {
			System.out.println(c.toStringTree());
		}
		*/
	}
}
