package de.lmu.ifi.pms.cep.dura.compiler.durac.collections;

import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.CONSTDEF;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.PREAMBLE;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraTree;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.TreeIterator;
import de.lmu.ifi.pms.cep.dura.compiler.durac.DuraCQueryCompiler;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraVisibilityModifier;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnknownConstantException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnspecificCompliationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraVisibilityException;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;


/**
 * Contains a mapping of user defined dura constants to TSA terms.
 * 
 * @author hausmann
 *
 */
public class ConstDefinitions extends HashMap<String, Definition<Term<?>>> {
	private ConstDefinitions() { }
	
	public ConstDefinitions(final DuraTree preamble, String package_name) {
		assert(preamble.getType() == PREAMBLE);
		
		for (DuraTree definition : new TreeIterator(preamble, CONSTDEF)) {
			String name = TSA_SyntaxPatterns.bufferFullName(package_name, definition.getText());
			DuraVisibilityModifier visibility = DuraVisibilityModifier.toVisibilityModifier(definition.getChild(0)); 
			Term<?> term = DuraCQueryCompiler.compileTerm(definition.getChild(1));
			
			this.put(name, new Definition<Term<?>>(name, visibility, term));
		}
	}
	
	public static ConstDefinitions join(List<ConstDefinitions> constants) {
		ConstDefinitions result = new ConstDefinitions();
		
		for (ConstDefinitions constant : constants) {
			assert(Collections.disjoint(result.keySet(), constant.keySet()));
			
			result.putAll(constant);
		}
		
		return result;
	}
	
	public Term<?> getTerm(String constName, Preamble preamble) throws DuraUnspecificCompliationException {
		Definition<Term<?>> definition = getDefinition(constName, preamble);
		
		if (definition.visibility==DuraVisibilityModifier.PUBLIC || definition.name.startsWith(preamble.package_name)) {
			return definition.definition;
		} else {
			throw new DuraVisibilityException(definition.name);
		}
	}
	
	private Definition<Term<?>> getDefinition(String constName, Preamble preamble) throws DuraUnknownConstantException {
		if (constName.contains(".")) {
			if (! this.containsKey(constName)) {
				throw new DuraUnknownConstantException(constName);
			}
			
			return this.get(constName);
		} else if (this.containsKey(TSA_SyntaxPatterns.bufferFullName(preamble.package_name, constName))){
			String full_name = TSA_SyntaxPatterns.bufferFullName(preamble.package_name, constName);
			
			return this.get(full_name);
		} else if (preamble.imports.containsKey(constName)) {
			String full_name = preamble.imports.get(constName);
			
			return this.get(full_name);
		}
		
		throw new DuraUnknownConstantException(constName);
	}
}
