package de.lmu.ifi.pms.cep.dura.compiler.durac.formula_transformations;

import java.util.Map;

import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.ConstDefinitions;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.Preamble;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.UnambiguousQueryVariables;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCExceptionWrapper;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnspecificCompliationException;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.Type;

public class FormulaNormalization extends VariableSubstitution {
	
	private final MapTypeAttributes typeAttributes;
	
	public FormulaNormalization(VariableSubstitution substitution, TemporalStreamSchema schema) {
		super(substitution);
		
		this.typeAttributes = new MapTypeAttributes(schema);
	}
	
	public FormulaNormalization(final UnambiguousQueryVariables variables, TemporalStreamSchema schema, Map<String, String> substitutePrefix, final ConstDefinitions constants, final Preamble preamble) {
		super(variables, substitutePrefix, constants, preamble);
		
		this.typeAttributes = new MapTypeAttributes(schema);
	}
	
	public FormulaNormalization(final UnambiguousQueryVariables variables, TemporalStreamSchema schema, final ConstDefinitions constants, final Preamble preamble) {
		super(variables, constants, preamble);
		
		this.typeAttributes = new MapTypeAttributes(schema);
	}
	
	public Term<? extends Type> apply(final Term<?> term) throws DuraUnspecificCompliationException {
		try {
			Term<?> result = term;
			
			result = super.apply(term);
			
			result = result.replaceAtomicTerm(this.typeAttributes);
			result = MapRewriteIntervalFunctions.instance.adapt(result);
			
			return MapTypeFunctions.instance.adapt(result);
		} catch (DuraCExceptionWrapper e) {
			throw e.cause;
		}
	}
	
	public ConditionFormula apply(final ConditionFormula formula) throws DuraUnspecificCompliationException {
		try {
			ConditionFormula result = formula;
	
			result = super.apply(result);
			
			result = result.replaceAtomicTerm(this.typeAttributes);
			result = result.replaceTerm(MapRewriteIntervalFunctions.instance);
			result = result.replaceAtomicFormula(MapRewriteTimeRelations.instance);
			result = result.replaceTerm(MapTypeFunctions.instance);
			
			return result;
		} catch (DuraCExceptionWrapper e) {
			throw e.cause;
		}
	}
}
