package de.lmu.ifi.pms.cep.event_mill.configuration;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLDecoder;
import java.util.Arrays;
import java.util.Enumeration;

import sun.misc.CompoundEnumeration;
import sun.misc.URLClassPath;

public class ExtendibleURLClassLoader extends URLClassLoader{

//	public ExtendibleURLClassLoader(final URL[] urls, final ClassLoader parent, final URLStreamHandlerFactory factory) {
//		super(urls, parent, factory);
//	}
	static{
		registerAsParallelCapable();
	}
	
	protected final ClassLoader parent;
	private final URLClassPath ucp;
	
	public final DriverReader driverReader;
	
	public static URL[] extendURLs(final URL[] urls, final ClassLoader parent){
		URL url = resourceURL("de.lmu.ifi.pms.cep.event_mill.configuration.DriverReaderImpl", parent);
		if(url != null){
			URL[] extendedUrls = Arrays.copyOf(urls, urls.length + 1);
			extendedUrls[urls.length] = url;
			return extendedUrls;
		}
		else{
			return urls;
		}		
	}
	
	public static URL resourceURL(final String className, final ClassLoader classLoader){
		int nameLength = className.split("\\.").length;
		URL url = classLoader.getResource(className.replace('.', '/')+".class");
		if(url != null){
			String protocol = url.getProtocol();
			if(protocol.equals("file")){
				File directory;
				try {
					directory = new File(URLDecoder.decode(url.getFile(),"UTF-8"));
				} catch (UnsupportedEncodingException e) {
					throw new Error(e);
				}
				for(; nameLength > 0; --nameLength){
					directory = directory.getParentFile();
				}		
				try {
					return new File(directory.getAbsolutePath() + "/").toURI().toURL();
				} catch (MalformedURLException e) {}
			}
			else if(protocol.equals("jar")){
				try{
					String file = URLDecoder.decode(url.getFile(),"UTF-8");
					if(file.startsWith("file:/") & file.contains(".jar!")){
						String jarFile = file.substring(6, file.indexOf('!'));
						try {
							return new File(new File(jarFile).getAbsolutePath()).toURI().toURL();
						} catch (MalformedURLException e) {}
					}
				} catch (UnsupportedEncodingException e) {
					throw new Error(e);
				}
			}
		}
		return null;
	}

	public ExtendibleURLClassLoader(final URL[] urls, final ClassLoader parent) {
		super(extendURLs(urls, parent), null);
		this.parent = parent;
		this.ucp = new URLClassPath(urls);
		try {
			this.driverReader = (DriverReader) this.loadClass("de.lmu.ifi.pms.cep.event_mill.configuration.DriverReaderImpl").newInstance();
			assert this.driverReader.getClass().getClassLoader() == this;
		} catch (Exception e) {
			throw new Error(e);
		}
	}
	
    @Override
	protected Class<?> findClass(final String name) throws ClassNotFoundException
	{
		if(name.equals("de.lmu.ifi.pms.cep.event_mill.configuration.DriverReaderImpl")){
			return super.findClass(name);
		}
		else{
			Class<?> c = null;
			try{
				return this.parent.loadClass(name);
			}
			catch (ClassNotFoundException e){}    		
			return super.findClass(name);
		}
	}

//	public ExtendibleURLClassLoader(final URL[] urls) {
//		super(urls);
//	}

	@Override
	public void addURL(final URL url) {
		super.addURL(url);
		this.ucp.addURL(url);
	}
	
	@Override
	public URL getResource(final String name) {
		URL resource = null;
		resource = this.parent.getResource(name);
		if(resource == null){
			resource = super.getResource(name);
		}
		return resource;
	}
	
	@Override
	public InputStream getResourceAsStream(final String name) {
		InputStream resource = null;
		resource = this.parent.getResourceAsStream(name);
		if(resource == null){
			resource = super.getResourceAsStream(name);
		}
		return resource;
	}
	
	@Override
	public Enumeration<URL> getResources(final String name) throws IOException {
		return new CompoundEnumeration<URL>(new Enumeration[]{this.parent.getResources(name), super.getResources(name)});
	}
	
}
