package de.lmu.ifi.pms.cep.event_mill.io;

import java.util.concurrent.ConcurrentSkipListMap;

import de.lmu.ifi.pms.cep.event_mill.io.catalog.DataCatalog;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;

import de.lmu.ifi.pms.java.util.adaption.Adapter;

public class MessageAdapter<MessageType1, MessageType2> implements Adapter<MessageType1, MessageType2>{
	
	public final Class<MessageType1> type1;
	public final Class<MessageType2> type2;
	
	final DataCatalog dataCatalog;
	
	protected final MessageFormatProvider<MessageType1> formatter1;
	protected final MessageFormatProvider<MessageType2> formatter2;
	
	protected final ConcurrentSkipListMap<String, SpecificMessageAdapter<MessageType1, MessageType2>> specificAdapters;

	public MessageAdapter(final DataCatalog dataCatalog,
			final MessageFormatProvider<MessageType1> formatter1,
			final MessageFormatProvider<MessageType2> formatter2) {
		super();
		this.dataCatalog = dataCatalog;
		this.formatter1 = formatter1;
		this.type1 = this.formatter1.type;
		this.formatter2 = formatter2;
		this.type2 = this.formatter2.type;
		this.specificAdapters = new ConcurrentSkipListMap<String, SpecificMessageAdapter<MessageType1,MessageType2>>();
	}

	@Override
	public MessageType2 adapt(final MessageType1 message) {
		String messageType = this.formatter1.messageType(message);
		SpecificMessageAdapter<MessageType1, MessageType2> specificAdapter = this.specificAdapters.get(messageType);
		if(specificAdapter == null){
			if(this.dataCatalog.exists(messageType)){
				specificAdapter = new SpecificMessageAdapter<MessageType1, MessageType2>(messageType, this.dataCatalog.attributeTypes(messageType), this.formatter1, this.formatter2);
			}
			else{
				return null;
			}
		}
		return specificAdapter.adapt(message);
	}
	
}
