package de.lmu.ifi.pms.cep.event_mill.io;

import java.util.Collection;
import java.util.NoSuchElementException;
import java.util.concurrent.ConcurrentLinkedQueue;

import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class MessageSourceSpliter<MessageType> extends MessageSink<MessageType> {
	
	FixedArrayNavigableMap<String, ConcurrentLinkedQueue<MessageType>> queues;
	MessageFormatProvider<MessageType> formatter;
	
	public MessageSourceSpliter(final FixedArrayNavigableSet<String> messageTypes, final MessageFormatProvider<MessageType> formatter) {
		super(formatter.type);
		@SuppressWarnings("unchecked")
		final ConcurrentLinkedQueue<MessageType>[] queues = (ConcurrentLinkedQueue<MessageType>[]) new ConcurrentLinkedQueue<?>[messageTypes.size()];
		for(int messageType = 0; messageType < queues.length; ++messageType){
			queues[messageType] = new ConcurrentLinkedQueue<MessageType>();
		}
		this.queues = new FixedArrayNavigableMap<String, ConcurrentLinkedQueue<MessageType>>(messageTypes, new FixedArray<ConcurrentLinkedQueue<MessageType>>(false, queues));
		
		this.formatter = formatter;
	}
	
	public SpecificBufferedMessageSource<MessageType> source(final String messageType){
		final int messageTypeIndex = this.queues.keys().indexOf(messageType);
		if(messageTypeIndex >= 0){
			final ConcurrentLinkedQueue<MessageType> queue = this.queues.values().get(messageTypeIndex);
			return new SpecificBufferedMessageSource<MessageType>(this.type, messageType) {
				
				MessageType current = null;
				
				@Override
				public MessageType current() {
					return this.current;
				}
				
				@Override
				public boolean next() {
					this.current = queue.poll();
					return this.current != null;
				}
				
				@Override
				public boolean nextBlock() {
					return true;
				}
				
				@Override
				public MessageType read() {
					return queue.poll();
				}
				
				@Override
				public MessageType[] readBlock() {
					MessageType message = queue.poll();
					if(message == null){
						return this.emptyBlock;
					}
					else{
						java.util.ArrayList<MessageType> block = new java.util.ArrayList<MessageType>();			
						while(message != null){
							block.add(message);
							message = queue.poll();
						}
						return block.toArray(this.newBlock(block.size()));
					}
				}
				
			};
		}
		else{
			throw new NoSuchElementException();///TODO better exception
		}
	}
	
	@Override
	public void writeBlock(final MessageType[] messages) {
		ConcurrentLinkedQueue<MessageType> queue;
		for(MessageType message: messages){
			queue = this.queues.get(this.formatter.messageType(message));
			if(queue != null){
				queue.offer(message);
			}
		}
	}

	@Override
	public void writeBlock(final Collection<MessageType> messages) {
		ConcurrentLinkedQueue<MessageType> queue;
		for(MessageType message: messages){
			queue = this.queues.get(this.formatter.messageType(message));
			if(queue != null){
				queue.offer(message);
			}
		}
	}
	
	@Override
	public void write(final MessageType message) {
		ConcurrentLinkedQueue<MessageType> queue = this.queues.get(this.formatter.messageType(message));
		if(queue != null){
			queue.offer(message);
		}
	}

}
