package de.lmu.ifi.pms.cep.event_mill.io.bus;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.io.MessageSink;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageGenerator;
import de.lmu.ifi.pms.cep.event_mill.io.stream.SinkStream;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class BusSinkStream<MessageType> extends SinkStream{
	
	static final int initialMessageBufferCapacity = 128;
	
	final java.util.ArrayList<MessageType> messageBuffer;
	
	final MessageGenerator<MessageType> messageGenerator;
	protected final MessageSink<MessageType> messageSink;
	public final int numberOfAttributes;
	
	public BusSinkStream(final String messageType, final FixedArrayNavigableMap<String, BasicTypeID> types, final MessageBus<MessageType> bus){
		super(messageType, types);
		this.messageSink = bus.sink();
		this.messageGenerator = bus.formatter().messageGenerator(messageType, types);
		this.numberOfAttributes = types.size();
		
		this.messageBuffer = new java.util.ArrayList<MessageType>(initialMessageBufferCapacity);
	}
	
	/*
	public BusSinkStream(final String messageType, final FixedArrayNavigableMap<String, BasicTypeID> types, final MessageSink<MessageType> sink) {
		super(messageType, types);
		this.messageGenerator = this.messageGenerator;
		this.messageSink = sink;
		this.numberOfAttributes = types.size();
		
		this.messageBuffer = new java.util.ArrayList<MessageType>(initialMessageBufferCapacity);
	}*/



	@Override
	public boolean nextBlock(){
		if(this.messageBuffer.size() > 0){
			this.messageSink.writeBlock(this.messageBuffer);
			this.messageBuffer.clear();
		}
		return true;
	}

	@Override
	public boolean next(){
		this.messageBuffer.add(this.messageGenerator.compose());
		this.messageGenerator.clearParameters();
		return true;
	}
	
	@Override
	public void close() {
		this.messageBuffer.clear();
	}

	@Override
	public void setString(final int parameterIndex, final String value) {
		this.messageGenerator.setString(parameterIndex, value);
	}

	@Override
	public void setBoolean(final int parameterIndex, final boolean value) {
		this.messageGenerator.setBoolean(parameterIndex, value);
	}

	@Override
	public void setByte(final int parameterIndex, final byte value) {
		this.messageGenerator.setByte(parameterIndex, value);
	}

	@Override
	public void setShort(final int parameterIndex, final short value) {
		this.messageGenerator.setShort(parameterIndex, value);
	}

	@Override
	public void setInt(final int parameterIndex, final int value) {
		this.messageGenerator.setInt(parameterIndex, value);
	}

	@Override
	public void setLong(final int parameterIndex, final long value) {
		this.messageGenerator.setLong(parameterIndex, value);
	}

	@Override
	public void setTimestamp(final int parameterIndex, final long value) {
		this.messageGenerator.setTimestamp(parameterIndex, value);
	}

	@Override
	public void setDuration(final int parameterIndex, final long value) {
		this.messageGenerator.setDuration(parameterIndex, value);
	}

	@Override
	public void setIdentifier(final int parameterIndex, final long value) {
		this.messageGenerator.setIdentifier(parameterIndex, value);
	}

	@Override
	public void setFloat(final int parameterIndex, final float value) {
		this.messageGenerator.setFloat(parameterIndex, value);
	}

	@Override
	public void setDouble(final int parameterIndex, final double value) {
		this.messageGenerator.setDouble(parameterIndex, value);
	}

	//logging
	protected static final Logger logger = LoggerFactory.getLogger(BusSinkStream.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}
	
}
