package de.lmu.ifi.pms.cep.event_mill.io.bus;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.io.MessageSourceSpliter;
import de.lmu.ifi.pms.cep.event_mill.io.catalog.DataCatalog;
import de.lmu.ifi.pms.cep.event_mill.io.stream.SinkStream;
import de.lmu.ifi.pms.cep.event_mill.io.stream.SourceStream;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIO;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;

public class BusStreamIO<MessageType> extends StreamIO {

	protected final DataCatalog dataCatalog;
	protected final MessageBus<MessageType> bus;
	protected final MessageSourceSpliter<MessageType> spliter;
	
	public BusStreamIO(final String identifier, final DataCatalog dataCatalog, final MessageBus<MessageType> bus) {
		super(identifier);
		this.dataCatalog = dataCatalog;
		this.bus = bus;
		this.spliter = new MessageSourceSpliter<MessageType>(dataCatalog.messageTypes(), bus.formatter());
	}

	@Override
	public FixedArrayNavigableSet<String> sinkNames() {
		return this.dataCatalog.messageTypes();
	}

	@Override
	protected SinkStream newSinkStream(final String name) {
		return new BusSinkStream<MessageType>(name, this.dataCatalog.attributeTypes(name), this.bus);
	}

	@Override
	public FixedArrayNavigableSet<String> sourceNames() {
		return this.dataCatalog.messageTypes();
	}

	@Override
	protected SourceStream newSourceStream(final String name) {
		return new BusSourceStream<MessageType>(name, this.dataCatalog.attributeTypes(name), this.spliter, this.bus);
	}

	//logging
	protected static final Logger logger = LoggerFactory.getLogger(BusStreamIO.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

}
