package de.lmu.ifi.pms.cep.event_mill.io.bus;

import java.lang.ref.WeakReference;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ConcurrentSkipListMap;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;

import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;


public class HeapBroadcastMessageBusProvider<MessageType> extends SimpleMessageBusProvider<MessageType> {
	
	@Element(name="format")
	final MessageFormatProvider<MessageType> formatter;
	
	final java.util.concurrent.ConcurrentSkipListMap<String, WeakReference<HeapBroadcastMessageBus<MessageType>>> buses;
	
	final ConcurrentSkipListMap<String, ConcurrentLinkedQueue<MessageType>> pipes;

	
	public HeapBroadcastMessageBusProvider(@Attribute(name="iri") final String identifier, @Element(name="format") final MessageFormatProvider<MessageType> formatter) {
		super(formatter.type, identifier);
		this.formatter = formatter;
		
		this.buses = new java.util.concurrent.ConcurrentSkipListMap<String, WeakReference<HeapBroadcastMessageBus<MessageType>>>();
		this.pipes = new java.util.concurrent.ConcurrentSkipListMap<String, ConcurrentLinkedQueue<MessageType>>();
	}
	
	@Override
	public boolean accepts(final String identifier) {
		return true;
	}
	
	@Override
	public synchronized MessageBus<MessageType> connect(final String identifier) {
		WeakReference<HeapBroadcastMessageBus<MessageType>> busReference = this.buses.get(identifier);
		if(busReference == null){
			return this.set(identifier);
		}
		else{
			HeapBroadcastMessageBus<MessageType> bus = busReference.get();
			if(bus == null){
				return this.set(identifier);
			}
			else{
				return bus;
			}
		}
		
	}
	
	@Override
	protected boolean isCompatible(final MessageBusProvider provider) {
		return super.isCompatible(provider) && provider.getClass().isInstance(this);
	}
	
	synchronized MessageBus<MessageType> set(final String identifier){
		HeapBroadcastMessageBus<MessageType> bus = new HeapBroadcastMessageBus<MessageType>(this.formatter, identifier, this);
		WeakReference<HeapBroadcastMessageBus<MessageType>> busReference = new WeakReference<HeapBroadcastMessageBus<MessageType>>(bus);
		this.buses.put(identifier, busReference);
		return bus;
	}
	
	synchronized void remove(final String identifier){
		this.buses.remove(identifier);
	}

}
