package de.lmu.ifi.pms.cep.event_mill.io.bus;

import java.lang.ref.WeakReference;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Root;
import org.simpleframework.xml.core.Resolve;

@Root(name="bus")
public abstract class MessageBusProvider {
	
	static final java.util.concurrent.ConcurrentSkipListMap<String, WeakReference<MessageBusProvider>> busProviders = new java.util.concurrent.ConcurrentSkipListMap<String, WeakReference<MessageBusProvider>>();
	
	static MessageBusProvider resolveProvider(final MessageBusProvider busProvider){
		synchronized (busProviders) {
			WeakReference<MessageBusProvider> busProviderReference = busProviders.get(busProvider.identifier);
			if(busProviderReference == null){
				registerProvider(busProvider);
				return busProvider;
			}
			else{
				MessageBusProvider cachedBusProvider = busProviderReference.get();
				if(cachedBusProvider == null){
					registerProvider(busProvider);
					return busProvider;
				}
				else{
					//TODO check compatibility?
					return cachedBusProvider;
				}
			}
		}
	}
	
	private static void registerProvider(final MessageBusProvider busProvider){
		synchronized (busProviders) {
			busProviders.put(busProvider.identifier, new WeakReference<MessageBusProvider>(busProvider));
		}
	}
	
	private static void removeProvider(final String identifier){
		synchronized (busProviders) {
			busProviders.remove(identifier);
		}
	}
	
	@Resolve
	public MessageBusProvider resolve(){
		MessageBusProvider resolvedProvider = resolveProvider(this);
		if(this.isCompatible(resolvedProvider)){
			return resolvedProvider;
		}
		else{
			throw new IllegalStateException("Incompatible bus provider with same identifier \"" + this.identifier + "\" already registered.");
		}
	}
	
	protected boolean isCompatible(final MessageBusProvider provider){
		return this.identifier.equals(provider.identifier) && this.getClass().isInstance(provider);
	}
	
	@Attribute(name="iri")
	public final String identifier;
	
	public MessageBusProvider(@Attribute(name="iri") final String identifier) {
		super();
		this.identifier = identifier;
	}
	
	public abstract boolean accepts(final String identifier);

	public abstract MessageBus<?> connect(final String identifier);
	
	void close(){
		removeProvider(this.identifier);
	}
	
	@Override
	protected void finalize() throws Throwable {
		this.close();
	}
	
}
