package de.lmu.ifi.pms.cep.event_mill.io.bus;

import java.util.Arrays;

import de.lmu.ifi.pms.cep.event_mill.io.MessageAdapter;
import de.lmu.ifi.pms.cep.event_mill.io.MessageSink;


public class WrapperMessageSink<MessageType1, MessageType2> extends MessageSink<MessageType1>{
	
	public final MessageAdapter<MessageType1, MessageType2> writeAdapter;
	final MessageSink<MessageType2> sink;

	public WrapperMessageSink(final MessageAdapter<MessageType1, MessageType2> writeAdapter, final MessageSink<MessageType2> sink) {
		super(writeAdapter.type1);
		this.writeAdapter = writeAdapter;
		this.sink = sink;
	}
	
	@Override
	public void writeBlock(final MessageType1[] block1) {
		MessageType2[] block2 = this.sink.newBlock(block1.length);
		MessageType2 message2;
		int index2 = 0;
		for(int index1 = 0; index1 < block1.length; ++index1){
			message2 = this.writeAdapter.adapt(block1[index1]);
			if(message2 != null){
				block2[index2] = message2;
				++index2;
			}
		}
		if(index2 < block2.length){
			this.sink.writeBlock(Arrays.copyOf(block2, index2));
		}
		else{
			this.sink.writeBlock(block2);
		}
	};
	
	@Override
	public void writeBlock(final java.util.Collection<MessageType1> block1) {
		MessageType2[] block2 = this.sink.newBlock(block1.size());
		MessageType2 message2;
		int index2 = 0;
		for(MessageType1 message1: block1){
			message2 = this.writeAdapter.adapt(message1);
			if(message2 != null){
				block2[index2] = message2;
				++index2;
			}
		}
		if(index2 < block2.length){
			this.sink.writeBlock(Arrays.copyOf(block2, index2));
		}
		else{
			this.sink.writeBlock(block2);
		}
	}

	@Override
	public void write(final MessageType1 message1) {
		MessageType2 message2 = this.writeAdapter.adapt(message1);
		if(message2 != null){
			this.sink.write(message2);
		}
	}
	
	@Override
	public void close() {
		this.sink.close();
	}
	
}
