package de.lmu.ifi.pms.cep.event_mill.io.format;

import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.*;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public class FixedColumnWidthReader extends MessageReader<String>{
	
	final static String defaultEscapeSequence = "\\";
	final static String defaultParameterStartSequence = "%{";
	final static String defaultParameterLengthDelimiterSequence = ":";
	final static String defaultParameterEndSequence = "}";
	final static ParameterFormatter defaultFormater = new ParameterFormatter();


	//final boolean relaxWhiteSpace;
	
	public final String pattern;
	String message;
	final String[] parameterValues;
	final int[] parameterValuesStarts;
	final int[] parameterValuesEnds;
	final int expextedMessageLength;
	
	public final String escapeSequence;
	public final String escapeRegEx;
	public final String parameterStartSequence;
	public final String parameterStartRegEx;
	public final String parameterLengthDelimiterSequence;
	public final String parameterLengthDelimiterRegEx;
	public final String parameterEndSequence;
	public final String parameterEndRegEx;
	
	public final String parameterRegEx;
	public final Pattern parameterPattern;
	
	public final String unquoteStringRegEx;
	public final Pattern unquoteStringPattern;
	
	public final String unquoteParameterRegEx;
	public final Pattern unquoteParameterPattern;
	
	public final ParameterFormatter formater;
	
	public String quoteString(final String s){
		return this.unquoteStringPattern.matcher(s).replaceAll(Matcher.quoteReplacement(this.escapeSequence) + "$1");
	}
	
	public String quoteParameter(final String parameter){
		return this.unquoteParameterPattern.matcher(parameter).replaceAll("$1");
	}
	
	public String unquoteString(final String s){
		return this.unquoteStringPattern.matcher(s).replaceAll("$1");
	}
	
	public String unquoteParameter(final String parameter){
		return this.unquoteParameterPattern.matcher(parameter).replaceAll("$1");
	}
	
	public FixedColumnWidthReader(final String pattern, final String[] parameters, final BasicTypeID[] types) {
		this(defaultFormater, pattern, parameters, types);
	}
	
	public FixedColumnWidthReader(final ParameterFormatter formater, final String pattern, final String[] parameters, final BasicTypeID[] types) {
		this(formater, pattern, new FixedArray<String>(parameters), new FixedArray<BasicTypeID>(types));
	}
	
	public FixedColumnWidthReader(final String pattern, final FixedArray<String> parameters, final FixedArray<BasicTypeID> types) {
		this(defaultFormater, pattern, parameters, types);
	}
	
	public FixedColumnWidthReader(final ParameterFormatter formater, final String pattern, final FixedArray<String> parameters, final FixedArray<BasicTypeID> types) {
		this(defaultEscapeSequence, defaultParameterStartSequence, defaultParameterLengthDelimiterSequence, defaultParameterEndSequence, formater, pattern, parameters, types);
	}
	
	public FixedColumnWidthReader(final String pattern, final FixedArrayNavigableMap<String, BasicTypeID> types) {
		this(defaultFormater, pattern, types);
	}
	
	public FixedColumnWidthReader(final ParameterFormatter formater, final String pattern, final FixedArrayNavigableMap<String, BasicTypeID> types) {
		this(formater, pattern, types.keys(), types.values());
	}
	
	public FixedColumnWidthReader(final String escapeSequence, final String parameterStartSequence, final String parameterLengthDelimiterSequence, final String parameterEndSequence, final String pattern, final String[] parameters, final BasicTypeID[] types) {
		this(escapeSequence, parameterStartSequence, parameterLengthDelimiterSequence, parameterEndSequence, defaultFormater, pattern, parameters, types);
	}
	
	public FixedColumnWidthReader(final String escapeSequence, final String parameterStartSequence, final String parameterLengthDelimiterSequence, final String parameterEndSequence, final ParameterFormatter formater, final String pattern, final String[] parameters, final BasicTypeID[] types) {
		this(escapeSequence, parameterStartSequence, parameterLengthDelimiterSequence, parameterEndSequence, defaultFormater, pattern, new FixedArray<String>(parameters), new FixedArray<BasicTypeID>(types));
	}
	
	public FixedColumnWidthReader(final String escapeSequence, final String parameterStartSequence, final String parameterLengthDelimiterSequence, final String parameterEndSequence, final String pattern, final FixedArrayNavigableMap<String, BasicTypeID> types) {
		this(escapeSequence, parameterStartSequence, parameterLengthDelimiterSequence, parameterEndSequence, defaultFormater, pattern, types);
	}
	
	public FixedColumnWidthReader(final String escapeSequence, final String parameterStartSequence, final String parameterLengthDelimiterSequence, final String parameterEndSequence, final ParameterFormatter formater, final String pattern, final FixedArrayNavigableMap<String, BasicTypeID> types) {
		this(escapeSequence, parameterStartSequence, parameterLengthDelimiterSequence, parameterEndSequence, formater, pattern, types.keys(), types.values());
	}
	
	public FixedColumnWidthReader(final String escapeSequence, final String parameterStartSequence, final String parameterLengthDelimiterSequence, final String parameterEndSequence, final String pattern, final FixedArray<String> parameters, final FixedArray<BasicTypeID> types) {
		this(escapeSequence, parameterStartSequence, parameterLengthDelimiterSequence, parameterEndSequence, defaultFormater, pattern, parameters, types);
	}

	public FixedColumnWidthReader(final String escapeSequence, final String parameterStartSequence, final String parameterLengthDelimiterSequence, final String parameterEndSequence, final ParameterFormatter formater, final String pattern, final FixedArray<String> parameters, final FixedArray<BasicTypeID> types) {
		super(String.class, parameters, types);
		//this.relaxWhiteSpace = relaxWhiteSpace;
		
		this.escapeSequence = escapeSequence;
		this.escapeRegEx = Pattern.quote(this.escapeSequence);
		this.parameterStartSequence = parameterStartSequence;
		this.parameterStartRegEx = Pattern.quote(this.parameterStartSequence);
		this.parameterLengthDelimiterSequence = parameterLengthDelimiterSequence;
		this.parameterLengthDelimiterRegEx = Pattern.quote(this.parameterLengthDelimiterSequence);
		this.parameterEndSequence = parameterEndSequence;
		this.parameterEndRegEx = Pattern.quote(this.parameterEndSequence);
		
		//TODO more generic 
		this.parameterRegEx = "(?<!" + this.escapeRegEx + ")" + this.parameterStartRegEx + "(?:" + "(.*?)" + this.parameterLengthDelimiterRegEx + ")??" + "(\\d+?)" + "(?<!" + this.escapeRegEx + ")" + this.parameterEndRegEx;

		this.parameterPattern = Pattern.compile(this.parameterRegEx);
		this.unquoteStringRegEx = this.escapeRegEx + "(" + this.escapeRegEx + "|" + this.parameterStartRegEx + ")";
		this.unquoteStringPattern = Pattern.compile(this.unquoteStringRegEx);
		this.unquoteParameterRegEx = this.escapeRegEx + "(" + this.escapeRegEx + "|" + this.parameterEndRegEx + ")";
		this.unquoteParameterPattern = Pattern.compile(this.unquoteParameterRegEx);
		
		this.formater = formater;
		
		this.pattern = pattern;
		
		this.parameterValues = new String[this.parameters.size()];
		this.parameterValuesStarts = new int[this.parameterValues.length];
		Arrays.fill(this.parameterValuesStarts, -1);
		this.parameterValuesEnds = new int[this.parameterValues.length];
		Arrays.fill(this.parameterValuesEnds, -1);
		
		
		Matcher parameterMatcher = this.parameterPattern.matcher(pattern);
		
		int inMessageLastEndIndex = 0;
		String parameter;
		int startIndex = 0;
		int parameterIndex;
		int parameterDelimiterLength;
		int parameterValueLength;
		while(parameterMatcher.find()){
			assert parameterMatcher.start() >= startIndex;

			parameter = parameterMatcher.group(1);
			
			parameterDelimiterLength = (parameterMatcher.start() - startIndex);
			
			//TODO more generic
			parameterValueLength = Integer.valueOf(parameterMatcher.group(2));			
			
			if(parameter != null){
				parameter = this.unquoteParameter(parameter);
				parameterIndex = this.indexOf(parameter) - 1;
				if(parameterIndex < 0){
					throw new IllegalArgumentException("Reference to undeclared parameter \"" + parameter + "\"");
				}
				else{
					if(this.parameterValuesStarts[parameterIndex] < 0){
						
						this.parameterValuesStarts[parameterIndex] = inMessageLastEndIndex + parameterDelimiterLength;
						this.parameterValuesEnds[parameterIndex] = this.parameterValuesStarts[parameterIndex] + parameterValueLength;
						
						startIndex = parameterMatcher.end();
						inMessageLastEndIndex = this.parameterValuesEnds[parameterIndex];
					}
				}			
			}
			else{				
				startIndex = parameterMatcher.end();
				inMessageLastEndIndex = inMessageLastEndIndex + parameterDelimiterLength + parameterValueLength;
			}
		}
		
		this.expextedMessageLength = inMessageLastEndIndex + (this.pattern.length() - startIndex); 
		
		//TODO assure that each parameter is set?
	}
	
	public void decompose(final String message, final String[] parameterValues){
		assert parameterValues.length == this.parameterValues.length;
		if(this.message != null){
			this.clearMessage();
		}
		if(message.length() == this.expextedMessageLength){
			this.message = message;
			for(int parameter = 0; parameter < this.parameterValues.length; ++parameter){
				this.parameterValues[parameter] = this.formater.matchParameter(message, this.parameterValuesStarts[parameter], this.parameterValuesEnds[parameter], this.types[parameter]);
			}
		}
		else{
			throw new IllegalArgumentException("Message has unexpected length. Expected length of message is " + this.expextedMessageLength);
		}
	}
	
	@Override
	public void decompose(final String message) {
		this.decompose(message, this.parameterValues);
	}

	@Override
	public void clearMessage() {
		this.message = null;
		Arrays.fill(this.parameterValues, null);
	}

	@Override
	public boolean isNull(final int parameterIndex) {
		return this.parameterValues[parameterIndex - 1] == null;
	}

	@Override
	public boolean getBoolean(final int parameterIndex) {
		return this.formater.parseBoolean(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public byte getByte(final int parameterIndex) {
		return this.formater.parseByte(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public short getShort(final int parameterIndex) {
		return this.formater.parseShort(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public int getInt(final int parameterIndex) {
		return this.formater.parseInt(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public long getLong(final int parameterIndex) {
		return this.formater.parseLong(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public long getTimestamp(final int parameterIndex) {
		return this.formater.parseTimestamp(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public long getDuration(final int parameterIndex) {
		return this.formater.parseDuration(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public long getIdentifier(final int parameterIndex) {
		return this.formater.parseIdentifier(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public float getFloat(final int parameterIndex) {
		return this.formater.parseFloat(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public double getDouble(final int parameterIndex) {
		return this.formater.parseDouble(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public String getString(final int parameterIndex) {
		return this.formater.parseString(this.parameterValues[parameterIndex - 1]);
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(FixedColumnWidthReader.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

	public static void main(final String[] args) {
		String pattern = " %{person.name:10}, %{5}%{person.address.street:12}, %{person.address.number:3}";
		
		FixedColumnWidthReader reader = new FixedColumnWidthReader(pattern, 
				new String[]{"person.name", "person.address.street", "person.address.number"},
				new BasicTypeID[]{	STRING, STRING, INT}
				);
		
		String message = " nobody    , ?????Valeria     , 15 ";
		
		reader.decompose(message);
		
		System.out.println(reader.getString(1));
		System.out.println(reader.getString(2));
		System.out.println(reader.getInt(3));
		
		reader.clearMessage();
	}

}
