package de.lmu.ifi.pms.cep.event_mill.io.format;

import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public abstract class MessageReader<MessageType> extends ParameterObject{
	
	public final Class<MessageType> type;

	public MessageReader(final Class<MessageType> type, final FixedArrayNavigableMap<String, BasicTypeID> types) {
		super(types);
		this.type = type;
	}

	public MessageReader(final Class<MessageType> type, final FixedArray<String> parameters, final FixedArray<BasicTypeID> types) {
		super(parameters, types);
		this.type = type;
	}

	public abstract void decompose(MessageType message);
	
	public abstract void clearMessage();
	
	public abstract boolean isNull(final int parameterIndex);
	
	public abstract boolean getBoolean(final int parameterIndex);
	
	public abstract byte getByte(final int parameterIndex);
	
	public abstract short getShort(final int parameterIndex);
	
	public abstract int getInt(final int parameterIndex);
	
	public abstract long getLong(final int parameterIndex);
	
	public abstract long getTimestamp(final int parameterIndex);
	
	public abstract long getDuration(final int parameterIndex);
	
	public abstract long getIdentifier(final int parameterIndex);
	
	public abstract float getFloat(final int parameterIndex);
	
	public abstract double getDouble(final int parameterIndex);
	
	public abstract String getString(final int parameterIndex);
	
	public String getString(final String parameter) {
		return this.getString(this.indexOf(parameter));
	}

	
	public boolean getBoolean(final String parameter) {
		return this.getBoolean(this.indexOf(parameter));
	}

	
	public byte getByte(final String parameter) {
		return this.getByte(this.indexOf(parameter));
	}

	
	public short getShort(final String parameter) {
		return this.getShort(this.indexOf(parameter));
	}

	
	public int getInt(final String parameter) {
		return this.getInt(this.indexOf(parameter));
	}

	
	public long getLong(final String parameter) {
		return this.getLong(this.indexOf(parameter));
	}
	
	public long getTimestamp(final String parameter) {
		return this.getTimestamp(this.indexOf(parameter));
	}
	
	public long getDuration(final String parameter) {
		return this.getDuration(this.indexOf(parameter));
	}
	
	public long getIdentifier(final String parameter) {
		return this.getIdentifier(this.indexOf(parameter));
	}

	
	public float getFloat(final String parameter) {
		return this.getFloat(this.indexOf(parameter));
	}

	
	public double getDouble(final String parameter) {
		return this.getDouble(this.indexOf(parameter));
	}

}
