package de.lmu.ifi.pms.cep.event_mill.io.format;

import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_TypeException;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.CompositeType;
import de.lmu.ifi.pms.cep.tsa.schema.NamedType;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.cep.tsa.schema.TypeID;

import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public abstract class StructuredStringMessageFormatProvider extends StringMessageFormatProvider {
	
	public StructuredStringMessageFormatProvider(final String identifier, final ParameterFormatter formatter) {
		super(identifier, formatter);
	}
	
	@Override
	protected String getGeneratorPattern(final String messageType, final FixedArrayNavigableMap<String, BasicTypeID> types){
		StringBuilder sb = new StringBuilder();
		try {
			this.getGeneratorPattern(messageType, CompositeType.get(basicTypes(types)), sb);
		} catch (TSA_TypeException e) {
			throw new IllegalArgumentException(e);
		}
		return sb.toString();
	}
	
	@Override
	public abstract String messageType(final String message);
	
	protected abstract void getMessageHead(final String messageType, final StringBuilder sb, boolean hasAttributes);
	
	protected abstract void getMessageFoot(final String messageType, final StringBuilder sb, boolean hasAttributes);
	
	protected void getGeneratorPattern(final String messageType, final CompositeType type, final StringBuilder sb){
		boolean hasAttributes = (type.attributes.size() > 0);
		int depth = 0;
		this.getMessageHead(messageType, sb, hasAttributes);
		if(hasAttributes){
			this.getGeneratorPattern(new StringBuilder(), type, sb, depth + 1);
		}
		this.getMessageFoot(messageType, sb, hasAttributes);
	}
	
	protected void getGeneratorPattern(final StringBuilder attributeName, final Type type, final StringBuilder sb, final int depth){
		if(type instanceof BasicType){
			this.getGeneratorPattern(attributeName, (BasicType) type, sb, depth);
		}
		else if(type instanceof CompositeType){
			this.getGeneratorPattern(attributeName, (CompositeType) type, sb, depth);
		}
		else{assert type instanceof NamedType<?>;
			this.getGeneratorPattern(attributeName, ((NamedType<?>) type).type, sb, depth);
		}
	}
	
	protected void getGeneratorPattern(final StringBuilder attributeName, final BasicType type, final StringBuilder sb, final int depth){
		sb.append(StringPatternGenerator.defaultParameterStartSequence);
		sb.append(attributeName.toString());
		sb.append(StringPatternGenerator.defaultParameterEndSequence);
	}
	
	protected abstract void getAttributePrefix(final String attribute_local_name, final StringBuilder sb, boolean hasSubAttributes, boolean hasValue, final int depth);
	
	protected abstract void getAttributePostfix(final String attribute_local_name, final StringBuilder sb, boolean hasSubAttributes, boolean hasValue, final int depth);
	
	protected abstract String getAttributeDelimiter(final int depth);
	
	protected void getGeneratorPattern(final StringBuilder attributeName, final CompositeType type, final StringBuilder sb, final int depth){
		FixedArrayNavigableMap<String, Type> attributes = type.attributes;
		int numberOfAttributes = attributes.size();
		if(numberOfAttributes > 0){
			final int attributeNameLength = attributeName.length();
			String subAttributeName;
			Type subAttributeType;
			for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
				subAttributeName = attributes.keys().get(attribute);
				subAttributeType = attributes.values().get(attribute);
				boolean hasSubAttributes = !subAttributeType.isEmpty() & subAttributeType.ID == TypeID.COMPOSITE | (subAttributeType.ID == TypeID.NAMED && ((NamedType<?>)subAttributeType).type.ID == TypeID.COMPOSITE );
				boolean hasValue = subAttributeType.ID == TypeID.BASIC | (subAttributeType.ID == TypeID.NAMED && ((NamedType<?>)subAttributeType).type.ID == TypeID.BASIC );
				
				this.getAttributePrefix(subAttributeName, sb, hasSubAttributes, hasValue, depth);
				if(hasSubAttributes | hasValue){
					TSA_SyntaxPatterns.prefixAttribute(attributeName, subAttributeName);
					this.getGeneratorPattern(attributeName, subAttributeType, sb, depth + 1);
					attributeName.setLength(attributeNameLength);
				}
				this.getAttributePostfix(subAttributeName, sb, hasSubAttributes, hasValue, depth);
				sb.append(this.getAttributeDelimiter(depth));
			}
			sb.setLength(sb.length() - this.getAttributeDelimiter(depth).length());
		}
	}
	
	static final int indentLength = 256;
	static final char[] indent = new char[indentLength]; 
	
	static{
		Arrays.fill(indent, '\t');
	}
	
	public static final void indentTab(final StringBuilder stringBuilder, int depth){
		while(depth > indentLength){
			stringBuilder.append(indent);
			depth -= indentLength;
		}
		if(depth > 0){
			stringBuilder.append(indent, 0, depth);
		}
	}
	
}
