package de.lmu.ifi.pms.cep.event_mill.sql;

import java.io.IOException;
import java.io.PrintStream;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import de.lmu.ifi.pms.cep.Version;
import de.lmu.ifi.pms.cep.compilation.CompilerFactory;
import de.lmu.ifi.pms.cep.event_mill.configuration.DEFAULTS;
import de.lmu.ifi.pms.cep.event_mill.configuration.SQL_ConnectionConfiguration;
import de.lmu.ifi.pms.cep.event_mill.configuration.SQL_DatabaseDriver;
import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableProgramState;
import de.lmu.ifi.pms.cep.event_mill.execution.program.StreamBufferStore;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.time.functions.TimeFunction;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public class SQL_StreamBufferStore extends StreamBufferStore<SQL_StreamBufferStore, SQL_Program_Spec> {
	
	public final SQL_DatabaseDriver driver;
	
	public final SQL_ConnectionConfiguration connectionConfiguration;
	final String meta_schema_name;
	
	final String db_start_command;
	final String db_stop_command;
	
	protected final CompilerFactory<DTSA_Program, SQL_Program_Spec> compiler;
	
	Process db_server;
	PrintStream db_input;
	
	
	private Connection connection;
	int maxSQLTableNameLength;
	int maxSQLAttributeNameLength;
	
	
	public synchronized Connection connection(){
		return this.connection;
	}
	
	private Statement statement;
	
	protected synchronized Statement statement(){
		return this.statement;
	}
	
	protected boolean usePreparedStatements;
	
	PreparedStatement addProgramSpec;
	PreparedStatement getProgramSpec;
	PreparedStatement getProgramSpecs;
	PreparedStatement removeProgramSpec;
	
	PreparedStatement addProgramInst;
	PreparedStatement getProgramInst;
	PreparedStatement getProgramInsts;
	PreparedStatement removeProgramInst;
	
	PreparedStatement getProgramInstSchema;
	
	PreparedStatement addProperty;
	
	public SQL_StreamBufferStore (final CompilerFactory<DTSA_Program, SQL_Program_Spec> compiler, final SQL_DatabaseDriver driver, final SQL_ConnectionConfiguration connection, final String meta_schema, final String db_start_command, final String db_stop_command){
		super(SQL_StreamBufferStore.class, SQL_Program_Spec.class);
		
		this.driver = driver;
		this.connectionConfiguration = connection;
		
		this.meta_schema_name = meta_schema;
		this.usePreparedStatements = this.connectionConfiguration.usePreparedStatements;
		
		this.db_start_command = db_start_command;
		this.db_stop_command = db_stop_command;
		
		this.compiler = (compiler == null ? DEFAULTS.defaultSQLSpecCompiler : compiler);
		
		this.connect();
		
		try {
			assert this.connectionConfiguration.catalog == null || this.connectionConfiguration.catalog.equals("") || this.connection.getMetaData().supportsCatalogsInTableDefinitions();
		} catch (SQLException e) {

		}
		
		if(this.isInizialized()){
			this.initStatements();
		}
		
	}
	
	@Override
	public CompilerFactory<DTSA_Program, SQL_Program_Spec> DTSASpecCompiler() {
		return this.compiler;
	}
	
	protected void connect(){		
		try{
			this.connection = this.driver.connect(this.connectionConfiguration);
		} catch(SQLException e){
			this.debug("Failed to connect", e);
			if(this.db_server != null){
				this.stopDB();
			}
			if(this.db_start_command != null){
				this.startDB();
			}
			else{
				this.error("Error on connecting to database!", e);
				throw new RuntimeException(e);
			}
		}
		this.info("Connected to database.");

		try{
			this.statement = this.connection.createStatement();
			this.maxSQLTableNameLength = this.connection.getMetaData().getMaxTableNameLength();
			if(this.maxSQLTableNameLength <= 0){
				this.maxSQLTableNameLength = Integer.MAX_VALUE;
			}
			this.maxSQLAttributeNameLength = this.connection.getMetaData().getMaxColumnNameLength();
			if(this.maxSQLAttributeNameLength <= 0){
				this.maxSQLAttributeNameLength = Integer.MAX_VALUE;
			}

		} catch(SQLException e){
			this.error("Error on creating statement!", e);
			throw new RuntimeException(e);
		}
	}
	
	protected void disconnect(){
		this.info("Disconnecting from database...");
		if(this.addProgramInst != null){			
			try { this.addProgramSpec.close(); this.addProgramSpec = null;} catch (SQLException e) {this.debug("",e);}
			try { this.getProgramSpec.close(); this.getProgramSpec = null;} catch (SQLException e) {this.debug("",e);}
			try { this.getProgramSpecs.close(); this.getProgramSpecs = null;} catch (SQLException e) {this.debug("",e);}
			try { this.removeProgramSpec.close(); this.removeProgramSpec = null;} catch (SQLException e) {this.debug("",e);}
			
			try { this.addProgramInst.close(); this.addProgramInst = null;} catch (SQLException e) {this.debug("",e);}
			try { this.getProgramInst.close(); this.getProgramInst = null;} catch (SQLException e) {this.debug("",e);}
			try { this.getProgramInsts.close(); this.getProgramInsts = null;} catch (SQLException e) {this.debug("",e);}
			try { this.removeProgramInst.close(); this.removeProgramInst = null;} catch (SQLException e) {this.debug("",e);}
			
			try { this.getProgramInstSchema.close(); this.getProgramInstSchema = null;} catch (SQLException e) {this.debug("",e);}
		}
		
		try { this.statement.close(); this.statement = null;} catch (SQLException e) {this.debug("",e);}
		try{ this.connection.close(); this.connection = null;} catch (SQLException e) {this.debug("",e);}
		this.info("Disconnected from database.");
	}
	
	public synchronized void reconnect(final Connection old_connection){
		if(this.connection == old_connection){
			this.disconnect();
			this.connect();
			this.initStatements();
		}
	}
	
	protected void startDB(){
		assert this.db_start_command != null;
		try {
			this.info("Immediate connection failed. Try starting database server.");
			this.db_server = Runtime.getRuntime().exec(this.db_start_command);
			this.db_input = new PrintStream(this.db_server.getOutputStream(), true);
		} catch (IOException e2) {
			this.error("Error on connecting to database!", e2);
			throw new RuntimeException(e2);
		}
		int tryCount = 0;
		while(this.db_server != null & this.connection == null){
			try {
				this.connection = this.driver.connect(this.connectionConfiguration);
			} catch (SQLException e2) {
			
				try{
					int exitValue = this.db_server.exitValue();
					this.debug("DB-Server terminated with exit code" + exitValue, e2);
					this.db_server = null;
					this.db_input.close();
					this.db_input = null;
				}
				catch(IllegalThreadStateException e){
					
				}
				
				this.debug("Failed to connect", e2);
				if(tryCount < 11){
					try {
						Thread.sleep(10000);
					} catch (InterruptedException e) {

					}
					++tryCount;
				}
				else{
					this.error("Error on connecting to database!", e2);
					throw new RuntimeException(e2);
				}
			}
		}
	}
	
	protected void stopDB(){
		if(this.db_server != null){ 
			if(this.db_stop_command != null){
				this.db_input.print(this.db_stop_command);
				try {
					Thread.sleep(1000);
				} catch (InterruptedException e) {

				}
			}
			
			try{
				this.db_server.exitValue();
			}
			catch (IllegalThreadStateException e) {
				this.db_server.destroy();
			}
			
			boolean terminated = false;
			while(!terminated){
				try {
					this.db_server.waitFor();
					terminated = true;
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			this.db_server = null;
			this.db_input.close();
			this.db_input = null;
		}
	}

	protected void initStatements(){
		while(true){
			try{
				this.addProgramSpec = SQL_Translation.prepareStatement(this.connection, SQL_Translation.sqlProgramSpecsAdd(this.meta_schema_name), this.usePreparedStatements);
				this.getProgramSpec = SQL_Translation.prepareStatement(this.connection, SQL_Translation.sqlGetProgramSpec(this.meta_schema_name), this.usePreparedStatements);
				this.getProgramSpecs = SQL_Translation.prepareStatement(this.connection, SQL_Translation.sqlGetProgramSpecs(this.meta_schema_name), this.usePreparedStatements);
				this.removeProgramSpec = SQL_Translation.prepareStatement(this.connection, SQL_Translation.sqlProgramSpecsRemove(this.meta_schema_name), this.usePreparedStatements);
				
				this.addProgramInst = SQL_Translation.prepareStatement(this.connection, SQL_Translation.sqlProgramInstsAdd(this.meta_schema_name), this.usePreparedStatements);
				this.getProgramInst = SQL_Translation.prepareStatement(this.connection, SQL_Translation.sqlGetProgramInst(this.meta_schema_name), this.usePreparedStatements);
				this.getProgramInsts = SQL_Translation.prepareStatement(this.connection, SQL_Translation.sqlGetProgramInsts(this.meta_schema_name), this.usePreparedStatements);
				this.removeProgramInst = SQL_Translation.prepareStatement(this.connection, SQL_Translation.sqlProgramInstsRemove(this.meta_schema_name), this.usePreparedStatements);
				
				this.getProgramInstSchema  = SQL_Translation.prepareStatement(this.connection, SQL_Translation.sqlGetProgramInst(this.meta_schema_name), this.usePreparedStatements);
				
				this.addProperty = SQL_Translation.prepareStatement(this.connection, SQL_Translation.sqlPropertyAdd(this.meta_schema_name), this.usePreparedStatements);
				
				break;
			} catch(SQLException e){
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
					continue;
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	protected boolean existsSchema(final String schema_name){
		while(true){
			try {
				DatabaseMetaData databaseMetaData = this.connection.getMetaData();
				boolean existsSchema =  SQL_Translation.existsSchema(databaseMetaData, this.connectionConfiguration.catalog, schema_name);
				return existsSchema;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	protected void createSchema(final String schema){
		while(true){
			try {
				this.statement.execute(SQL_Translation.sqlCreateSchema(schema));
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else{
					if(this.existsSchema(schema)){
						this.info("Schema \"{}\" already in use.", schema);
						throw new IllegalArgumentException("Schema \"" + schema + "\" already in use.");
					}
					this.error("Could not create schema \"" + schema + "\"", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	protected void dropSchema(final String schema){
		while(true){
			try {
				this.statement.execute(SQL_Translation.sqlDeleteSchema(schema));
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else{
					if(this.existsSchema(schema)){
						this.error("", e);
						throw new RuntimeException(e);
					}
					else{
						return;
					}
				}
			}
		}
	}
	
	@Override
	public synchronized void initialize(){

		String createPropertiesTable = SQL_Translation.sqlPropertiesTableDefinition(this.meta_schema_name);
		String createProgramSpecsTable = SQL_Translation.sqlProgramSpecsTableDefinition(this.meta_schema_name);
		String createProgramInstsTable = SQL_Translation.sqlProgramInstsTableDefinition(this.meta_schema_name);
		
		this.createSchema(this.meta_schema_name);
		
		try{
			while(true){
				try {
					this.statement.execute(createPropertiesTable);
					this.statement.getMoreResults();
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						this.reconnect(this.connection);
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else{
						throw e;
					}
				}
			}
			
			while(true){
				try {
					this.statement.execute(createProgramSpecsTable);
					this.statement.getMoreResults();
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						this.reconnect(this.connection);
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else{
						throw e;
					}
				}
			}
			
			while(true){
				try {
					this.statement.execute(createProgramInstsTable);
					this.statement.getMoreResults();
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						this.reconnect(this.connection);
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else{
						throw e;
					}
				}
			}
			
		}
		catch(RuntimeException e){
			this.dropSchema(this.meta_schema_name);
			throw e;
		}
		catch (Error e) {
			this.dropSchema(this.meta_schema_name);
			throw e;
		}
		catch(Throwable e){
			this.dropSchema(this.meta_schema_name);
			throw new RuntimeException(e);
		}
		
		try {
			//Create greater functions
			this.initGreatestFunctions();
			
			//Create least functions
			this.initLeastFunctions();
			
			this.initStatements();
			
			this.addProperty("version.status", Version.status);
			this.addProperty("version.major", Version.major);
			this.addProperty("version.minor", Version.minor);
			this.addProperty("version.revision", Version.revision);
		}
		catch(RuntimeException e){
			this.clearLeastFunctions();
			this.clearGreatestFunctions();
			this.dropSchema(this.meta_schema_name);
			throw e;
		}
		catch (Error e) {
			this.clearLeastFunctions();
			this.clearGreatestFunctions();
			this.dropSchema(this.meta_schema_name);
			throw e;
		}
		catch(Throwable e){
			this.clearLeastFunctions();
			this.clearGreatestFunctions();
			this.dropSchema(this.meta_schema_name);
			throw new RuntimeException(e);
		}
	}
	
	protected void addProperty(final String key, final String value){
		while(true){
			try {
				SQL_Translation.sqlSetPropertyAdd(this.addProperty, key, value);
				this.addProperty.executeUpdate();
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	protected void initGreatestFunctions(){
		while(true){
			try {
				this.driver.initGreatestFunctions(this.statement);
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
					this.clearGreatestFunctions();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	protected void clearGreatestFunctions(){
		while(true){
			try {
				this.driver.clearGreatestFunctions(this.statement);
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else{
					this.error("", e);
				}
			}
		}
	}
	
	protected void initLeastFunctions(){
		while(true){
			try {
				this.driver.initLeastFunctions(this.statement);
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
					this.clearLeastFunctions();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	protected void clearLeastFunctions(){
		while(true){
			try {
				this.driver.clearLeastFunctions(this.statement);
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else{
					this.error("", e);
				}
			}
		}
	}
	
	@Override
	public synchronized boolean isInizialized(){
		return this.existsSchema(this.meta_schema_name) 
				&& this.existsTable(this.meta_schema_name, SQL_Translation.propertiesTableName) 
		;
		//FIXME method for checking correct format
	}
	
	protected boolean existsTable(final String schema_name, final String table_name){
		while(true){
			try {
				DatabaseMetaData databaseMetaData = this.connection.getMetaData();
				boolean existsSchema =  SQL_Translation.existsTable(databaseMetaData, this.connectionConfiguration.catalog, schema_name, table_name);
				return existsSchema;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	@Override
	public synchronized void remove(){
		//Remove greater functions
		this.clearGreatestFunctions();
		//Remove least functions
		this.clearLeastFunctions();

		this.dropSchema(this.meta_schema_name);
	}
	
	@Override
	public synchronized FixedArrayNavigableSet<String> availableProgramSpecifications(){
		ResultSet programSpecs;
		while(true){
			try {
				programSpecs = this.getProgramSpecs.executeQuery();
				
				ArrayList<String> programsSpecsList = new ArrayList<String>();
				
				String fullName;
				while(programSpecs.next()){
					fullName = SQL_Translation.sqlProgramSpecsReadFullName(programSpecs);
					programsSpecsList.add(fullName);
				}
				
				return new FixedArrayNavigableSet<String>(TSA_SyntaxPatterns.programFullNameComparator, false, false, programsSpecsList.toArray(new String[programsSpecsList.size()]));
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
		
	}
	
	@Override
	public synchronized SQL_Program_Spec getProgramSpecification(final String namespace, final String simple_name){
		
		ResultSet programSpecs;
		boolean existProgramSpec;
		
		while(true){
			try {
				SQL_Translation.sqlSetGetProgramSpec(this.getProgramSpec, namespace, simple_name);

				programSpecs = this.getProgramSpec.executeQuery();

				this.getProgramSpec.clearParameters();
				existProgramSpec = programSpecs.next();	
			
				if(existProgramSpec){			
					String xml_spec;
					
					xml_spec = programSpecs.getString(1);
					
					SQL_Program_Spec program_spec = SQL_Program_Spec.deserialize(xml_spec);
					if(!program_spec.namespace.equals(namespace) || !program_spec.simple_name.equals(simple_name)){
						throw new RuntimeException("inconsistent naming!");
						//TODO better exception + handling
					}
					
					try {
						programSpecs.close();
					} catch (SQLException e) {
		
					}
					return program_spec;
				}
				else{
					IllegalArgumentException e = new IllegalArgumentException("No such program available: namespace = " + namespace + " , simple name = " + simple_name);
					this.debug("", e);
					throw e;
				}
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	@Override
	public synchronized void add(final SQL_Program_Spec program_spec){
		if(!this.existsProgSpec(program_spec.namespace, program_spec.simple_name)){
			while(true){
				try {
					SQL_Translation.sqlSetProgramSpecsAdd(this.addProgramSpec, program_spec);
					this.addProgramSpec.execute();
					try{
						this.removeProgramSpec.clearParameters();
					} catch (SQLException e) {
						
					}
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						this.reconnect(this.connection);
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
						this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
						this.initStatements();
					}
					else{
						this.error("", e);
						throw new RuntimeException(e);
					}
				}
			}
		}
		else{
			this.info("Program specification with name " + program_spec.fullName + " aready exists.");
			throw new IllegalArgumentException("Program specification with name " + program_spec.fullName + " aready exists.");
		}
	}
	
	@Override
	public synchronized boolean existsProgSpec(final String namespace, final String simple_name){
		try {
			while(true){
				try{
					SQL_Translation.sqlSetGetProgramSpec(this.getProgramSpec, namespace, simple_name);
					boolean returnedResultSet = this.getProgramSpec.execute();
					if(returnedResultSet){
						ResultSet result = this.getProgramSpec.getResultSet();
						boolean exists = result.next();
						result.close();
						this.getProgramSpec.clearParameters();
						return exists;
					}
					else{
						this.warn("Database bug: Select statement did not return a ResultSet");
						this.initStatements();
					}
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						this.reconnect(this.connection);
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
						this.debug("", e);
						this.initStatements();					
					}
					else{
						throw e;
					}
				}
			}
		} catch (SQLException e) {
			this.error("", e);
			throw new RuntimeException(e);
			//TODO better exception
		}		
	}
	
	public synchronized void remove(final SQL_Program_Spec program_spec){
		this.removeProgramSpec(program_spec.namespace, program_spec.simple_name);
	}
	
	@Override
	public synchronized void removeProgramSpec(final String namespace, final String simple_name){
		while(true){
			try {
				SQL_Translation.sqlSetProgramSpecsRemove(this.removeProgramSpec, namespace, simple_name);
				this.removeProgramSpec.execute();
				try{
					this.removeProgramSpec.clearParameters();
				} catch (SQLException e) {
	
				}
				
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	@Override
	public synchronized void clearProgramSpecs(){
		//TODO easier implementation
		for(String full_name: this.availableProgramSpecifications()){
			this.removeProgramSpec(full_name);
		}
	}
	
	@Override
	public synchronized void instanciate(final SQL_Program_Spec program_spec, final String instance, final String... schemas){
		this.instanciate(program_spec, instance, schemas[0], schemas[1], schemas[2], schemas[3], schemas[4]);
	}
	
	protected synchronized PreparedStatement prepare(final String sql) throws SQLException{
		while(true){
			try{
				return SQL_Translation.prepareStatement(this.connection, sql, this.usePreparedStatements);				
			} catch(SQLException e){
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
					continue;
				}
				else{
					this.error("Could not prepare statement", e);
					throw e;
				}
			}
		}
	}
	
	public synchronized void instanciate(final SQL_Program_Spec program_spec, final String instance, final String meta_schema, final String input_schema, final String working_schema, final String output_schema, final String log_schema){
		String[] schemas = new String[]{meta_schema, input_schema, working_schema, output_schema, log_schema};
		
		boolean addedProgramInstance = false;
		int schemaIndex = 0;
		
		try {

			while(true){
				try {
					SQL_Translation.sqlSetProgramInstsAdd(this.addProgramInst, program_spec, instance, program_spec.bufferMaxNumberOfMinTimes, program_spec.queryMaxNumberOfOutputMinTimes, program_spec.queryMaxNumberOfInputMinTimes, meta_schema, working_schema, input_schema, output_schema, log_schema);
					this.addProgramInst.execute();
					addedProgramInstance = true;
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						this.reconnect(this.connection);
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
						this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
						this.initStatements();
					}
					else{
						//TODO test for constraint violation // error code 23
						if(this.existsProgInst(program_spec.namespace, program_spec.simple_name, instance)){
							this.warn("", e);
							this.info("Program with name " + TSA_SyntaxPatterns.programInstFullName(program_spec.namespace, program_spec.simple_name, instance) + " aready exists.");
							throw new IllegalArgumentException("Program with name " + TSA_SyntaxPatterns.programInstFullName(program_spec.namespace, program_spec.simple_name, instance) + " aready exists.");
						}
						else{
							throw e;
						}
					}
				}
			}
			try {this.addProgramInst.clearParameters();} catch (SQLException e) {}

			//create schemas
			for(;schemaIndex < schemas.length; ++schemaIndex){
				this.createSchema(schemas[schemaIndex]);
			}
				
			//create tupleIdSequence
			while(true){
				try {
					this.statement.execute(SQL_Translation.sqlCreateTupleIdSequence(meta_schema));
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						this.reconnect(this.connection);
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else{
						this.error("Could not create tuple_id_sequence.", e);
						throw e;
					}
				}
			}
							
			//create program meta tables
			while(true){
				try {
					this.statement.execute(SQL_Translation.sqlCreateBufferMetaTable(meta_schema, program_spec));
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						this.reconnect(this.connection);
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else{
						this.error("Could not create buffer_meta_table.", e);
						throw e;
					}
				}
			}
			while(true){
				try {
					this.statement.execute(SQL_Translation.sqlCreateAttributesMetaTable(meta_schema));
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						this.reconnect(this.connection);
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else{
						this.error("Could not create attribute_meta_table.", e);
						throw e;
					}
				}
			}
			while(true){
				try {
					this.statement.execute(SQL_Translation.sqlCreateQueryMetaTable(meta_schema, program_spec));
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						this.reconnect(this.connection);
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else{
						this.error("Could not create query_meta_table.", e);
						throw e;
					}
				}
			}
				
				
				//create buffers and meta entries
				
				FixedArrayNavigableMap<String, String> sql_table_name_mapping = SQL_Translation.sql_table_name_mapping(program_spec.bufferNames(), this.maxSQLTableNameLength);
				FixedArrayNavigableMap<String, String> sql_attribute_name_mapping = SQL_Translation.sql_attribute_name_mapping(program_spec.attributeNames(), this.maxSQLAttributeNameLength);
				
				{//create buffers and buffer meta entries
					
					String addBufferMetaEntrySQL = SQL_Translation.sqlAddBufferMetaEntry(meta_schema, program_spec.bufferMaxNumberOfMinTimes);

					PreparedStatement addBufferMetaEntry = this.prepare(addBufferMetaEntrySQL);
					
					String addBufferAttributeMetaEntrySQL = SQL_Translation.sqlAddBufferAttributeMetaEntry(meta_schema);

					PreparedStatement addBufferAttributeMetaEntry = this.prepare(addBufferAttributeMetaEntrySQL);
					
					for(SQL_Buffer_Spec buffer_spec: program_spec.bufferSpecs()){
						
						//create database table for buffer
						while(true){
							try {
								this.statement.execute(SQL_Translation.sqlBufferDefinition(sql_table_name_mapping.get(buffer_spec.fullName), buffer_spec.type, buffer_spec.attributes, sql_attribute_name_mapping, meta_schema, working_schema, input_schema, output_schema, log_schema));
								break;
							} catch (SQLException e) {
								if(SQL_EXCEPTIONS.isConnectionFailure(e)){
									this.debug("Connection failure. Trying again.", e);
									this.reconnect(this.connection);
								}
								else if(SQL_EXCEPTIONS.isRollback(e)){
									this.debug("Statement rolled back. Trying again.", e);
								}
								else{
									this.error("Could not create table for buffer \"" + buffer_spec.fullName + "\" of type \"" + buffer_spec.type + "\".", e);
									throw e;
								}
							}
						}
						
						//create buffer meta entry
						while(true){
							try {
								SQL_Translation.sqlSetAddBufferMetaEntry(addBufferMetaEntry, buffer_spec, sql_table_name_mapping.get(buffer_spec.fullName), program_spec.bufferMaxNumberOfMinTimes);
								int updateCount = addBufferMetaEntry.executeUpdate();
								
								assert updateCount >= 0;
								addBufferMetaEntry.clearParameters();
								break;
							} catch (SQLException e) {
								if(SQL_EXCEPTIONS.isConnectionFailure(e)){
									this.debug("Connection failure. Trying again.", e);
									this.reconnect(this.connection);
									addBufferMetaEntry = this.prepare(addBufferMetaEntrySQL);
								}
								else if(SQL_EXCEPTIONS.isRollback(e)){
									this.debug("Statement rolled back. Trying again.", e);
								}
								else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
									this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
									addBufferMetaEntry = this.prepare(addBufferMetaEntrySQL);
								}
								else{
									this.error("", e);
									throw e;
								}
							}
						}
						
						FixedArrayNavigableMap<String, BasicType> attributes = buffer_spec.attributes;						
						for(int attribute = 0; attribute < attributes.keys().size(); ++attribute){
							//create attribute meta entry
							while(true){
								try {
									SQL_Translation.sqlSetAddBufferAttributeMetaEntry(addBufferAttributeMetaEntry, buffer_spec.type, buffer_spec.prog_package, buffer_spec.simple_name, attributes.keys().get(attribute), attributes.values().get(attribute).ID, sql_attribute_name_mapping.get(attributes.keys().get(attribute)));
									int updateCount = addBufferAttributeMetaEntry.executeUpdate();
									
									assert updateCount >= 0;
									addBufferAttributeMetaEntry.clearParameters();
									break;
								} catch (SQLException e) {
									if(SQL_EXCEPTIONS.isConnectionFailure(e)){
										this.debug("Connection failure. Trying again.", e);
										this.reconnect(this.connection);
										addBufferAttributeMetaEntry = this.prepare(addBufferAttributeMetaEntrySQL);
									}
									else if(SQL_EXCEPTIONS.isRollback(e)){
										this.debug("Statement rolled back. Trying again.", e);
									}
									else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
										this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
										addBufferAttributeMetaEntry = this.prepare(addBufferAttributeMetaEntrySQL);
									}
									else{
										this.error("", e);
										throw e;
									}
								}
							}
						}						
					}
					
					try{
						addBufferMetaEntry.clearParameters();
						addBufferMetaEntry.close();
					}
					catch (SQLException e) {
						this.trace("", e);
					}
					try{
						addBufferAttributeMetaEntry.clearParameters();
						addBufferAttributeMetaEntry.close();
					}
					catch (SQLException e) {
						this.trace("", e);
					}
				}
				
				{//create query meta entries
					
					String addQueryMetaEntrySQL = SQL_Translation.sqlAddQueryMetaEntry(meta_schema, program_spec);

					PreparedStatement addQueryMetaEntry = this.prepare(addQueryMetaEntrySQL);
					
					for(SQL_Query_Spec query_spec: program_spec.query_specs){
						//create query meta entry
						while(true){
							try{
								SQL_Translation.sqlSetAddQueryMetaEntry(addQueryMetaEntry, query_spec, program_spec.queryMaxNumberOfOutputMinTimes, program_spec.queryMaxNumberOfInputMinTimes, program_spec.buffer_specs, sql_table_name_mapping, sql_attribute_name_mapping, meta_schema, working_schema, input_schema, output_schema, log_schema);
								int updateCount = addQueryMetaEntry.executeUpdate();
								
								assert updateCount >= 0;
								break;
							} catch (SQLException e) {
								if(SQL_EXCEPTIONS.isConnectionFailure(e)){
									this.debug("Connection failure. Trying again.", e);
									this.reconnect(this.connection);
									addQueryMetaEntry = this.prepare(addQueryMetaEntrySQL);
								}
								else if(SQL_EXCEPTIONS.isRollback(e)){
									this.debug("Statement rolled back. Trying again.", e);
								}
								else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
									this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
									addQueryMetaEntry = this.prepare(addQueryMetaEntrySQL);
								}
								else{
									this.error("", e);
									throw e;
								}
							}			
						}
						try{addQueryMetaEntry.clearParameters();} catch (SQLException e) {}
					}
					
					try{
						addQueryMetaEntry.clearParameters();
						addQueryMetaEntry.close();
					}
					catch (SQLException e) {
						this.trace("", e);
					}
				}
				
				//set program state to created
				this.setProgramState(program_spec.namespace, program_spec.simple_name, instance, ExecutableProgramState.CREATED);
				
				//unlock program
				this.unlockProgram(program_spec.namespace, program_spec.simple_name, instance);
					
		} catch(Throwable t){
			if(addedProgramInstance){				
				--schemaIndex;
				for(;schemaIndex >= 0; --schemaIndex){
					this.dropSchema(schemas[schemaIndex]);
				}
				
				this.initStatements();
				
				//remove program entry
				while(true){
					try{
						SQL_Translation.sqlSetProgramInstsRemove(this.removeProgramInst, program_spec.namespace, program_spec.simple_name, instance);
						int updateCount = this.removeProgramInst.executeUpdate();
						assert updateCount >= 0 : "Database bug: delete statement returned negative update count.";
						break;
					} catch (SQLException e) {
						if(SQL_EXCEPTIONS.isConnectionFailure(e)){
							this.debug("Connection failure. Trying again.", e);
							this.reconnect(this.connection);
						}
						else if(SQL_EXCEPTIONS.isRollback(e)){
							this.debug("Statement rolled back. Trying again.", e);
						}
						else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
							this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
							this.initStatements();
						}
						else{
							this.error("", e);
							throw new RuntimeException(e);
						}
					}
				}
			}
			
			this.error("", t);
			try{
				throw t;
			}
			catch (SQLException e) { 
				throw new RuntimeException(e);
				//TODO better exception
			}
			catch (RuntimeException e) {
				throw e;
			}
			catch (Error e) {
				throw e;
			}
			catch(Throwable e){
				throw new RuntimeException(e);
			}
		}
	}
	
	protected synchronized void setProgramState(final String namespace, final String simple_name, final String instance, final ExecutableProgramState state) throws SQLException{
		while(true){
			try {
				int updateCount = this.statement.executeUpdate(SQL_Translation.sqlSetProgramInstState(this.meta_schema_name, namespace, simple_name, instance, state));
				assert updateCount == 1;
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else{
					this.error("Could not update the state of program \"" + TSA_SyntaxPatterns.programInstFullName(namespace, simple_name, instance) + "\".", e);
					throw e;
				}
			}
		}
	}
	
	protected synchronized void unlockProgram(final String namespace, final String simple_name, final String instance) throws SQLException{
		while(true){
			try {
				int updateCount = this.statement.executeUpdate(SQL_Translation.sqlUnlockProgramInst(this.meta_schema_name, namespace, simple_name, instance));
				assert updateCount == 1;
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else{
					this.error("Could not update the state of program \"" + TSA_SyntaxPatterns.programInstFullName(namespace, simple_name, instance) + "\".", e);
					throw e;
				}
			}
		}
	}
	
	protected synchronized boolean lockProgram(final String namespace, final String simple_name, final String instance) throws SQLException{
		while(true){
			try {
				int updateCount = this.statement.executeUpdate(SQL_Translation.sqlLockProgramInst(this.meta_schema_name, namespace, simple_name, instance));
				assert updateCount == 1 | updateCount == 0;
				return updateCount > 0;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else{
					this.error("Could not update the state of program \"" + TSA_SyntaxPatterns.programInstFullName(namespace, simple_name, instance) + "\".", e);
					throw e;
				}
			}
		}
	}
	
	@Override
	public synchronized boolean existsProgInst(final String namespace, final String simple_name, final String instance){
		while(true){
			try{
				SQL_Translation.sqlSetGetProgramInst(this.getProgramInst, namespace, simple_name, instance);
				ResultSet result = this.getProgramInst.executeQuery();
				boolean exists = result.next();
				try{result.close();} catch (SQLException e) {}
				try{this.getProgramInst.clearParameters();} catch (SQLException e) {}
				return exists;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}	
	}
	
	@Override
	public synchronized void removeProgramInst(final String namespace, final String simple_name, final String instance){
		try {
			if(this.lockProgram(namespace, simple_name, instance)){
				//remove schemas (removes all tables in schema, too)
				String[] schemas = null;
				while(true){
					try{
						SQL_Translation.sqlSetGetProgramInst(this.getProgramInstSchema, namespace, simple_name, instance);
						ResultSet results = this.getProgramInstSchema.executeQuery();
						if(results.next()){
							schemas = SQL_Translation.sqlProgramInstReadSchemas(results);
						}			
						break;
					} catch (SQLException e) {
						if(SQL_EXCEPTIONS.isConnectionFailure(e)){
							this.debug("Connection failure. Trying again.", e);
							this.reconnect(this.connection);
						}
						else if(SQL_EXCEPTIONS.isRollback(e)){
							this.debug("Statement rolled back. Trying again.", e);
						}
						else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
							this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
							this.initStatements();
						}
						else{
							this.error("", e);
							throw new RuntimeException(e);
						}
					}
				}
				
				if(schemas != null){
					
					for(String schema: schemas){
						while(this.existsSchema(schema)){
							try{
								this.statement.execute(SQL_Translation.sqlDeleteSchema(schema));
							} catch (SQLException e) {
								if(SQL_EXCEPTIONS.isConnectionFailure(e)){
									this.debug("Connection failure. Trying again.", e);
									this.reconnect(this.connection);
								}
								else if(SQL_EXCEPTIONS.isRollback(e)){
									this.debug("Statement rolled back. Trying again.", e);
								}
								else{
									this.warn("", e);
								}
							}
						}
					}
					
					this.initStatements();
					
					//remove program entry
					while(true){
						try{
							SQL_Translation.sqlSetProgramInstsRemove(this.removeProgramInst, namespace, simple_name, instance);
							int updateCount = this.removeProgramInst.executeUpdate();
							assert updateCount >= 0 : "Database bug: delete statement returned negative update count.";
							break;
						} catch (SQLException e) {
							if(SQL_EXCEPTIONS.isConnectionFailure(e)){
								this.debug("Connection failure. Trying again.", e);
								this.reconnect(this.connection);
							}
							else if(SQL_EXCEPTIONS.isRollback(e)){
								this.debug("Statement rolled back. Trying again.", e);
							}
							else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
								this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
								this.initStatements();
							}
							else{
								this.error("", e);
								throw new RuntimeException(e);
							}
						}
					}
				}

				try{this.getProgramInstSchema.clearParameters();} catch (SQLException e) {}
				try{this.removeProgramInst.clearParameters();} catch (SQLException e) {}
			}
			else{
				if(this.existsProgInst(namespace, simple_name, instance)){
					IllegalStateException e = new IllegalStateException("Program could not be locked.");
					this.debug("Program could not be locked. aquireLock() did not succeed.", e);
					throw e;
				}
			}
		} catch (SQLException e) {
			this.error("",e);
			throw new RuntimeException(e);
		}
	}
	
	@Override
	public synchronized FixedArrayNavigableSet<String> availablePrograms(){
		ResultSet programs;
		while(true){
			try {
				programs = this.getProgramInsts.executeQuery();
				
				ArrayList<String> programsList = new ArrayList<String>();
				
				String full_name;
				while(programs.next()){
					full_name = SQL_Translation.sqlProgramInstsReadFullName(programs);
					programsList.add(full_name);
				}

				try {
					programs.close();
				} catch (SQLException e) {

				}
				
				return new FixedArrayNavigableSet<String>(TSA_SyntaxPatterns.programInstFullNameComparator ,false, false, programsList.toArray(new String[programsList.size()]));
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	@Override
	public synchronized void clearProgramInsts(){
		super.clearProgramInsts();
	}
	
	@Override
	public synchronized SQL_ExecutableProgram getProgram(final String namespace, final String simple_name, final String instance){		
		//TODO close result sets
		int program_id;
		int bufferMaxNumberOfMinTimes;
		int queryMaxNumberOfInputMinTimes;
		int queryMaxNumberOfOutputMinTimes;
		String[] schemas;
		final String meta_schema;
		final String input_schema;
		final String working_schema;
		final String output_schema;
		final String log_schema;
		while(true){
			try {
				//remove schemas (removes all tables in schema, too)
				SQL_Translation.sqlSetGetProgramInst(this.getProgramInstSchema, namespace, simple_name, instance);
				ResultSet program = this.getProgramInstSchema.executeQuery();
				this.getProgramInstSchema.clearParameters();
				
				if(program.next()){
					program_id = SQL_Translation.sqlProgramInstReadId(program);
					bufferMaxNumberOfMinTimes = SQL_Translation.sqlProgramInstReadBufferMaxNumberOfMinTimes(program);
					queryMaxNumberOfOutputMinTimes = SQL_Translation.sqlProgramInstReadQueryMaxNumberOfOutputMinTimes(program);
					queryMaxNumberOfInputMinTimes = SQL_Translation.sqlProgramInstReadQueryMaxNumberOfInputMinTimes(program);
					schemas = SQL_Translation.sqlProgramInstReadSchemas(program);					
					break;
				}
				else{
					IllegalArgumentException e = new IllegalArgumentException("No such program available: namespace = " + namespace + " , simple name = " + simple_name);
					this.debug("", e);
					throw e;
				}
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
		
		meta_schema = schemas[0];
		input_schema = schemas[1];
		working_schema = schemas[2];
		output_schema = schemas[3];
		log_schema = schemas[4];
		
		while(true){
			try {
					
				SQL_ExecutableProgram executable_program = new SQL_ExecutableProgram(program_id, namespace, simple_name, instance, meta_schema, input_schema, working_schema, output_schema, log_schema, this, false);
				
				ResultSet buffers;
				
				buffers = this.statement.executeQuery(SQL_Translation.sqlGetBufferMetaEntries(meta_schema, bufferMaxNumberOfMinTimes));
				
				int buffer_id;
				String buffer_prog_package;
				String buffer_simple_name;
				DTSA_BufferType buffer_type;
				String buffer_table_name;
				FixedArrayNavigableSet<String> buffer_minTimes;				
				
				while(buffers.next()){
					buffer_id = SQL_Translation.sqlReadBufferMetaEntryId(buffers);
					buffer_prog_package = SQL_Translation.sqlReadBufferMetaEntryProgPackage(buffers);
					buffer_simple_name = SQL_Translation.sqlReadBufferMetaEntrySimpleName(buffers);					
					buffer_type = SQL_Translation.sqlReadBufferMetaEntryType(buffers);
					buffer_table_name = SQL_Translation.sqlReadBufferMetaEntryTable(buffers);
					buffer_minTimes = SQL_Translation.sqlReadBufferMetaEntryMinTimes(buffers);
					
					PreparedStatement getAttributeEntries = SQL_Translation.prepareStatement(this.connection, SQL_Translation.sqlGetAttributeMetaEntries(meta_schema), this.usePreparedStatements);
					
					SQL_Translation.sqlSetGetAttributeMetaEntries(getAttributeEntries, buffer_type, TSA_SyntaxPatterns.bufferFullName(buffer_prog_package, buffer_simple_name));
					ResultSet attributes = getAttributeEntries.executeQuery();
									
					new SQL_StreamBuffer(executable_program, buffer_id, buffer_prog_package, buffer_simple_name, SQL_Translation.sqlReadAttributeMetaEntriesSQLTypeMapping(attributes), buffer_minTimes, buffer_type, buffer_table_name);
				}
				
				try {buffers.close();} catch (SQLException e1) {}
				
				ResultSet queries;
				
				queries = this.statement.executeQuery(SQL_Translation.sqlGetQueryMetaEntries(meta_schema, queryMaxNumberOfOutputMinTimes, queryMaxNumberOfInputMinTimes));
				
				int query_id;
				String query_prog_package;
				DTSA_QueryType query_type;
				boolean isMonotonic;
				DTSA_BufferType query_input_type;
				FixedArrayNavigableSet<String> query_inputs;
				String query_output;
				String query_sql_eval_statement;
				FixedArrayNavigableSet<String> query_referencedClocks;
				TimeFunction query_min_times_computation;
				TimeFunction query_min_time_clock_increases_computation;
				long lastRound;
				long[] inputMinTimes;
							
				while(queries.next()){
					query_id = SQL_Translation.sqlReadQueryMetaEntryId(queries);
					query_prog_package = SQL_Translation.sqlReadQueryMetaEntryProgPackage(queries);
					query_type = SQL_Translation.sqlReadQueryMetaEntryType(queries);
					isMonotonic = SQL_Translation.sqlReadQueryMetaEntryIsMonotonic(queries);
					query_input_type = SQL_Translation.sqlReadQueryMetaEntryInputType(queries);
					query_inputs = SQL_Translation.sqlReadQueryMetaEntryInputs(queries);
					query_output = SQL_Translation.sqlReadQueryMetaEntryOutput(queries);
					query_sql_eval_statement = SQL_Translation.sqlReadQueryMetaEntrySQL_Eval_Statement(queries);
					query_referencedClocks = SQL_Translation.sqlReadQueryMetaEntryReferencedClocks(queries);
					query_min_times_computation = SQL_Translation.sqlReadQueryMetaEntryMinTimesComputation(queries);
					query_min_time_clock_increases_computation = SQL_Translation.sqlReadQueryMetaEntryMinTimeClockIncreasesComputation(queries);
					lastRound = SQL_Translation.sqlReadQueryMetaEntryLastRound(queries);
					inputMinTimes = SQL_Translation.sqlReadQueryMetaEntryInputMinTimes(queries);
					
					//TODO consistency checks
					
					
					
					new SQL_ExecutableQuery(executable_program, query_id, query_type, query_prog_package, isMonotonic, query_sql_eval_statement, query_referencedClocks, query_min_times_computation, query_min_time_clock_increases_computation, query_input_type, query_inputs, query_output, lastRound, inputMinTimes);
				}
				
				try {queries.close();} catch (SQLException e) {}
				
				return executable_program;
		
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	@Override
	public void close(){
		try{ this.disconnect(); } catch (Exception e) {}				
	}
	
	@Override
	protected void finalize() throws Throwable{
		try{ this.disconnect(); } catch (Exception e) {}		
	}

}
