package de.lmu.ifi.pms.cep.event_mill.sql;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.SQLFeatureNotSupportedException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.simpleframework.xml.core.Persister;

import de.lmu.ifi.pms.cep.COMMON_CONSTANTS;
import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableProgramState;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.ClockFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID;
import de.lmu.ifi.pms.cep.tsa.conditions.DurationValue;
import de.lmu.ifi.pms.cep.tsa.conditions.Greatest;
import de.lmu.ifi.pms.cep.tsa.conditions.Least;
import de.lmu.ifi.pms.cep.tsa.conditions.MinTime;
import de.lmu.ifi.pms.cep.tsa.conditions.Minus;
import de.lmu.ifi.pms.cep.tsa.conditions.Plus;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TupleFunction;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.time.functions.AddDuration;
import de.lmu.ifi.pms.cep.tsa.time.functions.ConstantTimeFunction;
import de.lmu.ifi.pms.cep.tsa.time.functions.IdentityTimeFunction;
import de.lmu.ifi.pms.cep.tsa.time.functions.MapTimeFunction;
import de.lmu.ifi.pms.cep.tsa.time.functions.ParallelCompositeTimeFunction;
import de.lmu.ifi.pms.cep.tsa.time.functions.SequencialCompositeTimeFunction;
import de.lmu.ifi.pms.cep.tsa.time.functions.TimeConstants;
import de.lmu.ifi.pms.cep.tsa.time.functions.TimeFunction;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedEnumMap;


public class SQL_Translation
{
	public static final long block_counter_initial_value = 0l;
	
	
	public final static String schema_name_RegEx = "[a-zA-Z_][a-zA-Z0-9_]*";
	public final static String table_name_RegEx = "[a-zA-Z_][a-zA-Z0-9_]*";
	public final static String attribute_name_RegEx = "[a-zA-Z_][a-zA-Z0-9_]*";
	public final static String sequence_name_RegEx = "[a-zA-Z_][a-zA-Z0-9_]*";
	
	public final static String engine_meta_schema_name_RegEx = schema_name_RegEx;
	
	public final static String program_schema_name_RegEx = schema_name_RegEx;
	
	//public final static String input_last_minTime_parameter_local_regEx = "(input[.]last):" + TSA_SyntaxPatterns.LocalMinTimeNameRegEx;
	public final static String input_last_minTime_parameter_qualified_regEx = "(input[.]last):" + TSA_SyntaxPatterns.QualifiedMinTimeNameRegEx;
	//public final static String input_now_minTime_parameter_local_regEx = "(input[.]now):" + TSA_SyntaxPatterns.LocalMinTimeNameRegEx;
	public final static String input_now_minTime_parameter_qualified_regEx = "(input[.]now):" + TSA_SyntaxPatterns.QualifiedMinTimeNameRegEx;
	public final static String output_last_minTime_parameter_qualified_regEx = "(output[.]last):" + TSA_SyntaxPatterns.QualifiedMinTimeNameRegEx;
	public final static String output_now_minTime_parameter_qualified_regEx = "(output[.]now):" + TSA_SyntaxPatterns.QualifiedMinTimeNameRegEx;
	public final static String clock_now_parameter_regEx = "(clock[.]now):" + "(" + TSA_SyntaxPatterns.ClockNameRegEx + ")";
	public final static Pattern clock_parameter_Pattern = Pattern.compile(clock_now_parameter_regEx);
	public final static Pattern clock_parameter_inSQL_Pattern = Pattern.compile("/[*]'(?:"+ clock_now_parameter_regEx +")'[*]/[?]");
	//public final static String parameter_local_regEx = input_last_minTime_parameter_local_regEx + "|" + input_now_minTime_parameter_local_regEx + "|" + clock_last_minTime_parameter_regEx + "|" + clock_now_minTime_parameter_regEx + "|" + clock_now_parameter_regEx;
	//public final static Pattern parameter_local_Pattern = Pattern.compile(parameter_local_regEx);
	public final static String parameter_qualified_regEx = input_last_minTime_parameter_qualified_regEx+ "|" + input_now_minTime_parameter_qualified_regEx+ "|" + output_last_minTime_parameter_qualified_regEx+ "|" + output_now_minTime_parameter_qualified_regEx + "|" + clock_now_parameter_regEx + "|" + "(query).((?:id)|(?:round))";
	public final static Pattern parameter_qualified_Pattern = Pattern.compile(parameter_qualified_regEx);
	
	//public final static String parameter_inSQL_local_regEx = "/[*](?:" + parameter_local_regEx +")[*]/[?]";
	//public final static Pattern parameter_inSQL_local_Pattern = Pattern.compile(parameter_inSQL_local_regEx);
	public final static String parameter_inSQL_qualified_regEx = "/[*]'(" + parameter_qualified_regEx +")'[*]/[?]";
	public final static Pattern parameter_inSQL_qualified_Pattern = Pattern.compile(parameter_inSQL_qualified_regEx);
	
	public static String sqlCreateSchema(final String schema_name){
		if(schema_name.matches(schema_name_RegEx)){		
			return String.format(
					"CREATE SCHEMA \"%1$s\";"
			,schema_name		
			);
		}
		else{
			throw new IllegalArgumentException(String.format("Illegal SQL schema name: %1$s", schema_name));
		}
	}
	
	public static String sqlDeleteSchema(final String schema_name){
		if(schema_name.matches(schema_name_RegEx)){		
			return String.format(
					"DROP SCHEMA \"%1$s\";"
			,schema_name		
			);
		}
		else{
			throw new IllegalArgumentException(String.format("Illegal SQL schema name: %1$s", schema_name));
		}
	}
	
	public static boolean existsSchema(final DatabaseMetaData databaseMetaData, final String catalog, final String schema_name) throws SQLException{
		while(true){
			try{
				ResultSet schemas = databaseMetaData.getSchemas(catalog, schema_name);
				return schemas.next();
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isRollback(e)){
					continue;
				}
				else{
					throw e;
				}
			}
		}
	}
	
	
	public static boolean existsTable(final DatabaseMetaData databaseMetaData, final String catalog, final String schema_name, final String table_name) throws SQLException{
		while(true){
			try{
				//TODO get right table type from driver
				ResultSet tables = databaseMetaData.getTables(catalog, schema_name, table_name, new String[]{"TABLE"});
				return tables.next();
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isRollback(e)){
					continue;
				}
				else{
					throw e;
				}
			}
		}
	}
	
	public static String sqlInsertStatement(final String program_meta_schema, final String table_schema, final String table, final int buffer_id, final int bulkSize, final String block_attribute, final String... attributes){
		assert table_schema.matches(schema_name_RegEx);
		assert table.matches(table_name_RegEx);
		String getBlockCounterStatement = SQL_Translation.sqlGetBufferBlockCounter(program_meta_schema, buffer_id);
		getBlockCounterStatement = getBlockCounterStatement.substring(0, getBlockCounterStatement.length() - 1);//remove ;
		
		StringBuilder statement = new StringBuilder();
		statement.append("INSERT INTO \"");
		statement.append(table_schema);
		statement.append("\".\"");
		statement.append(table);
		statement.append("\"");
		if(block_attribute != null | attributes.length >= 0){
			statement.append(" \n");
			statement.append("	(");
			
			//block attribute
			if(block_attribute != null){
				statement.append("\"");
				statement.append(block_attribute);
				statement.append("\"");
				statement.append(", ");
			}
			
			//main attributes
			
			for(int attribute = 0; attribute < attributes.length; ++attribute){
				statement.append("\"");
				assert attributes[attribute].matches(attribute_name_RegEx);
				statement.append(attributes[attribute]);
				statement.append("\"");
				statement.append(", ");
			}
			statement.setLength(statement.length() - ", ".length());
			statement.append("	) \n");
			statement.append("VALUES \n");
			for(int size = 0; size < bulkSize; ++size){
				statement.append("	(");
				//block attribute
				if(block_attribute != null){
					statement.append("(");
					statement.append(getBlockCounterStatement);
					statement.append(")");
					statement.append(", ");
				}
				
				//main attributes
				for(int attribute = 0; attribute < attributes.length; ++attribute){
					statement.append("?");
					statement.append(", ");
				}
				statement.setLength(statement.length() - ", ".length());
				statement.append("	), \n");
			}
			statement.setLength(statement.length() - ", \n".length());
			statement.append(" \n");
		}
		statement.append(";");
		return statement.toString();
	}
	
	public static String sqlReadRangeStatement(final String schema, final String table, final String rangeAttribute, final boolean startInclusive, final boolean endInclusive, final String... attributes){
		assert schema.matches(schema_name_RegEx);
		assert table.matches(table_name_RegEx);
		StringBuilder statement = new StringBuilder();
		statement.append("SELECT ");
		for(int attribute = 0; attribute < attributes.length; ++attribute){
			statement.append("\"");
			assert attributes[attribute].matches(attribute_name_RegEx);
			statement.append(attributes[attribute]);
			statement.append("\", ");
		}
		statement.setLength(statement.length() - ", ".length());
		statement.append(" \n");
		statement.append("FROM \"");
		statement.append(schema);
		statement.append("\".\"");
		statement.append(table);
		statement.append("\" \n");
		statement.append("WHERE ");
		statement.append("? ");
		if(startInclusive){
			statement.append("<=");	
		}
		else{
			statement.append("<");
		}
		statement.append(" \"");
		statement.append(rangeAttribute);
		statement.append("\" AND \"");
		statement.append(rangeAttribute);
		statement.append("\" ");
		if(endInclusive){
			statement.append("<=");	
		}
		else{
			statement.append("<");
		}		
		statement.append(" ?");
		statement.append(" \n");
		statement.append(";");
		return statement.toString();
	}
	
	public static PreparedStatement prepareStatement(final Connection connection, final String sql, final boolean usePreparedStatements) throws SQLException{
		if(usePreparedStatements){
			return connection.prepareStatement(sql);
		}
		else{
			return new SimulatedPreparedStatement(connection, sql);
		}
	}
	
	public static final String propertiesTableName = "properties";
	public static final String programSpecsTableName = "program_specs";
	public static final String programInstsTableName = "program_insts";
	
	public static final String programBufferMetaTableName = "buffers";
	public static final String programBufferAttributeMetaTableName = "attributes";
	public static final String programQueryMetaTableName = "queries";
	public static final String programInputStreamsMetaTableName = "inputs";
	
	public static final String tupleIdSeqenceName = COMMON_CONSTANTS.TUPLE_ID_CLOCK_NAME;
	
	static{
		assert programSpecsTableName.matches(table_name_RegEx);
		assert programInstsTableName.matches(table_name_RegEx);
		
		assert programBufferMetaTableName.matches(table_name_RegEx);
		assert programQueryMetaTableName.matches(table_name_RegEx);
	}
	
	public static String sqlPropertiesTableDefinition(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		String tableDefinition = 
				"CREATE TABLE \"" + engine_meta_schema_name + "\".\"" + propertiesTableName + "\" (\n" +  
				"	\"key\" CLOB PRIMARY KEY,\n" + 
				"	\"value\" CLOB NOT NULL" +  
				");"
				;
		return tableDefinition;
	}
	
	public static String sqlPropertiesTableDeletion(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		return "DROP TABLE \"" + engine_meta_schema_name + "\".\"" + propertiesTableName + "\";";
	}
	
	public static String sqlPropertyAdd(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return	"INSERT INTO \"" + engine_meta_schema_name + "\".\"" + propertiesTableName + "\" " + "\n" +
				"	(\"key\", \"value\")\n" +
				"VALUES \n" +
				"	( ?, ?)\n" +
				";"
		;
	}
	
	public static void sqlSetPropertyAdd(final PreparedStatement statement, final String key, final String value) throws SQLException{
		statement.setString(1, key);
		statement.setString(2, value);
	}
	
	public static String sqlGetProperty(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return	"SELECT \"value\" FROM \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.propertiesTableName + "\" " + "\n" + 
				"WHERE \n"+
				"   \"key\" = ? \n" +  
				";"
		;
	}
	
	public static void sqlSetGetProperty(final PreparedStatement statement, final String key) throws SQLException{
		statement.setString(1, key);
	}
	
	public static String sqlProgramSpecsTableDefinition(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		String tableDefinition = 
				"CREATE TABLE \"" + engine_meta_schema_name + "\".\"" + programSpecsTableName + "\" (\n" + 
				"	\"id\" INT AUTO_INCREMENT PRIMARY KEY,\n" + 
				"	\"namespace\" CLOB NOT NULL,\n" + 
				"	\"simple_name\" CLOB NOT NULL,\n" + 
				"	\"description\" CLOB NOT NULL DEFAULT '',\n" + 
				"	\"xml_spec\" CLOB NOT NULL,\n" +   
				"	CONSTRAINT \"program_spec_unique_name\" UNIQUE (\"namespace\", \"simple_name\")\n" + 
				");"
				;
		return tableDefinition;
	}
	
	public static String sqlProgramSpecsTableDeletion(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		return "DROP TABLE \"" + engine_meta_schema_name + "\".\"" + programSpecsTableName + "\";";
	}
	
	public static String sqlProgramSpecsAdd(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return	"INSERT INTO \"" + engine_meta_schema_name + "\".\"" + programSpecsTableName + "\" " + "\n" +
				"	(\"namespace\", \"simple_name\", \"description\", \"xml_spec\")\n" +
				"VALUES \n" +
				"	( ?, ?, ?, ?)\n" +
				";"
		;
	}
	
	public static void sqlSetProgramSpecsAdd(final PreparedStatement statement, final SQL_Program_Spec program_spec) throws SQLException{
		statement.setString(1, program_spec.namespace);
		statement.setString(2, program_spec.simple_name);
		statement.setString(3, program_spec.description);
		statement.setString(4, SQL_Program_Spec.serialize(program_spec));
	}
	
	public static String sqlGetProgramSpec(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return	"SELECT \"xml_spec\" FROM \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programSpecsTableName + "\" " + "\n" + 
				"WHERE \n"+
				"   \"namespace\" = ? \n" + 
				"AND \n" + 
				"   \"simple_name\" = ? \n" + 
				";"
		;
	}
	
	public static void sqlSetGetProgramSpec(final PreparedStatement statement, final String namespace, final String simple_name) throws SQLException{
		statement.setString(1, namespace);
		statement.setString(2, simple_name);
	}
	
	public static String sqlGetProgramSpecs(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return	"SELECT \"namespace\", \"simple_name\", \"description\" FROM \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programSpecsTableName + "\" " + "\n" + 
				"ORDER BY \"namespace\" ASC, \"simple_name\" ASC;"
		;
	}
	
	public static String sqlProgramSpecsReadFullName(final ResultSet programSpecs) throws SQLException{
		String namespace = programSpecs.getString("namespace");
		String simple_name = programSpecs.getString("simple_name");
		return TSA_SyntaxPatterns.programFullName(namespace, simple_name);
	}

	public static String sqlProgramSpecsRemove(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return	"DELETE FROM \"" + engine_meta_schema_name + "\".\"" + programSpecsTableName + "\" " + "\n" +
				"WHERE \n" +
				"	\"namespace\" = ? \n" +
				"	AND \n" +
				"	\"simple_name\" = ? \n" +
				";"
		;
	}
	
	public static void sqlSetProgramSpecsRemove(final PreparedStatement statement, final String namespace, final String simple_name) throws SQLException{
		assert namespace != null;
		statement.setString(1, namespace);
		statement.setString(2, simple_name);
	}
	
	public static String sqlProgramInstsTableDefinition(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		String tableDefinition = 
				"CREATE TABLE \"" + engine_meta_schema_name + "\".\"" + programInstsTableName + "\" (\n" + 
				"	\"id\" INT AUTO_INCREMENT PRIMARY KEY,\n" + 
				"	\"spec_ref\" INT REFERENCES \"" + engine_meta_schema_name +"\".\"" + programSpecsTableName + "\" ON UPDATE CASCADE ON DELETE CASCADE,\n" + 
				"	\"instance\" CLOB NOT NULL DEFAULT '',\n" + 
				"	\"buffer_max_number_of_min_times\" INT, \n" +
				"	\"query_max_number_of_output_min_times\" INT, \n" +
				"	\"query_max_number_of_input_min_times\" INT, \n" +
				"	\"meta_schema\" CLOB NOT NULL,\n" + 
				"	\"input_schema\" CLOB NOT NULL,\n" + 
				"	\"working_schema\" CLOB NOT NULL,\n" + 
				"	\"output_schema\" CLOB NOT NULL,\n" + 
				"	\"log_schema\"  CLOB NOT NULL,\n" + 
				"	\"state\" CLOB NOT NULL, \n" + 
				"	\"lock\" CLOB, \n" +
				"	CONSTRAINT \"program_unique_name\" UNIQUE (\"spec_ref\", \"instance\")\n" + 
				");"
				;
		return tableDefinition;
	}
	
	public static String sqlProgramInstsTableDeletion(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return "DROP TABLE \"" + engine_meta_schema_name + "\".\"program_insts\";";
	}
	
	public static String sqlProgramInstsAdd(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return	"INSERT INTO \"" + engine_meta_schema_name + "\".\"" + programInstsTableName + "\" " + "\n" +
				"	(\"spec_ref\", \"instance\", \"buffer_max_number_of_min_times\", \"query_max_number_of_output_min_times\", \"query_max_number_of_input_min_times\", \"meta_schema\", \"input_schema\", \"working_schema\", \"output_schema\", \"log_schema\", \"state\", \"lock\")\n" +
				"VALUES \n" +
				"	( " +
						"(" +
						"SELECT \"id\" FROM \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programSpecsTableName + "\" " + "\n" + 
						"WHERE \n"+
						"   \"namespace\" = ? \n" + 
						"AND \n" + 
						"   \"simple_name\" = ? \n" + 
				        ")" +
				"       , ?, ?, ?, ?, ?, ?, ?, ?, ?, '"+ ExecutableProgramState.CREATING.name() +"', 'LOCKED')\n" +
				";"
		;
	}
	
	public static void sqlSetProgramInstsAdd(final PreparedStatement statement, final SQL_Program_Spec program_spec, final String instance, final int bufferMaxNumberOfMinTimes, final int queryMaxNumberOfOutputMinTimes, final int queryMaxNumberOfInputMinTimes,  final String meta_schema, final String working_schema, final String input_schema, final String output_schema, final String log_schema) throws SQLException{
		int param = 0;
		statement.setString(++param, program_spec.namespace);
		statement.setString(++param, program_spec.simple_name);
		statement.setString(++param, instance);
		statement.setInt(++param, bufferMaxNumberOfMinTimes);
		statement.setInt(++param, queryMaxNumberOfOutputMinTimes);
		statement.setInt(++param, queryMaxNumberOfInputMinTimes);
		statement.setString(++param, meta_schema);
		statement.setString(++param, input_schema);
		statement.setString(++param, working_schema);
		statement.setString(++param, output_schema);
		statement.setString(++param, log_schema);
	}
	
	public static String sqlGetProgramInst(final String engine_meta_schema_name, final String program_full_name){
		String[] names = TSA_SyntaxPatterns.programInstFullName(program_full_name);
		return sqlGetProgramInst(engine_meta_schema_name, names[0], names[1], names[2]);
	}
	
	public static String sqlGetProgramInst(final String engine_meta_schema_name, final String namespace, final String simple_name, final String instance){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		assert namespace.matches(TSA_SyntaxPatterns.ProgramNamespaceRegEx);
		assert simple_name.matches(TSA_SyntaxPatterns.ProgramNameRegEx);
		assert instance.matches(TSA_SyntaxPatterns.program_instance_name_RegEx);
		
		return	"SELECT \"insts\".\"id\" AS \"id\", \"insts\".\"buffer_max_number_of_min_times\", \"query_max_number_of_output_min_times\", \"query_max_number_of_input_min_times\", \"insts\".\"meta_schema\" AS \"meta_schema\", \"insts\".\"input_schema\" AS \"input_schema\", \"insts\".\"working_schema\" AS \"working_schema\", \"insts\".\"output_schema\" AS \"output_schema\", \"insts\".\"log_schema\" AS \"log_schema\", \"insts\".\"state\" AS \"state\" FROM \n" + 
				"   \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programSpecsTableName + "\" AS \"specs\", " + "\n" +
				"   \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programInstsTableName + "\" AS \"insts\"" + "\n" + 
				"WHERE \n"+
				"   \"specs\".\"namespace\" = '" + namespace + "' \n" + 
				"AND \n" + 
				"   \"specs\".\"simple_name\" = '" + simple_name + "' \n" +
				"AND \n" + 
				"   \"insts\".\"instance\" = '" + instance + "' \n" +
				"AND \n" + 
				"   \"insts\".\"state\" <> '" + ExecutableProgramState.CREATING.name() + "'\n" +
				"AND \n" + 
				"   \"insts\".\"state\" <> '" + ExecutableProgramState.DELETING.name() + "'\n" +
				"AND \n" +
				"   \"specs\".\"id\" = \"insts\".\"spec_ref\" \n" +
				";"
		;
	}
	
	public static String sqlGetProgramInst(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return	"SELECT \"insts\".\"id\" AS \"id\", \"insts\".\"buffer_max_number_of_min_times\", \"query_max_number_of_output_min_times\", \"query_max_number_of_input_min_times\", \"insts\".\"meta_schema\" AS \"meta_schema\", \"insts\".\"input_schema\" AS \"input_schema\", \"insts\".\"working_schema\" AS \"working_schema\", \"insts\".\"output_schema\" AS \"output_schema\", \"insts\".\"log_schema\" AS \"log_schema\", \"insts\".\"state\" AS \"state\" FROM \n" + 
				"   \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programSpecsTableName + "\" AS \"specs\", " + "\n" +
				"   \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programInstsTableName + "\" AS \"insts\"" + "\n" + 
				"WHERE \n"+
				"   \"specs\".\"namespace\" = ? \n" + 
				"AND \n" + 
				"   \"specs\".\"simple_name\" = ? \n" +
				"AND \n" + 
				"   \"insts\".\"instance\" = ? \n" +
				"AND \n" + 
				"   \"insts\".\"state\" <> '" + ExecutableProgramState.CREATING.name() + "'\n" +
				"AND \n" + 
				"   \"insts\".\"state\" <> '" + ExecutableProgramState.DELETING.name() + "'\n" +
				"AND \n" +
				"   \"specs\".\"id\" = \"insts\".\"spec_ref\" \n" +
				";"
		;
	}
	
	public static void sqlSetGetProgramInst(final PreparedStatement statement, final String full_name) throws SQLException{
		String[] names = TSA_SyntaxPatterns.programInstFullName(full_name);
		sqlSetGetProgramInst(statement, names[0], names[1], names[2]);
	}
	
	public static void sqlSetGetProgramInst(final PreparedStatement statement, final String namespace, final String simple_name, final String instance) throws SQLException{
		assert namespace.matches(TSA_SyntaxPatterns.ProgramNamespaceRegEx);
		assert simple_name.matches(TSA_SyntaxPatterns.ProgramNameRegEx);
		assert instance.matches(TSA_SyntaxPatterns.program_instance_name_RegEx);
		
		statement.setString(1, namespace);
		statement.setString(2, simple_name);
		statement.setString(3, instance);
	}
	
	public static int sqlProgramInstReadId(final ResultSet schemas) throws SQLException{
		return schemas.getInt("id");
	}
	
	public static int sqlProgramInstReadBufferMaxNumberOfMinTimes(final ResultSet resultSet) throws SQLException{
		return resultSet.getInt("buffer_max_number_of_min_times");
	}
	
	public static int sqlProgramInstReadQueryMaxNumberOfOutputMinTimes(final ResultSet resultSet) throws SQLException{
		return resultSet.getInt("query_max_number_of_output_min_times");
	}
	
	public static int sqlProgramInstReadQueryMaxNumberOfInputMinTimes(final ResultSet resultSet) throws SQLException{
		return resultSet.getInt("query_max_number_of_input_min_times");
	}
	
	public static String[] sqlProgramInstReadSchemas(final ResultSet schemas) throws SQLException{
		return new String[]{schemas.getString("meta_schema"), schemas.getString("input_schema"), schemas.getString("working_schema"), schemas.getString("output_schema"), schemas.getString("log_schema")};
	}
	
	public static ExecutableProgramState sqlProgramInstReadState(final ResultSet state) throws SQLException{
		return ExecutableProgramState.valueOf(state.getString("state"));
	}
	
	public static String sqlLockProgramInst(final String engine_meta_schema_name, final String namespace, final String simple_name, final String instance){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		assert TSA_SyntaxPatterns.isValidNamespace(namespace);
		assert TSA_SyntaxPatterns.isValidProgramSimpleName(simple_name);
		assert TSA_SyntaxPatterns.isValidProgramInstanceName(instance);
		
		return	"UPDATE \n" + 
		"   \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programInstsTableName + "\"" + "\n" +
		"SET \"lock\" = '" + "LOCKED" + "' \n" +
		"WHERE \n"+
		"   \"instance\" = '" + instance + "'\n" +
		" AND \n" +
		"   \"spec_ref\" = 	(" +
						"SELECT \"id\" FROM \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programSpecsTableName + "\" " + "\n" + 
						"WHERE \n"+
						"   \"namespace\" = '" + namespace + "'\n" + 
						"AND \n" + 
						"   \"simple_name\" = '" + simple_name + "'\n" + 
						") \n" +
		" AND \n" +
		"   \"lock\" IS NULL" +
		";"
		;
	}
	
	public static String sqlUnlockProgramInst(final String engine_meta_schema_name, final String namespace, final String simple_name, final String instance){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		assert TSA_SyntaxPatterns.isValidNamespace(namespace);
		assert TSA_SyntaxPatterns.isValidProgramSimpleName(simple_name);
		assert TSA_SyntaxPatterns.isValidProgramInstanceName(instance);
		
		return	"UPDATE \n" + 
		"   \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programInstsTableName + "\"" + "\n" +
		"SET \"lock\" = " + "NULL" + " \n" +
		"WHERE \n"+
		"   \"instance\" = '" + instance + "'\n" +
		" AND \n" +
		"   \"spec_ref\" = 	(" +
						"SELECT \"id\" FROM \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programSpecsTableName + "\" " + "\n" + 
						"WHERE \n"+
						"   \"namespace\" = '" + namespace + "'\n" + 
						"AND \n" + 
						"   \"simple_name\" = '" + simple_name + "'\n" + 
						") \n" +
		" AND \n" +
		"   \"lock\" IS NOT NULL" +
		";"
		;
	}
	
	public static String sqlLockProgramInst(final String engine_meta_schema_name, final int id){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return	"UPDATE \n" + 
		"   \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programInstsTableName + "\"" + "\n" +
		"SET \"lock\" = '" + "LOCKED" + "' \n" +
		"WHERE \n"+
		"   \"id\" = '" + id + "'\n" +
		" AND \n" +
		"   \"lock\" IS NULL" +
		";"
		;
	}
	
	public static String sqlUnlockProgramInst(final String engine_meta_schema_name, final int id){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return	"UPDATE \n" + 
		"   \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programInstsTableName + "\"" + "\n" +
		"SET \"lock\" = " + "NULL" + " \n" +
		"WHERE \n"+
		"   \"id\" = '" + id + "'\n" +
		" AND \n" +
		"   \"lock\" IS NOT NULL" +
		";"
		;
	}
	
	public static String sqlGetProgramInstState(final String engine_meta_schema_name, final int id){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return	"SELECT \"state\" \n" +
				"FROM \n" + 
				"   \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programInstsTableName + "\"" + "\n" + 
				"WHERE \n"+
				"   \"id\" = '" + id + "' \n" +
				";"
		;
	}
	
	public static String sqlSetProgramInstState(final String engine_meta_schema_name, final String namespace, final String simple_name, final String instance, final ExecutableProgramState state){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		assert TSA_SyntaxPatterns.isValidNamespace(namespace);
		assert TSA_SyntaxPatterns.isValidProgramSimpleName(simple_name);
		assert TSA_SyntaxPatterns.isValidProgramInstanceName(instance);
		
		return	"UPDATE \n" + 
		"   \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programInstsTableName + "\"" + "\n" +
		"SET \"state\" = '" + state.name() + "' \n" +
		"WHERE \n"+
		"   \"instance\" = '" + instance + "'\n" +
		" AND \n" +
		"   \"spec_ref\" = 	(" +
						"SELECT \"id\" FROM \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programSpecsTableName + "\" " + "\n" + 
						"WHERE \n"+
						"   \"namespace\" = '" + namespace + "'\n" + 
						"AND \n" + 
						"   \"simple_name\" = '" + simple_name + "'\n" + 
						") \n" +
		";"
		;
	}
	
	public static String sqlSetProgramInstState(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return	"UPDATE \n" + 
		"   \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programInstsTableName + "\"" + "\n" +
		"SET \"state\" = ? \n" +
		"WHERE \n"+
		"   \"instance\" = ?\n" +
		" AND \n" +
		"   \"spec_ref\" = 	(" +
						"SELECT \"id\" FROM \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programSpecsTableName + "\" " + "\n" + 
						"WHERE \n"+
						"   \"namespace\" = ? \n" + 
						"AND \n" + 
						"   \"simple_name\" = ? \n" + 
						") \n" +
		";"
		;
	}
	
	public static void sqlSetSetProgramInstState(final PreparedStatement statement, final String namespace, final String simple_name, final String instance, final ExecutableProgramState state) throws SQLException{
		assert TSA_SyntaxPatterns.isValidNamespace(namespace);
		assert TSA_SyntaxPatterns.isValidProgramSimpleName(simple_name);
		assert TSA_SyntaxPatterns.isValidProgramInstanceName(instance);
		int parameter = 0;
		statement.setString(++parameter, instance);
		statement.setString(++parameter, namespace);
		statement.setString(++parameter, simple_name);
		statement.setString(++parameter, state.name());
	}
	
	public static String sqlSetProgramInstState(final String engine_meta_schema_name, final int id){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return	"UPDATE \n" + 
				"   \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programInstsTableName + "\"" + "\n" +
				"SET \"state\" = ? \n" +
				"WHERE \n"+
				"   \"id\" = '" + id + "' \n" +
				";"
		;
	}
	
	public static void sqlSetSetProgramInstState(final PreparedStatement statement, final ExecutableProgramState state) throws SQLException{
		int parameter = 0;
		statement.setString(++parameter, state.name());
	}
	
	public static String sqlGetProgramInsts(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);

		return	"SELECT \"specs\".\"namespace\" AS \"namespace\", \"specs\".\"simple_name\" AS \"simple_name\", \"insts\".\"instance\" AS \"instance\" " +
				"FROM \n" +
				"   \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programSpecsTableName + "\" AS \"specs\", " + "\n" +
				"   \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programInstsTableName + "\" AS \"insts\"" + "\n" +
				"WHERE \n"+
				"   \"specs\".\"id\" = \"insts\".\"spec_ref\"" +
				"AND \n" + 
				"   \"insts\".\"state\" <> '" + ExecutableProgramState.CREATING.name() + "'\n" +
				"AND \n" + 
				"   \"insts\".\"state\" <> '" + ExecutableProgramState.DELETING.name() + "'\n" +
				"ORDER BY \"namespace\" ASC, \"simple_name\" ASC, \"instance\" ASC \n" +
				";"
		;
	}
	
	public static String sqlProgramInstsReadFullName(final ResultSet programInsts) throws SQLException{
		String namespace = programInsts.getString("namespace");
		String simple_name = programInsts.getString("simple_name");
		String instance = programInsts.getString("instance");
		return TSA_SyntaxPatterns.programInstFullName(namespace, simple_name, instance);
	}
	
	public static String sqlProgramInstsRemove(final String engine_meta_schema_name){
		assert engine_meta_schema_name.matches(engine_meta_schema_name_RegEx);
		
		return	"DELETE FROM \"" + engine_meta_schema_name + "\".\"" + programInstsTableName + "\" " + "\n" +
				"WHERE \n" +
				"   \"instance\" = ? \n" +
				" AND \n" +
				"   \"spec_ref\" = (\n" +
				"                      SELECT \"id\" FROM \"" + engine_meta_schema_name + "\".\"" + SQL_Translation.programSpecsTableName + "\" " + "\n" + 
				"                      WHERE \n"+
				"                         \"namespace\" = ? \n" + 
				"                      AND \n" + 
				"                         \"simple_name\" = ? \n" + 
		        "                    ) \n" +
				";"
		;
	}
	
	public static void sqlSetProgramInstsRemove(final PreparedStatement statement, final String namespace, final String simple_name, final String instance) throws SQLException{
		statement.setString(1, instance);
		statement.setString(2, namespace);
		statement.setString(3, simple_name);
	}
	
	public static String SQL_BasicTableName(final String tsa_qualifiedBufferName){
		assert tsa_qualifiedBufferName.matches(TSA_SyntaxPatterns.BufferFullNameRegEx);
		
		String sql_tableName = tsa_qualifiedBufferName.replaceAll("[_]", "__").replaceAll("[.]", "_o_").replaceAll("-", "_d_").replaceAll("[$]", "_D_");
		
		assert sql_tableName.matches(table_name_RegEx);
		return sql_tableName;
	}
	
	public static FixedArrayNavigableMap<String, String> sql_table_name_mapping(final FixedArrayNavigableSet<String> tsa_buffer_names, final int maxSQLNameLength){
		//FIXME hidden names? Dot to the beginning?
		int maxSQLLongNameSize = 0;
		String[] sqlNames = new String[tsa_buffer_names.size()];
		for(int name = 0; name < sqlNames.length; ++name){
			sqlNames[name] = SQL_BasicTableName(tsa_buffer_names.get(name));
			maxSQLLongNameSize = (maxSQLLongNameSize < sqlNames[name].length() ? sqlNames[name].length() : maxSQLLongNameSize);
		}
		if(maxSQLLongNameSize > maxSQLNameLength){
			int idLength = String.valueOf(sqlNames.length).length();
			if(idLength + 1 > maxSQLNameLength){
				throw new RuntimeException("Encoding of type names to SQL table names is impossible. The maximum table name length is too short.");
			}
			else{
				for(int name = 0; name < sqlNames.length; ++name){
					String longSqlName = sqlNames[name];
					int startPostFix = longSqlName.length() - (maxSQLNameLength - 1 - idLength);
					if(startPostFix < 0){
						startPostFix = 0;
					}
					sqlNames[name] = longSqlName.substring(0, 1) + String.format("%" + idLength + "d", name) + longSqlName.substring(startPostFix, longSqlName.length());
					assert sqlNames[name].length() <= maxSQLNameLength;
					maxSQLLongNameSize = (maxSQLLongNameSize < sqlNames[name].length() ? sqlNames[name].length() : maxSQLLongNameSize);
				}
			}
		}
		return new FixedArrayNavigableMap<String, String>(tsa_buffer_names, new FixedArray<String>(false, sqlNames));
	}
	
	public static String TSA_BufferName(final String sql_basicTableName){
		assert sql_basicTableName.matches(table_name_RegEx);
		
		String tsa_bufferName = sql_basicTableName.replaceAll("_D_","\\$").replaceAll("_d_", "-").replaceAll("_o_", ".").replaceAll("__", "_");
		
		assert TSA_SyntaxPatterns.isValidBufferFullName(tsa_bufferName);
		return tsa_bufferName;
	}
	
	public static FixedArrayNavigableMap<String, String> sql_attribute_name_mapping(final FixedArrayNavigableSet<String> tsa_attribute_names, final int maxSQLNameLength){
		//FIXME hidden names? Dot to the beginning?
		int maxSQLLongNameSize = 0;
		String[] sqlNames = new String[tsa_attribute_names.size()];
		for(int name = 0; name < sqlNames.length; ++name){
			sqlNames[name] = SQL_BasicTableName(tsa_attribute_names.get(name));
			maxSQLLongNameSize = (maxSQLLongNameSize < sqlNames[name].length() ? sqlNames[name].length() : maxSQLLongNameSize);
		}
		if(maxSQLLongNameSize > maxSQLNameLength){
			int idLength = String.valueOf(sqlNames.length).length();
			if(idLength + 1 > maxSQLNameLength){
				throw new RuntimeException("Encoding of type names to SQL table names is impossible. The maximum table name length is too short.");
			}
			else{
				for(int name = 0; name < sqlNames.length; ++name){
					String longSqlName = sqlNames[name];
					int startPostFix = longSqlName.length() - (maxSQLNameLength - 1 - idLength);
					if(startPostFix < 0){
						startPostFix = 0;
					}
					sqlNames[name] = longSqlName.substring(0, 1) + String.format("%" + idLength + "d", name) + longSqlName.substring(startPostFix, longSqlName.length());
					assert sqlNames[name].length() <= maxSQLNameLength;
					maxSQLLongNameSize = (maxSQLLongNameSize < sqlNames[name].length() ? sqlNames[name].length() : maxSQLLongNameSize);
				}
			}
		}
		return new FixedArrayNavigableMap<String, String>(tsa_attribute_names, new FixedArray<String>(false, sqlNames));
	}
	
	public static String SQL_BasicAttributeName(final String tsa_attributeName){
		assert tsa_attributeName.matches(TSA_SyntaxPatterns.AttributeNameRegEx);
		
		String sql_attributeName = tsa_attributeName.replaceAll("[_]", "__").replaceAll("[.][.]", "_oo_").replaceAll("[.]", "_o_").replaceAll("-", "_d_");
		
		assert sql_attributeName.matches(attribute_name_RegEx);
		return sql_attributeName;
	}
	
	public static String SQLTypeSpec(final BasicType basicType, final String program_meta_schema){
		return SQLTypeSpec(basicType.ID, program_meta_schema);
	}

	//FIXME translate default attributes and Primary key, foreign key and not null constaints
	public static String SQLTypeSpec(final BasicTypeID basicType, final String program_meta_schema){
		switch (basicType) {
		case TIMESTAMP: return SQLType(basicType) + " NOT NULL"; 
		case DURATION: return SQLType(basicType) + " NOT NULL";
		case INT: return SQLType(basicType) + " NOT NULL";
		case LONG: return SQLType(basicType) + " NOT NULL";
		case FLOAT: return SQLType(basicType) + " NOT NULL";
		case DOUBLE: return SQLType(basicType) + " NOT NULL";
		case BOOLEAN: return SQLType(basicType) + " NOT NULL";
		case STRING: return SQLType(basicType) + " NOT NULL";
		case IDENTIFIER: return SQLType(basicType) + " NOT NULL DEFAULT NEXT VALUE FOR \"" + program_meta_schema + "\".\"" + tupleIdSeqenceName + "\"";//FIXME more generic, NOT NULL and DEFAULTS aus constraints
		//FIXME better treatment of default values
		case REFERENCE: return SQLType(basicType) + " NOT NULL";
		}
		throw new Error();
		//TODO better exception
	}
	
	public static String SQLType(final BasicTypeID basicType){
		switch (basicType) {
		case TIMESTAMP: return "BIGINT";
		case DURATION: return "BIGINT";
		case INT: return "INT";
		case LONG: return "BIGINT";
		case FLOAT: return "REAL";
		case DOUBLE: return "DOUBLE";
		case BOOLEAN: return "BOOLEAN";
		case STRING: return "CLOB";
		case IDENTIFIER: return "BIGINT";
		case REFERENCE: return "BIGINT";
		}
		throw new Error();
		//TODO better exception
	}
	
	public static String sqlTypeSpecification(final String program_meta_schema, final String sqlAttributeName, final BasicType basicType){
		return "\"" + sqlAttributeName + "\" " + SQLTypeSpec(basicType, program_meta_schema);
	}
	
	public static String sqlCreateTupleIdSequence(final String program_meta_schema){
		return "CREATE SEQUENCE \"" + program_meta_schema + "\".\"" + tupleIdSeqenceName + "\" AS BIGINT START WITH 0;";
	}
	
	public static String sqlGetTupleIdSequenceValue(final String program_meta_schema){
		return "SELECT get_value_for('" + program_meta_schema + "','" + tupleIdSeqenceName + "');"; 
	}
	
	public static String sqlCreateBufferMetaTable(final String program_meta_schema, final SQL_Program_Spec program_spec){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String tableDefinition = 
		"CREATE TABLE \"" + program_meta_schema + "\".\"" + programBufferMetaTableName + "\" (\n" + 
		"	\"id\" INT AUTO_INCREMENT PRIMARY KEY,\n" +  
		"	\"prog_package\" CLOB NOT NULL,\n" + 
		"	\"simple_name\" CLOB NOT NULL,\n" +
		"	\"min_times_number\" INT NOT NULL, \n" +
		"	\"type\" CLOB NOT NULL,\n" +
		"	\"table\" CLOB NOT NULL,\n" +
		"	\"block_counter\" BIGINT NOT NULL";
		
		for(int i=1; i <= program_spec.bufferMaxNumberOfMinTimes; ++i){
			tableDefinition += ", \n" +
		"	\"min_time_name_" + Integer.toString(i) + "\" CLOB"/* + ", \n" +
		"	\"min_time_write_value_" + Integer.toString(i) + "\" BIGINT, \n" +
		"	\"min_time_read_value_" + Integer.toString(i) + "\" BIGINT"*/;
		}
		
		tableDefinition +=	", \n" +
		"	CONSTRAINT \"unique_names\" UNIQUE (\"type\", \"prog_package\", \"simple_name\"), " + "\n" +
		"	CONSTRAINT \"unique_table\" UNIQUE (\"type\", \"table\") " + "\n" +
		");"
		;
		return tableDefinition;
	}
	
	public static String sqlAddBufferMetaEntry(final String program_meta_schema, final int maxNumberOfMinTimes){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String sql_statement = 
				"INSERT INTO \"" + program_meta_schema + "\".\"" + programBufferMetaTableName + "\" \n" +
				"   ( \"prog_package\", \"simple_name\", \"min_times_number\", \"type\", \"table\", \"block_counter\" ";
		
		for(int i=1; i <= maxNumberOfMinTimes; ++i){
			sql_statement += 	", \"min_time_name_" + Integer.toString(i) + "\"" /*+
			                    ", \"min_time_write_value_" + Integer.toString(i) + "\"" +
								", \"min_time_read_value_" + Integer.toString(i) + "\""*/;
		}
		
			sql_statement +=
				") \n" +
				"VALUES \n" +
				"   ( ?, ?, ?, ?, ?, ?";
			
		for(int i=1; i <= maxNumberOfMinTimes; ++i){
			sql_statement += 	", ?";//", ?, ?, ?";
		}
		
			sql_statement += ") \n" +
				";"
		;
		return sql_statement;
	}

	public static void sqlSetAddBufferMetaEntry(final PreparedStatement statement, final SQL_Buffer_Spec buffer_spec, final String sql_name, final int max_buffer_min_times) throws SQLException{
		
		int parameter = 0;
		statement.setString(++parameter, buffer_spec.prog_package);
		statement.setString(++parameter, buffer_spec.simple_name);
		statement.setInt(++parameter, buffer_spec.min_times.size());
		statement.setString(++parameter, buffer_spec.type.name());
		statement.setString(++parameter, sql_name);
		statement.setLong(++parameter, block_counter_initial_value);
		
		int left_min_times = max_buffer_min_times;
		
		for(String min_time: buffer_spec.min_times){
			statement.setString(++parameter, min_time);
			/*statement.setLong(i, SQL_CONSTANTS.BIGINT.MIN_VALUE);
			++i;
			statement.setLong(i, SQL_CONSTANTS.BIGINT.MIN_VALUE);
			++i;*/
			--left_min_times;
		}
		for(;left_min_times > 0; --left_min_times){
			statement.setNull(++parameter, java.sql.Types.CLOB);
			/*statement.setNull(i, java.sql.Types.BIGINT);
			++i;
			statement.setNull(i, java.sql.Types.BIGINT);
			++i;*/
		}
		
	}
	
	public static String sqlGetBufferId(final String program_meta_schema, final DTSA_BufferType type, final String buffer_full_name){
		String[] names = TSA_SyntaxPatterns.bufferFullName(buffer_full_name);
		return sqlGetBufferId(program_meta_schema, type, names[0], names[1]);
	}
	
	public static String sqlGetBufferId(final String program_meta_schema, final DTSA_BufferType type, final String prog_package, final String simple_name){
		assert prog_package.matches(TSA_SyntaxPatterns.PackageNameRegEx);
		assert simple_name.matches(TSA_SyntaxPatterns.BufferNameRegEx);
		return "SELECT \"id\" \n" + 
		"FROM \"" + program_meta_schema + "\".\"" + programBufferMetaTableName + "\" \n" + 
		"WHERE \"type\" = '" + type.name() + "' AND \"prog_package\" = '" + prog_package + "' AND \"simple_name\" = '" + simple_name + "' \n" + 
		";";
	}
	
	public static String sqlGetBufferBlockCounter(final String program_meta_schema, final int buffer_id){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		return "SELECT \"block_counter\" FROM \"" +  program_meta_schema + "\".\"" + programBufferMetaTableName + "\" WHERE \"id\" = '" + buffer_id + "';";
	}
	
	public static String sqlIncrementBlockCounter(final String program_meta_schema, final int buffer_id){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		return "UPDATE \"" +  program_meta_schema + "\".\"" + programBufferMetaTableName + "\" SET \"block_counter\" = \"block_counter\" + 1 WHERE \"id\" = '" + buffer_id + "';";
	}
	
	public static String sqlGetBufferMetaEntry(final String program_meta_schema, final int bufferMaxNumberOfMinTimes){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String sql_statement = 
			"SELECT \"id\", \"prog_package\", \"simple_name\", \"min_times_number\", \"type\", \"table\"";
		
		for(int i=1; i <= bufferMaxNumberOfMinTimes; ++i){
			sql_statement += 	", \"min_time_name_" + Integer.toString(i) + "\"";
		}
		
		sql_statement += " \n" +
			"FROM \"" + program_meta_schema + "\".\"" + programBufferMetaTableName + "\" \n" +
			"WHERE \"type\" = ? AND \"prog_package\" = ? AND \"simple_name\" = ?\n" +
			";";
		
		return	sql_statement;
		
	}
	
	public static String sqlGetBufferSimpleMetaEntry(final String program_meta_schema){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String sql_statement = 
			"SELECT \"id\", \"prog_package\", \"simple_name\", \"type\", \"table\"";
		
		sql_statement += " \n" +
			"FROM \"" + program_meta_schema + "\".\"" + programBufferMetaTableName + "\" \n" +
			"WHERE \"type\" = ? AND \"prog_package\" = ? AND \"simple_name\" = ?\n" +
			";";
		
		return	sql_statement;
		
	}
	
	public static void sqlSetGetBufferMetaEntry(final PreparedStatement statement, final DTSA_BufferType type, final String full_name) throws SQLException{
		String[] names = TSA_SyntaxPatterns.bufferFullName(full_name);
		sqlSetGetBufferMetaEntry(statement, type, names[0], names[1]);
	}
	
	public static void sqlSetGetBufferMetaEntry(final PreparedStatement statement, final DTSA_BufferType type, final String prog_package, final String simple_name) throws SQLException{
		int parameter = 0;
		statement.setString(++parameter, type.name());
		statement.setString(++parameter, prog_package);
		statement.setString(++parameter, simple_name);		
	}
	
	public static String sqlGetBufferMetaEntries(final String program_meta_schema, final int bufferMaxNumberOfMinTimes){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String sql_statement = 
			"SELECT \"id\", \"prog_package\", \"simple_name\", \"min_times_number\", \"type\", \"table\"";
		
		for(int i=1; i <= bufferMaxNumberOfMinTimes; ++i){
			sql_statement += 	", \"min_time_name_" + Integer.toString(i) + "\"";
		}
		
		sql_statement += " \n" +
			"FROM \"" + program_meta_schema + "\".\"" + programBufferMetaTableName + "\" \n" +
			"ORDER BY \"type\", \"prog_package\" ASC, \"simple_name\" ASC \n" +
			";";
		
		return	sql_statement;
		
	}
	
	public static String sqlGetBufferMetaSimpleEntriesOfType(final String program_meta_schema){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String sql_statement = 
			"SELECT \"id\", \"prog_package\", \"simple_name\", \"type\", \"table\"";
		
		sql_statement += " \n" +
			"FROM \"" + program_meta_schema + "\".\"" + programBufferMetaTableName + "\" \n" +
			"WHERE \"type\" = ? \n" +
			"ORDER BY \"prog_package\" ASC, \"simple_name\" ASC \n" +
			";";
		
		return	sql_statement;
		
	}
	
	public static void sqlSetGetBufferMetaEntriesOfType(final PreparedStatement statement, final DTSA_BufferType type) throws SQLException{
		int parameter = 0;
		statement.setString(++parameter, type.name());
	}
	
	public static int sqlReadBufferMetaEntryId(final ResultSet resultSet) throws SQLException{
		return resultSet.getInt("id");
	}
	
	public static String sqlReadBufferMetaEntryProgPackage(final ResultSet resultSet) throws SQLException{
		return resultSet.getString("prog_package");
	}
	
	public static String sqlReadBufferMetaEntrySimpleName(final ResultSet resultSet) throws SQLException{
		return resultSet.getString("simple_name");
	}
	
	public static int sqlReadBufferMetaEntryNumberOfMinTimes(final ResultSet resultSet) throws SQLException{
		return resultSet.getInt("min_times_number");
	}
	
	public static DTSA_BufferType sqlReadBufferMetaEntryType(final ResultSet resultSet) throws SQLException{
		return DTSA_BufferType.valueOf(resultSet.getString("type"));
	}
	
	public static String sqlReadBufferMetaEntryTable(final ResultSet resultSet) throws SQLException{
		return resultSet.getString("table");
	}
	
	
	public static FixedArrayNavigableSet<String> sqlReadBufferMetaEntryMinTimes(final ResultSet resultSet) throws SQLException{
		int numberOfMinTimes = sqlReadBufferMetaEntryNumberOfMinTimes(resultSet);
		String[] minTimes = new String[numberOfMinTimes];
		for(int minTime = 1; minTime <= numberOfMinTimes; ++minTime){
			minTimes[minTime - 1] = resultSet.getString("min_time_name_" + Integer.toString(minTime));
		}
		return new FixedArrayNavigableSet<String>(false, minTimes);
	}
	
	public static FixedArrayNavigableSet<String> sqlReadBufferMetaEntriesNames(final ResultSet resultSet) throws SQLException{
		ArrayList<String> tsa_names = new ArrayList<String>();
		while(resultSet.next()){
			tsa_names.add(TSA_SyntaxPatterns.bufferFullName(sqlReadBufferMetaEntryProgPackage(resultSet), sqlReadBufferMetaEntrySimpleName(resultSet)));
		}
		try{
			if(resultSet.getType() != ResultSet.TYPE_FORWARD_ONLY){
				resultSet.first();
			}
		}
		catch (SQLFeatureNotSupportedException e) {
		
		}
		catch (SQLException e) {
		
		}
		return new FixedArrayNavigableSet<String>(TSA_SyntaxPatterns.bufferFullNameComparator, false, false, tsa_names.toArray(new String[0]));
	}

	public static FixedArrayNavigableMap<String, String> sqlReadBufferMetaEntriesSQLNameMapping(final ResultSet resultSet) throws SQLException{
		ArrayList<String> tsa_names = new ArrayList<String>();
		ArrayList<String> sql_names = new ArrayList<String>();
		while(resultSet.next()){
			tsa_names.add(TSA_SyntaxPatterns.bufferFullName(sqlReadBufferMetaEntryProgPackage(resultSet), sqlReadBufferMetaEntrySimpleName(resultSet)));
			sql_names.add(sqlReadBufferMetaEntryTable(resultSet));
		}
		try{
			if(resultSet.getType() != ResultSet.TYPE_FORWARD_ONLY){
				resultSet.first();
			}
		}
		catch (SQLFeatureNotSupportedException e) {
		
		}
		catch (SQLException e) {
		
		}
		return new FixedArrayNavigableMap<String, String>(new FixedArrayNavigableSet<String>(TSA_SyntaxPatterns.bufferFullNameComparator, false, false, tsa_names.toArray(new String[0])), new FixedArray<String>(false, sql_names.toArray(new String[0])));
	}
	
	public static String sqlCreateAttributesMetaTable(final String program_meta_schema){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String tableDefinition = 
		"CREATE TABLE \"" + program_meta_schema + "\".\"" + programBufferAttributeMetaTableName + "\" (\n" + 
		"	\"id\" INT AUTO_INCREMENT PRIMARY KEY,\n" +  
		"	\"buffer_id\" INT REFERENCES \"" + program_meta_schema +"\".\"" + programBufferMetaTableName + "\" ON UPDATE CASCADE ON DELETE CASCADE,\n" + 
		"	\"name\" CLOB NOT NULL,\n" +
		"	\"type\" CLOB NOT NULL,\n" +
		"	\"sql_name\" CLOB NOT NULL";
		
		tableDefinition +=	", \n" +
		"	CONSTRAINT \"unique_attribute_names\" UNIQUE (\"buffer_id\", \"name\"), " + "\n" +
		"	CONSTRAINT \"unique_sql_attribute_names\" UNIQUE (\"buffer_id\", \"sql_name\") " + "\n" +
		");"
		;
		return tableDefinition;
	}
	
	public static String sqlAddBufferAttributeMetaEntry(final String program_meta_schema){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String sql_statement = 
				"INSERT INTO \"" + program_meta_schema + "\".\"" + programBufferAttributeMetaTableName + "\" \n" +
				"   ( \"buffer_id\", \"name\", \"type\", \"sql_name\" ";
		
			sql_statement +=
				") \n" +
				"VALUES \n" +
				"   ( \n" +
				"		( \n" +
				"			SELECT \"id\" AS \"buffer_id\" \n" +
				"			FROM \"" + program_meta_schema + "\".\"" + programBufferMetaTableName + "\" \n" +
				"			WHERE \"type\" = ? AND \"prog_package\" = ? AND \"simple_name\" = ?" +
				"		) \n" +
				", ?, ?, ?) \n" +
				";"
		;
		return sql_statement;
	}

	public static void sqlSetAddBufferAttributeMetaEntry(final PreparedStatement statement, final DTSA_BufferType buffer_type, final String buffer_package_name, final String buffer_simple_name, final String name, final BasicTypeID type, final String sql_name) throws SQLException{
		int parameter = 0;
		statement.setString(++parameter, buffer_type.name());
		statement.setString(++parameter, buffer_package_name);
		statement.setString(++parameter, buffer_simple_name);
		statement.setString(++parameter, name);
		statement.setString(++parameter, type.name());
		statement.setString(++parameter, sql_name);
	}
	
	public static String sqlGetAttributeMetaEntry(final String program_meta_schema){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String sql_statement = 
			"SELECT \"attributes\".\"id\" AS \"id\", \"attributes\".\"name\" AS \"name\", \"attributes\".\"type\" AS \"type\", \"attributes\".\"sql_name\" AS \"sql_name\"";
		
		sql_statement += " \n" +
			"FROM \n" +
			"	\"" + program_meta_schema + "\".\"" + programBufferMetaTableName + "\" AS \"buffers\",\n" +
			"	\"" + program_meta_schema + "\".\"" + programBufferAttributeMetaTableName + "\" AS \"attributes\"\n" +
			"WHERE \"buffers\".\"id\" = \"attributes\".\"buffer_id\" AND \"buffers\".\"type\" = ? AND \"buffers\".\"prog_package\" = ? AND \"buffers\".\"simple_name\" = ?  AND \"attributes\".\"name\" = ? \n" +
			";";
		
		return	sql_statement;		
	}
	
	public static void sqlSetGetAttributeMetaEntry(final PreparedStatement statement, final DTSA_BufferType buffer_type, final String buffer_name, final String name) throws SQLException{
		String[] names = TSA_SyntaxPatterns.bufferFullName(buffer_name);
		sqlSetGetAttributeMetaEntry(statement, buffer_type, names[0], names[1], name);
	}
	
	public static void sqlSetGetAttributeMetaEntry(final PreparedStatement statement, final DTSA_BufferType buffer_type, final String buffer_prog_package, final String buffer_simple_name, final String name) throws SQLException{
		int parameter = 0;
		statement.setString(++parameter, buffer_type.name());
		statement.setString(++parameter, buffer_prog_package);
		statement.setString(++parameter, buffer_simple_name);
		statement.setString(++parameter, name);
	}
	
	public static String sqlGetAttributeMetaEntries(final String program_meta_schema){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String sql_statement = 
			"SELECT \"attributes\".\"id\" AS \"id\", \"attributes\".\"name\" AS \"name\", \"attributes\".\"type\" AS \"type\", \"attributes\".\"sql_name\" AS \"sql_name\"";
		
		sql_statement += " \n" +
			"FROM \n" +
			"	\"" + program_meta_schema + "\".\"" + programBufferMetaTableName + "\" AS \"buffers\", \n" +
			"	\"" + program_meta_schema + "\".\"" + programBufferAttributeMetaTableName + "\" AS \"attributes\" \n" +
			"WHERE \"buffers\".\"id\" = \"attributes\".\"buffer_id\" AND \"buffers\".\"type\" = ? AND \"buffers\".\"prog_package\" = ? AND \"buffers\".\"simple_name\" = ?\n" +
			"ORDER BY \"attributes\".\"name\" ASC \n" +
			";";
		
		return	sql_statement;	
	}
	
	public static void sqlSetGetAttributeMetaEntries(final PreparedStatement statement, final DTSA_BufferType buffer_type, final String buffer_name) throws SQLException{
		String[] names = TSA_SyntaxPatterns.bufferFullName(buffer_name);
		sqlSetGetAttributeMetaEntries(statement, buffer_type, names[0], names[1]);
	}
	
	public static void sqlSetGetAttributeMetaEntries(final PreparedStatement statement, final DTSA_BufferType buffer_type, final String buffer_prog_package, final String buffer_simple_name) throws SQLException{
		int parameter = 0;
		statement.setString(++parameter, buffer_type.name());
		statement.setString(++parameter, buffer_prog_package);
		statement.setString(++parameter, buffer_simple_name);
	}
	
	public static int sqlReadAttributeMetaEntryId(final ResultSet resultSet) throws SQLException{
		return resultSet.getInt("id");
	}
	
	public static String sqlReadAttributeMetaEntryName(final ResultSet resultSet) throws SQLException{
		return resultSet.getString("name");
	}
	
	public static BasicTypeID sqlReadAttributeMetaEntryType(final ResultSet resultSet) throws SQLException{
		return BasicTypeID.valueOf(resultSet.getString("type"));
	}
	
	public static String sqlReadAttributeMetaEntrySQLName(final ResultSet resultSet) throws SQLException{
		return resultSet.getString("sql_name");
	}
	
	public static FixedArrayNavigableSet<String> sqlReadAttributeMetaEntriesNames(final ResultSet resultSet) throws SQLException{
		ArrayList<String> tsa_names = new ArrayList<String>();
		while(resultSet.next()){
			tsa_names.add(sqlReadAttributeMetaEntryName(resultSet));
		}
		try{
			if(resultSet.getType() != ResultSet.TYPE_FORWARD_ONLY){
				resultSet.first();
			}
		}
		catch (SQLFeatureNotSupportedException e) {
		
		}
		catch (SQLException e) {
		
		}
		return new FixedArrayNavigableSet<String>(false, false, tsa_names.toArray(new String[0]));
	}
	
	public static FixedArrayNavigableMap<String, String> sqlReadAttributeMetaEntriesSQLNameMapping(final ResultSet resultSet) throws SQLException{
		ArrayList<String> tsa_names = new ArrayList<String>();
		ArrayList<String> sql_names = new ArrayList<String>();
		while(resultSet.next()){
			tsa_names.add(sqlReadAttributeMetaEntryName(resultSet));
			sql_names.add(sqlReadAttributeMetaEntrySQLName(resultSet));
		}
		try{
			if(resultSet.getType() != ResultSet.TYPE_FORWARD_ONLY){
				resultSet.first();
			}
		}
		catch (SQLFeatureNotSupportedException e) {
		
		}
		catch (SQLException e) {
		
		}
		return new FixedArrayNavigableMap<String, String>(new FixedArrayNavigableSet<String>(false, false, tsa_names.toArray(new String[0])), new FixedArray<String>(false, sql_names.toArray(new String[0])));
	}
	
	public static FixedArrayNavigableMap<String, BasicTypeID> sqlReadAttributeMetaEntriesSQLTypeMapping(final ResultSet resultSet) throws SQLException{
		ArrayList<String> tsa_names = new ArrayList<String>();
		ArrayList<BasicTypeID> types = new ArrayList<BasicTypeID>();
		while(resultSet.next()){
			tsa_names.add(sqlReadAttributeMetaEntryName(resultSet));
			types.add(sqlReadAttributeMetaEntryType(resultSet));
		}
		try{
			if(resultSet.getType() != ResultSet.TYPE_FORWARD_ONLY){
				resultSet.first();
			}
		}
		catch (SQLFeatureNotSupportedException e) {
		
		}
		catch (SQLException e) {
		
		}
		return new FixedArrayNavigableMap<String, BasicTypeID>(new FixedArrayNavigableSet<String>(false, false, tsa_names.toArray(new String[0])), new FixedArray<BasicTypeID>(false, types.toArray(new BasicTypeID[0])));
	}

	public static String sqlCreateQueryMetaTable(final String program_meta_schema, final SQL_Program_Spec program_spec){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String tableDefinition = 
		"CREATE TABLE \"" + program_meta_schema + "\".\"" + programQueryMetaTableName + "\" (\n" + 
		"	\"id\" INT AUTO_INCREMENT PRIMARY KEY,\n" +  
		"	\"prog_package\" CLOB NOT NULL,\n" + 
		"	\"type\" CLOB NOT NULL,\n" +
		"	\"is_monotonic\" BOOLEAN NOT NULL DEFAULT FALSE,\n" +
		"	\"number_of_output_min_times\" INT NOT NULL,\n" +
		"	\"number_of_input_min_times\" INT NOT NULL,\n" +
		"	\"is_active\" BOOLEAN DEFAULT TRUE NOT NULL, \n" +
		"	\"input_type\" CLOB NOT NULL,\n" +
		"	\"inputs\" CLOB NOT NULL,\n" +
		"	\"output\" CLOB NOT NULL,\n" +
		"	\"sql_eval_statement\" CLOB NOT NULL, \n" +
		"	\"referenced_clocks\" CLOB NOT NULL,\n" +
		"	\"min_times_computation\" CLOB NOT NULL,\n" +
		"	\"min_time_clock_increases_computation\" CLOB NOT NULL,\n" +
		"	\"last_round\" BIGINT NOT NULL";
		
		for(int i=1; i <= program_spec.queryMaxNumberOfOutputMinTimes; ++i){
			tableDefinition +=  ", \n" +
		"	\"output_min_time_name_" + Integer.toString(i) + "\" CLOB, \n" +
		"	\"output_min_time_value_" + Integer.toString(i) + "\" BIGINT"; 
		}
		
		for(int i=1; i <= program_spec.queryMaxNumberOfInputMinTimes; ++i){
			tableDefinition +=  ", \n" +
		"	\"input_min_time_name_" + Integer.toString(i) + "\" CLOB, \n" +
		"	\"input_min_time_value_" + Integer.toString(i) + "\" BIGINT"; 
		}
		
		tableDefinition +=  " \n);"
		;
		return tableDefinition;		
	}
	
	
	public static String sqlAddQueryMetaEntry(final String program_meta_schema, final SQL_Program_Spec program_spec){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String sql_statement = 
				"INSERT INTO \"" + program_meta_schema + "\".\"" + programQueryMetaTableName + "\" \n" +
				"   ( \"prog_package\", \"type\", \"is_monotonic\", \"number_of_output_min_times\", \"number_of_input_min_times\", \"input_type\", \"inputs\", \"output\", \"sql_eval_statement\", \"referenced_clocks\", \"min_times_computation\", \"min_time_clock_increases_computation\", \"last_round\" ";
		
		for(int i=1; i <= program_spec.queryMaxNumberOfOutputMinTimes; ++i){
			sql_statement += 	", \"output_min_time_name_" + Integer.toString(i) + "\"" +
								", \"output_min_time_value_" + Integer.toString(i) + "\"";
		}
		for(int i=1; i <= program_spec.queryMaxNumberOfInputMinTimes; ++i){
			sql_statement += 	", \"input_min_time_name_" + Integer.toString(i) + "\"" +
								", \"input_min_time_value_" + Integer.toString(i) + "\"";
		}
			sql_statement +=
				") \n" +
				"VALUES \n" +
				"   ( ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?";
			
		for(int i=1; i <= program_spec.queryMaxNumberOfOutputMinTimes; ++i){
			sql_statement += 	", ?, ?";
		}
		for(int i=1; i <= program_spec.queryMaxNumberOfInputMinTimes; ++i){
			sql_statement += 	", ?, ?";
		}
			sql_statement += ") \n" +
				";"
		;
		return sql_statement;
	}

	public static void sqlSetAddQueryMetaEntry(final PreparedStatement statement, final SQL_Query_Spec query_spec, final int max_output_min_times, final int max_input_min_times, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, SQL_Buffer_Spec>> buffer_specs, final FixedArrayNavigableMap<String, String> sql_table_name_mapping, final FixedArrayNavigableMap<String, String> sql_attribute_name_mapping, final String meta_schema, final String working_schema, final String input_schema, final String output_schema, final String log_schema) throws SQLException{
		Persister persister = new Persister();
		de.lmu.ifi.pms.java.util.string.StringBuilder sb = new de.lmu.ifi.pms.java.util.string.StringBuilder();
		
		String inputs;
		try {
			persister.write(query_spec.inputBuffers, sb);
		} catch (Exception e) {
			throw new Error(e);
		}
		inputs = sb.toString();
		
		sb.clear();
		String referencedClocks;
		try {
			persister.write(query_spec.referencedClocks, sb);
		} catch (Exception e) {
			throw new Error(e);
		}
		referencedClocks = sb.toString();
		
		String output = query_spec.output;
		String sql_eval_statement = instanciateSQLStatement(query_spec, sql_table_name_mapping, sql_attribute_name_mapping, meta_schema, working_schema, input_schema, output_schema, log_schema);		
		String min_times_computation = instanciateOutputMinTimesFunction(query_spec, buffer_specs.get(query_spec.inputBuffersType)).serialize();
		String min_time_clock_increases_computation = instanciateOutputMinTimeClockIncreasesFunction(query_spec, buffer_specs.get(query_spec.inputBuffersType)).serialize();//FIXME
		int param = 0;
		statement.setString(++param, query_spec.prog_package);
		statement.setString(++param, query_spec.type.name());
		statement.setBoolean(++param, query_spec.isMonotonic);
		statement.setInt(++param, query_spec.numberOfOutputMinTimes);
		
		int queryNumberOfInputMinTimes = 0;			
		for(String inputBuffer: query_spec.inputBuffers){
			queryNumberOfInputMinTimes += buffer_specs.get(query_spec.inputBuffersType).get(inputBuffer).min_times.size();
		}		
		
		statement.setInt(++param, queryNumberOfInputMinTimes);
		statement.setString(++param, query_spec.inputBuffersType.name());
		statement.setString(++param, inputs);
		statement.setString(++param, output);
		statement.setString(++param, sql_eval_statement);
		statement.setString(++param, referencedClocks);
		statement.setString(++param, min_times_computation);
		statement.setString(++param, min_time_clock_increases_computation);
		
		statement.setLong(++param, 0);
		
		int left_output_min_times = max_output_min_times;
		for(String output_min_time: query_spec.outputMinTimes.keys()){
			statement.setString(++param, output_min_time);
			statement.setLong(++param, SQL_CONSTANTS.BIGINT.MIN_VALUE);
			--left_output_min_times;
		}
		for(;left_output_min_times > 0; --left_output_min_times){
			statement.setNull(++param, java.sql.Types.CLOB);
			statement.setNull(++param, java.sql.Types.BIGINT);
		}
		
		int left_input_min_times = max_input_min_times;
		for(String inputBuffer: query_spec.inputBuffers){
			for(String min_time: buffer_specs.get(query_spec.inputBuffersType).get(inputBuffer).min_times){
				statement.setString(++param, TSA_SyntaxPatterns.attributeQualifiedName(inputBuffer, min_time));
				statement.setLong(++param, SQL_CONSTANTS.BIGINT.MIN_VALUE);
				--left_input_min_times;
			}
		}
		for(;left_input_min_times > 0; --left_input_min_times){
			statement.setNull(++param, java.sql.Types.CLOB);
			statement.setNull(++param, java.sql.Types.BIGINT);
		}
		
	}
	
	public static String sqlGetQueryMetaEntries(final String program_meta_schema, final int maxNumberOfOutputMinTimes, final int maxNumberOfInputMinTimes){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
	String sql_statement = 
			"SELECT \"id\", \"prog_package\", \"type\", \"is_monotonic\", \"number_of_output_min_times\", \"number_of_input_min_times\", \"input_type\", \"inputs\", \"output\", \"sql_eval_statement\", \"referenced_clocks\", \"min_times_computation\", \"min_time_clock_increases_computation\", \"last_round\"";
	
	for(int i=1; i <= maxNumberOfOutputMinTimes; ++i){
		sql_statement += 	", \"output_min_time_name_" + Integer.toString(i) + "\"" +
		                    ", \"output_min_time_value_" + Integer.toString(i) + "\"";
	}
	for(int i=1; i <= maxNumberOfInputMinTimes; ++i){
		sql_statement += 	", \"input_min_time_name_" + Integer.toString(i) + "\"" +
							", \"input_min_time_value_" + Integer.toString(i) + "\"";
	}
		
	sql_statement += " \n" + 
			"FROM \"" + program_meta_schema + "\".\"" + programQueryMetaTableName + "\" \n" +
			"ORDER BY \"type\" ASC, \"prog_package\" ASC, \"output\" ASC \n" +
			";";
	return sql_statement;
	}
	
	public static String sqlGetOutputQueryMetaEntry(final String program_meta_schema, final int maxNumberOfOutputMinTimes, final int maxNumberOfInputMinTimes){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
	String sql_statement = 
			"SELECT \"id\", \"prog_package\", \"type\", \"is_monotonic\", \"number_of_output_min_times\", \"number_of_input_min_times\", \"inputs\", \"output\", \"sql_eval_statement\", \"referenced_clocks\", \"min_times_computation\", \"min_time_clock_increases_computation\", \"last_round\"";
	
	for(int i=1; i <= maxNumberOfOutputMinTimes; ++i){
		sql_statement += 	", \"output_min_time_name_" + Integer.toString(i) + "\"" +
		                    ", \"output_min_time_value_" + Integer.toString(i) + "\"";
	}
	for(int i=1; i <= maxNumberOfInputMinTimes; ++i){
		sql_statement += 	", \"input_min_time_name_" + Integer.toString(i) + "\"" +
							", \"input_min_time_value_" + Integer.toString(i) + "\"";
	}
		
	sql_statement += " \n" + 
			"FROM \"" + program_meta_schema + "\".\"" + programQueryMetaTableName + "\" \n" +
			"WHERE \"type\" = '" + DTSA_QueryType.OUTPUT.name() + "' AND \"output\" = ? \n" +
			";";
	return sql_statement;
	}
	
	public static void sqlSetGetOutputQueryMetaEntry(final PreparedStatement statement, final String buffer_name) throws SQLException{
		assert TSA_SyntaxPatterns.isValidBufferFullName(buffer_name);
		int parameter = 0;
		statement.setString(++parameter, buffer_name);
	}
	
	public static FixedArrayNavigableMap<String, String> sqlReadQueryMetaEntryOutputMinTimesValueSQLNameMapping(final ResultSet resultSet) throws SQLException{
		int numberOfMinTimes = sqlReadQueryMetaEntryNumberOfOutputMinTimes(resultSet);
		String[] minTimes = new String[numberOfMinTimes];
		String[] minTimesValueSQLNameMapping = new String[numberOfMinTimes];
		for(int minTime = 1; minTime <= numberOfMinTimes; ++minTime){
			minTimes[minTime - 1] = resultSet.getString("output_min_time_name_" + Integer.toString(minTime));
			minTimesValueSQLNameMapping[minTime - 1] = "output_min_time_value_" + Integer.toString(minTime);
		}
		return new FixedArrayNavigableMap<String, String>(minTimes, minTimesValueSQLNameMapping);
	}

	public static int sqlReadQueryMetaEntryId(final ResultSet resultSet) throws SQLException{
		return resultSet.getInt("id");
	}

	public static String sqlReadQueryMetaEntryProgPackage(final ResultSet resultSet) throws SQLException{
		return resultSet.getString("prog_package");
	}

	public static DTSA_QueryType sqlReadQueryMetaEntryType(final ResultSet resultSet) throws SQLException{
		return DTSA_QueryType.valueOf(resultSet.getString("type"));
	}
	
	public static boolean sqlReadQueryMetaEntryIsMonotonic(final ResultSet resultSet) throws SQLException{
		return resultSet.getBoolean("is_monotonic");
	}
	
	public static int sqlReadQueryMetaEntryNumberOfOutputMinTimes(final ResultSet resultSet) throws SQLException{
		return resultSet.getInt("number_of_output_min_times");
	}
	
	public static int sqlReadQueryMetaEntryNumberOfInputMinTimes(final ResultSet resultSet) throws SQLException{
		return resultSet.getInt("number_of_input_min_times");
	}
	
	public static long[] sqlReadQueryMetaEntryOutputMinTimes(final ResultSet resultSet) throws SQLException{
		int numberOfOutputMinTimes = sqlReadQueryMetaEntryNumberOfOutputMinTimes(resultSet);
		long[] outputMinTimes = new long[numberOfOutputMinTimes];
		for(int minTime = 1; minTime <= numberOfOutputMinTimes; ++minTime){
			outputMinTimes[minTime - 1] = resultSet.getLong("output_min_time_value_" + Integer.toString(minTime));
		}
		return outputMinTimes;
	}

	public static long[] sqlReadQueryMetaEntryInputMinTimes(final ResultSet resultSet) throws SQLException{
		int numberOfInputMinTimes = sqlReadQueryMetaEntryNumberOfInputMinTimes(resultSet);
		long[] inputMinTimes = new long[numberOfInputMinTimes];
		for(int minTime = 1; minTime <= numberOfInputMinTimes; ++minTime){
			inputMinTimes[minTime - 1] = resultSet.getLong("input_min_time_value_" + Integer.toString(minTime));
		}
		return inputMinTimes;
	}
	
	public static DTSA_BufferType sqlReadQueryMetaEntryInputType(final ResultSet resultSet) throws SQLException{
		return DTSA_BufferType.valueOf(resultSet.getString("input_type"));
	}
	
	//TODO make more stable agianst wrong input
	@SuppressWarnings("unchecked")
	public static FixedArrayNavigableSet<String> sqlReadQueryMetaEntryInputs(final ResultSet resultSet) throws SQLException{
		Persister persister = new Persister();
		try {
			return persister.read(FixedArrayNavigableSet.class, resultSet.getString("inputs"));
		} catch (Exception e) {
			if(e instanceof SQLException){
				throw (SQLException)e;
			}
			else{
				throw new Error(e);
			}
		}
	}
	
	public static String sqlReadQueryMetaEntryOutput(final ResultSet resultSet) throws SQLException{
		return resultSet.getString("output");
	}
	
	public static String sqlReadQueryMetaEntrySQL_Eval_Statement(final ResultSet resultSet) throws SQLException{
		return resultSet.getString("sql_eval_statement");
	}
	
	//TODO make more stable agianst wrong input
	@SuppressWarnings("unchecked")
	public static FixedArrayNavigableSet<String> sqlReadQueryMetaEntryReferencedClocks(final ResultSet resultSet) throws SQLException{
		Persister persister = new Persister();
		try {
			return persister.read(FixedArrayNavigableSet.class, resultSet.getString("referenced_clocks"));
		} catch (Exception e) {
			if(e instanceof SQLException){
				throw (SQLException)e;
			}
			else{
				throw new Error(e);
			}
		}
	}
	
	public static TimeFunction sqlReadQueryMetaEntryMinTimesComputation(final ResultSet resultSet) throws SQLException{
		return TimeFunction.deserialize(resultSet.getString("min_times_computation"));
	}
	
	public static TimeFunction sqlReadQueryMetaEntryMinTimeClockIncreasesComputation(final ResultSet resultSet) throws SQLException{
		return TimeFunction.deserialize(resultSet.getString("min_time_clock_increases_computation"));
	}
	
	public static long sqlReadQueryMetaEntryLastRound(final ResultSet resultSet) throws SQLException{
		return resultSet.getLong("last_round");
	}

	public static String sqlUpdateQueryMetaEntry(final String program_meta_schema, final int query_id, final int numberOfOutputMinTimes, final int numberOfInputMinTimes, final int numberOfReferencedClocks){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		//FIXME Query for the case where there are neither inout nor output min times
		
		String sql_statement = 
				"UPDATE \"" + program_meta_schema + "\".\"" + programQueryMetaTableName + "\" \n" +
				"SET ";
				
		sql_statement += 	
				"	\"last_round\" = ?, \n";
		
		for(int i=1; i <= numberOfOutputMinTimes; ++i){
			sql_statement += 	
				"	\"output_min_time_value_" + Integer.toString(i) + "\" = ?, \n";
		}
		for(int i=1; i <= numberOfInputMinTimes; ++i){
			sql_statement += 	
				"	\"input_min_time_value_" + Integer.toString(i) + "\" = ?, \n";
		}
		sql_statement = sql_statement.substring(0, sql_statement.length() - ", \n".length()) + " \n";
		
		sql_statement += "WHERE \"id\" = '" + query_id + "' \n" +
				";";
		return sql_statement;
	}
	
	public static void sqlSetUpdateQueryMetaEntry(final PreparedStatement statement, final long currentRound, final long[] currentOutputMinTimes, final long[] currentInputMinTimes) throws SQLException{
		int parameterMetaEntry = 1;
		statement.setLong(parameterMetaEntry, currentRound);
		++parameterMetaEntry;
		
		for(int outputMinTime = 0; outputMinTime < currentOutputMinTimes.length; ++outputMinTime){
			statement.setLong(parameterMetaEntry, currentOutputMinTimes[outputMinTime]);
			++parameterMetaEntry;
		}
		
		for(int inputMinTime = 0; inputMinTime < currentInputMinTimes.length; ++inputMinTime){
			statement.setLong(parameterMetaEntry, currentInputMinTimes[inputMinTime]);
			++parameterMetaEntry;
		}
	}
	
	
	
	public static String sqlCreateInputStreamsMetaTable(final String program_meta_schema, final SQL_Program_Spec program_spec){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String tableDefinition = 
		"CREATE TABLE \"" + program_meta_schema + "\".\"" + programInputStreamsMetaTableName + "\" (\n" + 
		"	\"id\" INT AUTO_INCREMENT PRIMARY KEY,\n" +  
		"	\"buffer_ref\" INT \"" + program_meta_schema +"\".\"" + programBufferMetaTableName + "\" ON UPDATE CASCADE ON DELETE CASCADE\n";
		
		for(int i=1; i <= program_spec.bufferMaxNumberOfMinTimes; ++i){//TODO inputBufferMaxNumberOfMinTimes
			tableDefinition += ", \n" +
		"	\"min_time_name_" + Integer.toString(i) + "\" CLOB" + ", \n" +
		"	\"min_time_write_value_" + Integer.toString(i) + "\" BIGINT" /*+ ", \n" +
		"	\"min_time_read_value_" + Integer.toString(i) + "\" BIGINT"*/;
		}

		;
		return tableDefinition;
	}
	
	public static String sqlAddInputStreamMetaEntry(final String program_meta_schema, final int maxNumberOfMinTimes){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String sql_statement = 
				"INSERT INTO \"" + program_meta_schema + "\".\"" + programInputStreamsMetaTableName + "\" \n" +
				"   ( \"buffer_ref\", ";
		
		for(int i=1; i <= maxNumberOfMinTimes; ++i){
			sql_statement += 	"\"min_time_write_value_" + Integer.toString(i) + "\"";
		}
		
			sql_statement +=
				") \n" +
				"VALUES \n" +
				"   ( ";
		
			sql_statement += "(\n" +
				"   SELECT \"id\"\n" +
				"   FROM\n" +
				"       \"" + program_meta_schema + "\".\"" + programBufferMetaTableName + "\"" + 
				"   WHERE\n" +
				"       \"type\" = '" + DTSA_BufferType.INPUT.name() + "'" +
				"   AND" +
				"       \"prog_package\" = ?" +
				"   AND" +
				"       \"simple_name\" = ?" +
				"\n)";
			
		for(int i=1; i <= maxNumberOfMinTimes; ++i){
			sql_statement += 	", ?";
			sql_statement += 	", ?";
		}
		
			sql_statement += ") \n" +
				";"
		;
		return sql_statement;
	}

	public static void sqlSetAddInputStreamMetaEntry(final PreparedStatement statement, final String prog_package, final String simple_name, final String[] min_times, final int max_buffer_min_times) throws SQLException{
		assert max_buffer_min_times > min_times.length;
		
		int parameter = 0;
		statement.setString(++parameter, prog_package);
		statement.setString(++parameter, simple_name);
		
		for(int min_time_index = 0; min_time_index < min_times.length; ++min_time_index){
			statement.setString(++parameter, min_times[min_time_index]);
			statement.setLong(++parameter, SQL_CONSTANTS.BIGINT.MIN_VALUE);
		}
		for(int min_time_index = min_times.length; min_time_index < max_buffer_min_times; ++min_time_index){
			statement.setNull(++parameter, java.sql.Types.CLOB);
			statement.setNull(++parameter, java.sql.Types.BIGINT);
		}
		
	}
	
	/*FIXME
	public static String sqlGetInputMetaEntry(final String program_meta_schema, final int numberOfMinTimes){
		assert program_meta_schema.matches(program_schema_name_RegEx);
		
		String sql_statement = 
			"SELECT \"id\", \"prog_package\", \"simple_name\", \"min_times_number\", \"type\", \"table\"";
		
		for(int i=1; i <= bufferMaxNumberOfMinTimes; ++i){
			sql_statement += 	", \"min_time_name_" + Integer.toString(i) + "\"";
		}
		
		sql_statement += " \n" +
			"FROM \"" + program_meta_schema + "\".\"" + programBufferMetaTableName + "\" \n" +
			"WHERE \"type\" = ? AND \"prog_package\" = ? AND \"simple_name\" = ?\n" +
			";";
		
		return	sql_statement;
		
	}
	*/
	
	public static String sqlBufferDefinition(final String sql_name, final DTSA_BufferType type, final FixedArrayNavigableMap<String, BasicType> tsa_attributes,  final FixedArrayNavigableMap<String, String> tsa_sql_mapping, final String meta_schema, final String working_schema, final String input_schema, final String output_schema, final String log_schema){
		String[] sql_attribute_names = new String[tsa_attributes.size()];
		BasicType[] attribute_types = new BasicType[sql_attribute_names.length];
		for(int attribute = 0; attribute < sql_attribute_names.length; ++attribute){
			assert tsa_sql_mapping.keys().contains(tsa_attributes.keys().get(attribute));
			sql_attribute_names[attribute] = tsa_sql_mapping.get(tsa_attributes.keys().get(attribute));
			attribute_types[attribute] = tsa_attributes.values().get(attribute);
		}
		final FixedArrayNavigableMap<String, BasicType> sql_attributes = new FixedArrayNavigableMap<String, BasicType>(sql_attribute_names, attribute_types);
		return sqlBufferDefinition(sql_name, type, sql_attributes, meta_schema, working_schema, input_schema, output_schema, log_schema);
	}

	public static String sqlBufferDefinition(final String sql_name, final DTSA_BufferType type, final FixedArrayNavigableMap<String, BasicType> sql_attributes, final String meta_schema, final String working_schema, final String input_schema, final String output_schema, final String log_schema){
		final String schema;
		switch (type) {
		case INPUT:
			schema = input_schema;
			break;
		case WORKING:
			schema = working_schema;
			break;
		case OUTPUT:
			schema = output_schema;
			break;
		case LOG:
			schema = log_schema;
			break;
		default:
			throw new Error();
		}
		return sqlTableDefinition(meta_schema, schema, sql_name, sql_attributes);
	}
	
	public static String sqlTableDefinition(final String program_meta_schema, final String SQL_schema_name, final String SQl_table_name, final FixedArrayNavigableMap<String, BasicType> sql_attributes){
		assert SQL_schema_name.matches(schema_name_RegEx);
		assert SQl_table_name.matches(table_name_RegEx);
		
		String tableDefinition = "CREATE TABLE \"" + SQL_schema_name + "\".\"" + SQl_table_name + "\" (" + "\n";
		for(int index = 0; index < sql_attributes.size(); ++index){
			tableDefinition += "    " + sqlTypeSpecification(program_meta_schema, sql_attributes.keys().get(index), sql_attributes.values().get(index)) + "," + "\n";
		}
		tableDefinition = tableDefinition.substring(0, tableDefinition.length() - 2);
		tableDefinition += "\n" + ");";
		return tableDefinition;
	}
	
	public static FixedArrayNavigableSet<String> referencedClocks(final String sql_statement){
		Matcher clockParameterMatcher = clock_parameter_inSQL_Pattern.matcher(sql_statement);
		ArrayList<String> referencedClocks = new ArrayList<String>();
		while(clockParameterMatcher.find()){
			if(clockParameterMatcher.group(1)!=null){
				referencedClocks.add(clockParameterMatcher.group(2));	
			}
			else{
				throw new Error();
			}
		}
		return new FixedArrayNavigableSet<String>(referencedClocks);
	}
	
	public static String instanciateSQLStatement(final SQL_Query_Spec query_spec, final FixedArrayNavigableMap<String, String> sql_table_name_mapping, final FixedArrayNavigableMap<String, String> sql_attribute_name_mapping, final String meta_schema, final String working_schema, final String input_schema, final String output_schema, final String log_schema){
		String instanciated_SQL_PreparedQuery = query_spec.SQL_PreparedQuery;
		
		Pattern schema = Pattern.compile("/[*]'[$]table'[*]/\"(" + schema_name_RegEx + ")\".\"(" + table_name_RegEx + ")\"/[*]'[$]'[*]/");
		Matcher schemaMatcher = schema.matcher(instanciated_SQL_PreparedQuery);
		StringBuffer sb = new StringBuffer(instanciated_SQL_PreparedQuery.length() * 2);
		while(schemaMatcher.find()){
			String schema_name;
			String schema_var = schemaMatcher.group(1);
			if(schema_var.equals("meta_schema")){
				schema_name = meta_schema;
			}
			else if(schema_var.equals("input_schema")){
				schema_name = input_schema;
			}
			else if(schema_var.equals("working_schema")){
				schema_name = working_schema;
			}
			else if(schema_var.equals("output_schema")){
				schema_name = output_schema;
			}
			else if(schema_var.equals("log_schema")){
				schema_name = log_schema;
			}
			else{
				throw new IllegalArgumentException();
				//TODO better exception
			}
			String buffer_name = SQL_Translation.TSA_BufferName(schemaMatcher.group(2));
			String table_name;
			int buffer_spec_index = sql_table_name_mapping.keys().indexOf(buffer_name);			
			if(buffer_spec_index >= 0){
				table_name = sql_table_name_mapping.get(buffer_name);
			}
			else{
				throw new RuntimeException("Query refers buffer that has no sql mapping: " + buffer_name);
				//TODO better exception
			}
			schemaMatcher.appendReplacement(sb, Matcher.quoteReplacement("\"" +  schema_name + "\".\"" + table_name + "\""));
		}
		schemaMatcher.appendTail(sb);
		instanciated_SQL_PreparedQuery = sb.toString();
		
		//FIXME treat attribute names with limited lengths
		
		//FIXME treatment of clock and other functions that need a database specific implementation
		Pattern sequence = Pattern.compile("/[*]'[$]seq'[*]/\"(" + schema_name_RegEx + ")\".\"(" + sequence_name_RegEx + ")\"/[*]'[$]'[*]/");
		Matcher sequenceMatcher = sequence.matcher(instanciated_SQL_PreparedQuery);
		sb.setLength(0);
		while(sequenceMatcher.find()){
			String schema_name;
			String schema_var = sequenceMatcher.group(1);
			if(schema_var.equals("meta_schema")){
				schema_name = meta_schema;
			}
			else if(schema_var.equals("input_schema")){
				schema_name = input_schema;
			}
			else if(schema_var.equals("working_schema")){
				schema_name = working_schema;
			}
			else if(schema_var.equals("output_schema")){
				schema_name = output_schema;
			}
			else if(schema_var.equals("log_schema")){
				schema_name = log_schema;
			}
			else{
				throw new IllegalArgumentException();
				//TODO better exception
			}
			String sequence_name = sequenceMatcher.group(2);			
			//FIXME treat sequence names with limited lengths
			sequenceMatcher.appendReplacement(sb, "\"" +  schema_name + "\".\"" + sequence_name + "\"");
		}
		sequenceMatcher.appendTail(sb);
		instanciated_SQL_PreparedQuery = sb.toString();
		

		
		return instanciated_SQL_PreparedQuery;
	}
	
	public static String[] parametersOfInstanciatedSQLStatement(final String SQL_PreparedQuery){
		Matcher parameterMatcher = parameter_inSQL_qualified_Pattern.matcher(SQL_PreparedQuery);
		ArrayList<String> parameters = new ArrayList<String>();
		while(parameterMatcher.find()){
			parameters.add(parameterMatcher.group(1));
		}
		return parameters.toArray(new String[parameters.size()]);
	}
	
	public static String[] parameterSplitedInstanciatedSQLStatement(final String SQL_PreparedQuery){
		Matcher parameterMatcher = parameter_inSQL_qualified_Pattern.matcher(SQL_PreparedQuery);
		ArrayList<String> parameterSplitedInstanciatedSQLStatement = new ArrayList<String>();
		int startIndex = 0;
		while(parameterMatcher.find()){
			assert parameterMatcher.end() >= startIndex;
			parameterSplitedInstanciatedSQLStatement.add(SQL_PreparedQuery.substring(startIndex, parameterMatcher.end()-1));
			startIndex = parameterMatcher.end();
		}
		parameterSplitedInstanciatedSQLStatement.add(SQL_PreparedQuery.substring(startIndex, SQL_PreparedQuery.length()));
		return parameterSplitedInstanciatedSQLStatement.toArray(new String[parameterSplitedInstanciatedSQLStatement.size()]);
	}
	
	public static FixedArrayNavigableSet<String> referencedClocks(final FixedArrayNavigableMap<String, Term<?>> outputMinTimes){
		ArrayList<String> referencedClocks = new ArrayList<String>();
		for(Term<?> time: outputMinTimes.values()){
			referencedClocks(time, referencedClocks);
		}
		return new FixedArrayNavigableSet<String>(referencedClocks);
	}
	
	protected static void referencedClocks(final Term<?> time, final ArrayList<String> referencedClocks){
		if(time instanceof ClockFunction){
			ClockFunction clock = (ClockFunction) time;
			referencedClocks.add(clock.name);	
		}
		else if(time instanceof MinTime){
			
		}
		else if(time instanceof DurationValue){
			
		}
		else if(time instanceof ConditionFormula && ((ConditionFormula)time).ID == ConditionFormulaID.BOTTOM){
			
		}
		else if(time instanceof ConditionFormula && ((ConditionFormula)time).ID == ConditionFormulaID.TOP){
			
		}
		else if(time instanceof TupleFunction){
			TupleFunction<?> function = (TupleFunction<?>) time;
			for(Term<?> argument: function.arguments()){
				referencedClocks(argument, referencedClocks);
			}
		}
		else{
			throw new IllegalArgumentException();
			//TODO better exception
		}
	}
	
	public static TimeFunction instanciateOutputMinTimesFunction(final SQL_Query_Spec query_spec, final FixedArrayNavigableMap<String, SQL_Buffer_Spec> buffer_specs){
		ArrayList<Term<?>> inputMapping = new ArrayList<Term<?>>();
		TimeFunction minTimes = instanciateTimeFunction(query_spec.outputMinTimes, query_spec.inputs, inputMapping);

		int queryNumberOfMinTimes = 0;
		int[] numberOfPreceedingMinTimes = new int[query_spec.inputBuffers.size()];
		for(int index = 0; index < numberOfPreceedingMinTimes.length; ++index){
			numberOfPreceedingMinTimes[index] = queryNumberOfMinTimes;
			queryNumberOfMinTimes += buffer_specs.get(query_spec.inputBuffers.get(index)).min_times.size();
		}
		
		
		int[] mapping = new int[minTimes.argumentArity()];
		for(int index = 0; index < mapping.length; ++index){
			//FIXME not only input min times but also clocks
			Term<?> parameter = inputMapping.get(index);
			if(parameter instanceof MinTime){
				MinTime minTime = (MinTime) parameter;
				int bufferIndex = query_spec.inputBuffers.indexOf(minTime.buffer_name);
				assert bufferIndex >= 0;//TODO validate
				
				assert buffer_specs.has(minTime.buffer_name);
				SQL_Buffer_Spec buffer_spec = buffer_specs.get(minTime.buffer_name);
				assert buffer_spec.min_times.contains(minTime.name);
				int minTimeIndex = buffer_spec.min_times.indexOf(minTime.name);
				
				mapping[index] = numberOfPreceedingMinTimes[bufferIndex] + minTimeIndex;
			}
			else{
				throw new Error();
			}
			
		}
		
		return new SequencialCompositeTimeFunction(minTimes, new MapTimeFunction(queryNumberOfMinTimes, mapping));
	}
	
	public static TimeFunction instanciateOutputMinTimeClockIncreasesFunction(final SQL_Query_Spec query_spec, final FixedArrayNavigableMap<String, SQL_Buffer_Spec> buffer_specs){

		ArrayList<Term<?>> inputMapping = new ArrayList<Term<?>>();
		TimeFunction minTimes = instanciateTimeFunction(query_spec.outputMinTimeClockIncreases, query_spec.inputs, inputMapping);
				
		final int queryNumberOfClockMinTimes = query_spec.referencedClocks.size();
		
		int[] mapping = new int[minTimes.argumentArity()];
		for(int index = 0; index < mapping.length; ++index){
			//FIXME not only input min times but also clocks
			Term<?> parameter = inputMapping.get(index);
			
			if(parameter instanceof ClockFunction){
				ClockFunction clock = (ClockFunction) parameter;				
				int clock_index = query_spec.referencedClocks.indexOf(clock.name);
				if(clock_index >= 0){
					mapping[index] = clock_index;
				}
				else{
					throw new IllegalArgumentException();
					//TODO better exception 
				}
			}
			else{
				throw new Error("Parameter \n" + parameter + "\nis not a clock function.");
			}
			
		}
		
		return new SequencialCompositeTimeFunction(minTimes, new MapTimeFunction(queryNumberOfClockMinTimes, mapping));
	}
	
	public static TimeFunction instanciateTimeFunction(final FixedArrayNavigableMap<String, Term<?>> outputMinTimes, final FixedArrayNavigableMap<String, String> inputs, final ArrayList<Term<?>> inputMapping){		
		assert inputMapping.isEmpty();
		TimeFunction[] minTimes = new TimeFunction[outputMinTimes.size()];
		for(int index = 0; index < minTimes.length; ++index){
			minTimes[index] = instanciateTimeFunction(outputMinTimes.values().get(index), inputs, inputMapping);
		}		
		return new ParallelCompositeTimeFunction(minTimes);
	}
	
	protected static TimeFunction instanciateTimeFunction(final Term<?> time, final FixedArrayNavigableMap<String, String> inputs, final ArrayList<Term<?>> inputMapping){
		if(time instanceof MinTime){
			MinTime minTime = (MinTime) time;
			//FIXME ?
			inputMapping.add(minTime);
			return IdentityTimeFunction.instance;
		}
		else if(time instanceof ClockFunction){
			ClockFunction clock = (ClockFunction) time;
			assert TSA_SyntaxPatterns.isValidClockName(clock.name);
			//FIXME ?
			inputMapping.add(clock);
			return IdentityTimeFunction.instance;
		}
		else if(time instanceof ConditionFormula && ((ConditionFormula)time).ID == ConditionFormulaID.BOTTOM){
			return new ConstantTimeFunction((TimeConstants.MINUS_INFINITY > SQL_CONSTANTS.BIGINT.MIN_VALUE ? TimeConstants.MINUS_INFINITY : SQL_CONSTANTS.BIGINT.MIN_VALUE));
		}
		else if(time instanceof ConditionFormula && ((ConditionFormula)time).ID == ConditionFormulaID.TOP){
			return new ConstantTimeFunction((TimeConstants.PLUS_INFINITY < SQL_CONSTANTS.BIGINT.MAX_VALUE ? TimeConstants.PLUS_INFINITY : SQL_CONSTANTS.BIGINT.MAX_VALUE));
		}
		else if(time instanceof Plus){
			Plus<?> plus = (Plus<?>) time;
			if(plus.arg(1) instanceof DurationValue){
				DurationValue arg1 = (DurationValue) plus.arg(1);
				TimeFunction arg0 = instanciateTimeFunction(plus.arg(0), inputs, inputMapping);
				if(arg0 == IdentityTimeFunction.instance){
					return new AddDuration(arg1.value);
				}
				else{
					return new SequencialCompositeTimeFunction(new AddDuration(arg1.value), arg0);
				}
			}
			else{
				throw new IllegalArgumentException();
				//TODO better exception
			}
		}
		else if(time instanceof Minus){
			Minus<?> minus = (Minus<?>) time;
			if(minus.arg(1) instanceof DurationValue){
				DurationValue arg1 = (DurationValue) minus.arg(1);
				TimeFunction arg0 = instanciateTimeFunction(minus.arg(0), inputs, inputMapping);
				if(arg0 == IdentityTimeFunction.instance){
					return new AddDuration(-arg1.value);
				}
				else{
					return new SequencialCompositeTimeFunction(new AddDuration(-arg1.value), arg0);
				}
			}
			else{
				throw new IllegalArgumentException();
				//TODO better exception
			}
		}
		else if(time instanceof Least<?>){
			Least<?> min = (Least<?>) time;
			TimeFunction[] subFunctions = new TimeFunction[min.arity()];
			for(int index = 0; index < subFunctions.length; ++index){
				subFunctions[index] = instanciateTimeFunction(min.arg(index), inputs, inputMapping);
			}
			return new SequencialCompositeTimeFunction(new de.lmu.ifi.pms.cep.tsa.time.functions.Min(min.arity()), new ParallelCompositeTimeFunction(subFunctions));
		}
		else if(time instanceof Greatest){
			Greatest<?> max = (Greatest<?>) time;
			TimeFunction[] subFunctions = new TimeFunction[max.arity()];
			for(int index = 0; index < subFunctions.length; ++index){
				subFunctions[index] = instanciateTimeFunction(max.arg(index), inputs, inputMapping);
			}
			return new SequencialCompositeTimeFunction(new de.lmu.ifi.pms.cep.tsa.time.functions.Max(max.arity()), new ParallelCompositeTimeFunction(subFunctions));
		}
		else{
			throw new IllegalArgumentException();
			//TODO better exception
		}
	}
	
	public static boolean isPossibleInputBuffersType(final DTSA_QueryType query_type, final DTSA_BufferType buffer_type){
		switch (query_type) {
		case INIT:
			return buffer_type == DTSA_BufferType.INPUT | buffer_type == DTSA_BufferType.WORKING;
		case INPUT:
			return buffer_type == DTSA_BufferType.INPUT;
		case WORKING:
			return buffer_type == DTSA_BufferType.WORKING;
		case OUTPUT:
			return buffer_type == DTSA_BufferType.WORKING;
		case LOG:
			return buffer_type == DTSA_BufferType.WORKING;
		case INPUT_GARBAGE:
			return buffer_type == DTSA_BufferType.INPUT;
		default:
			throw new Error();
		}
	}
	
	public static DTSA_BufferType outputBufferType(final DTSA_QueryType type){
		switch (type) {
		case INIT:
			return DTSA_BufferType.WORKING;
		case INPUT:
			return DTSA_BufferType.WORKING;
		case WORKING:
			return DTSA_BufferType.WORKING;
		case OUTPUT:
			return DTSA_BufferType.OUTPUT;
		case LOG:
			return DTSA_BufferType.LOG;
		case INPUT_GARBAGE:
			return null;
		default:
			throw new Error();
		}
	}

	public static String escape(final String s){
		//FIXME move to Driver
		return s.replaceAll("(['\\\\])", "\\\\$1");
	}
	
	public static String sqlReadBufferContent(final String schema, final String table_name){
		assert schema.matches(schema_name_RegEx);
		assert table_name.matches(table_name_RegEx);
		return "SELECT * FROM \"" + schema + "\".\"" + table_name + "\";";
	}
	
	public static String sqlReadBufferContent(final String schema, final String table_name, final String query_id_attribute, final int query_id){
		assert schema.matches(schema_name_RegEx);
		assert table_name.matches(table_name_RegEx);
		assert query_id_attribute.matches(attribute_name_RegEx);
		return "SELECT * FROM \"" + schema + "\".\"" + table_name + "\" WHERE \"" + query_id_attribute + "\" = " + query_id + ";";
	}
	
	public static String sqlReadBufferContent(final String schema, final String table_name, final String query_id_attribute, final String round_attribute, final int query_id, final long round){
		assert schema.matches(schema_name_RegEx);
		assert table_name.matches(table_name_RegEx);
		assert query_id_attribute.matches(attribute_name_RegEx);
		assert round_attribute.matches(attribute_name_RegEx);
		return "SELECT * FROM \"" + schema + "\".\"" + table_name + "\" WHERE \"" + query_id_attribute + "\" = " + query_id + " AND \"" + round_attribute + "\" = " + round + ";";
	}
	
	public static String sqlRecoverQuery(final String schema, final String table_name, final String query_id_attribute, final String round_attribute, final int query_id, final long last_successful_round){
		assert schema.matches(schema_name_RegEx);
		assert table_name.matches(table_name_RegEx);
		assert query_id_attribute.matches(attribute_name_RegEx);
		assert round_attribute.matches(attribute_name_RegEx);
		return "DELETE FROM \"" + schema + "\".\"" + table_name + "\" WHERE \"" + query_id_attribute + "\" = " + query_id + " AND \"" + round_attribute + "\" > " + last_successful_round + ";";		
	}

}
