package de.lmu.ifi.pms.cep.event_mill.sql;

import java.io.InputStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.net.URL;
import java.sql.Array;
import java.sql.Blob;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.Date;
import java.sql.NClob;
import java.sql.ParameterMetaData;
import java.sql.PreparedStatement;
import java.sql.Ref;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.RowId;
import java.sql.SQLException;
import java.sql.SQLXML;
import java.sql.Statement;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.Arrays;
import java.util.Calendar;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.lmu.ifi.pms.development.UnimplementedMethodError;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;


public class SimulatedPreparedStatement extends SQL_StatementWrapper implements PreparedStatement {
	
	final String[] parameterValues;
	final String[] parameterSplitedStatement;
	final StringBuilder sb;
	
	//parameterSplitedStatement should contain ''
	
	public SimulatedPreparedStatement(final Connection connection, final String sql)throws SQLException {
		this(connection.createStatement(), sql);
	}
	
	final static String comment = "/\\*(?:[^*]|\\*[^/])*+\\*/";
	
	final static String string = "(?:'(?:[^'\\\\]|\\\\\\\\|\\\\')*+')";
	
	final static String normal = "(?:[^'/?]|/[^*])*+";
	
	public final static String parameterRegex = "\\G(?:" + normal + "|" + comment + "|" + string + ")*(\\?)";
	public final static Pattern parameterPattern = Pattern.compile(parameterRegex);

	public SimulatedPreparedStatement(final Statement statement, final String sql) {
		super(statement);
		Matcher parameterMatcher = parameterPattern.matcher(sql);
		ArrayList<String> parameterSplitedSQL = new ArrayList<String>();
		int startIndex = 0;
		if(parameterMatcher.find()){
			assert parameterMatcher.end() >= startIndex;
			parameterSplitedSQL.add(sql.substring(startIndex, parameterMatcher.end()-1));
			startIndex = parameterMatcher.end();
						
			while(parameterMatcher.find()){
				assert parameterMatcher.end() >= startIndex;
				parameterSplitedSQL.add(sql.substring(startIndex, parameterMatcher.end()-1));
				startIndex = parameterMatcher.end();
			}
			parameterSplitedSQL.add(sql.substring(startIndex, sql.length()));
			this.parameterSplitedStatement = parameterSplitedSQL.toArray(new String[parameterSplitedSQL.size()]);
		}
		else{
			this.parameterSplitedStatement = new String[]{sql};
		}
		this.parameterValues = new String[this.parameterSplitedStatement.length - 1];
		
		assert this.parameterSplitedStatement.length == this.parameterValues.length + 1;
		this.sb = new StringBuilder();
	}
	
	
	public String executableStatement(){
		if(this.sb.length() == 0){
			for(int parameter = 0; parameter < this.parameterValues.length; ++parameter){
				this.sb.append(this.parameterSplitedStatement[parameter]);
				String value = this.parameterValues[parameter];
				if(value == null){
					this.sb.append("NULL");
				}
				else{
					this.sb.append("'");
					this.sb.append(value);
					this.sb.append("'");
				}
			}
			this.sb.append(this.parameterSplitedStatement[this.parameterValues.length]);
		}
		return this.sb.toString();
	}

	@Override
	public ResultSet executeQuery() throws SQLException {
		return this.statement.executeQuery(this.executableStatement());
	}

	@Override
	public int executeUpdate() throws SQLException {
		return this.statement.executeUpdate(this.executableStatement());
	}
	
	@Override
	public boolean execute() throws SQLException {
		return this.statement.execute(this.executableStatement());
	}
	
	@Override
	public void clearParameters() throws SQLException {
		Arrays.fill(this.parameterValues, null);
		this.sb.setLength(0);
	}

	@Override
	public void setNull(final int parameterIndex, final int sqlType) throws SQLException {
		this.parameterValues[parameterIndex - 1] = null;
		this.sb.setLength(0);
	}

	@Override
	public void setBoolean(final int parameterIndex, final boolean x) throws SQLException {
		this.parameterValues[parameterIndex - 1] = String.valueOf(x);
		this.sb.setLength(0);
	}

	@Override
	public void setByte(final int parameterIndex, final byte x) throws SQLException {
		this.parameterValues[parameterIndex - 1] = String.valueOf(x);
		this.sb.setLength(0);
	}

	@Override
	public void setShort(final int parameterIndex, final short x) throws SQLException {
		this.parameterValues[parameterIndex - 1] = String.valueOf(x);
		this.sb.setLength(0);
	}

	@Override
	public void setInt(final int parameterIndex, final int x) throws SQLException {
		this.parameterValues[parameterIndex - 1] = String.valueOf(x);
		this.sb.setLength(0);
	}

	@Override
	public void setLong(final int parameterIndex, final long x) throws SQLException {
		this.parameterValues[parameterIndex - 1] = String.valueOf(x);
		this.sb.setLength(0);
	}

	@Override
	public void setFloat(final int parameterIndex, final float x) throws SQLException {
		this.parameterValues[parameterIndex - 1] = String.valueOf(x);
		this.sb.setLength(0);
	}

	@Override
	public void setDouble(final int parameterIndex, final double x) throws SQLException {
		this.parameterValues[parameterIndex - 1] = String.valueOf(x);
		this.sb.setLength(0);
	}

	@Override
	public void setBigDecimal(final int parameterIndex, final BigDecimal x) throws SQLException {
		this.parameterValues[parameterIndex - 1] = String.valueOf(x);
		this.sb.setLength(0);
	}

	@Override
	public void setString(final int parameterIndex, final String x) throws SQLException {
		this.parameterValues[parameterIndex - 1] = SQL_Translation.escape(x);
		this.sb.setLength(0);
	}

	@Override
	public void setBytes(final int parameterIndex, final byte[] x) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setDate(final int parameterIndex, final Date x) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setTime(final int parameterIndex, final Time x) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setTimestamp(final int parameterIndex, final Timestamp x) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setAsciiStream(final int parameterIndex, final InputStream x, final int length) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Deprecated
	@Override
	public void setUnicodeStream(final int parameterIndex, final InputStream x, final int length) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setBinaryStream(final int parameterIndex, final InputStream x, final int length) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setObject(final int parameterIndex, final Object x, final int targetSqlType) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setObject(final int parameterIndex, final Object x) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void addBatch() throws SQLException {

	}

	@Override
	public void setCharacterStream(final int parameterIndex, final Reader reader, final int length) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setRef(final int parameterIndex, final Ref x) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setBlob(final int parameterIndex, final Blob x) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setClob(final int parameterIndex, final Clob x) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setArray(final int parameterIndex, final Array x) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public ResultSetMetaData getMetaData() throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setDate(final int parameterIndex, final Date x, final Calendar cal) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setTime(final int parameterIndex, final Time x, final Calendar cal) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setTimestamp(final int parameterIndex, final Timestamp x, final Calendar cal) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setNull(final int parameterIndex, final int sqlType, final String typeName) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setURL(final int parameterIndex, final URL x) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public ParameterMetaData getParameterMetaData() throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setRowId(final int parameterIndex, final RowId x) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setNString(final int parameterIndex, final String value) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setNCharacterStream(final int parameterIndex, final Reader value, final long length) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setNClob(final int parameterIndex, final NClob value) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setClob(final int parameterIndex, final Reader reader, final long length) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setBlob(final int parameterIndex, final InputStream inputStream, final long length) throws SQLException {
		throw new UnimplementedMethodError();
	}

	@Override
	public void setNClob(final int parameterIndex, final Reader reader, final long length) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setSQLXML(final int parameterIndex, final SQLXML xmlObject) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setObject(final int parameterIndex, final Object x, final int targetSqlType, final int scaleOrLength) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setAsciiStream(final int parameterIndex, final InputStream x, final long length) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setBinaryStream(final int parameterIndex, final InputStream x, final long length) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setCharacterStream(final int parameterIndex, final Reader reader, final long length) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setAsciiStream(final int parameterIndex, final InputStream x) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setBinaryStream(final int parameterIndex, final InputStream x) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setCharacterStream(final int parameterIndex, final Reader reader) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setNCharacterStream(final int parameterIndex, final Reader value) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setClob(final int parameterIndex, final Reader reader) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	@Override
	public void setBlob(final int parameterIndex, final InputStream inputStream) throws SQLException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setNClob(final int parameterIndex, final Reader reader) throws SQLException {
		throw new UnimplementedMethodError();
		////FIXME implement
	}

	public static void main(final String[] args) {
		Matcher m = parameterPattern.matcher(SQL_Translation.sqlProgramInstsAdd("meta"));
		while(m.find()){
			System.out.println(m.group(0));
			System.out.println("-------");
		}
	}

}
