package de.lmu.ifi.pms.cep.event_mill.sql.creation;

import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;

import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;



public final class SQL_QueryBuildHelper {
	
	static final int indentLength = 256;
	static final char[] indent = new char[indentLength]; 
	
	static{
		Arrays.fill(indent, '\t');
	}

	private SQL_QueryBuildHelper() {
		// TODO Auto-generated constructor stub
	}
	
	public static final void newLineAndIndentTab(final StringBuilder stringBuilder, final int depth){
		stringBuilder.append("\n");
		indentTab(stringBuilder, depth);
	}
	
	public static final void indentTab(final StringBuilder stringBuilder, int depth){
		while(depth > indentLength){
			stringBuilder.append(indent);
			depth -= indentLength;
		}
		if(depth > 0){
			stringBuilder.append(indent, 0, depth);
		}
	}

	public static String toString(final ResultSet resultSet, final int limit) throws SQLException{
		StringBuilder sb = new StringBuilder();
		write(resultSet, limit, sb);
		return sb.toString();
	}

	public static void write(final ResultSet resultSet, final int limit, final StringBuilder sb) throws SQLException
	{
        ResultSetMetaData metaData = resultSet.getMetaData();
        int colCount = metaData.getColumnCount();
		ArrayList<String[]> result = new ArrayList<String[]>();
        int[] maxColSize = new int[colCount];
        String[] rowBuffer = new String[colCount];
        result.add(rowBuffer);
        String entryBuffer;
        for (int col = 0, dbCol = 1; col < colCount; ++col, ++dbCol){
        	rowBuffer[col] = metaData.getColumnName(dbCol) + ":" +
                          metaData.getColumnTypeName(dbCol);
        	maxColSize[col] = Math.max(maxColSize[col], rowBuffer[col].length());
        }
        
        //place holder for horizontal line --------
        rowBuffer = new String[colCount];
        result.add(rowBuffer);
 
        for (int i = 0; resultSet.next() && i != limit; i++) {
            rowBuffer = new String[colCount];
            result.add(rowBuffer);
            for (int col = 0, dbCol = 1; col < colCount; ++col, ++dbCol) {
            	entryBuffer = resultSet.getString(dbCol);
            	rowBuffer[col] = (entryBuffer == null ? "null" : entryBuffer);
            	maxColSize[col] = Math.max(maxColSize[col], rowBuffer[col].length());
            }
        }
        
        int globalMaxColSize = de.lmu.ifi.pms.java.math.Math.max(maxColSize);
        
        rowBuffer = result.get(1);
        String hline = repeatString("-", globalMaxColSize);
        
        //setting horizontal line ---------
    	for (int col = 0; col < colCount; ++col){
    		rowBuffer[col] = hline.substring(0, maxColSize[col]);
    	}
        
        String whitespace = whitespace(globalMaxColSize);
        int numberOfTuples = result.size() - 2;
        
        writeLine(colCount, result.get(1), maxColSize, whitespace, sb);
        writeTuple(colCount, result.get(0), maxColSize, whitespace, sb);
        writeLine(colCount, result.get(1), maxColSize, whitespace, sb);
        
        String[] row;
        for(int rowIndex = 2; rowIndex < numberOfTuples + 2; ++rowIndex)
        {
        	row = result.get(rowIndex);
        	writeTuple(colCount, row, maxColSize, whitespace, sb);
        }
        
        writeLine(colCount, result.get(1), maxColSize, whitespace, sb);
        sb.append(numberOfTuples);
        sb.append(" tuples");
        sb.append("\n");
	}
	
	static void writeTuple(final int colCount, final String[] row, final int[] maxColSize, final String whitespace, final StringBuilder sb){
    	sb.append("| ");
    	for (int col = 0; col < colCount; ++col){
    		
    		sb.append(row[col]);
    		sb.append(whitespace.substring(0,maxColSize[col] - row[col].length()));
    		sb.append(" | ");
    	}
    	sb.setLength(sb.length() - " ".length());
    	sb.append("\n");
	}
	
	static void writeLine(final int colCount, final String[] row, final int[] maxColSize, final String whitespace, final StringBuilder sb){
    	sb.append("+-");
    	for (int col = 0; col < colCount; ++col){
    		
    		sb.append(row[col]);
    		sb.append(whitespace.substring(0,maxColSize[col] - row[col].length()));
    		sb.append("-+-");
    	}
    	sb.setLength(sb.length() - "-".length());
    	sb.append("\n");
	}
	
	public static String whitespace(final int count)
	{
		return repeatString(" ", count);
	}
	
	public static String repeatString(final String s, final int count)
	{
		int finalLength = s.length() * count;
		String result = s;
		while(result.length() < finalLength){
			result+=result;
		}
		return result.substring(0, finalLength);
	}
}
