package de.lmu.ifi.pms.cep.event_mill.sql.optimization.dtsa;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.DTSA_Query;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;
import de.lmu.ifi.pms.cep.tsa.DTSA_StreamBuffer;
import de.lmu.ifi.pms.cep.tsa.TSA_CompositeExpression;
import de.lmu.ifi.pms.cep.tsa.TSA_Expression;
import de.lmu.ifi.pms.cep.tsa.TSA_TemporalStreamExpression;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.Parameter;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_RuntimeException;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ImbeddingOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ProjectionOperator;
import de.lmu.ifi.pms.cep.tsa.optimization.DTSA_Optimizer;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedEnumMap;


public class DTSA_RecoveryEnhacer extends DTSA_Optimizer {
	
	public final static String tsa_recovery_query_id = ".tsa.recovery.query_id"; 
	public final static String tsa_recovery_round = ".tsa.recovery.round";
	
	@Override
	public DTSA_Program optimize(final DTSA_Program program) {
		try {
			return this.optimizeImpl(program);
		} catch (TSA_Exception e) {
			throw new TSA_RuntimeException(e);
		}
	}

	public DTSA_Program optimizeImpl(final DTSA_Program program) throws TSA_Exception {
		//FIXME assign ids aready to DTSA queries and hardcode id into query expression
		
		final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers = program.streamBuffers;	
		final FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>> queries = program.queries;
		
		@SuppressWarnings("unchecked")
		final FixedEnumMap<DTSA_BufferType, ArrayList<String>> newStreamBufferNamesList = new FixedEnumMap<DTSA_BufferType, ArrayList<String>>(DTSA_BufferType.class, new ArrayList<String>(), new ArrayList<String>(), new ArrayList<String>(), new ArrayList<String>());
		@SuppressWarnings("unchecked")
		final FixedEnumMap<DTSA_BufferType, ArrayList<DTSA_StreamBuffer>> newStreamBuffersList = new FixedEnumMap<DTSA_BufferType, ArrayList<DTSA_StreamBuffer>>(DTSA_BufferType.class, new ArrayList<DTSA_StreamBuffer>(), new ArrayList<DTSA_StreamBuffer>(), new ArrayList<DTSA_StreamBuffer>(), new ArrayList<DTSA_StreamBuffer>());
		@SuppressWarnings("unchecked")
		final FixedEnumMap<DTSA_QueryType, ArrayList<DTSA_Query>> newQueriesList = new FixedEnumMap<DTSA_QueryType, ArrayList<DTSA_Query>>(DTSA_QueryType.class, new ArrayList<DTSA_Query>(), new ArrayList<DTSA_Query>(), new ArrayList<DTSA_Query>(), new ArrayList<DTSA_Query>(), new ArrayList<DTSA_Query>(), new ArrayList<DTSA_Query>());
		
		for(DTSA_BufferType type: DTSA_BufferType.values()){
			FixedArray<DTSA_StreamBuffer> typeBuffers = streamBuffers.get(type).values();
			ArrayList<String> typeNewStreamBufferNamesList = newStreamBufferNamesList.get(type);
			ArrayList<DTSA_StreamBuffer> typeNewStreamBuffersList = newStreamBuffersList.get(type);
			for(DTSA_StreamBuffer buffer: typeBuffers){
				DTSA_StreamBuffer newBuffer = this.optimize(buffer);
				typeNewStreamBufferNamesList.add(newBuffer.full_name);
				typeNewStreamBuffersList.add(newBuffer);
			}
		}
		
		@SuppressWarnings("unchecked")
		final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> newStreamBuffers = new FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>>(DTSA_BufferType.class,
																													new FixedArrayNavigableMap<String, DTSA_StreamBuffer>(newStreamBufferNamesList.get(DTSA_BufferType.INPUT).toArray(new String[0]), newStreamBuffersList.get(DTSA_BufferType.INPUT).toArray(new DTSA_StreamBuffer[0])),
																													new FixedArrayNavigableMap<String, DTSA_StreamBuffer>(newStreamBufferNamesList.get(DTSA_BufferType.WORKING).toArray(new String[0]), newStreamBuffersList.get(DTSA_BufferType.WORKING).toArray(new DTSA_StreamBuffer[0])),
																													new FixedArrayNavigableMap<String, DTSA_StreamBuffer>(newStreamBufferNamesList.get(DTSA_BufferType.OUTPUT).toArray(new String[0]), newStreamBuffersList.get(DTSA_BufferType.OUTPUT).toArray(new DTSA_StreamBuffer[0])),
																													new FixedArrayNavigableMap<String, DTSA_StreamBuffer>(newStreamBufferNamesList.get(DTSA_BufferType.LOG).toArray(new String[0]), newStreamBuffersList.get(DTSA_BufferType.LOG).toArray(new DTSA_StreamBuffer[0]))
																													);
		
		for(DTSA_QueryType type: DTSA_QueryType.values()){
			FixedArray<DTSA_Query> typeQueries = queries.get(type);
			ArrayList<DTSA_Query> typeNewQueriesList = newQueriesList.get(type);
			for(DTSA_Query query: typeQueries){
				try {
					typeNewQueriesList.add(this.optimize(query, newStreamBuffers));
				} catch (TSA_Exception e) {
					throw new TSA_RuntimeException(e);
				}
			}
		}
		
		@SuppressWarnings("unchecked")
		final FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>> newQueries = new FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>>(DTSA_QueryType.class,
				new FixedArray<DTSA_Query>(newQueriesList.get(DTSA_QueryType.INIT).toArray(new DTSA_Query[0])),
				new FixedArray<DTSA_Query>(newQueriesList.get(DTSA_QueryType.INPUT).toArray(new DTSA_Query[0])),
				new FixedArray<DTSA_Query>(newQueriesList.get(DTSA_QueryType.WORKING).toArray(new DTSA_Query[0])),
				new FixedArray<DTSA_Query>(newQueriesList.get(DTSA_QueryType.OUTPUT).toArray(new DTSA_Query[0])),
				new FixedArray<DTSA_Query>(newQueriesList.get(DTSA_QueryType.LOG).toArray(new DTSA_Query[0])),
				new FixedArray<DTSA_Query>(newQueriesList.get(DTSA_QueryType.INPUT_GARBAGE).toArray(new DTSA_Query[0]))
				);;
		
		try {
			return new DTSA_Program(program.namespace, program.simple_name, program.description, newStreamBuffers, newQueries, program.origin);
		} catch (TSA_Exception e) {
			throw new TSA_RuntimeException(e);
		}
	}

	public DTSA_StreamBuffer optimize(final DTSA_StreamBuffer streamBuffer) throws TSA_Exception {
		if(streamBuffer.type == DTSA_BufferType.INPUT){
			return streamBuffer;
		}
		else{
			TemporalStreamSchema extendedSchema;
			try {
				extendedSchema = streamBuffer.schema.extend(new FixedArrayNavigableMap<String, Type>(new String[]{tsa_recovery_query_id, tsa_recovery_round}, new Type[]{BasicTypeID.LONG.type, BasicTypeID.LONG.type}));
			} catch (TSA_SchemaException e) {
				throw new TSA_Exception(streamBuffer.origin, e);
			}
			return new DTSA_StreamBuffer(streamBuffer.prog_package, streamBuffer.simple_name, streamBuffer.description, streamBuffer.type, extendedSchema, streamBuffer.min_times.keys(), streamBuffer.origin);
		}
	}
	
	public DTSA_Query optimize(final DTSA_Query query, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception {
		TSA_ImbeddingOperator recoveryAttributesDefinition = new TSA_ImbeddingOperator(query.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(new BasicTypeAttribute[]{BasicTypeAttribute.create(tsa_recovery_query_id, BasicTypeID.LONG), BasicTypeAttribute.create(tsa_recovery_round, BasicTypeID.LONG)}, new Term<?>[]{new Parameter("query.id"/*TODO more generic*/, BasicTypeID.LONG), new Parameter("query.round"/*TODO more generic*/, BasicTypeID.LONG)}));
		return new DTSA_Query(query.prog_package, new TSA_CompositeExpression(recoveryAttributesDefinition, this.optimize(query.expression, streamBuffers)), streamBuffers.get(query.output.type).get(query.output.full_name), query.type, query.outputMinTimes, query.outputMinTimeClockIncreases, query.description, query.origin);
	}
	
	public TSA_Expression optimize(final TSA_Expression tsa_expression, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception {
		
		TSA_Expression newExpression;
		if(tsa_expression instanceof TSA_TemporalStreamExpression){
			newExpression = this.optimize((TSA_TemporalStreamExpression) tsa_expression, streamBuffers);
		}
		else{assert tsa_expression instanceof TSA_CompositeExpression;
			newExpression = this.optimize((TSA_CompositeExpression) tsa_expression, streamBuffers);
		}
		return newExpression;
	}
	
	public TSA_Expression optimize(final TSA_CompositeExpression expression, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception {
		TSA_Expression[] newSubExpressions = new TSA_Expression[expression.subExpressions.size()];
		
		for(int subExpression = 0; subExpression < newSubExpressions.length; ++subExpression){
			newSubExpressions[subExpression] = this.optimize(expression.subExpressions.get(subExpression), streamBuffers);
		}
		
		return new TSA_CompositeExpression(expression.operator, new FixedArray<TSA_Expression>(false, newSubExpressions));
	}
	
	public TSA_Expression optimize(final TSA_TemporalStreamExpression expression, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception {
		if(expression.type == DTSA_BufferType.INPUT){
			return expression;
		}
		else{
			TSA_ProjectionOperator inputProjection = new TSA_ProjectionOperator(expression.origin, expression.schema.attributes.keys());//TODO why this???
			return new TSA_CompositeExpression(inputProjection, expression);
		}
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(DTSA_RecoveryEnhacer.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

}
