 package de.lmu.ifi.pms.cep.tsa;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementMap;

import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public class DTSA_Query
{
	@Attribute(name="package", required=true)
	public final String prog_package;
	@Attribute(name="type", required=true)
	public final DTSA_QueryType type;
	public final DTSA_BufferType input_type;
	@Element(name="output", required=true)
	public final DTSA_StreamBuffer output;
	@Element(name="expression", required=true)
	public final TSA_Expression expression;
	
	public final boolean isStatic;	
	public final boolean isMonotonic;
	
	
	public final FixedArrayNavigableMap<String, Term<BasicType>> outputMinTimes;
	
	@ElementMap(entry="min-time-function", inline=true, key="min-time", attribute=true, required=false)
	public final java.util.HashMap<String, Term<BasicType>> getOutputMinTimes(){
		return new java.util.HashMap<String, Term<BasicType>>(this.outputMinTimes.javaMap());
	}
	
	public final FixedArrayNavigableMap<String, Term<BasicType>> outputMinTimeClockIncreases;
	
	@ElementMap(entry="min-time-clock-increase-function", inline=true, key="min-time", attribute=true, required=false)
	public final java.util.HashMap<String, Term<BasicType>> getOutputMinTimeClockIncreases(){
		return new java.util.HashMap<String, Term<BasicType>>(this.outputMinTimeClockIncreases.javaMap());
	}
	
	@Element(name="description", required=false)
	public final String description;
	
	public final FixedCollection<String> origin;
	
	@ElementList(entry="origin", inline=true, required=false)
	java.util.ArrayList<String> getOriginList(){
		return new java.util.ArrayList<String>(this.origin.javaCollection());
	}
	
	public DTSA_Query(@Attribute(name="package", required=true) final String prog_package
			, @Element(name="expression", required=true) final TSA_Expression expression
			, @Element(name="output", required=true) final DTSA_StreamBuffer output
			, @Attribute(name="type", required=true) final DTSA_QueryType type
			, @ElementMap(entry="min-time-function", inline=true, key="min-time", attribute=true, required=false) final java.util.HashMap<String, Term<BasicType>> outputMinTimes
			, @ElementMap(entry="min-time-clock-increase-function", inline=true, key="min-time", attribute=true, required=false) final java.util.HashMap<String, Term<BasicType>> outputMinTimeClockIncreaseDefinitions
			, @Element(name="description", required=false) final String description
			, @ElementList(entry="origin", inline=true, required=false) final java.util.ArrayList<String> origin
		) throws TSA_Exception {
		this(prog_package, expression, output, type
				, (outputMinTimes == null ? new FixedArrayNavigableMap<String, Term<BasicType>>() : new FixedArrayNavigableMap<String, Term<BasicType>>(outputMinTimes))
				, (outputMinTimeClockIncreaseDefinitions == null ? new FixedArrayNavigableMap<String, Term<BasicType>>() : new FixedArrayNavigableMap<String, Term<BasicType>>(outputMinTimeClockIncreaseDefinitions))
				, (description == null ? "" : description)
				, (origin == null ? new FixedArray<String>(): new FixedArray<String>(origin)));
	}

	public DTSA_Query(final String prog_package, final TSA_Expression expression,
			final DTSA_StreamBuffer output, final DTSA_QueryType type,
			final FixedArrayNavigableMap<String, Term<BasicType>> outputMinTimes,
			final FixedArrayNavigableMap<String, Term<BasicType>> outputMinTimeClockIncreaseDefinitions,
			final String description, final FixedCollection<String> origin) throws TSA_Exception {
		super();
		this.prog_package = prog_package;
		assert output.type == DTSA_BufferType.WORKING | output.type == DTSA_BufferType.OUTPUT | output.type == DTSA_BufferType.LOG;
		this.type = type;
		
		FixedArray<TSA_TemporalStreamExpression> inputs = expression.inputs(); 
		if(inputs.size() > 0){		
			this.input_type = inputs.get(0).type;
			for(int input = 1; input < inputs.size(); ++input){
				if(!(inputs.get(input).type == this.input_type)){
					throw new TSA_Exception(origin, "All inputs of a query must be of the same type. Mixing type \"" + this.input_type + "\" and type \"" + inputs.get(input).type + "\" is not allowed.");
				}
			}
		}
		else{
			this.input_type = DTSA_BufferType.INPUT;
		}
		
		this.output = output;
		this.expression = expression;
		this.isStatic = this.expression.isStatic;
		this.isMonotonic = expression.isMonotonic();
		this.outputMinTimes = outputMinTimes;
		this.outputMinTimeClockIncreases = outputMinTimeClockIncreaseDefinitions;
		if(outputMinTimes.size() != this.outputMinTimeClockIncreases.size()){
			throw new IllegalArgumentException();
			//TODO better exception
		}
		this.description = description;
		
		this.origin = origin;
	}
	
	public void writeTo(final StringBuilder sb, final int indentDepth){
		indentTab(sb, indentDepth);
		sb.append("DERIVE ");
		sb.append(this.output.type.toString());
		sb.append(" ");
		sb.append(this.output.full_name);
		sb.append(" BY\n");
		this.expression.writeTo(sb, indentDepth + 1);
		sb.append("\n");
		indentTab(sb, indentDepth);
		sb.append("END");
	}
	
	public void writeTo(final StringBuilder sb){
		this.writeTo(sb, 0);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}

}
