package de.lmu.ifi.pms.cep.tsa;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.conditions.MinTime;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_TypeException;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


@Root(name="buffer")
public class DTSA_StreamBuffer {
	
	@Attribute(name="package", required=true)
	public final String prog_package;
	@Attribute(name="simple_name", required=true)
	public final String simple_name;
	public final String full_name;
	
	@Element(name="description", required=false)
	public final String description;
	
	@Element(name="type", required=true)
	public final DTSA_BufferType type;
	
	@Element(name="schema", required=true)
	public final TemporalStreamSchema schema;
	public final boolean isStatic;
	
	public final FixedArrayNavigableMap<String, MinTime> min_times;
	
	@ElementList(name="min-times", entry="min-time", inline=true, required=false, type=String.class)
	java.util.ArrayList<String> getMinTime(){
		return new java.util.ArrayList<String>(this.min_times.keys().javaCollection());
	}
	
	public final FixedCollection<String> origin;
	
	@ElementList(name="origins", entry="origin", inline=true, required=false, type=String.class)
	java.util.ArrayList<String> getOriginList(){
		return new java.util.ArrayList<String>(this.origin.javaCollection());
	}
	
	public DTSA_StreamBuffer(@Attribute(name="package", required=true) final String prog_package
			, @Attribute(name="simple_name", required=true) final String simple_name
			, @Element(name="description", required=false) final String description
			, @Element(name="type", required=true) final DTSA_BufferType type
			, @Element(name="schema", required=true) final TemporalStreamSchema schema
			, @ElementList(name="min-times", entry="min-time", inline=true, required=false) final java.util.ArrayList<String> min_times
			, @ElementList(name="origins", entry="origin", inline=true, required=false) final java.util.ArrayList<String> origin
		) throws TSA_Exception {
		this(prog_package, simple_name, (description == null ? "" : description), type, schema, (min_times == null ? new FixedArrayNavigableSet<String>() : new FixedArrayNavigableSet<String>(min_times)), (origin == null ? new FixedArray<String>() : new FixedArray<String>(origin)));
	}
	
	public DTSA_StreamBuffer(final String prog_package, final String simple_name,
			final String description, final DTSA_BufferType type,
			final TemporalStreamSchema schema,
			final FixedArrayNavigableSet<String> min_times, final FixedCollection<String> origin) throws TSA_Exception {
		this(new String[]{prog_package, simple_name}, description, type, schema, min_times, origin);		
	}
	
	public DTSA_StreamBuffer(final String full_name, final String description, final DTSA_BufferType type,
			final TemporalStreamSchema schema,
			final FixedArrayNavigableSet<String> min_times, final FixedCollection<String> origin) throws TSA_Exception {
		this(TSA_SyntaxPatterns.bufferFullName(full_name), description, type, schema, min_times, origin);
	}

	protected DTSA_StreamBuffer(final String[] full_name, final String description, final DTSA_BufferType type,
			final TemporalStreamSchema schema,
			final FixedArrayNavigableSet<String> min_times, final FixedCollection<String> origin) throws TSA_Exception {
		super();
		this.prog_package = full_name[0];
		this.simple_name = full_name[1];
		this.full_name = TSA_SyntaxPatterns.bufferFullName(this.prog_package, this.simple_name);
		this.description = description;
		this.type = type;
		this.schema = schema;
		this.isStatic = this.schema.isStatic;
		
		MinTime[] min_time_terms = new MinTime[min_times.size()];
		
		for(int min_time_index = 0; min_time_index < min_time_terms.length; ++min_time_index){
			String name = min_times.get(min_time_index);
			try {
				assert this.schema.typeOf(name) instanceof BasicType;
				min_time_terms[min_time_index] = new MinTime(this.full_name, name, (BasicType) this.schema.typeOf(name));
			} catch (TSA_TypeException e) {
				throw new TSA_Exception(origin, e);
			}
		}
		
		this.min_times = new FixedArrayNavigableMap<String, MinTime>(min_times, new FixedArray<MinTime>(false, min_time_terms));
		
		this.origin = origin;
	}
	
	public void writeTo(final StringBuilder sb, final int indentDepth, final boolean indentFirstLine, final boolean finish){
		if(indentFirstLine){
			indentTab(sb, indentDepth);
		}
		if(this.prog_package != ""){
			sb.append("PACKAGE ");
			sb.append(this.prog_package);
			sb.append("\n\n");
			indentTab(sb, indentDepth);
		}
		sb.append(this.type.toString());
		sb.append("\n");
		indentTab(sb, indentDepth);
		this.schema.writeTo(sb, indentDepth, false, this.simple_name, finish);
	}
	
	public void writeTo(final StringBuilder sb, final int indentDepth){
		this.writeTo(sb, indentDepth, false, true);
	}
	
	public void writeTo(final StringBuilder sb){
		this.writeTo(sb, 0);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}

}
