package de.lmu.ifi.pms.cep.tsa;

import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;


@Root(name="expression")
public abstract class TSA_Expression {
	
	public final boolean isStatic;
	
	public TSA_Expression(final boolean isStatic) {
		super();
		this.isStatic = isStatic;
	}

	public abstract TemporalStreamSchema schema();
	
	public abstract TSA_Expression replaceSchema(TemporalStreamSchema newSchema) throws TSA_Exception;
	
	public FixedArray<TSA_TemporalStreamExpression> inputs(){
		ArrayList<TSA_TemporalStreamExpression> inputs = new ArrayList<TSA_TemporalStreamExpression>();
		this.inputs(inputs);
		return new FixedArray<TSA_TemporalStreamExpression>(false, inputs.toArray(new TSA_TemporalStreamExpression[inputs.size()]));
	}
	
	protected abstract void inputs(final ArrayList<TSA_TemporalStreamExpression> inputs);
	
	public abstract boolean isMonotonic();
	
	public boolean isValid(){
		try{
			this.check();
			//TODO maybe more efficient implementation as isValid does not need to collect exceptions
			return true;
		}
		catch(TSA_Exception e){
			return false;
		}
	}
	
	public abstract void check() throws TSA_Exception;
	
	public abstract void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth);
	
	public void writeTo(final StringBuilder sb, final int indentDepth){
		this.writeTo(sb, indentDepth, indentDepth);
	}
	
	public void writeTo(final StringBuilder sb){
		this.writeTo(sb, 0);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}

}
