package de.lmu.ifi.pms.cep.tsa;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementUnion;

import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.schema.ActionStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.EventStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.StateStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;


public class TSA_TemporalStreamExpression extends TSA_Expression {
	
	public final FixedCollection<String> origin;
	
	@ElementList(entry="origin", inline=true)
	java.util.ArrayList<String> getOriginList(){
		return new java.util.ArrayList<String>(this.origin.javaCollection());
	}
	
	public final String prog_package;
	public final String simple_name;
	
	@Attribute(name="name")
	public final String full_name;
	
	@Attribute(name="type")
	public final DTSA_BufferType type;
	
	@ElementUnion({
		@Element(name="event-schema", type=EventStreamSchema.class),
		@Element(name="state-schema", type=StateStreamSchema.class),
		@Element(name="action-schema", type=ActionStreamSchema.class),
		@Element(name="schema", type=TemporalStreamSchema.class)
	})
	public final TemporalStreamSchema schema;
	
	public TSA_TemporalStreamExpression(final String prog_package, final String simple_name, final DTSA_BufferType type, final TemporalStreamSchema schema, final String origin) {
		this(new String[]{prog_package, simple_name}, type, schema, new FixedArray<String>(false, origin));
	}
	
	public TSA_TemporalStreamExpression(final String prog_package, final String simple_name, final DTSA_BufferType type, final TemporalStreamSchema schema, final FixedCollection<String> origin) {
		this(new String[]{prog_package, simple_name}, type, schema, origin);
	}
	
	public TSA_TemporalStreamExpression(final String full_name, final DTSA_BufferType type, final TemporalStreamSchema schema, final String origin) {
		this(TSA_SyntaxPatterns.bufferFullName(full_name), type, schema, new FixedArray<String>(false, origin));
	}
	
	public TSA_TemporalStreamExpression(@Attribute(name="name") final String full_name
			, @Attribute(name="type") final DTSA_BufferType type
			, @Element(name="schema", type=TemporalStreamSchema.class) final TemporalStreamSchema schema
			, @ElementList(entry="origin", inline=true) final java.util.ArrayList<String> origin) {
		this(full_name, type, schema, new FixedArray<String>(origin));
	}
	
	public TSA_TemporalStreamExpression(final String full_name, final DTSA_BufferType type, final TemporalStreamSchema schema, final FixedCollection<String> origin) {
		this(TSA_SyntaxPatterns.bufferFullName(full_name), type, schema, origin);
	}
	
	protected TSA_TemporalStreamExpression(final String[] full_name, final DTSA_BufferType type, final TemporalStreamSchema schema, final FixedCollection<String> origin) {
		super(schema.isStatic);
		this.prog_package = full_name[0];
		this.simple_name = full_name[1];
		this.full_name = TSA_SyntaxPatterns.bufferFullName(this.prog_package, this.simple_name);
		assert type == DTSA_BufferType.INPUT | type == DTSA_BufferType.WORKING;
		this.type = type;
		this.schema = schema;
		//FIXME to this.schema = TemporalStreamSchema.create(schema.ID, schema.attributes, schema.tupleConstraints(), ConditionFormula.and(schema.relationConstraints()).focus(AbstractFilter.notFilter(DefaultDefinition.implicitAttributeValuesSpecifications)).subformulas());
		
		this.origin = origin;
	}
	
	@Override
	public TSA_Expression replaceSchema(final TemporalStreamSchema newSchema) {
		return new TSA_TemporalStreamExpression(this.prog_package, this.simple_name, this.type, newSchema, this.origin); 
	}

	@Override
	public TemporalStreamSchema schema() {
		return this.schema;
	}
	
	@Override
	public boolean isMonotonic(){
		return true;
	}
	
	@Override
	public void check() throws TSA_Exception {

	}
	
	@Override
	protected void inputs(final ArrayList<TSA_TemporalStreamExpression> inputs) {
		inputs.add(this);
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth) {
		indentTab(sb, indentDepthFirstLine);
		sb.append(this.type.toString());
		sb.append(" ");
		sb.append(this.full_name);
	}
	
	@Override
	public String toString() {
		return String.format("%s %s : %s", this.type.toString(), this.full_name, this.schema.toString());
	}

}
