package de.lmu.ifi.pms.cep.tsa.compiler;

import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.Term.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.*;
import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.*;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.COMMON_CONSTANTS;
import de.lmu.ifi.pms.cep.event_mill.configuration.DEFAULTS;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.DTSA_Query;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;
import de.lmu.ifi.pms.cep.tsa.DTSA_StreamBuffer;
import de.lmu.ifi.pms.cep.tsa.TSA_CompositeExpression;
import de.lmu.ifi.pms.cep.tsa.TSA_Expression;
import de.lmu.ifi.pms.cep.tsa.TSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_Query;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.TSA_TemporalStream;
import de.lmu.ifi.pms.cep.tsa.TSA_TemporalStreamExpression;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTermID;
import de.lmu.ifi.pms.cep.tsa.conditions.Attribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID;
import de.lmu.ifi.pms.cep.tsa.conditions.BooleanAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BooleanTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.ClockFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DefaultDefinition;
import de.lmu.ifi.pms.cep.tsa.conditions.DisjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DurationValue;
import de.lmu.ifi.pms.cep.tsa.conditions.IdentifierAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.MinTime;
import de.lmu.ifi.pms.cep.tsa.conditions.Parameter;
import de.lmu.ifi.pms.cep.tsa.conditions.StreamBoundConstraint;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.Variable;
import de.lmu.ifi.pms.cep.tsa.conditions.adaptor.RenameAttributes;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_RuntimeException;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_BasicOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ComplexOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ImbeddingOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ProjectionOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_RenameOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_SelectionOperator;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.cep.tsa.schema.TypeID;
import de.lmu.ifi.pms.cep.tsa.time.StreamBoundBasicAttributeFilter;
import de.lmu.ifi.pms.cep.tsa.time.analysis.ArithmeticAnalyser;
import de.lmu.ifi.pms.cep.tsa.time.analysis.ArithmeticAnalyserFactory;
import de.lmu.ifi.pms.cep.tsa.time.analysis.SimpleArithmeticAnalyzerFactory;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.filter.AbstractFilter;
import de.lmu.ifi.pms.java.util.filter.Filter;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedEnumMap;


public class TSA_DefaultDifferencer extends TSA_Differencer {
	
	static final BasicTypeAttribute _o_id_attribute = attr(DEFAULTS._o_id, IDENTIFIER);
	static final BasicTypeAttribute _o_block_attribute = attr(DEFAULTS._o_block, IDENTIFIER);
	//static final String id = DuraDefaultAttributeNameMapping.identifierName;
	static final ClockFunction tuple_id_seq = new ClockFunction(COMMON_CONSTANTS.TUPLE_ID_CLOCK_NAME, IDENTIFIER); 
	
	static final String isNewAttributeName = ".tsa.isNew";
	static final String isNewAttributeAliasName = ".tsa.isNew.old";
	static final BooleanAttribute isNewAttribute = BooleanAttribute.create(isNewAttributeName);
	static final BooleanAttribute isNewAliasAttribute = BooleanAttribute.create(isNewAttributeAliasName);
	
	static final AbstractFilter<ConditionFormula> streamBoundFilter = new StreamBoundBasicAttributeFilter();
	
	final ArithmeticAnalyserFactory identifierAnalyserFactory;
	final ArithmeticAnalyserFactory temporalAnalyserFactory;
	
	public TSA_DefaultDifferencer(){
		this(new SimpleArithmeticAnalyzerFactory(IDENTIFIER), new SimpleArithmeticAnalyzerFactory(TIMESTAMP));
	}
	
	public TSA_DefaultDifferencer(final ArithmeticAnalyserFactory identifierAnalyserFactory, final ArithmeticAnalyserFactory temporalAnalyserFactory) {
		this.identifierAnalyserFactory = identifierAnalyserFactory;
		this.temporalAnalyserFactory = temporalAnalyserFactory;
	}
	
	@Override
	public DTSA_Program differenciate(final TSA_Program program) {
		try {
			return this.differenciateImpl(program);
		} catch (TSA_Exception e) {
			throw new TSA_RuntimeException(e);
		}
	}

	public DTSA_Program differenciateImpl(final TSA_Program program) throws TSA_Exception {
		try{
			assert program.isValid();
			ArrayList<String> inputStreamBufferNames = new ArrayList<String>();
			ArrayList<DTSA_StreamBuffer> inputStreamBuffers = new ArrayList<DTSA_StreamBuffer>();
			ArrayList<String> workingStreamBufferNames = new ArrayList<String>();
			ArrayList<DTSA_StreamBuffer> workingStreamBuffers = new ArrayList<DTSA_StreamBuffer>();
			ArrayList<String> outputStreamBufferNames = new ArrayList<String>();
			ArrayList<DTSA_StreamBuffer> outputStreamBuffers = new ArrayList<DTSA_StreamBuffer>();
			ArrayList<String> logStreamBufferNames = new ArrayList<String>();
			ArrayList<DTSA_StreamBuffer> logStreamBuffers = new ArrayList<DTSA_StreamBuffer>();
			ArrayList<TSA_Query> initQueries = new ArrayList<TSA_Query>();
			ArrayList<TSA_Query> inputQueries = new ArrayList<TSA_Query>();
			ArrayList<TSA_Query> workingQueries = new ArrayList<TSA_Query>();
			ArrayList<TSA_Query> outputQueries = new ArrayList<TSA_Query>();
			ArrayList<TSA_Query> logQueries = new ArrayList<TSA_Query>();
			
			for(TSA_TemporalStream temporalStream: program.temporalStreams.values()){
				try{
					DTSA_BufferType type = DTSA_BufferType.INPUT;
					if(temporalStream.types.contains(type)){
		
						final TemporalStreamSchema schema = temporalStream.schema;
						
						ConditionFormula implicitSpecificationsFormula = new ConjunctiveConditionFormula(schema.relationConstraints()).focus(DefaultDefinition.implicitAttributeValuesSpecifications).flaten();
						assert implicitSpecificationsFormula == ConditionFormula.Top
						| implicitSpecificationsFormula == ConditionFormula.Bottom //TODO should not happen: means contradiction in constraints and should be expected and treated earlier
						| implicitSpecificationsFormula instanceof BinaryComparissonFormula 
						| (implicitSpecificationsFormula instanceof ConjunctiveConditionFormula & !implicitSpecificationsFormula.containsDisjunction() & !implicitSpecificationsFormula.containsNegation());
						assert schema.basicAttributes.keys().containsAll(implicitSpecificationsFormula.AttributeNames());
						DefaultDefinition[] implicitSpecifications = implicitSpecificationsFormula.collectFormulas(DefaultDefinition.implicitAttributeValuesSpecifications).toArray(new DefaultDefinition[0]);
												
							ArrayList<TypedAttribute<?>> implicitAttributes = new ArrayList<TypedAttribute<?>>(implicitSpecifications.length);
							ArrayList<String> implicitAttributeNames = new ArrayList<String>(implicitSpecifications.length);
							ArrayList<Term<?>> definition = new ArrayList<Term<?>>(implicitSpecifications.length);
							
							for (int specification = 0; specification < implicitSpecifications.length; specification++) {
								assert implicitSpecifications[specification].attribute.isTyped;
								assert implicitSpecifications[specification].attribute.ID == AtomicTermID.ATTRIBUTE;
								TypedAttribute<?> implicitAttribute = (TypedAttribute<?>) implicitSpecifications[specification].attribute;
		
								assert implicitSpecifications[specification].definition.isTyped;
								implicitAttributes.add(implicitAttribute);
								implicitAttributeNames.add(implicitAttribute.name);
								definition.add(implicitSpecifications[specification].definition);
							}
						
						DTSA_StreamBuffer inputStream;
						TemporalStreamSchema newSchema;
						TSA_Expression query_expression;
						if(implicitAttributes.size() > 0){
							final FixedArrayNavigableSet<String> explicitAttributeNames = FixedArrayNavigableSet.difference(schema.basicAttributes.keys(), new FixedArrayNavigableSet<String>(false, true, implicitAttributeNames.toArray(new String[0])));
							final AbstractFilter<AtomicConditionFormula> explicitAttributeConstraintsFilter = new AbstractFilter<AtomicConditionFormula>(){
								@Override
								public boolean accepts(final AtomicConditionFormula atom) {
									if(explicitAttributeNames.containsAll(atom.AttributeNames())){
										return true;
									}
									return false;
								}
							};
							if(explicitAttributeNames.size() > 0){
								ConditionFormula newTupleConstraints = new ConjunctiveConditionFormula(schema.tupleConstraints()).focus(explicitAttributeConstraintsFilter).flaten();
								ConditionFormula newRelationConstraints = new ConjunctiveConditionFormula(schema.relationConstraints()).focus(AbstractFilter.notFilter(DefaultDefinition.implicitAttributeValuesSpecifications)).focus(explicitAttributeConstraintsFilter).flaten();
								try {
									newSchema = TemporalStreamSchema.create(schema.ID, schema.basicAttributes, (newTupleConstraints instanceof ConjunctiveConditionFormula ? ((ConjunctiveConditionFormula)newTupleConstraints).subformulas() : new FixedArray<ConditionFormula>(newTupleConstraints)), (newRelationConstraints instanceof ConjunctiveConditionFormula ? ((ConjunctiveConditionFormula)newRelationConstraints).subformulas() : new FixedArray<ConditionFormula>(newRelationConstraints)));
								} catch (TSA_SchemaException e) {
									throw new TSA_Exception(temporalStream.origin, e);
								}
			
								try {
									newSchema = newSchema.retain(explicitAttributeNames);
								} catch (TSA_SchemaException e) {
									throw new TSA_Exception(temporalStream.origin, e);
								}
							}
							else{
								try {
									newSchema = TemporalStreamSchema.create(schema.ID, new FixedArrayNavigableMap<String, Type>(), new FixedArray<ConditionFormula>(), new FixedArray<ConditionFormula>());
								} catch (TSA_SchemaException e) {
									throw new TSA_Exception(temporalStream.origin, e);
								} 
							}
							
							if(!temporalStream.schema.isStatic){
								try{
									//FIXME .id not stream bound
									newSchema = newSchema.extend(new FixedArrayNavigableMap<String, Type>(new String[]{DEFAULTS._o_block, DEFAULTS._o_id}, new Type[]{IDENTIFIER.type, IDENTIFIER.type})).constraint(new FixedArray<ConditionFormula>(new StreamBoundConstraint(_o_block_attribute)));
								} catch (TSA_SchemaException e) {
									throw new TSA_Exception(temporalStream.origin, e);
								} 
							}
							
							inputStream = this.differentiate(temporalStream.full_name, temporalStream.description, type, newSchema, temporalStream.origin);
							query_expression = new TSA_TemporalStreamExpression(inputStream.full_name, DTSA_BufferType.INPUT, newSchema, temporalStream.origin);					
							
							TSA_ImbeddingOperator imbeddingOperator = new TSA_ImbeddingOperator(temporalStream.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(implicitAttributes.toArray(new TypedAttribute<?>[0]), definition.toArray(new Term[0])));
							query_expression = new TSA_CompositeExpression(imbeddingOperator, query_expression);
						}
						else{
							if(!temporalStream.schema.isStatic){
								try{
									newSchema = schema.extend(new FixedArrayNavigableMap<String, Type>(new String[]{DEFAULTS._o_id}, new Type[]{IDENTIFIER.type})).constraint(new FixedArray<ConditionFormula>(new StreamBoundConstraint(attr(DEFAULTS._o_id, IDENTIFIER.type))));
								} catch (TSA_SchemaException e) {
									throw new TSA_Exception(temporalStream.origin, e);
								} 
							}
							else{
								newSchema = schema;
							}
							inputStream = this.differentiate(temporalStream.full_name, temporalStream.description, type, newSchema, temporalStream.origin);
							query_expression = new TSA_TemporalStreamExpression(inputStream.full_name, DTSA_BufferType.INPUT, newSchema, temporalStream.origin);
						}
						
						query_expression = new TSA_CompositeExpression(new TSA_ProjectionOperator(temporalStream.origin, schema.attributes.keys()), query_expression);
						
						inputStreamBufferNames.add(inputStream.full_name);
						inputStreamBuffers.add(inputStream);
						if(temporalStream.types.contains(DTSA_BufferType.WORKING)){
							if(temporalStream.isStatic){
								initQueries.add(new TSA_Query(inputStream.prog_package, temporalStream.full_name, temporalStream.schema, DTSA_BufferType.WORKING, query_expression, "Initial query for type " + inputStream.full_name, temporalStream.origin));
							}
							else{
								inputQueries.add(new TSA_Query(inputStream.prog_package, temporalStream.full_name, temporalStream.schema, DTSA_BufferType.WORKING, query_expression, "Input query for type " + inputStream.full_name, temporalStream.origin));
							}
						}
					}
				}
				catch (TSA_Exception e) {
					throw e;
				}
				catch (RuntimeException e) {
					throw new TSA_Exception(temporalStream.origin, e);
				}
			}
			for(TSA_TemporalStream temporalStream: program.temporalStreams.values()){
				try{
					DTSA_BufferType type = DTSA_BufferType.WORKING;
					if(temporalStream.types.contains(type)){
						TemporalStreamSchema schema;
						try {
							schema = TemporalStreamSchema.create(temporalStream.schema.ID, temporalStream.schema.basicAttributes, temporalStream.schema.tupleConstraints(), and(temporalStream.schema.relationConstraints()).focus(AbstractFilter.notFilter(DefaultDefinition.implicitAttributeValuesSpecifications)).subformulas());
						} catch (TSA_SchemaException e) {
							throw new TSA_Exception(temporalStream.origin, e);
						}
						
						DTSA_StreamBuffer workingStream = this.differentiate(temporalStream.full_name, temporalStream.description, type, schema, temporalStream.origin);
						workingStreamBufferNames.add(workingStream.full_name);
						workingStreamBuffers.add(workingStream);
					}
				}
				catch (TSA_Exception e) {
					throw e;
				}
				catch (RuntimeException e) {
					throw new TSA_Exception(temporalStream.origin, e);
				}
			}
			for(TSA_TemporalStream temporalStream: program.temporalStreams.values()){
				try{
					DTSA_BufferType type = DTSA_BufferType.OUTPUT;
					if(temporalStream.types.contains(type)){
						DTSA_StreamBuffer outputStreamBuffer;
						try {
							outputStreamBuffer = this.differentiate(temporalStream.full_name, temporalStream.description, type, temporalStream.schema.extend(new FixedArrayNavigableMap<String, Type>(new String[]{DEFAULTS._o_id}, new Type[]{IDENTIFIER.type})).constraint(new FixedArray<ConditionFormula>(new StreamBoundConstraint(_o_id_attribute))), temporalStream.origin);
						} catch (TSA_SchemaException e) {
							throw new TSA_Exception(temporalStream.origin, e);
						}
						
						TSA_Expression query_expression = new TSA_TemporalStreamExpression(temporalStream.full_name, DTSA_BufferType.WORKING, temporalStream.schema, temporalStream.origin);
						query_expression = new TSA_CompositeExpression(new TSA_ImbeddingOperator(temporalStream.origin, new FixedArrayNavigableMap<BasicTypeAttribute, Term<?>>(new BasicTypeAttribute[]{_o_id_attribute}, new Term<?>[]{tuple_id_seq})), query_expression);
						
						outputQueries.add(new TSA_Query(outputStreamBuffer.prog_package, outputStreamBuffer.full_name, outputStreamBuffer.schema, DTSA_BufferType.OUTPUT, query_expression, "Output query for type " + temporalStream.full_name, temporalStream.origin));
						outputStreamBufferNames.add(outputStreamBuffer.full_name);
						outputStreamBuffers.add(outputStreamBuffer);
					}
				}
				catch (TSA_Exception e) {
					throw e;
				}
				catch (RuntimeException e) {
					throw new TSA_Exception(temporalStream.origin, e);
				}
			}
			for(TSA_TemporalStream temporalStream: program.temporalStreams.values()){
				try{
					DTSA_BufferType type = DTSA_BufferType.LOG;
					if(temporalStream.types.contains(type)){
						DTSA_StreamBuffer logStreamBuffer;
						try {
							logStreamBuffer = this.differentiate(temporalStream.full_name, temporalStream.description, type, temporalStream.schema.extend(new FixedArrayNavigableMap<String, Type>(new String[]{DEFAULTS._o_id}, new Type[]{IDENTIFIER.type})).constraint(new FixedArray<ConditionFormula>(new StreamBoundConstraint(_o_id_attribute))), temporalStream.origin);
						} catch (TSA_SchemaException e) {
							throw new TSA_Exception(temporalStream.origin, e);
						}
						
						TSA_Expression query_expression = new TSA_TemporalStreamExpression(temporalStream.full_name, DTSA_BufferType.WORKING, temporalStream.schema, temporalStream.origin);
						query_expression = new TSA_CompositeExpression(new TSA_ImbeddingOperator(temporalStream.origin, new FixedArrayNavigableMap<BasicTypeAttribute, Term<?>>(new BasicTypeAttribute[]{_o_id_attribute}, new Term<?>[]{tuple_id_seq})), query_expression);
						
						logQueries.add(new TSA_Query(logStreamBuffer.prog_package, logStreamBuffer.full_name, logStreamBuffer.schema, DTSA_BufferType.OUTPUT, query_expression, "Output query for type " + temporalStream.full_name, temporalStream.origin));
						
						logStreamBufferNames.add(logStreamBuffer.full_name);
						logStreamBuffers.add(logStreamBuffer);
					}
				}
				catch (TSA_Exception e) {
					throw e;
				}
				catch (RuntimeException e) {
					throw new TSA_Exception(temporalStream.origin, e);
				}
			}
			
			@SuppressWarnings("unchecked")
			FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers = new FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String,DTSA_StreamBuffer>>(DTSA_BufferType.class,
					new FixedArrayNavigableMap<String,DTSA_StreamBuffer>(new FixedArrayNavigableSet<String>(false, false, inputStreamBufferNames.toArray(new String[0])), new FixedArray<DTSA_StreamBuffer>(false, inputStreamBuffers.toArray(new DTSA_StreamBuffer[0]))),		
					new FixedArrayNavigableMap<String,DTSA_StreamBuffer>(new FixedArrayNavigableSet<String>(false, false, workingStreamBufferNames.toArray(new String[0])), new FixedArray<DTSA_StreamBuffer>(false, workingStreamBuffers.toArray(new DTSA_StreamBuffer[0]))),
					new FixedArrayNavigableMap<String,DTSA_StreamBuffer>(new FixedArrayNavigableSet<String>(false, false, outputStreamBufferNames.toArray(new String[0])), new FixedArray<DTSA_StreamBuffer>(false, outputStreamBuffers.toArray(new DTSA_StreamBuffer[0]))),
					new FixedArrayNavigableMap<String,DTSA_StreamBuffer>(new FixedArrayNavigableSet<String>(false, false, logStreamBufferNames.toArray(new String[0])), new FixedArray<DTSA_StreamBuffer>(false, logStreamBuffers.toArray(new DTSA_StreamBuffer[0])))
			);
			
			for(TSA_Query query: program.queries){
				if(query.isStatic){
					//TODO query may only refer to input tables that are explicit (only input modifier) or to the non default attributes of an implicit input
					initQueries.add(query);
				}
				else{
					workingQueries.add(query);
				}
			}
			
			ArrayList<DTSA_Query> dtsa_initQueries = new ArrayList<DTSA_Query>();
			ArrayList<DTSA_Query> dtsa_inputQueries = new ArrayList<DTSA_Query>();
			ArrayList<DTSA_Query> dtsa_workingQueries = new ArrayList<DTSA_Query>();
			ArrayList<DTSA_Query> dtsa_outputQueries = new ArrayList<DTSA_Query>();
			ArrayList<DTSA_Query> dtsa_logQueries = new ArrayList<DTSA_Query>();
			ArrayList<DTSA_Query> dtsa_garbageQueries = new ArrayList<DTSA_Query>();
			for(TSA_Query query: initQueries){
				DTSA_QueryType type = DTSA_QueryType.INIT;
				dtsa_inputQueries.add(this.differentiate(query, type, streamBuffers));
			}
			for(TSA_Query query: inputQueries){
				DTSA_QueryType type = DTSA_QueryType.INPUT;
				dtsa_inputQueries.add(this.differentiate(query, type, streamBuffers));
			}
			for(TSA_Query query: workingQueries){
				DTSA_QueryType type = DTSA_QueryType.WORKING;
				dtsa_workingQueries.add(this.differentiate(query, type, streamBuffers));
			}
			for(TSA_Query query: outputQueries){
				DTSA_QueryType type = DTSA_QueryType.OUTPUT;
				dtsa_outputQueries.add(this.differentiate(query, type, streamBuffers));
			}
			for(TSA_Query query: logQueries){
				DTSA_QueryType type = DTSA_QueryType.LOG;
				dtsa_logQueries.add(this.differentiate(query, type, streamBuffers));
			}
			
			@SuppressWarnings("unchecked")
			FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>> queries = new FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>>(DTSA_QueryType.class,
					new FixedArray<DTSA_Query>(false, dtsa_initQueries.toArray(new DTSA_Query[0])),
					new FixedArray<DTSA_Query>(false, dtsa_inputQueries.toArray(new DTSA_Query[0])),
					new FixedArray<DTSA_Query>(false, dtsa_workingQueries.toArray(new DTSA_Query[0])),
					new FixedArray<DTSA_Query>(false, dtsa_outputQueries.toArray(new DTSA_Query[0])),
					new FixedArray<DTSA_Query>(false, dtsa_logQueries.toArray(new DTSA_Query[0])),
					new FixedArray<DTSA_Query>(false, dtsa_garbageQueries.toArray(new DTSA_Query[0]))
			);
			
			return new DTSA_Program(program.namespace, program.simple_name, program.description, streamBuffers, queries, program.origin);
		}
		catch (TSA_Exception e) {
			throw e;
		}
		catch (RuntimeException e) {
			throw new TSA_Exception(program.origin, e);
		}
	}
	
	public DTSA_BufferType outputType(final DTSA_QueryType query_type){
		switch (query_type) {
		case INIT: return DTSA_BufferType.WORKING;
		case INPUT: return DTSA_BufferType.WORKING;
		case WORKING: return DTSA_BufferType.WORKING;
		case OUTPUT: return DTSA_BufferType.OUTPUT;
		case LOG: return DTSA_BufferType.LOG;
		default:
			throw new IllegalArgumentException();
		}
	}
	
	protected ConditionFormula outputMinTimeDefinitionFormula(final MinTime outputMinTime, final ConditionFormula minTimeFormula, final ArithmeticAnalyser temporalAnalyser, final FixedArrayNavigableMap<String, MinTime> referencedAttributes, final FixedArrayNavigableMap<String, MinTime> referencedVariables){
		assert !minTimeFormula.containsNegation();
		String outputAttributeName = outputMinTime.name;
		
		final BasicTypeID type = outputMinTime.type.ID;
		Filter<ConditionFormula> typeFilter = new AbstractFilter<ConditionFormula>() {
			
			@Override
			public boolean accepts(final ConditionFormula element) {
				assert element instanceof BinaryComparissonFormula;
				BinaryComparissonFormula bcf = (BinaryComparissonFormula) element;
				assert bcf.firstTerm() instanceof MinTime;
				MinTime minTime = (MinTime) bcf.firstTerm();
				assert minTime.type.equals(bcf.secondTerm().type);
				return minTime.type.ID == type;
			}
			
		};
		
		final ConditionFormula typeFilteredMinTimeFormula = not(minTimeFormula).focus(typeFilter).subformula().flaten();
		
		if(typeFilteredMinTimeFormula.ID == ConditionFormulaID.TOP){
			return Top;
		}
		else{		
			for(ConjunctiveConditionFormula possibleStreamBounds: typeFilteredMinTimeFormula.DNFiteration()){
				boolean isStreamBound = true;
				ArrayList<ConditionFormula> outputMinTimeDefinitionFormulaPart = new ArrayList<ConditionFormula>();
				for(ConditionFormula atom: possibleStreamBounds.subformulas()){
					assert atom instanceof BinaryComparissonFormula;
					BinaryComparissonFormula streamBound = (BinaryComparissonFormula) atom;
					boolean isAttribute = (streamBound.secondTerm() instanceof BasicTypeAttribute);
					String streamBoundAttribute;
					MinTime streamBoundMinTime;
					if(isAttribute){
						streamBoundAttribute = ((Attribute<?>) streamBound.secondTerm()).name;
						assert referencedAttributes.has(streamBoundAttribute);
						streamBoundMinTime = referencedAttributes.get(streamBoundAttribute);
					}
					else{
						assert streamBound.secondTerm() instanceof BasicTypeVariable;
						streamBoundAttribute = ((Variable<?>) streamBound.secondTerm()).name;
						assert referencedVariables.has(streamBoundAttribute);
						streamBoundMinTime = referencedVariables.get(streamBoundAttribute);
					}
					
					long distance = temporalAnalyser.distance(streamBoundAttribute, isAttribute, outputAttributeName, true);
					if(distance < ArithmeticAnalyser.PLUS_INFTY){
						long constant = ArithmeticAnalyser.constant(distance);
						if(constant == 0){
							outputMinTimeDefinitionFormulaPart.add(new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS_EQUAL, outputMinTime, streamBoundMinTime));
						}
						else{
							outputMinTimeDefinitionFormulaPart.add(new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS_EQUAL, outputMinTime, apply(MINUS, TIMESTAMP.type, streamBoundMinTime, new DurationValue(constant))));
						}
					}
					else{
						isStreamBound = false;
						break;
					}
				}
				if(isStreamBound){
					//FIXME alternatives! now just the first possible one
					return and(outputMinTimeDefinitionFormulaPart).flaten();
				}
			}
			return Bottom;
		}
	}
	
	@SuppressWarnings("unchecked")
	protected static Term<BasicType> outputMinTimeDefinition(final ConditionFormula outputMinTimeDefinitionFormula){
		assert outputMinTimeDefinitionFormula.isFlatened;
		switch(outputMinTimeDefinitionFormula.ID){
			case ATOMIC: assert outputMinTimeDefinitionFormula instanceof BinaryComparissonFormula;
				BinaryComparissonFormula bcf = (BinaryComparissonFormula) outputMinTimeDefinitionFormula;
				assert bcf.operator == BinaryComparissonOperatorID.LESS_EQUAL;
				assert bcf.secondTerm().isTyped & ((Term<Type>)bcf.secondTerm()).type.ID == TypeID.BASIC;
				return (Term<BasicType>) bcf.secondTerm();
			case CONJUNCTIVE:{
				ConjunctiveConditionFormula conj = (ConjunctiveConditionFormula) outputMinTimeDefinitionFormula;
				Term<BasicType>[] subDefinitions = (Term<BasicType>[])new Term<?>[conj.subformulas().size()];
				for(int subFormula = 0; subFormula < subDefinitions.length; ++subFormula){
					subDefinitions[subFormula] = outputMinTimeDefinition(conj.subformulas().get(subFormula));
				}
				assert subDefinitions.length > 0;
				BasicType type = subDefinitions[0].type;
				return apply(LEAST, type, subDefinitions);
			}
			case DISJUNCTIVE:{
				DisjunctiveConditionFormula disj = (DisjunctiveConditionFormula) outputMinTimeDefinitionFormula;
				Term<BasicType>[] subDefinitions = (Term<BasicType>[])new Term<?>[disj.subformulas().size()];
				for(int subFormula = 0; subFormula < subDefinitions.length; ++subFormula){
					subDefinitions[subFormula] = outputMinTimeDefinition(disj.subformulas().get(subFormula));
				}
				assert subDefinitions.length > 0;
				BasicType type = subDefinitions[0].type;
				return apply(GREATEST, type, subDefinitions);
			}
			case NEGATED: throw new Error("Unexpected Case" + outputMinTimeDefinitionFormula);
			case BOTTOM: return Bottom;
			case TOP: return Top;
		}
		throw new Error("Unexpected Case: " + outputMinTimeDefinitionFormula);
	}
	
	protected Term<BasicType> outputMinTimeClockIncreaseDefinition(final MinTime outputMinTime, final ArithmeticAnalyser temporalAnalyser, final FixedArrayNavigableMap<String, ClockFunction> clockAttributes, final FixedArrayNavigableMap<String, ClockFunction> clockVariables){
		String outputAttributeName = outputMinTime.name;
		ArrayList<Term<BasicType>> increasingClocks = new ArrayList<Term<BasicType>>();
		for(String clockAttribute: clockAttributes.keys()){
			long distance = temporalAnalyser.distance(clockAttribute, true, outputAttributeName, true);
			if(distance < ArithmeticAnalyser.PLUS_INFTY){
				long constant = ArithmeticAnalyser.constant(distance);
				ClockFunction clock = clockAttributes.get(clockAttribute);
				if(constant == 0){
					increasingClocks.add(clock);
				}
				else{
					increasingClocks.add(apply(MINUS, clock.type, clock, new DurationValue(constant)));
				}
			}
		}
		for(String clockVariable: clockVariables.keys()){
			long distance = temporalAnalyser.distance(clockVariable, false, outputAttributeName, true);
			if(distance < ArithmeticAnalyser.PLUS_INFTY){
				long constant = ArithmeticAnalyser.constant(distance);
				ClockFunction clock = clockVariables.get(clockVariable);
				if(constant == 0){
					increasingClocks.add(clock);
				}
				else{
					increasingClocks.add(apply(MINUS, clock.type, clock, new DurationValue(constant)));
				}
			}
		}
		
		if(increasingClocks.size() == 0){
			return Bottom;
		}
		else if(increasingClocks.size() == 1){
			return increasingClocks.first();
		}
		else{
			return apply(GREATEST, increasingClocks.first().type, increasingClocks.toArray(new Term<?>[0]));	
		}
	}
	
	protected final static Filter<ConditionFormula> minTimesFilter = new AbstractFilter<ConditionFormula>() {
		
		@Override
		public boolean accepts(final ConditionFormula element) {
			if(element instanceof BinaryComparissonFormula){
				BinaryComparissonFormula bcf = (BinaryComparissonFormula) element;
				if(bcf.firstTerm() instanceof MinTime){
					MinTime minTime = (MinTime) bcf.firstTerm();
					if(minTime.type.ID == IDENTIFIER | minTime.type.ID == TIMESTAMP){
						if(bcf.secondTerm() instanceof BasicTypeAttribute){
							BasicTypeAttribute attribute = (BasicTypeAttribute) bcf.secondTerm();
							return attribute.ID == minTime.type.ID;
						}
						if(bcf.secondTerm() instanceof BasicTypeVariable){
							BasicTypeVariable v2 = (BasicTypeVariable) bcf.secondTerm();
							return v2.ID == minTime.type.ID;
						}
					}
				}
			}
			return false;
		}
		
	};
	
	protected final static Filter<ConditionFormula> atomicFormulaFilter = new AbstractFilter<ConditionFormula>() {
		@Override
		public boolean accepts(final ConditionFormula element) {
			return element.ID == ConditionFormulaID.ATOMIC;
		}
	};
	
	public DTSA_Query differentiate(final TSA_Query tsa_query, final DTSA_QueryType type, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception{
		try{
			assert (tsa_query.isStatic & type == DTSA_QueryType.INIT) | (!tsa_query.isStatic & type != DTSA_QueryType.INIT); 
			
			final DTSA_BufferType output_type = this.outputType(type);
			
			final DTSA_StreamBuffer output = streamBuffers.get(output_type).get(tsa_query.output_full_name);
			
			final boolean isMonotonic = tsa_query.isMonotonic();
			
			TSA_Expression differenciatedExpression = this.differentiate(tsa_query.expression, type, isMonotonic, streamBuffers);
			
			ConditionFormula minTimeFormula = and(differenciatedExpression.schema().relationConstraints()).focus(minTimesFilter).flaten();
			assert tsa_query.isStatic | minTimeFormula.ID != ConditionFormulaID.TOP: "Top is only allowed for static relations";
			
			final FixedArrayNavigableMap<String, MinTime> referencedIdentifierAttributes;
			final FixedArrayNavigableMap<String, MinTime> referencedTimestampAttributes;
			final FixedArrayNavigableMap<String, MinTime> referencedIdentifierVariables;
			final FixedArrayNavigableMap<String, MinTime> referencedTimestampVariables;
			{
				AbstractCollection<ConditionFormula> minTimeInequations = minTimeFormula.collectFormulas(atomicFormulaFilter);
				
				ArrayList<String> referencedIdentifierAttributeNamesList = new ArrayList<String>();
				ArrayList<String> referencedTimestampAttributeNamesList = new ArrayList<String>();
				ArrayList<MinTime> referencedIdentifierAttributeMinTimeList = new ArrayList<MinTime>();
				ArrayList<MinTime> referencedTimestampAttributeMinTimeList = new ArrayList<MinTime>();
				ArrayList<String> referencedIdentifierVariableNamesList = new ArrayList<String>();
				ArrayList<String> referencedTimestampVariableNamesList = new ArrayList<String>();
				ArrayList<MinTime> referencedIdentifierVariableMinTimeList = new ArrayList<MinTime>();
				ArrayList<MinTime> referencedTimestampVariableMinTimeList = new ArrayList<MinTime>();
				
				for(ConditionFormula minTimeInequationFormula: minTimeInequations){
					BinaryComparissonFormula minTimeInequation = (BinaryComparissonFormula) minTimeInequationFormula;
					MinTime minTime = (MinTime) minTimeInequation.firstTerm();
					if(minTimeInequation.secondTerm() instanceof BasicTypeAttribute){
						BasicTypeAttribute attribute = (BasicTypeAttribute) minTimeInequation.secondTerm();
						if(attribute.ID == IDENTIFIER){
							referencedIdentifierAttributeNamesList.add(attribute.name);
							referencedIdentifierAttributeMinTimeList.add(minTime);
						}
						else{assert attribute.ID == TIMESTAMP;
							referencedTimestampAttributeNamesList.add(attribute.name);
							referencedTimestampAttributeMinTimeList.add(minTime);
						}
					}
					else{assert minTimeInequation.secondTerm() instanceof BasicTypeVariable;					
						BasicTypeVariable variable = (BasicTypeVariable) minTimeInequation.secondTerm();
						if(variable.ID == IDENTIFIER){
							referencedIdentifierVariableNamesList.add(variable.name);
							referencedIdentifierVariableMinTimeList.add(minTime);
						}
						else{assert variable.ID == TIMESTAMP;
							referencedTimestampVariableNamesList.add(variable.name);
							referencedTimestampVariableMinTimeList.add(minTime);
						}
					}
				}
				
				referencedIdentifierAttributes = new FixedArrayNavigableMap<String, MinTime>(referencedIdentifierAttributeNamesList.toArray(new String[0]), referencedIdentifierAttributeMinTimeList.toArray(new MinTime[0]));
				referencedTimestampAttributes = new FixedArrayNavigableMap<String, MinTime>(referencedTimestampAttributeNamesList.toArray(new String[0]), referencedTimestampAttributeMinTimeList.toArray(new MinTime[0]));
				referencedIdentifierVariables = new FixedArrayNavigableMap<String, MinTime>(referencedIdentifierVariableNamesList.toArray(new String[0]), referencedIdentifierVariableMinTimeList.toArray(new MinTime[0]));
				referencedTimestampVariables = new FixedArrayNavigableMap<String, MinTime>(referencedTimestampVariableNamesList.toArray(new String[0]), referencedTimestampVariableMinTimeList.toArray(new MinTime[0]));
			}
			
			Filter<ConditionFormula> clockFunctionsFilter = new AbstractFilter<ConditionFormula>() {
				
				@Override
				public boolean accepts(final ConditionFormula element) {
					if(element instanceof BinaryComparissonFormula){
						BinaryComparissonFormula bcf = (BinaryComparissonFormula) element;
						if(bcf.secondTerm() instanceof ClockFunction){
							return true;
						}
					}
					return false;
				}
				
			};
			
			final FixedArrayNavigableMap<String, ClockFunction> clockIdentifierAttributes;
			final FixedArrayNavigableMap<String, ClockFunction> clockTimestampAttributes;
			final FixedArrayNavigableMap<String, ClockFunction> clockIdentifierVariables;
			final FixedArrayNavigableMap<String, ClockFunction> clockTimestampVariables;
			{
				AbstractCollection<ConditionFormula> clockEquations = and(differenciatedExpression.schema().tupleConstraints()).collectFormulas(clockFunctionsFilter);
				
				ArrayList<String> clockIdentifierAttributeNamesList = new ArrayList<String>();
				ArrayList<String> clockTimestampAttributeNamesList = new ArrayList<String>();
				ArrayList<ClockFunction> clockIdentifierAttributeClockFunctionList = new ArrayList<ClockFunction>();
				ArrayList<ClockFunction> clockTimestampAttributeClockFunctionList = new ArrayList<ClockFunction>();
				ArrayList<String> clockIdentifierVariableNamesList = new ArrayList<String>();
				ArrayList<String> clockTimestampVariableNamesList = new ArrayList<String>();
				ArrayList<ClockFunction> clockIdentifierVariableClockFunctionList = new ArrayList<ClockFunction>();
				ArrayList<ClockFunction> clockTimestampVariableClockFunctionList = new ArrayList<ClockFunction>();
				
				for(ConditionFormula minTimeInequationFormula: clockEquations){
					BinaryComparissonFormula clockEquation = (BinaryComparissonFormula) minTimeInequationFormula;
					ClockFunction clock = (ClockFunction) clockEquation.secondTerm();
					if(clockEquation.firstTerm() instanceof BasicTypeAttribute){
						BasicTypeAttribute attribute = (BasicTypeAttribute) clockEquation.firstTerm();
						if(attribute.ID == IDENTIFIER){
							clockIdentifierAttributeNamesList.add(attribute.name);
							clockIdentifierAttributeClockFunctionList.add(clock);
						}
						else{assert attribute.ID == TIMESTAMP;
							clockTimestampAttributeNamesList.add(attribute.name);
							clockTimestampAttributeClockFunctionList.add(clock);
						}
					}
					else{assert clockEquation.firstTerm() instanceof BasicTypeVariable;					
						BasicTypeVariable variable = (BasicTypeVariable) clockEquation.firstTerm();
						if(variable.ID == IDENTIFIER){
							clockIdentifierVariableNamesList.add(variable.name);
							clockIdentifierVariableClockFunctionList.add(clock);
						}
						else{assert variable.ID == TIMESTAMP;
							clockTimestampVariableNamesList.add(variable.name);
							clockTimestampVariableClockFunctionList.add(clock);
						}
					}
				}
				
				clockIdentifierAttributes = new FixedArrayNavigableMap<String, ClockFunction>(clockIdentifierAttributeNamesList.toArray(new String[0]), clockIdentifierAttributeClockFunctionList.toArray(new ClockFunction[0]));
				clockTimestampAttributes = new FixedArrayNavigableMap<String, ClockFunction>(clockTimestampAttributeNamesList.toArray(new String[0]), clockTimestampAttributeClockFunctionList.toArray(new ClockFunction[0]));
				clockIdentifierVariables = new FixedArrayNavigableMap<String, ClockFunction>(clockIdentifierVariableNamesList.toArray(new String[0]), clockIdentifierVariableClockFunctionList.toArray(new ClockFunction[0]));
				clockTimestampVariables = new FixedArrayNavigableMap<String, ClockFunction>(clockTimestampVariableNamesList.toArray(new String[0]), clockTimestampVariableClockFunctionList.toArray(new ClockFunction[0]));
			}
			
			final FixedArrayNavigableMap<String, MinTime> outputIdentifierAttributes;
			final FixedArrayNavigableMap<String, MinTime> outputTimestampAttributes;
			final FixedArrayNavigableMap<String, MinTime> outputMinTimes;
			
			{
				outputMinTimes = output.min_times;
				ArrayList<String> outputIdentifierAttributeNamesList = new ArrayList<String>();
				ArrayList<String> outputTimestampAttributeNamesList = new ArrayList<String>();
				ArrayList<MinTime> outputIdentifierAttributeMinTimeList = new ArrayList<MinTime>();
				ArrayList<MinTime> outputTimestampAttributeMinTimeList = new ArrayList<MinTime>();
				
				for(int outputMinTime = 0; outputMinTime < outputMinTimes.size(); ++outputMinTime){
					MinTime minTime = outputMinTimes.values().get(outputMinTime);
					BasicTypeAttribute correspondingOutputAttribute = (BasicTypeAttribute) output.schema.getAttribute(outputMinTimes.keys().get(outputMinTime));
					if(correspondingOutputAttribute.ID == IDENTIFIER){
						outputIdentifierAttributeNamesList.add(correspondingOutputAttribute.name);
						outputIdentifierAttributeMinTimeList.add(minTime);
					}
					else{assert correspondingOutputAttribute.ID == TIMESTAMP;
						outputTimestampAttributeNamesList.add(correspondingOutputAttribute.name);
						outputTimestampAttributeMinTimeList.add(minTime);					
					}
				}
				
				outputIdentifierAttributes = new FixedArrayNavigableMap<String, MinTime>(new FixedArrayNavigableSet<String>(false, outputIdentifierAttributeNamesList.toArray(new String[0])),new FixedArray<MinTime>(false, outputIdentifierAttributeMinTimeList.toArray(new MinTime[0])));
				outputTimestampAttributes = new FixedArrayNavigableMap<String, MinTime>(new FixedArrayNavigableSet<String>(false, outputTimestampAttributeNamesList.toArray(new String[0])),new FixedArray<MinTime>(false, outputTimestampAttributeMinTimeList.toArray(new MinTime[0])));
			}
			
			@SuppressWarnings("unchecked")
			final ArithmeticAnalyser identifierAnalyser = this.identifierAnalyserFactory.make(and(differenciatedExpression.schema().tupleConstraints()), FixedArrayNavigableSet.merge(referencedIdentifierAttributes.keys(), clockIdentifierAttributes.keys(), outputIdentifierAttributes.keys()), FixedArrayNavigableSet.merge(referencedIdentifierVariables.keys(), clockIdentifierVariables.keys()));
			
			ConditionFormula[] outputIdentifierMinTimesDefinitionFormulas = new ConditionFormula[outputIdentifierAttributes.size()];
			for(int outputMinTimeIndex = 0; outputMinTimeIndex < outputIdentifierMinTimesDefinitionFormulas.length; ++outputMinTimeIndex){
				MinTime outputMinTime = outputIdentifierAttributes.values().get(outputMinTimeIndex);
				outputIdentifierMinTimesDefinitionFormulas[outputMinTimeIndex] = this.outputMinTimeDefinitionFormula(outputMinTime, minTimeFormula, identifierAnalyser, referencedIdentifierAttributes, referencedIdentifierVariables);
			}
			
			@SuppressWarnings("unchecked")
			final ArithmeticAnalyser temporalAnalyser = this.temporalAnalyserFactory.make(and(differenciatedExpression.schema().tupleConstraints()), FixedArrayNavigableSet.merge(referencedTimestampAttributes.keys(), clockTimestampAttributes.keys(), outputTimestampAttributes.keys()), FixedArrayNavigableSet.merge(referencedTimestampVariables.keys(), clockTimestampVariables.keys()));
			
			ConditionFormula[] outputTimestampMinTimesDefinitionFormulas = new ConditionFormula[outputTimestampAttributes.size()];
			for(int outputMinTimeIndex = 0; outputMinTimeIndex < outputTimestampMinTimesDefinitionFormulas.length; ++outputMinTimeIndex){
				MinTime outputMinTime = outputTimestampAttributes.values().get(outputMinTimeIndex);
				outputTimestampMinTimesDefinitionFormulas[outputMinTimeIndex] = this.outputMinTimeDefinitionFormula(outputMinTime, minTimeFormula, temporalAnalyser, referencedTimestampAttributes, referencedTimestampVariables);
			}
			
			final FixedArrayNavigableMap<String, Term<BasicType>> outputIdentifierMinTimeDefinitions;
			{
					@SuppressWarnings("unchecked")
					Term<BasicType>[] definitions = (Term<BasicType>[])new Term<?>[outputIdentifierMinTimesDefinitionFormulas.length];
					for(int outputMinTime = 0; outputMinTime < outputIdentifierMinTimesDefinitionFormulas.length; ++outputMinTime){
						definitions[outputMinTime] = outputMinTimeDefinition(outputIdentifierMinTimesDefinitionFormulas[outputMinTime]);
					}
					outputIdentifierMinTimeDefinitions = new FixedArrayNavigableMap<String, Term<BasicType>>(outputIdentifierAttributes.keys(), new FixedArray<Term<BasicType>>(false, definitions));
			}
			
			final FixedArrayNavigableMap<String, Term<BasicType>> outputTimestampMinTimeDefinitions;
			{
					@SuppressWarnings("unchecked")
					Term<BasicType>[] definitions = (Term<BasicType>[])new Term<?>[outputTimestampMinTimesDefinitionFormulas.length];
					for(int outputMinTime = 0; outputMinTime < outputTimestampMinTimesDefinitionFormulas.length; ++outputMinTime){
						definitions[outputMinTime] = outputMinTimeDefinition(outputTimestampMinTimesDefinitionFormulas[outputMinTime]);
					}
					outputTimestampMinTimeDefinitions = new FixedArrayNavigableMap<String, Term<BasicType>>(outputTimestampAttributes.keys(), new FixedArray<Term<BasicType>>(false, definitions));
			}
			
			final FixedArrayNavigableMap<String, Term<BasicType>> outputMinTimeDefinitions = FixedArrayNavigableMap.merge(outputIdentifierMinTimeDefinitions, outputTimestampMinTimeDefinitions);
			
			/*
			@SuppressWarnings("unchecked")
			ArrayList<Term>[] outputMinTimePreDefinitions = new ArrayList[outputMinTimeNames.size()];
			for(int outputMinTime = 0; outputMinTime < outputMinTimePreDefinitions.length; ++outputMinTime){
				outputMinTimePreDefinitions[outputMinTime] = new ArrayList<Term>();
			}*/
			
			final FixedArrayNavigableMap<String, Term<BasicType>> outputIdentifierMinTimeClockIncreaseDefinitions;
			{
				@SuppressWarnings("unchecked")
				Term<BasicType>[] definitions = (Term<BasicType>[])new Term<?>[outputIdentifierMinTimesDefinitionFormulas.length];
				for(int outputMinTimeIndex = 0; outputMinTimeIndex < outputIdentifierMinTimesDefinitionFormulas.length; ++outputMinTimeIndex){
					MinTime outputMinTime = outputIdentifierAttributes.values().get(outputMinTimeIndex);
					definitions[outputMinTimeIndex] = this.outputMinTimeClockIncreaseDefinition(outputMinTime, identifierAnalyser, clockIdentifierAttributes, clockIdentifierVariables);	
				}
				outputIdentifierMinTimeClockIncreaseDefinitions = new FixedArrayNavigableMap<String, Term<BasicType>>(outputIdentifierAttributes.keys(), new FixedArray<Term<BasicType>>(false, definitions));
			}
			final FixedArrayNavigableMap<String, Term<BasicType>> outputTimestampMinTimeClockIncreaseDefinitions;
			{
				@SuppressWarnings("unchecked")
				Term<BasicType>[] definitions = (Term<BasicType>[])new Term<?>[outputTimestampMinTimesDefinitionFormulas.length];
				for(int outputMinTimeIndex = 0; outputMinTimeIndex < outputTimestampMinTimesDefinitionFormulas.length; ++outputMinTimeIndex){
					MinTime outputMinTime = outputTimestampAttributes.values().get(outputMinTimeIndex);
					definitions[outputMinTimeIndex] = this.outputMinTimeClockIncreaseDefinition(outputMinTime, temporalAnalyser, clockTimestampAttributes, clockTimestampVariables);	
				}
				outputTimestampMinTimeClockIncreaseDefinitions = new FixedArrayNavigableMap<String, Term<BasicType>>(outputTimestampAttributes.keys(), new FixedArray<Term<BasicType>>(false, definitions));
			}
			
			final FixedArrayNavigableMap<String, Term<BasicType>> outputMinTimeClockIncreaseDefinitions = FixedArrayNavigableMap.merge(outputIdentifierMinTimeClockIncreaseDefinitions, outputTimestampMinTimeClockIncreaseDefinitions);
			
			boolean hasNonClockOutputStreamBound = false; 
			//check definitions of output minTimes
			for(int outputMinTimeIndex = 0; outputMinTimeIndex < outputMinTimeDefinitions.size(); ++outputMinTimeIndex){
				Term<BasicType> definition = outputMinTimeDefinitions.values().get(outputMinTimeIndex);
				Term<BasicType> clockIncreaseDefinition = outputMinTimeClockIncreaseDefinitions.values().get(outputMinTimeIndex);
	
				if(definition instanceof ConditionFormula && ((ConditionFormula)definition).ID == ConditionFormulaID.BOTTOM){
					if(clockIncreaseDefinition instanceof ConditionFormula && ((ConditionFormula)clockIncreaseDefinition).ID == ConditionFormulaID.BOTTOM){
						throw new TSA_Exception(tsa_query.origin , "Output stream bound " + outputMinTimeDefinitions.keys().get(outputMinTimeIndex) + " is invalid.");
					}
				}
				else{
					hasNonClockOutputStreamBound = true;
				}
			}
			
			TSA_Expression incrementExpression;
			
			if(type == DTSA_QueryType.INIT){
				incrementExpression = differenciatedExpression;
			}
			else if(isMonotonic){		
				
				boolean hasIsNewAttribute = differenciatedExpression.schema().hasAttribute(isNewAttributeName);
				ConditionFormula isNewFormula = isNewFormula(hasIsNewAttribute, referencedIdentifierAttributes);
				assert isNewFormula.ID != ConditionFormulaID.TOP; 
				final TSA_SelectionOperator isNewSelection = new TSA_SelectionOperator(tsa_query.origin, isNewFormula);
	
				if(hasIsNewAttribute){
					incrementExpression =  new TSA_CompositeExpression(new TSA_ProjectionOperator(tsa_query.origin, tsa_query.expression.schema().basicAttributes.keys()), new TSA_CompositeExpression(isNewSelection, differenciatedExpression));
				}
				else{
					incrementExpression = new TSA_CompositeExpression(isNewSelection, differenciatedExpression);
				}
				
	
			}
			else{
				assert type != DTSA_QueryType.INIT;
				assert type != DTSA_QueryType.INPUT;
				
				if(!tsa_query.isStatic & !hasNonClockOutputStreamBound){
					//at least one outputMinTimeDefinition must be different from Bottom
					throw new TSA_Exception(tsa_query.origin, "The temporal relations are too weak to derive an incremental evaluation expression for the non-static query.");
					//TODO better Exception
				}
				assert !tsa_query.isStatic | !hasNonClockOutputStreamBound;//a static query may only derive stream bounds out of clocks
				
				ArrayList<ConditionFormula> incrementLowerBoundConditions = new ArrayList<ConditionFormula>();
				ArrayList<ConditionFormula> incrementUpperBoundConditions = new ArrayList<ConditionFormula>();
				
				for(int outputMinTimeIndex = 0; outputMinTimeIndex < outputMinTimeDefinitions.size(); ++outputMinTimeIndex){
					Term<BasicType> definition = outputMinTimeDefinitions.values().get(outputMinTimeIndex);
					if(!(definition instanceof ConditionFormula && ((ConditionFormula)definition).ID == ConditionFormulaID.BOTTOM)){
						MinTime minTime = outputMinTimes.values().get(outputMinTimeIndex);
						BasicTypeAttribute correspondingAttribute = (BasicTypeAttribute) output.schema.getAttribute(minTime.name);
						
						Parameter lowerBound = new Parameter("output.last:" + minTime.qualified_name, minTime.type);
						Parameter upperBound = new Parameter("output.now:" + minTime.qualified_name, minTime.type);
						
						incrementUpperBoundConditions.add(new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS_EQUAL, correspondingAttribute, upperBound));
						incrementLowerBoundConditions.add(new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS, lowerBound, correspondingAttribute));
					}
				}			
				
				TSA_SelectionOperator incrementSelection = new TSA_SelectionOperator(tsa_query.origin, and(incrementLowerBoundConditions).flaten(), or(incrementUpperBoundConditions).flaten());
	
				incrementExpression = new TSA_CompositeExpression(incrementSelection, differenciatedExpression);
			}
	
			return new DTSA_Query(tsa_query.prog_package, incrementExpression, output, type, outputMinTimeDefinitions, outputMinTimeClockIncreaseDefinitions, tsa_query.description, tsa_query.origin);
		}
		catch (TSA_Exception e) {
			throw e;
		}
		catch (RuntimeException e) {
			throw new TSA_Exception(tsa_query.origin, e);
		}
	}
	
	public DTSA_StreamBuffer differentiate(final String full_name, final String description, final DTSA_BufferType type, final TemporalStreamSchema schema, final FixedCollection<String> origin) throws TSA_Exception{
		try{		
			FixedArrayNavigableSet<String> minTimes = and(and(schema.relationConstraints()).collectFormulas(streamBoundFilter)).AttributeNames();
					
			TemporalStreamSchema differentiatedSchema = schema;
			return new DTSA_StreamBuffer(full_name, description, type, differentiatedSchema, minTimes, origin);
		}
		catch (TSA_Exception e) {
			throw e;
		}
		catch (RuntimeException e) {
			throw new TSA_Exception(origin, e);
		}
	}
	
	public TSA_Expression differentiate(final TSA_Expression tsa_expression, final DTSA_QueryType type, final boolean monotonic, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception{
		if(tsa_expression instanceof TSA_TemporalStreamExpression){
			return this.differentiate((TSA_TemporalStreamExpression) tsa_expression, streamBuffers);
		}
		else{assert tsa_expression instanceof TSA_CompositeExpression;
			return this.differentiate((TSA_CompositeExpression) tsa_expression, type, monotonic, streamBuffers);
		}
				
	}
	
	protected static ConditionFormula isNewFormula(final boolean hasIsNewAttribute, final FixedArrayNavigableMap<String, MinTime> idAttributes){
		ConditionFormula isNewFormula = Top;
		if(idAttributes.size() > 0){
			BinaryComparissonFormula[] lowerBoundConditions = new BinaryComparissonFormula[idAttributes.size()];
			BinaryComparissonFormula[] upperBoundConditions = new BinaryComparissonFormula[lowerBoundConditions.length];
			for (int discardedIdAttribute = 0; discardedIdAttribute < lowerBoundConditions.length; ++discardedIdAttribute) {
				IdentifierAttribute attribute = IdentifierAttribute.create(idAttributes.keys().get(discardedIdAttribute));
				MinTime minTime = idAttributes.values().get(discardedIdAttribute);
				
				Parameter lowerBound = new Parameter("input.last:" + minTime.qualified_name, minTime.type);
				Parameter upperBound = new Parameter("input.now:" + minTime.qualified_name, minTime.type);
				lowerBoundConditions[discardedIdAttribute] = new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS, lowerBound, attribute); 
				upperBoundConditions[discardedIdAttribute] = new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS_EQUAL, attribute, upperBound);
			}
			isNewFormula = and(new DisjunctiveConditionFormula(lowerBoundConditions), new ConjunctiveConditionFormula(upperBoundConditions));
		}
		if(hasIsNewAttribute){
			isNewFormula = 	and(new BooleanTerm(isNewAttribute) ,isNewFormula);
		}
		return isNewFormula.flaten();
	}
	
	public TSA_Expression differentiate(final TSA_CompositeExpression tsa_compositeExpression, final DTSA_QueryType type, final boolean monotonic, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception{
		try{
			switch (tsa_compositeExpression.operator.ID) {
			case BASIC: 
			{
				switch (((TSA_BasicOperator)tsa_compositeExpression.operator).ID) {
				case CROSSPRODUCT: 
				{				
					assert tsa_compositeExpression.subExpressions.size() > 0;
					boolean containsAnyIsNewAttribute = false;
					ArrayList<BooleanTerm> isNewAttributes = new ArrayList<BooleanTerm>();
					TSA_Expression[] differentiatedSubExpressions = new TSA_Expression[tsa_compositeExpression.subExpressions.size()];
					for (int subExpression = 0; subExpression < differentiatedSubExpressions.length; ++subExpression) {
						differentiatedSubExpressions[subExpression] = this.differentiate(tsa_compositeExpression.subExpressions.get(subExpression), type, monotonic, streamBuffers);
						if(differentiatedSubExpressions[subExpression].schema().hasAttribute(isNewAttributeName)){
							containsAnyIsNewAttribute = true;
							String isNewAttributeNewName = isNewAttributeName + subExpression;
							isNewAttributes.add(new BooleanTerm(BooleanAttribute.create(isNewAttributeNewName)));
							differentiatedSubExpressions[subExpression] = new TSA_CompositeExpression(new TSA_RenameOperator(tsa_compositeExpression.operator.origin, new FixedArrayNavigableMap<String, String>(new String[]{isNewAttributeName}, new String[]{isNewAttributeNewName})), differentiatedSubExpressions[subExpression]);
						}
					}
					if(containsAnyIsNewAttribute){
						TSA_ImbeddingOperator isNewImbedding = new TSA_ImbeddingOperator(tsa_compositeExpression.operator.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(new BooleanAttribute[]{isNewAttribute}, new Term<?>[]{and(isNewAttributes)}));
						TSA_ProjectionOperator isNewProjection = new TSA_ProjectionOperator(tsa_compositeExpression.operator.origin, FixedArrayNavigableSet.merge(tsa_compositeExpression.schema().basicAttributes.keys(), new FixedArrayNavigableSet<String>(isNewAttributeName)));
						return new TSA_CompositeExpression(isNewProjection, new TSA_CompositeExpression(isNewImbedding, new TSA_CompositeExpression(tsa_compositeExpression.operator, new FixedArray<TSA_Expression>(false, differentiatedSubExpressions))));
					}
					else{
						return new TSA_CompositeExpression(tsa_compositeExpression.operator, new FixedArray<TSA_Expression>(false, differentiatedSubExpressions));
					}
				}
				case UNION: 
				{				
					assert tsa_compositeExpression.subExpressions.size() > 0;
					TSA_Expression[] differentiatedSubExpressions = new TSA_Expression[tsa_compositeExpression.subExpressions.size()];
					for (int subExpression = 0; subExpression < differentiatedSubExpressions.length; ++subExpression) {
						differentiatedSubExpressions[subExpression] = this.differentiate(tsa_compositeExpression.subExpressions.get(subExpression), type, monotonic, streamBuffers);
					}
					return new TSA_CompositeExpression(tsa_compositeExpression.operator, new FixedArray<TSA_Expression>(false, differentiatedSubExpressions));
				}
				case GROUPING:
				{
					assert tsa_compositeExpression.subExpressions.size() == 1;
					assert monotonic == false;
					return new TSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers));
				}
				case IMBEDDING:
				{
					assert tsa_compositeExpression.subExpressions.size() == 1;
					return new TSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers));
				}
				case PROJECTION:
				{ 
					assert tsa_compositeExpression.subExpressions.size() == 1;	
					TSA_Expression differenciatedSubExpression = this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers);
					TSA_ProjectionOperator projection = (TSA_ProjectionOperator) tsa_compositeExpression.operator;
					
					if(monotonic & type != DTSA_QueryType.INIT){
						AbstractFilter<ConditionFormula> identifierAttributeMinTimeFilter = new AbstractFilter<ConditionFormula>() {
							
							@Override
							public boolean accepts(final ConditionFormula element) {
								if(element instanceof BinaryComparissonFormula){
									BinaryComparissonFormula bcf = (BinaryComparissonFormula) element;
									if(bcf.firstTerm() instanceof MinTime){
										MinTime minTime = (MinTime) bcf.firstTerm();
										if(minTime.type.ID == IDENTIFIER){
											if(bcf.secondTerm() instanceof BasicTypeAttribute){
												BasicTypeAttribute attribute = (BasicTypeAttribute) bcf.secondTerm();
												return attribute.ID == minTime.type.ID;
											}
										}
									}
								}
								return false;
							}
							
						};
						
						final FixedArrayNavigableMap<String, MinTime> referencedIdentifierAttributes;
						{
							ConditionFormula minTimeFormula = not(and(differenciatedSubExpression.schema().relationConstraints()).focus(minTimesFilter)).focus(identifierAttributeMinTimeFilter).subformula().flaten();
							//FIXME assert minTimeFormula.ID != ConditionFormulaID.TOP;
							//FIXME assert minTimeFormula.ID != ConditionFormulaID.BOTTOM;
							/*FIXME couldn't this be if there is already a isNewAttribute that covers all inputs? 
							introduce monotonic stream bounds check domination by monotonic stream bounds, introduce isNew if last monotonic stream bound (or equal attribute) is discarded. 
							In this case BOTTOM may possibly occur. 
							Better: if stream bound is is used in isNew then replace streamBound constraint by TOP. Then TOP would be allowed. But Bottom would not be allowed and ould not occur if implementation is correct;
							*/
							AbstractCollection<ConditionFormula> minTimeInequations = minTimeFormula.collectFormulas(atomicFormulaFilter);
							
							ArrayList<String> referencedIdentifierAttributeNamesList = new ArrayList<String>();
							ArrayList<MinTime> referencedIdentifierAttributeMinTimeList = new ArrayList<MinTime>();
							
							for(ConditionFormula minTimeInequationFormula: minTimeInequations){
								BinaryComparissonFormula minTimeInequation = (BinaryComparissonFormula) minTimeInequationFormula;
								MinTime minTime = (MinTime) minTimeInequation.firstTerm();
								assert minTimeInequation.secondTerm() instanceof BasicTypeAttribute;
								BasicTypeAttribute attribute = (BasicTypeAttribute) minTimeInequation.secondTerm();
								assert attribute.ID == IDENTIFIER;
								referencedIdentifierAttributeNamesList.add(attribute.name);
								referencedIdentifierAttributeMinTimeList.add(minTime);
							}
							
							referencedIdentifierAttributes = new FixedArrayNavigableMap<String, MinTime>(referencedIdentifierAttributeNamesList.toArray(new String[0]), referencedIdentifierAttributeMinTimeList.toArray(new MinTime[0]));
						}
		
						final FixedArrayNavigableSet<String> remainingIdentifierAttributes;
						{
							FixedArrayNavigableMap<String, BasicType> attributes = tsa_compositeExpression.schema().basicAttributes;
							ArrayList<String> remainingIdAttributesList = new ArrayList<String>();
							for(int attribute = 0; attribute < attributes.size(); ++attribute){
								if(attributes.values().get(attribute).ID == IDENTIFIER){
									remainingIdAttributesList.add(attributes.keys().get(attribute));
								}
							}
							
							remainingIdentifierAttributes = new FixedArrayNavigableSet<String>(false, remainingIdAttributesList);
						}
						
						final FixedArrayNavigableMap<String, MinTime> discardedIdentifierAttributes;
						{
							//TODO more efficient by setDifference with FixedArrayNavigableSet ?
							ArrayList<String> discardedIdAttributesList = new ArrayList<String>();
							ArrayList<MinTime> correspondingMinTimesList = new ArrayList<MinTime>();
							for (int minTimeFormula = 0; minTimeFormula < referencedIdentifierAttributes.size(); ++minTimeFormula) {
								String name = referencedIdentifierAttributes.keys().get(minTimeFormula);
								if(!remainingIdentifierAttributes.contains(name)){
									discardedIdAttributesList.add(name);
									correspondingMinTimesList.add(referencedIdentifierAttributes.values().get(minTimeFormula));
								}
							}
							discardedIdentifierAttributes = new FixedArrayNavigableMap<String, MinTime>(new FixedArrayNavigableSet<String>(false, false, discardedIdAttributesList.toArray(new String[0])), new FixedArray<MinTime>(false, correspondingMinTimesList.toArray(new MinTime[0])));
						}
						 
						boolean hasIsNewAttribute = differenciatedSubExpression.schema().hasAttribute(isNewAttributeName);
						
						TSA_ProjectionOperator relaxedProjection = new TSA_ProjectionOperator(projection.origin, FixedArrayNavigableSet.merge(projection.attributes, new FixedArrayNavigableSet<String>(isNewAttributeName)));
	
						if(discardedIdentifierAttributes.size() > 0){
							
							ArithmeticAnalyser identifierAnalyser = this.identifierAnalyserFactory.make(and(differenciatedSubExpression.schema().tupleConstraints()), FixedArrayNavigableSet.merge(remainingIdentifierAttributes, discardedIdentifierAttributes.keys()), new FixedArrayNavigableSet<String>());
							
							FixedArrayNavigableMap<String, String> aliasMapping;
							final FixedArrayNavigableMap<String, MinTime> nonReplaceableDiscardedIdentifierAttributes;
							
							{
								ArrayList<String> replaceableDiscardedAttributesList = new ArrayList<String>();
								ArrayList<String> replaceableDiscardedAttributesAliasList = new ArrayList<String>();
								ArrayList<String> nonReplaceableDiscardedAttributesList = new ArrayList<String>();
								ArrayList<MinTime> nonReplaceableDiscardedAttributesMinTimeList = new ArrayList<MinTime>();
								
								for(int discardedIdentifierAttribute = 0; discardedIdentifierAttribute < discardedIdentifierAttributes.size(); ++discardedIdentifierAttribute){
									String discardedAttribute = discardedIdentifierAttributes.keys().get(discardedIdentifierAttribute);
									boolean replaceable = false;
									for(int remainingIdentifierAttribute = 0; remainingIdentifierAttribute <remainingIdentifierAttributes.size(); ++remainingIdentifierAttribute){
										String remainingAttribute = remainingIdentifierAttributes.get(remainingIdentifierAttribute);
										long distance1 = identifierAnalyser.distance(discardedAttribute, true, remainingAttribute, true);
										long distance2 = identifierAnalyser.distance(remainingAttribute, true, discardedAttribute, true);
										if(ArithmeticAnalyser.constant(distance1) == 0 & !ArithmeticAnalyser.minusEpsilon(distance1) & ArithmeticAnalyser.constant(distance2) == 0 & !ArithmeticAnalyser.minusEpsilon(distance2)){
											replaceable = true;
											replaceableDiscardedAttributesList.add(discardedAttribute);
											replaceableDiscardedAttributesAliasList.add(remainingAttribute);
											break;
										}
									}
									if(!replaceable){
										nonReplaceableDiscardedAttributesList.add(discardedAttribute);
										nonReplaceableDiscardedAttributesMinTimeList.add(discardedIdentifierAttributes.values().get(discardedIdentifierAttribute));
									}
								}
								
								aliasMapping = new FixedArrayNavigableMap<String, String>(new FixedArrayNavigableSet<String>(false, replaceableDiscardedAttributesList), new FixedArray<String>(replaceableDiscardedAttributesAliasList));
								nonReplaceableDiscardedIdentifierAttributes = new FixedArrayNavigableMap<String, MinTime>(new FixedArrayNavigableSet<String>(false, nonReplaceableDiscardedAttributesList), new FixedArray<MinTime>(nonReplaceableDiscardedAttributesMinTimeList));
							}
							
							TemporalStreamSchema modifiedSchema;
							try {
								modifiedSchema = TemporalStreamSchema.create(differenciatedSubExpression.schema().ID, differenciatedSubExpression.schema().basicAttributes, differenciatedSubExpression.schema().tupleConstraints(), and(differenciatedSubExpression.schema().relationConstraints()).replaceAtomicTerm(new RenameAttributes(aliasMapping)).subformulas());
							} catch (TSA_SchemaException e) {
								throw new TSA_Exception(projection.origin, e);
							}
							
							TSA_Expression schemaModifiedDifferentiatedSubExpression = differenciatedSubExpression.replaceSchema(modifiedSchema);
							
							if(nonReplaceableDiscardedIdentifierAttributes.size() > 0){
								ConditionFormula isNewFormula = isNewFormula(hasIsNewAttribute, nonReplaceableDiscardedIdentifierAttributes);
															
								if(hasIsNewAttribute){
									FixedArrayNavigableMap<String, String> mapping = new FixedArrayNavigableMap<String, String>(new String[]{isNewAttributeName}, new String[]{isNewAttributeAliasName});
									try {
										isNewFormula = isNewFormula.replaceAtomicTerm(new RenameAttributes(mapping));
									} catch (TSA_SchemaException e) {
										throw new TSA_Exception(projection.origin, e);
									}
									TSA_ImbeddingOperator isNewImbedding = new TSA_ImbeddingOperator(projection.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(new BooleanAttribute[]{isNewAttribute}, new Term<?>[]{isNewFormula}));
									return new TSA_CompositeExpression(relaxedProjection, new TSA_CompositeExpression(isNewImbedding, new TSA_CompositeExpression(new TSA_RenameOperator(projection.origin, mapping), schemaModifiedDifferentiatedSubExpression)));
								}
								else{
									TSA_ImbeddingOperator isNewImbedding = new TSA_ImbeddingOperator(projection.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(new BooleanAttribute[]{isNewAttribute}, new Term<?>[]{isNewFormula}));
									return new TSA_CompositeExpression(relaxedProjection, new TSA_CompositeExpression(isNewImbedding, schemaModifiedDifferentiatedSubExpression));
								}
							}
							else{
								if(hasIsNewAttribute){								
									return new TSA_CompositeExpression(relaxedProjection, schemaModifiedDifferentiatedSubExpression);
								}
								else{
									return new TSA_CompositeExpression(projection, schemaModifiedDifferentiatedSubExpression);
								}
							}		
						}
						else{
							if(hasIsNewAttribute){
								return new TSA_CompositeExpression(relaxedProjection, differenciatedSubExpression);
							}
							else{
								return new TSA_CompositeExpression(projection, differenciatedSubExpression);
							}				
						}
					}
					else{
						assert !monotonic | type == DTSA_QueryType.INIT;
						assert !differenciatedSubExpression.schema().hasAttribute(isNewAttributeName);
						return new TSA_CompositeExpression(projection, differenciatedSubExpression);
					}
					
				}
				case SELECTION:
				{
					assert tsa_compositeExpression.subExpressions.size() == 1;
					return new TSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers));
				}	
				case SETDIFFERENCE:
				{
					assert tsa_compositeExpression.subExpressions.size() == 2;
					assert monotonic == false;
					return new TSA_CompositeExpression(tsa_compositeExpression.operator, new FixedArray<TSA_Expression>(false, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers), this.differentiate(tsa_compositeExpression.subExpressions.get(1), type, monotonic, streamBuffers)));
				}
				default:
					throw new Error("Unexpected case");
				}
			}
			case COMPLEX: 
			{
				switch (((TSA_ComplexOperator)tsa_compositeExpression.operator).ID) {
				case RENAME: {
					assert tsa_compositeExpression.subExpressions.size() == 1;
					TSA_Expression differenciatedSubExpression = this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers);
					TSA_RenameOperator renameOperator = (TSA_RenameOperator) tsa_compositeExpression.operator;
					boolean isPrefix = renameOperator.mapping.keys().contains("");
					
					if(isPrefix & differenciatedSubExpression.schema().hasAttribute(isNewAttributeName)){
						FixedArrayNavigableSet<String> attributeNames = differenciatedSubExpression.schema().basicAttributes.keys();
						String[] newAttributeNames = new String[attributeNames.size()];
						String prefix = renameOperator.mapping.get("");
						
						for(int attribute = 0; attribute < newAttributeNames.length; ++attribute){
							newAttributeNames[attribute] = TSA_SyntaxPatterns.prefixAttribute(prefix, attributeNames.get(attribute));
						}
						
						FixedArrayNavigableMap<String, String> modifiedMapping = new FixedArrayNavigableMap<String, String>(attributeNames, new FixedArray<String>(false, newAttributeNames)); 
						TSA_RenameOperator modifiedRenameOperator = new TSA_RenameOperator(renameOperator.origin, modifiedMapping);
						return new TSA_CompositeExpression(modifiedRenameOperator, differenciatedSubExpression);
					}
					else{
						//rename operator is not allowed to explicitly affect any attribute starting with .tsa
						return new TSA_CompositeExpression(tsa_compositeExpression.operator, differenciatedSubExpression);
					}
				}
				case GROUPING_AND_AGGREGATION: {
					assert tsa_compositeExpression.subExpressions.size() == 1;
					assert monotonic == false;
					return new TSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers));
				}
				case DISTINCT_PROJECTION: {
					assert tsa_compositeExpression.subExpressions.size() == 1;
					assert monotonic == false;
					return new TSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers));
				}
				case ANTI_SEMI_JOIN: {
					assert tsa_compositeExpression.subExpressions.size() == 2;
					assert monotonic == false;
					return new TSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers));
				}
				case SEMI_JOIN: {
					assert tsa_compositeExpression.subExpressions.size() == 2;
					assert monotonic == false;
					return new TSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers));
				}
				default:
					throw new Error("Unexpected case");
				}
			}
			case COMPOSITE: 
			{
				//FIXME
				throw new Error("Not implemented case");
				//break;
			}
			default:
				throw new Error("Unexpected case");
			}
		}
		catch (TSA_Exception e) {
			throw e;
		}
		catch (RuntimeException e) {
			throw new TSA_Exception(tsa_compositeExpression.operator.origin, e);
		}
	}

	
	public TSA_TemporalStreamExpression differentiate(final TSA_TemporalStreamExpression tsa_temporalStreamExpression, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception{
		try{
			final DTSA_StreamBuffer dtsa_streamBuffer = streamBuffers.get(tsa_temporalStreamExpression.type).get(tsa_temporalStreamExpression.full_name);		
			
			if(dtsa_streamBuffer == null){
				throw new TSA_Exception(tsa_temporalStreamExpression.origin, "The expression refers to an non-existing stream: " + tsa_temporalStreamExpression.type + ":" + tsa_temporalStreamExpression.full_name);
			}
			
			//FIXME and why didn't the validity check find this?
			TemporalStreamSchema schema = dtsa_streamBuffer.schema;
			
			ConditionFormula streamBoundFormula = and(tsa_temporalStreamExpression.schema.relationConstraints()).focus(streamBoundFilter).flaten();
			if(streamBoundFormula.containsDisjunction()){
				throw new TSA_Exception(dtsa_streamBuffer.origin);//FIXME paper?
			}
			if(streamBoundFormula.containsNegation()){
				throw new TSA_Exception(dtsa_streamBuffer.origin, "Stream bound formulas my not contain negation.");
			}
			if(streamBoundFormula.ID == ConditionFormulaID.BOTTOM){
				throw new TSA_Exception(dtsa_streamBuffer.origin);
			}
			if(streamBoundFormula.ID == ConditionFormulaID.TOP){
				return new TSA_TemporalStreamExpression(dtsa_streamBuffer.prog_package, dtsa_streamBuffer.simple_name, dtsa_streamBuffer.type, dtsa_streamBuffer.schema, dtsa_streamBuffer.origin);
			}
			else{
			
			Adapter<AtomicConditionFormula, AtomicConditionFormula> streamBoundInequationAdapter = new Adapter<AtomicConditionFormula, AtomicConditionFormula>() {
				
				@Override
				public AtomicConditionFormula adapt(final AtomicConditionFormula source) {
					if(streamBoundFilter.accepts(source)){
						StreamBoundConstraint streamBound = (StreamBoundConstraint) source;
	
						BasicTypeAttribute attribute = (BasicTypeAttribute) streamBound.atomicTerm;
						String name = attribute.name;
						BasicType type = attribute.type;
	
						return new BinaryComparissonFormula(BinaryComparissonOperatorID.GREATER_EQUAL, new MinTime(dtsa_streamBuffer.full_name, name, type), attribute);
					}
					else{
						return source;
					}
				}
				
			};
			
			DisjunctiveConditionFormula streamBoundInequations = or(streamBoundFormula.DNFiteration().iterator().next().subformulas()).replaceAtomicFormula(streamBoundInequationAdapter);
			
			TemporalStreamSchema differentiatedSchema;
			try {
				differentiatedSchema = TemporalStreamSchema.create(schema.ID, schema.basicAttributes, schema.tupleConstraints(), and(and(schema.relationConstraints()).focus(AbstractFilter.notFilter(streamBoundFilter)), streamBoundInequations).subformulas());
			} catch (TSA_SchemaException e) {
				throw new TSA_Exception(tsa_temporalStreamExpression.origin, e);
			}
			
			return new TSA_TemporalStreamExpression(dtsa_streamBuffer.prog_package, dtsa_streamBuffer.simple_name, dtsa_streamBuffer.type, differentiatedSchema, dtsa_streamBuffer.origin);
			}
		}
		catch (TSA_Exception e) {
			throw e;
		}
		catch (RuntimeException e) {
			throw new TSA_Exception(tsa_temporalStreamExpression.origin, e);
		}
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(TSA_DefaultDifferencer.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

}
