package de.lmu.ifi.pms.cep.tsa.conditions;

import de.lmu.ifi.pms.cep.tsa.schema.AType;
import de.lmu.ifi.pms.cep.tsa.schema.NoType;


public abstract class AggregationFunction<TypeClass extends AType> extends Function<TypeClass> {
	
	@SuppressWarnings("hiding")
	public final AggregationFunctionID ID;
	
	public AggregationFunction(final AggregationFunctionID ID, final TypeClass type, final Term<?> argument) {
		this(ID, ID.toString(), type, argument);
	}
	
	AggregationFunction(final AggregationFunctionID ID, final TypeClass type, final Term<?>... arguments) {
		this(ID, ID.toString(), type, arguments);
	}
	
	AggregationFunction(final AggregationFunctionID ID, final TypeClass type, final java.util.ArrayList<Term<?>> arguments) {
		this(ID, ID.toString(), type, arguments);
	}
	
	AggregationFunction(final AggregationFunctionID ID, final String name, final TypeClass type, final Term<?>... arguments) {
		super(FunctionID.AGGREGATION, name, type, 1, 1, arguments);
		this.ID = ID;
	}
	
	AggregationFunction(final AggregationFunctionID ID, final String name, final TypeClass type, final java.util.ArrayList<Term<?>> arguments) {
		this(ID, name, type, arguments.toArray(new Term<?>[arguments.size()]));
	}
	
	public AggregationFunction(final AggregationFunctionID ID, final String name, final TypeClass type, final Term<?> argument) {
		this(ID, name, type, new Term<?>[]{argument});
	}
	
	@Override
	protected int minArity() {
		return 1;
	}
	
	@Override
	protected int maxArity() {
		return 1;
	}

	@Override
	public
	final AggregationFunction<TypeClass> newTerm(final Term<?>... arguments){
		return this.newTerm(arguments[0]);
	}
	
	protected abstract AggregationFunction<TypeClass> newTerm(final Term<?> argument);
	
	public static AggregationFunction<NoType> apply(final AggregationFunctionID ID, final Term<?> argument){
		return apply(ID, NoType.instance, argument);
	}
	
	public static <TypeClass extends AType> AggregationFunction<TypeClass> apply(final AggregationFunctionID ID, final TypeClass type, final Term<?> argument){
		switch (ID) {
		case AVG:
			return new Average<TypeClass>(type, argument);
		case COUNT:
			return new Count<TypeClass>(type, argument);
		case MAX:
			return new Max<TypeClass>(type, argument);
		case MIN:
			return new Min<TypeClass>(type, argument);
		case SUM:
			return new Sum<TypeClass>(type, argument);
		case USER:
			throw new IllegalArgumentException("Does not support user functions.");
		default:
			throw new Error("Unexpected case: " + ID.toString());
		}
	}
}
