package de.lmu.ifi.pms.cep.tsa.conditions;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.filter.Filter;
import de.lmu.ifi.pms.java.util.tupel.Single;

public abstract class AtomicConditionFormula extends ConditionFormula{

	public AtomicConditionFormula(final boolean isNormalized, final boolean isFlatened) {
		this(ConditionFormulaID.ATOMIC, isNormalized, isFlatened);
	}
	
	protected AtomicConditionFormula(final ConditionFormulaID ID, final boolean isNormalized, final boolean isFlatened) {
		super(ID, isNormalized, isFlatened);
	}
	
	@Override
	protected ConditionFormula normalizeImpl() {
		return this;
	}
	
	@Override
	protected ConditionFormula flatenImpl() {
		return this;
	}

	@Override
	public int DNFsize(){
		return 1;
	}
	
	@Override
	protected AbstractCollection<ConditionFormula> DNFiteration(final boolean positivePolarity) {
		if(positivePolarity){
			return new Single<ConditionFormula, ConditionFormula>(this);
		}
		else{
			return new Single<ConditionFormula, ConditionFormula>(new NegatedConditionFormula(this));
		}
	}

	@Override
	public DisjunctiveConditionFormula DNF() {
		return new DisjunctiveConditionFormula(new ConjunctiveConditionFormula(this));
	}
	
	@Override
	public AbstractCollection<ConjunctiveConditionFormula> DNFiteration() {
		return new Single<ConjunctiveConditionFormula,ConjunctiveConditionFormula>(new ConjunctiveConditionFormula(this));
	}
	
	@Override
	public int CNFsize() {
		return 1;
	}
	
	@Override
	protected AbstractCollection<ConditionFormula> CNFiteration(final boolean positivePolarity) {
		if(positivePolarity){
			return new Single<ConditionFormula, ConditionFormula>(this);
		}
		else{
			return new Single<ConditionFormula, ConditionFormula>(new NegatedConditionFormula(this));
		}
	}

	@Override
	public ConjunctiveConditionFormula CNF() {
		return new ConjunctiveConditionFormula(new DisjunctiveConditionFormula(this));
	}
	
	@Override
	public AbstractCollection<DisjunctiveConditionFormula> CNFiteration() {
		return new Single<DisjunctiveConditionFormula,DisjunctiveConditionFormula>(new DisjunctiveConditionFormula(this));
	}
	
	@Override
	public boolean containsNegation() {
		return false;
	}
	@Override
	public boolean containsConjunction() {
		return false;
	}
	@Override
	public boolean containsDisjunction() {
		return false;
	}
	
	@Override
	protected ConditionFormula focus(final Filter<? super AtomicConditionFormula> topic, final boolean positivePolarity) {
		if(topic.accepts(this)){
			return this;
		}
		else{
			if(positivePolarity){
				return ConditionFormula.Top;
			}
			else{
				return ConditionFormula.Bottom;
			}
		}
	}
	
	@Override
	public ConditionFormula replaceAtomicFormula(final Adapter<AtomicConditionFormula, ? extends ConditionFormula> replacement) {
		return replacement.adapt(this);
	}
	
	@Override
	public abstract ConditionFormula replaceAtomicTerm(Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement);
	
	@Override
	public abstract ConditionFormula replaceTerm(Adapter<Term<?>, ? extends Term<?>> replacement);
	
	protected int compareToImpl(final AtomicConditionFormula o) {
		return this.getClass().getName().compareTo(o.getClass().getName());
	}
	
	@Override
	protected boolean equalsImpl(final Object o) {
		if(this.getClass().isInstance(o)){
			return this.compareToImpl(((AtomicConditionFormula)o)) == 0;
		}
		else{
			return false;
		}
	}
	
	@Override
	public int hashCode() {
		return this.getClass().hashCode();
	}
	
}
