package de.lmu.ifi.pms.cep.tsa.conditions;

import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

public abstract class BasicTypeAttribute extends TypedAttribute<BasicType> {
	
	@SuppressWarnings("hiding")
	public final BasicTypeID ID; 
	
	protected BasicTypeAttribute(final String name, final BasicTypeID typeID) {
		super(name, typeID.type);
		this.ID = typeID;
	}
	
	public static BasicTypeAttribute create(final String name, final BasicType type){
		return create(name, type.ID);
	}
	
	
	
	public static BasicTypeAttribute create(final String name, final BasicTypeID typeID){
		switch (typeID) {
		case TIMESTAMP:
			return new TimestampAttribute(name);
		case DURATION:
			return new DurationAttribute(name);
		case BOOLEAN:
			return new BooleanAttribute(name);
		case INT:
			return new IntAttribute(name);
		case LONG:
			return new LongAttribute(name);
		case FLOAT:
			return new FloatAttribute(name);
		case DOUBLE:
			return new DoubleAttribute(name);
		case STRING:
			return new StringAttribute(name);
		case IDENTIFIER:
			return new IdentifierAttribute(name);
		case REFERENCE:
			return new ReferenceAttribute(name);
		}
		throw new Error();
		//TODO better exception
	}

	@Override
	public BasicTypeAttribute rename(final String newName) {
		return BasicTypeAttribute.create(newName, this.ID);
	}
}
