package de.lmu.ifi.pms.cep.tsa.conditions;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;

import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;



public class ClockFunction extends Function<BasicType> {
	
	@Attribute(name="name")
	String getName(){
		return this.name;
	}
	
	ClockFunction(@Attribute(name="name") final String name, @Element(name="type", required=false)final BasicType type, @ElementList(entry="term", inline=true, required=false, empty=false) final java.util.ArrayList<Term<?>> arguments) {
		this(name, type);
		if(arguments != null && arguments.size() > 0){
			throw new IllegalArgumentException();//FIXME better exception
		}
	}
	
	public ClockFunction(final String name, final BasicTypeID typeID) {
		this(name, typeID.type);
	}
	
	public ClockFunction(final String name, final BasicType type) {
		super(FunctionID.CLOCK, name, type, 0, 0, new Term[0]);
		if(!TSA_SyntaxPatterns.isValidClockName(name)){
			throw new IllegalArgumentException("Invalid clock name: " + name);
		}
		//TODO allow only IDENTIFIER and TIMESTAMP?
	}
	
	@Override
	protected int minArity() {
		return 0;
	}
	
	@Override
	protected int maxArity() {
		return 0;
	}

	@Override
	public ClockFunction newTerm(final Term<?>... arguments) {
		return new ClockFunction(this.name, this.type);
	}
	
	@Override
	public String toString() {
		return super.toString() + " : " + this.type.toString();
	}
}
