package de.lmu.ifi.pms.cep.tsa.conditions;

import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID.*;

import org.simpleframework.xml.ElementList;

import de.lmu.ifi.pms.java.lang.iteration.ComposedIterator;
import de.lmu.ifi.pms.java.lang.iteration.FixedCrossProductArrayIterator;
import de.lmu.ifi.pms.java.lang.iteration.Iterator;
import de.lmu.ifi.pms.java.util.adaption.FixedIteratorAdapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;


public class ConjunctiveConditionFormula extends CompositeConditionFormula<ConjunctiveConditionFormula> {
	
	public ConjunctiveConditionFormula(
			@ElementList(entry="condition",inline=true,required=false,empty=false)//FIXME inline
			final java.util.ArrayList<ConditionFormula> subformulas) {
		this((subformulas == null ? new ConditionFormula[0] : subformulas.toArray(new ConditionFormula[subformulas.size()])));
	}
	
	public ConjunctiveConditionFormula(final java.util.Collection<? extends ConditionFormula> subformulas) {
		this(subformulas.toArray(new ConditionFormula[subformulas.size()]));
	}
	
	public ConjunctiveConditionFormula(final Collection<? extends ConditionFormula> subformulas) {
		this(subformulas.toArray(new ConditionFormula[subformulas.size()]));
	}

	public ConjunctiveConditionFormula(final ConditionFormula... subformulas) {
		this(initialIsNormalized(subformulas), subformulas);
	}
	
	ConjunctiveConditionFormula(final boolean isNormalized, final ConditionFormula... subformulas) {
		this(isNormalized, initialIsFlatened(isNormalized, subformulas), subformulas);
	}
	
	ConjunctiveConditionFormula(final boolean isNormalized, final boolean isFlatened, final ConditionFormula... subformulas) {
		super(ConditionFormulaID.CONJUNCTIVE, isNormalized, isFlatened, subformulas);
	}
	
	protected static boolean initialIsNormalized(final ConditionFormula... subformulas){
		if(subformulas.length < 2){
			return false;
		}
		else{
			for(ConditionFormula subformula: subformulas){
				if(subformula.ID == TOP | subformula.ID == BOTTOM | !subformula.isNormalized){
					return false;
				}
			}
			return true;
		}
	}
	
	
	protected static boolean initialIsFlatened(final boolean isNormalized, final ConditionFormula... subformulas){
		if(isNormalized){
			for(ConditionFormula subformula: subformulas){
				if(!subformula.isFlatened || subformula.ID == CONJUNCTIVE){
					return false;
				}
			}
			return true;
		}
		else{
			return false;
		}
	}
	
	@Override
	protected ConditionFormula normalizeImpl() {
		ArrayList<ConditionFormula> new_subformulas = new ArrayList<ConditionFormula>(this.subformulas.length);
		for(ConditionFormula subformula: this.subformulas){
			ConditionFormula normalizedSubFormula = subformula.normalize();			
			if(normalizedSubFormula.ID == TOP){
				
			}
			else if(normalizedSubFormula.ID == BOTTOM){
				return Bottom;
			}
			else{
				new_subformulas.add(normalizedSubFormula);
			}
		}
		if(new_subformulas.size() == 0){
			return Top;
		}		
		else if(new_subformulas.size() == 1){
			return new_subformulas.first();
		}
		else{
			return new ConjunctiveConditionFormula(true, new_subformulas.toArray(new ConditionFormula[new_subformulas.size()]));
		}
	}
	
	@Override
	protected ConditionFormula flatenImpl() {
		assert this.isNormalized;
		assert !this.isFlatened;
		
		ArrayList<ConditionFormula> new_subformulas = new ArrayList<ConditionFormula>(this.subformulas.length);
		this.flatenImpl(new_subformulas);
		
		return new ConjunctiveConditionFormula(true, true, new_subformulas.toArray(new ConditionFormula[new_subformulas.size()]));
	}
	
	protected void flatenImpl(final ArrayList<ConditionFormula> new_subformulas){
		for(ConditionFormula c: this.subformulas){				
			if(c.ID == CONJUNCTIVE){
				((ConjunctiveConditionFormula)c).flatenImpl(new_subformulas);
			}
			else{
				ConditionFormula flatened = c.flaten();
				assert flatened.ID != TOP;
				assert flatened.ID != BOTTOM;
				
				new_subformulas.add(flatened);		
			}
		}
	}

	@Override
	public int DNFsize() {
		int size = 1;
		for(ConditionFormula c: this.subformulas){
			size *= c.DNFsize();
		}
		return size;
	}

	@Override
	protected AbstractCollection<ConditionFormula> DNFiteration(final boolean positivePolarity) {
		@SuppressWarnings("unchecked")
		final AbstractCollection<ConditionFormula>[] iterations = new AbstractCollection[this.subformulas.length];
		int size = 0;
		for(int i = 0; i < iterations.length; ++i){
			iterations[i] = this.subformulas[i].DNFiteration(positivePolarity);
			size *= iterations[i].size();
		}
		final int fixedSize = size;
		return new AbstractCollection<ConditionFormula>() {
			//FIXME also use an iterative representation for the Conjunctions (Disjunctions), i.e. no arrrays, e.g. AbstractCollection<AbstractCollection<ConditionFormula>>
			@Override
			public Iterator<ConditionFormula> iterator() {
				return new FixedIteratorAdapter<ConditionFormula[], ConditionFormula>(new FixedCrossProductArrayIterator<ConditionFormula>(ConditionFormula.class, iterations)) {

					@Override
					public ConditionFormula adapt(final ConditionFormula[] subformulas) {
						if(positivePolarity){
							return new ConjunctiveConditionFormula(subformulas);
						}
						else{
							return new DisjunctiveConditionFormula(subformulas);
						}
					}
					
				}; 
			}

			@Override
			public int size() {
				return fixedSize;
			}
		};
	}
	
	@Override
	public int CNFsize() {
		int size = 0;
		for(ConditionFormula c: this.subformulas){
			size += c.DNFsize();
		}
		return size;
	}

	@Override
	protected AbstractCollection<ConditionFormula> CNFiteration(final boolean positivePolarity) {
		@SuppressWarnings("unchecked")
		final AbstractCollection<ConditionFormula>[] iterations = new AbstractCollection[this.subformulas.length];
		int size = 0;
		for(int i = 0; i < iterations.length; ++i){
			iterations[i] = this.subformulas[i].CNFiteration(positivePolarity);
			size += iterations[i].size();
		}
		final int fixedSize = size;
		return new AbstractCollection<ConditionFormula>() {

			@Override
			public Iterator<ConditionFormula> iterator() {
				return new ComposedIterator<ConditionFormula>(iterations);
			}

			@Override
			public int size() {
				return fixedSize;
			}
		};
	}
	
	@Override
	public boolean containsNegation() {
		for(ConditionFormula subFormula: this.subformulas){
			if(subFormula.containsNegation()){
				return true;
			}
		}
		return false;
	}
	@Override
	public boolean containsConjunction() {
		return true;
	}
	@Override
	public boolean containsDisjunction() {
		for(ConditionFormula subFormula: this.subformulas){
			if(subFormula.containsDisjunction()){
				return true;
			}
		}
		return false;
	}
	
	@Override
	protected ConjunctiveConditionFormula newFormula(final ConditionFormula[] subformulas) {
		return new ConjunctiveConditionFormula(subformulas);
	}

}
