package de.lmu.ifi.pms.cep.tsa.conditions;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;


@Root(name="min-time")
public class MinTime extends AtomicTerm<BasicType> {
	
	@Attribute(name="buffer")
	public final String buffer_name;
	@Attribute(name="name")
	public final String name;
	public final String qualified_name;
	
	@Element(name="type", required=true)
	protected final BasicType getType(){
		return this.type;
	}
	
	public MinTime(final String buffer_name, final String name, final BasicTypeID typeID) {
		this(buffer_name, name, typeID.type);
	}

	public MinTime(@Attribute(name="buffer") final String buffer_name, @Attribute(name="name") final String name, @Element(name="type", required=true) final BasicType type) {
		super(AtomicTermID.MIN_TIME, type);
		//FIXME
		if(!TSA_SyntaxPatterns.isValidBufferFullName(buffer_name)){
			throw new IllegalArgumentException("Invalid buffer name: " + buffer_name);
		}
		if(!TSA_SyntaxPatterns.MinTimeNamePattern.matcher(name).matches()){
			throw new IllegalArgumentException("Invalid min-time name: " + name);
		}
		this.buffer_name = buffer_name;
		this.name = name;
		this.qualified_name = TSA_SyntaxPatterns.minTimeQualifiedName(buffer_name, name);
		//TODO allow only IDENTIFIER and TIMESTAMP?
	}
	
	public MinTime rename(final String newName){
		return new MinTime(this.buffer_name, newName, this.type);
	}
	
	@Override
	protected int compareToImpl(final Term<?> o) {
		int cmp;
		if(o instanceof MinTime){
			cmp = this.name.compareTo(((MinTime)o).name);
			if(cmp == 0){
				cmp = super.compareToImpl(o);
			}
		}
		else{
			cmp = super.compareToImpl(o);
			assert cmp != 0;
		}
		return cmp;
	}
	
	@Override
	protected boolean equalsImpl(final Object o) {
		if(this.getClass() == o.getClass()){
			return this.name.equals(((MinTime)o).name);
		}
		else{
			return false;
		}
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.name.hashCode();
	}

	
	@Override
	public String toString() {
		return this.qualified_name;
	}
}
