package de.lmu.ifi.pms.cep.tsa.conditions;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID.*;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;

public class NegatedConditionFormula extends ConditionFormula {

	protected final ConditionFormula subformula;

	public NegatedConditionFormula(final ConditionFormula subformula) {
		this(initialIsNormalized(subformula), subformula);
	}
	
	NegatedConditionFormula(final boolean isNormalized, final ConditionFormula subformula) {
		this(isNormalized, initialIsFlatened(isNormalized, subformula), subformula);		
	}
	
	NegatedConditionFormula(final boolean isNormalized, final boolean isFlatened, final ConditionFormula subformula) {
		super(NEGATED, isNormalized, isFlatened);
		assert !(isNormalized & !isFlatened & subformula.isFlatened);
		this.subformula = subformula;
	}
	
	protected static boolean initialIsNormalized(final ConditionFormula subformula){
		return subformula.isNormalized & !(subformula.ID == NEGATED | subformula.ID == TOP | subformula.ID == BOTTOM);
	}
	
	protected static boolean initialIsFlatened(final boolean isNormalized, final ConditionFormula subformula){
		return isNormalized & subformula.isFlatened;
	}

	public ConditionFormula subformula(){
		return this.subformula;
	}
	
	@Override
	protected ConditionFormula normalizeImpl() {
		assert !this.isNormalized;
		switch (this.subformula.ID) {
		case TOP: return Bottom;
		case BOTTOM: return Top;
		case NEGATED: return ((NegatedConditionFormula)this.subformula).subformula.normalize();
		case ATOMIC:
		case CONJUNCTIVE:
		case DISJUNCTIVE:
			assert !this.subformula.isNormalized;
			ConditionFormula normalizedSubFormula = this.subformula.normalizeImpl();
			switch (normalizedSubFormula.ID) {
			case ATOMIC:
				return new NegatedConditionFormula(true, normalizedSubFormula);
			case BOTTOM:
				return ConditionFormula.Top;
			case CONJUNCTIVE:
				return new NegatedConditionFormula(true, normalizedSubFormula);
			case DISJUNCTIVE:
				return new NegatedConditionFormula(true, normalizedSubFormula);
			case NEGATED:
				return ((NegatedConditionFormula)normalizedSubFormula).subformula;
			case TOP:
				return ConditionFormula.Bottom;
			}
			//	default:
			//		break;
		}
		throw new Error("Forgotten Case");
	}

	@Override
	protected ConditionFormula flatenImpl() {
		assert this.isNormalized;
		assert !this.isFlatened;
		assert this.subformula.ID != NEGATED;
		assert this.subformula.isNormalized;
		assert !this.subformula.isFlatened;
		return new NegatedConditionFormula(true, true, this.subformula.flatenImpl());
	}

	@Override
	public int DNFsize() {
		return this.subformula.CNFsize();
	}

	@Override
	protected AbstractCollection<ConditionFormula> DNFiteration(final boolean positivePolarity) {
		return this.subformula.CNFiteration(!positivePolarity);
	}

	@Override
	public int CNFsize() {
		return this.subformula.DNFsize();
	}

	@Override
	protected AbstractCollection<ConditionFormula> CNFiteration(final boolean positivePolarity) {
		return this.subformula.DNFiteration(!positivePolarity);
	}
	
	@Override
	public boolean containsNegation() {
		return true;
	}
	@Override
	public boolean containsConjunction() {
		return this.subformula.containsConjunction();
	}
	@Override
	public boolean containsDisjunction() {
		return this.subformula.containsDisjunction();
	}
	
	@Override
	public NegatedConditionFormula focus(final Filter<? super AtomicConditionFormula> topic){
		return this.focus(topic, true);
	}
	
	@Override
	protected NegatedConditionFormula focus(final Filter<? super AtomicConditionFormula> topic, final boolean positivePolarity) {
		ConditionFormula subformula = this.subformula.focus(topic, !positivePolarity);
		if(subformula == this.subformula){
			return this;
		}
		else{
			return new NegatedConditionFormula(subformula);
		}
	}

	@Override
	public NegatedConditionFormula replaceAtomicFormula(final Adapter<AtomicConditionFormula, ? extends ConditionFormula> replacement) {
		ConditionFormula subformula = this.subformula.replaceAtomicFormula(replacement);
		if(subformula == this.subformula){
			return this;
		}
		else{
			return new NegatedConditionFormula(subformula);
		}
	}
	
	@Override
	public NegatedConditionFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		ConditionFormula subformula = this.subformula.replaceAtomicTerm(replacement);
		if(subformula == this.subformula){
			return this;
		}
		else{
			return new NegatedConditionFormula(subformula);
		}
	}
	
	@Override
	public NegatedConditionFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		ConditionFormula subformula = this.subformula.replaceTerm(replacement);
		if(subformula == this.subformula){
			return this;
		}
		else{
			return new NegatedConditionFormula(subformula);
		}
	}
	
	@Override
	protected void collectFormulas(final Filter<ConditionFormula> filter, final ArrayList<ConditionFormula> formulas) {
		super.collectFormulas(filter, formulas);
		this.subformula.collectFormulas(filter, formulas);
	}
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		this.subformula.collect(filter, terms);
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append(this.ID);
		sb.append("{\n");
		this.subformula.writeTo(sb, indentDepth + 1);
		sb.append("\n");
		indentTab(sb, indentDepth);
		sb.append("}");
	}

}
