package de.lmu.ifi.pms.cep.tsa.conditions;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.core.Replace;
import org.simpleframework.xml.core.Resolve;

import de.lmu.ifi.pms.cep.tsa.schema.AType;
import de.lmu.ifi.pms.cep.tsa.schema.NoType;
import de.lmu.ifi.pms.cep.tsa.schema.Type;


public class Variable<TypeClass extends AType> extends AtomicTerm<TypeClass>{
	
	@Attribute(name="name")
	public final String name;
	
	@Element(name="type", required=false)
	protected final TypeClass getType(){
		if(this.type == NoType.instance){
			return null;
		}
		else{
			assert this.type instanceof Type;
			return this.type;
		}
	}
	
	@SuppressWarnings("unused")
	@Resolve
	private Variable<?> resolve(){
		if(this.type == null){
			return new UntypedVariable(this.name);
		}
		else{
			assert this.type != NoType.instance;
			return TypedVariable.create(this.name, (Type)this.type);
		}
	}
	
	@Replace
	Variable<TypeClass> replace(){
		return new Variable<TypeClass>(this.name, this.type);
	}

	public Variable(@Attribute(name="name") final String name, @Element(name="type", required=false)final TypeClass type) {
		super(AtomicTermID.VARIABLE, type);
		this.name = name;
	}
	
	public Variable<TypeClass> rename(final String newName){
		throw new AbstractMethodError();
	}
	
	@Override
	protected int compareToImpl(final Term<?> o) {
		int cmp;
		if(o instanceof Variable<?>){
			cmp = this.name.compareTo(((Variable<?>)o).name);
			if(cmp == 0){
				cmp = super.compareToImpl(o);
			}
		}
		else{
			cmp = super.compareToImpl(o);
			assert cmp != 0;
		}
		return cmp;
	}
	
	@Override
	protected boolean equalsImpl(final Object o) {
		if(this.getClass().isInstance(o)){
			return this.name.equals(((Variable<?>)o).name);
		}
		else{
			return false;
		}
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.name.hashCode();
	}
	
	@Override
	public String toString() {
		return "var " + this.name;
	}

}
