package de.lmu.ifi.pms.cep.tsa.operators;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;
import static de.lmu.ifi.pms.cep.tsa.conditions.Term.attr;

import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementMap;

import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.NamedTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.Record;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.schema.CompositeType;
import de.lmu.ifi.pms.cep.tsa.schema.NamedType;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.cep.tsa.schema.TypeID;

import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public class TSA_ImbeddingOperator extends TSA_BasicOperator
{
	public final FixedArrayNavigableMap<? extends TypedAttribute<?>, ? extends Term<?>> definitions;
	
	//Only BasicType attributes or NamesType<BasicType> attributes
	public final FixedArrayNavigableMap<? extends TypedAttribute<?>, Term<?>> simpleDefinitions;
	
	@ElementMap(entry="definition", key="attribute", value="value", inline=true)
	java.util.HashMap<TypedAttribute<?>, Term<?>> getDefinitions(){
		return new java.util.HashMap<TypedAttribute<?>, Term<?>>(this.definitions.javaMap());
	}
	
	public final FixedArray<ConditionFormula> derivedConstraints;
	
	public TSA_ImbeddingOperator(final String origin, final FixedArrayNavigableMap<? extends TypedAttribute<?>, ? extends Term<?>> definitionsOfNewAttributes) throws TSA_Exception{
		this(new FixedArray<String>(false, origin), definitionsOfNewAttributes);
	}
	
	@SuppressWarnings("unchecked")
	public TSA_ImbeddingOperator(@ElementList(entry="origin", inline=true) final java.util.ArrayList<String> origin
			, @ElementMap(entry="definition", key="attribute", value="value", inline=true, keyType=TypedAttribute.class, valueType=Term.class) final java.util.HashMap<? extends TypedAttribute<?>, ? extends Term<?>> definitionsOfNewAttributes
		) throws TSA_Exception {
		this(new FixedArray<String>(origin), new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>((java.util.Map<TypedAttribute<?>, Term<?>>)definitionsOfNewAttributes));
	}
	
	public TSA_ImbeddingOperator(final FixedCollection<String> origin, final FixedArrayNavigableMap<? extends TypedAttribute<?>, ? extends Term<?>> definitionsOfNewAttributes) throws TSA_Exception
	{
		super(TSA_BasicOperatorID.IMBEDDING, origin);
		//FIXME check attributes and definitions
		this.definitions = definitionsOfNewAttributes;
		ArrayList<TypedAttribute<?>> attributes = new ArrayList<TypedAttribute<?>>();
		ArrayList<Term<?>> simpleDefinitions = new ArrayList<Term<?>>();
		final int size = this.definitions.size();
		for(int newAttribute = 0; newAttribute < size; ++newAttribute){
			if(definitionsOfNewAttributes.keys().get(newAttribute).type.equals(definitionsOfNewAttributes.values().get(newAttribute).type)){
				simpleDefinitions(definitionsOfNewAttributes.keys().get(newAttribute), definitionsOfNewAttributes.values().get(newAttribute), attributes, simpleDefinitions);
			}
			else{
				throw new TSA_Exception(origin, "Type of attribute %s and the definition do not correspond.", definitionsOfNewAttributes.keys().get(newAttribute));
				//simpleDefinitions(definitionsOfNewAttributes.keys().get(newAttribute), definitionsOfNewAttributes.values().get(newAttribute), attributes, simpleDefinitions);
			}
		}
		
		this.simpleDefinitions = new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(attributes, simpleDefinitions);
		
		ConditionFormula[] derivedConstraints = new ConditionFormula[this.definitions.size()];
		for(int newAttribute = 0; newAttribute < derivedConstraints.length; ++newAttribute){
			derivedConstraints[newAttribute] = new BinaryComparissonFormula(BinaryComparissonOperatorID.EQUAL, definitionsOfNewAttributes.keys().get(newAttribute), definitionsOfNewAttributes.values().get(newAttribute));
		}
		
		this.derivedConstraints = new FixedArray<ConditionFormula>(false, derivedConstraints);
	}
	
	public static void simpleDefinitions(final TypedAttribute<?> attribute, final Term<?> definition, final ArrayList<TypedAttribute<?>> attributes, final ArrayList<Term<?>> simpleDefinitions){
		//FIXME activate assert attribute.type.equals(definition.type);
		if(attribute instanceof BasicTypeAttribute || (attribute instanceof NamedTypeAttribute<?> && ((NamedTypeAttribute<?>)attribute).type.ID == TypeID.BASIC)){
			attributes.add(attribute);
			simpleDefinitions.add(definition);
		}
		else {
			simpleDefinitions(attribute.name, attribute.type, definition.replaceAtomicTerm(TemporalStreamSchema.complexAttributeEliminator), attributes, simpleDefinitions);
		}
	}
	
	public static void simpleDefinitions(final String name, final Type type, final Term<?> definition, final ArrayList<TypedAttribute<?>> attributes, final ArrayList<Term<?>> simpleDefinitions){
		//FIXME activate assert type.equals(definition.type);
		if(type.ID == TypeID.BASIC || (type.ID == TypeID.NAMED && ((NamedType<?>)type).type.ID == TypeID.BASIC)){
			attributes.add(attr(name, type));
			simpleDefinitions.add(definition);
		}
		else{assert type.ID == TypeID.COMPOSITE || (type.ID == TypeID.NAMED && ((NamedType<?>)type).type.ID == TypeID.COMPOSITE);
			if(definition instanceof Record<?>){
				CompositeType compositeType = (CompositeType)(type.ID == TypeID.COMPOSITE ? type : ((NamedType<?>)type).type);
				Record<?> recordDefinition = (Record<?>) definition;
				assert FixedArrayNavigableSet.symmetricDifference(compositeType.attributes.keys(), recordDefinition.arguments.keys()).size() == 0;
				int size = compositeType.attributes.size();
				for(int arg = 0; arg < size; ++arg){
					simpleDefinitions(TSA_SyntaxPatterns.prefixAttribute(name, compositeType.attributes.keys().get(arg)), compositeType.attributes.values().get(arg), recordDefinition.arg(arg),attributes, simpleDefinitions);
				}
			}
			else{
				attributes.add(attr(name, type));
				simpleDefinitions.add(definition);
			}
		}
	}
	
	@Override
	public int inputArity() {
		return 1;
	}

	@Override
	public TemporalStreamSchema apply(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		this.checkInputs(inputs);
		assert inputs.size() == 1;
		
		TemporalStreamSchema input = inputs.iterator().next();
		
		FixedArrayNavigableSet<String> newAttributeNames = new FixedArrayNavigableSet<String>(false, false, new CollectionAdapter<TypedAttribute<?>, String>(this.definitions.keys()){
			@Override
			public String adapt(final TypedAttribute<?> source) {
				return source.name;
			}
		}.toArray(new String[this.definitions.size()]));//Attributes sorted by name
		
		Type[] newAttributeTypes = new Type[newAttributeNames.size()];
		for(int attribute = 0; attribute < newAttributeTypes.length; ++attribute){
			newAttributeTypes[newAttributeNames.indexOf(this.definitions.keys().get(attribute).name)] = this.definitions.keys().get(attribute).type;
		}
		
		FixedArrayNavigableMap<String, Type> newAttributes = new FixedArrayNavigableMap<String, Type>(newAttributeNames, new FixedArray<Type>(false, newAttributeTypes));
		
		try {
			return input.extend(newAttributes).restrict(this.derivedConstraints);
		} catch (TSA_SchemaException e) {
			throw new TSA_Exception(this.origin, e);
		}
	}
	
	
	@Override
	public boolean isMonotonic(){
		return true;
	}
	
	@Override
	public void checkInputs(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		super.checkInputs(inputs);
		//FIXME definitions do not use undefined attributes, do not contain variables 
	}
	
	@Override
	public void check(final Collection<TemporalStreamSchema> inputs) {
		//FIXME implement
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth) {
		indentTab(sb, indentDepthFirstLine);
		sb.append(this.ID.toString());
		if(this.derivedConstraints.size() > 0){
			sb.append("[");
			for(ConditionFormula derivedConstraint: this.derivedConstraints){
				derivedConstraint.writeTo(sb, 0);
				sb.append(", ");
			}
			sb.setLength(sb.length() - ", ".length());
			sb.append("]");
		}
		else{
			sb.append("[]");
		}
	}
	
}
