package de.lmu.ifi.pms.cep.tsa.operators;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementMap;

import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;

import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public class TSA_RenameOperator extends TSA_ComplexOperator
{
	public final FixedArrayNavigableMap<String, String> mapping;
	
	@ElementMap(entry="map", key="from", value="to", inline=true)
	java.util.HashMap<String, String> getMapping(){
		return new java.util.HashMap<String,String>(this.mapping.javaMap());
	}
	
	public TSA_RenameOperator(final String origin, final FixedArrayNavigableMap<String, String> mapping) throws TSA_Exception {
		this(new FixedArray<String>(false, origin), mapping);
	}
	
	protected static FixedArrayNavigableMap<String, String> resolveMapping(final java.util.HashMap<String, String> mapping){
		if(mapping == null){
			return new FixedArrayNavigableMap<String, String>();
		}
		else{
			if(mapping.containsKey(null)){
				mapping.put("", mapping.get(null));
				mapping.remove(null);
			}
			return new FixedArrayNavigableMap<String, String>(mapping);
		}
	}
	
	public TSA_RenameOperator(@ElementList(entry="origin", inline=true) final java.util.ArrayList<String> origin
			, @ElementMap(entry="map", key="from", value="to", inline=true) final java.util.HashMap<String, String> mapping) throws TSA_Exception {
		this(new FixedArray<String>(origin), resolveMapping(mapping));
	}
	
	public TSA_RenameOperator(final FixedCollection<String> origin, final FixedArrayNavigableMap<String, String> mapping) throws TSA_Exception {
		super(TSA_ComplexOperatorID.RENAME, origin);
		try {
			TemporalStreamSchema.checkNameMappingStatic(mapping);
		} catch (TSA_SchemaException e) {
			throw new TSA_Exception(origin, e);
		}
		
		this.mapping = mapping;
	
	}
	
	@Override
	public int inputArity() {
		return 1;
	}

	@Override
	public TemporalStreamSchema apply(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		this.checkInputs(inputs);
		assert inputs.size() == 1;
		TemporalStreamSchema input = inputs.iterator().next();

		try {
			return input.rename(this.mapping);
		} catch (TSA_SchemaException e) {
			throw new TSA_Exception(this.origin, e);
		}		
	}
		
	@Override
	public boolean isMonotonic(){
		return true;
	}
	
	@Override
	public void checkInputs(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		super.checkInputs(inputs);
		//FIXME implement
	}
	
	@Override
	public void check(final Collection<TemporalStreamSchema> inputs) {
		//FIXME implement
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth) {
		indentTab(sb, indentDepthFirstLine);
		sb.append(this.ID.toString());
		if(this.mapping.size() > 0){
			sb.append("[");
			for(int entry = 0; entry < this.mapping.size(); ++entry){
				sb.append(this.mapping.keys().get(entry));				
				sb.append(" -> ");
				sb.append(this.mapping.values().get(entry));
				sb.append(", ");
			}
			sb.setLength(sb.length() - ", ".length());
			sb.append("]");
		}
		else{
			sb.append("[]");
		}
	}
	
}
