package de.lmu.ifi.pms.cep.tsa.operators;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;

import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;

import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;


public class TSA_SelectionOperator extends TSA_BasicOperator
{	
	public final FixedArray<ConditionFormula> conditions;
	
	@Element(name="conditions")
	ConjunctiveConditionFormula getConditionFormula(){
		return new ConjunctiveConditionFormula(this.conditions);
	}
	
	public TSA_SelectionOperator(final String origin, final ConditionFormula... conditions){
		this(new FixedArray<String>(origin), conditions);
	}
	
	public TSA_SelectionOperator(@ElementList(entry="origin", inline=true) final java.util.ArrayList<String> origin
			,@Element(name="conditions") final ConjunctiveConditionFormula conditions){
		this(new FixedArray<String>(origin), conditions.subformulas());
	}
	
	public TSA_SelectionOperator(final FixedCollection<String> origin, final ConditionFormula... conditions){
		this(origin, new FixedArray<ConditionFormula>(conditions));
	}
	
	public TSA_SelectionOperator(final String origin, final FixedArray<ConditionFormula> conditions){
		this(new FixedArray<String>(origin), conditions);
	}
	
	public TSA_SelectionOperator(final FixedCollection<String> origin, final FixedArray<ConditionFormula> conditions)
	{
		super(TSA_BasicOperatorID.SELECTION, origin);

		this.conditions = conditions;
	}
	
	@Override
	public int inputArity() {
		return 1;
	}

	@Override
	public TemporalStreamSchema apply(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		this.checkInputs(inputs);
		assert inputs.size() == 1;
		TemporalStreamSchema input = inputs.iterator().next();
		try {
			return input.restrict(this.conditions);
		} catch (TSA_SchemaException e) {
			throw new TSA_Exception(this.origin, e);
		}
	}
	
	
	@Override
	public boolean isMonotonic(){
		return true;
	}
	
	@Override
	public void checkInputs(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {		
		super.checkInputs(inputs);
	}
	
	@Override
	public void check(final Collection<TemporalStreamSchema> inputs) {
		//FIXME implement
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth) {
		indentTab(sb, indentDepthFirstLine);
		sb.append(this.ID.toString());
		if(this.conditions.size() > 0){
			sb.append("[\n");
			for(ConditionFormula condition: this.conditions){
				condition.writeTo(sb, indentDepth + 1);
				sb.append(",\n");
			}
			sb.setLength(sb.length() - ", ".length());
			sb.append("\n");
			indentTab(sb, indentDepth);
			sb.append("]");
		}
		else{
			sb.append("[]");
		}
	}
	
}
