package de.lmu.ifi.pms.cep.tsa.schema;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementMap;

import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.map.Map;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class ActionStreamSchema extends TemporalStreamSchema {
	
	public ActionStreamSchema(final FixedArrayNavigableMap<String, ? extends Type> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException {
		super(TemporalStreamSchemaTypeID.ACTION, attributes, tupleConstraints, relationConstraints, addTypeTupleConstraints);
	}

	public ActionStreamSchema(final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints) throws TSA_SchemaException {
		this(new FixedArrayNavigableMap<String, Type>(attributes), tupleConstraints, relationConstraints, true);
	}

	protected ActionStreamSchema(
			  @ElementMap(name="attributes") final java.util.HashMap<String, ? extends Type> attributes
			, @Element(name="tuple-constraints") final ConjunctiveConditionFormula tupleConstraints
			, @Element(name="relation-constraints") final ConjunctiveConditionFormula relationConstraints
		) throws TSA_SchemaException {
		this(new FixedArrayNavigableMap<String, Type>(attributes), tupleConstraints.subformulas() , relationConstraints.subformulas(), false);
	}
	
	protected ActionStreamSchema(final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException {
		this(new FixedArrayNavigableMap<String, Type>(attributes), tupleConstraints, relationConstraints, addTypeTupleConstraints);
	}
	
	@Override
	public ActionStreamSchema newStreamSchema(final FixedArrayNavigableMap<String, ? extends Type> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException{
		return new ActionStreamSchema(attributes, tupleConstraints, relationConstraints, addTypeTupleConstraints);
	}
}
