package de.lmu.ifi.pms.cep.tsa.time;

import static de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.Term.apply;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.*;
import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.*;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.Attribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeValue;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DisjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DurationValue;
import de.lmu.ifi.pms.cep.tsa.conditions.Greatest;
import de.lmu.ifi.pms.cep.tsa.conditions.IdentifierValue;
import de.lmu.ifi.pms.cep.tsa.conditions.IntValue;
import de.lmu.ifi.pms.cep.tsa.conditions.Least;
import de.lmu.ifi.pms.cep.tsa.conditions.LongValue;
import de.lmu.ifi.pms.cep.tsa.conditions.Minus;
import de.lmu.ifi.pms.cep.tsa.conditions.NegatedConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.Plus;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TermID;
import de.lmu.ifi.pms.cep.tsa.conditions.TimestampAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.TimestampValue;
import de.lmu.ifi.pms.cep.tsa.conditions.TupleFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID;
import de.lmu.ifi.pms.cep.tsa.conditions.Value;
import de.lmu.ifi.pms.cep.tsa.conditions.Variable;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.cep.tsa.schema.TypeID;
import de.lmu.ifi.pms.cep.tsa.time.constraints.ConjunctiveTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.DisjunctiveTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.NegatedTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.SimpleTemporalFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.TemporalMinMax;
import de.lmu.ifi.pms.cep.tsa.time.constraints.TemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.BinaryComparisonOperator;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.MinMax;
import de.lmu.ifi.pms.java.math.Math;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.AbstractFilter;
import de.lmu.ifi.pms.java.util.filter.Filter;

public class ArithmeticRelationFormulaExtractor {
		
	public static class AdapterFilter extends AbstractFilter<AtomicConditionFormula>{
		
		final Adapter<AtomicConditionFormula, ?> adapter;
		
		public AdapterFilter(final Adapter<AtomicConditionFormula, ?> localType) {
			super();
			this.adapter = localType;
		}

		@Override
		public boolean accepts(final AtomicConditionFormula focusedFormula) {
			return this.adapter.adapt(focusedFormula) != null;
		}
		
	}
	
	final Type periodType;
	final Type durationType;
	final Value<?> periodZero;
	final Value<?> durationZero;
	final Filter<AtomicConditionFormula> focus;
	public Adapter<AtomicConditionFormula, ConditionFormula> allLeftInequations;
	
	public final Adapter<Term<?>, Term<?>> pushConstantUp;

	public final Adapter<AtomicConditionFormula, ConditionFormula> periodSpliter;
	
	protected static final DurationValue zeroDuration = new DurationValue(0);
	protected static final LongValue zeroLong = new LongValue(0);
	
	protected static final IdentifierValue zeroIdentifierPeriod = new IdentifierValue(0);
	protected static final TimestampValue zeroTimestampPeriod = new TimestampValue(0);
	
	public Adapter<AtomicConditionFormula, ConditionFormula> minMaxEliminator;
	
	public Adapter<AtomicConditionFormula, ConditionFormula> leftToRightInequations;
	
	public final Adapter<Term<?>, Term<?>> zeroAdditionRemover;
	
	final Adapter<AtomicConditionFormula, TemporalRelationFormula> localAdapter;
	

	public ArithmeticRelationFormulaExtractor(final Type periodType) {
		super();
		this.periodType = periodType;
		this.durationType = this.durationType(this.periodType);
		this.periodZero = this.zeroPeriod(this.periodType);
		this.durationZero = this.zeroDuration(this.durationType);
		this.focus = new AbstractFilter<AtomicConditionFormula>() {
			
			@Override
			public boolean accepts(final AtomicConditionFormula atom) {
				if(atom instanceof TemporalRelationFormula){
					return true;
				}
				else if(atom instanceof BinaryComparissonFormula){
					BinaryComparissonFormula bcf = (BinaryComparissonFormula) atom;
					if((bcf.firstTerm().type.equals(periodType) & bcf.secondTerm().type.equals(periodType)) | (bcf.firstTerm().type.equals(ArithmeticRelationFormulaExtractor.this.durationType) & bcf.secondTerm().type.equals(ArithmeticRelationFormulaExtractor.this.durationType))){
						for(Term<?> term: atom.collect(AbstractFilter.<Term<?>>allFilter())){
							if(term.ID == TermID.ATOMIC){
								if((term instanceof Attribute<?>) | (term instanceof Variable<?>)){
									if(!term.type.equals(periodType)){
										return false;
									}
								}
								else if(term instanceof Value<?>){
									if(!term.type.equals(ArithmeticRelationFormulaExtractor.this.durationType)){
										return false;
									}
								}
								else{
									return false;
								}
							}
							else if(term instanceof TupleFunction<?>){
								TupleFunction<?> function = (TupleFunction<?>) term;								 
								if(function.ID == PLUS){
									//FIXME only duration + duration or period + duration (note that subterms are already examined, because all terms of the formula are examined)
								}
								else if(function.ID == MINUS){
									//FIXME only period - period or period - duration, or duration - duration(note that subterms are already examined, because all terms of the formula are examined)
								}
								else if(function.ID == LEAST| function.ID == GREATEST){								
									//FIXME only all period or all duration (note that subterms are already examined, because all terms of the formula are examined)
									
								}
								else{
									return false;
								}
							}
							else{
								return false;
							}
						}
						return true;
					}
				}
				return false;
			}
		};
		this.allLeftInequations = new Adapter<AtomicConditionFormula, ConditionFormula>() {
			
			@Override
			public ConditionFormula adapt(final AtomicConditionFormula source) {
				if(source instanceof BinaryComparissonFormula){
					BinaryComparissonFormula bcf = (BinaryComparissonFormula) source;
					assert (bcf.firstTerm().type.equals(ArithmeticRelationFormulaExtractor.this.periodType) & bcf.secondTerm().type.equals(ArithmeticRelationFormulaExtractor.this.periodType)) | (bcf.firstTerm().type.equals(ArithmeticRelationFormulaExtractor.this.durationType) & bcf.secondTerm().type.equals(ArithmeticRelationFormulaExtractor.this.durationType));
					return new BinaryComparissonFormula(bcf.operator, apply(MINUS, ArithmeticRelationFormulaExtractor.this.durationType, bcf.firstTerm(), bcf.secondTerm()), ArithmeticRelationFormulaExtractor.this.durationZero);
				}
				return source;
			}
			
		};
		this.pushConstantUp = new Adapter<Term<?>, Term<?>>() {
			
			@Override
			public Term<?> adapt(final Term<?> source) {
				if(source instanceof Value<?>){
					return source;
				}
				else if(source instanceof AtomicTerm<?>){
					AtomicTerm<?> atomicTerm = (AtomicTerm<?>) source;
					assert atomicTerm.type.equals(ArithmeticRelationFormulaExtractor.this.periodType) | atomicTerm.type.equals(ArithmeticRelationFormulaExtractor.this.durationType);
					return apply(PLUS, atomicTerm.type, atomicTerm, ArithmeticRelationFormulaExtractor.this.zeroDuration((Type)atomicTerm.type));
				}
				else if(source instanceof TupleFunction<?>){
					TupleFunction<?> function = (TupleFunction<?>) source;
					Term<?>[] normalizedArguments = new Term<?>[function.arity()];
					boolean allValues = true;
					for(int arg = 0; arg < normalizedArguments.length; ++arg){
						normalizedArguments[arg] = this.adapt(function.arg(arg));
						if(!(normalizedArguments[arg] instanceof Value<?>)){
							allValues = false;
						}
					}
					if(allValues){
						return ArithmeticRelationFormulaExtractor.this.compute(function);
					}
					else{//at least one of the arguments is not a value
						//all arguments are normalized
						switch (function.ID) {
						case PLUS: {
							assert function.type.equals(ArithmeticRelationFormulaExtractor.this.periodType) | function.type.equals(ArithmeticRelationFormulaExtractor.this.durationType);
							if(normalizedArguments[0] instanceof Value<?>){
								return this.adapt(TupleFunction.apply(PLUS, function.type, normalizedArguments[1], normalizedArguments[0]));
							}
							else{assert normalizedArguments[0] instanceof Plus<?>; 
								Plus<?> arg1 = (Plus<?>) normalizedArguments[0];
								assert arg1.arg(1).type.equals(ArithmeticRelationFormulaExtractor.this.durationType);
								if(normalizedArguments[1] instanceof Value<?>){
									assert normalizedArguments[1].type.equals(ArithmeticRelationFormulaExtractor.this.durationType);
									return apply(PLUS, function.type, arg1.arg(0), ArithmeticRelationFormulaExtractor.this.compute(apply(function.ID, ArithmeticRelationFormulaExtractor.this.durationType, arg1.arg(1), normalizedArguments[1])));								
								}
								else{ assert normalizedArguments[1] instanceof Plus<?>;
									Plus<?> arg2 = (Plus<?>) normalizedArguments[1];
									return apply(PLUS, function.type, apply(function.ID, function.type, arg1.arg(0), arg2.arg(0)), ArithmeticRelationFormulaExtractor.this.compute(apply(function.ID, ArithmeticRelationFormulaExtractor.this.durationType, arg1.arg(1), arg2.arg(1))));
								}
							}
						}
						case MINUS: {
							assert function.type.equals(ArithmeticRelationFormulaExtractor.this.periodType) | function.type.equals(ArithmeticRelationFormulaExtractor.this.durationType);						
							if(normalizedArguments[0] instanceof Value<?>){
								assert normalizedArguments[1] instanceof Plus;
								assert normalizedArguments[1].isTyped;
								return this.adapt(TupleFunction.apply(PLUS, function.type, apply(MINUS, ArithmeticRelationFormulaExtractor.this.zeroPeriod((Type)normalizedArguments[1].type), normalizedArguments[1]), normalizedArguments[0]));
							}
							else{assert normalizedArguments[0] instanceof Plus<?>; 
								Plus<?> arg1 = (Plus<?>) normalizedArguments[0];						
								if(normalizedArguments[1] instanceof Value<?>){
									return apply(PLUS, function.type, arg1.arg(0), ArithmeticRelationFormulaExtractor.this.compute(apply(function.ID, arg1.arg(1), normalizedArguments[1])));								
								}
								else{ assert normalizedArguments[1] instanceof Plus<?>;
									Plus<?> arg2 = (Plus<?>) normalizedArguments[1];
									return apply(PLUS, function.type, apply(function.ID, function.type, arg1.arg(0), arg2.arg(0)), ArithmeticRelationFormulaExtractor.this.compute(apply(function.ID, arg1.arg(1), arg2.arg(1))));
								}
							}
						}
						case GREATEST: 	case LEAST: {
							ArrayList<Term<?>> nonValueTermsList = new ArrayList<Term<?>>();
							ArrayList<Term<?>> valueTermsList = new ArrayList<Term<?>>();
							//combine all constant values
							for(int arg = 0; arg < normalizedArguments.length; ++arg){
								if(normalizedArguments[arg] instanceof Value<?>){
									valueTermsList.add(normalizedArguments[arg]);
								}
								else{
									nonValueTermsList.add(normalizedArguments[arg]);
								}
							}						
							assert nonValueTermsList.size() > 0;
							if(valueTermsList.size() > 0){
								Value<?> value = ArithmeticRelationFormulaExtractor.this.compute(apply(function.ID, function.type, valueTermsList.toArray(new Term<?>[0])));
								Term<?>[] nonValueTerms = nonValueTermsList.toArray(new Term<?>[0]);
								for(int arg = 0; arg < nonValueTerms.length; ++arg){
									assert nonValueTerms[arg] instanceof Plus<?>;
									Plus<?> nonValueTerm = (Plus<?>) nonValueTerms[arg];
									nonValueTerms[arg] = apply(PLUS, nonValueTerm.type, nonValueTerm.arg(0), ArithmeticRelationFormulaExtractor.this.compute(apply(MINUS, value.type, nonValueTerm.arg(1), value)));
								}
								return apply(PLUS, function.type, apply(function.ID, function.type, nonValueTerms), value);
							}
							else{
								assert function.isTyped;
								return apply(PLUS, function.type, apply(function.ID, function.type, normalizedArguments), ArithmeticRelationFormulaExtractor.this.zeroDuration((Type)function.type));
							}	
						}
						default:
							throw new IllegalArgumentException();
						}
					}
				}
				return source;
			}		
			
		};
		this.periodSpliter = new Adapter<AtomicConditionFormula, ConditionFormula>() {
			
			@Override
			public ConditionFormula adapt(final AtomicConditionFormula source) {
				if(source instanceof BinaryComparissonFormula){					
					BinaryComparissonFormula bcf = (BinaryComparissonFormula) source;
					assert bcf.secondTerm().equals(ArithmeticRelationFormulaExtractor.this.durationZero);
					assert bcf.firstTerm().type.equals(ArithmeticRelationFormulaExtractor.this.durationType);
					assert bcf.firstTerm() instanceof Plus<?>;
					Plus<?> plus = (Plus<?>) bcf.firstTerm();
					assert plus.arg(1) instanceof Value<?>;
					assert plus.arg(1).type.equals(ArithmeticRelationFormulaExtractor.this.durationType);
					assert plus.arg(0).type.equals(ArithmeticRelationFormulaExtractor.this.durationType);
					assert plus.arg(0) instanceof Minus<?>;
					
					Value<?> constant = (Value<?>) plus.arg(1);
					Minus<?> periodDifference = (Minus<?>) plus.arg(0);
					assert periodDifference.arg(0).type.equals(periodType);
					assert periodDifference.arg(1).type.equals(periodType);
					
					return new BinaryComparissonFormula(bcf.operator, periodDifference.arg(0), apply(PLUS, periodType, periodDifference.arg(1), ArithmeticRelationFormulaExtractor.this.compute(apply(MINUS, ArithmeticRelationFormulaExtractor.this.durationType, ArithmeticRelationFormulaExtractor.this.durationZero, constant))));
				}
				return source;
			}
			
		};
		this.minMaxEliminator = new Adapter<AtomicConditionFormula, ConditionFormula>() {
			
			@Override
			public ConditionFormula adapt(final AtomicConditionFormula source) {
				if(source instanceof BinaryComparissonFormula){
					BinaryComparissonFormula bcf = (BinaryComparissonFormula) source;
					assert bcf.firstTerm().type.equals(periodType);
					assert bcf.secondTerm().type.equals(periodType);
					assert bcf.secondTerm() instanceof Plus<?>;
					Plus<?> secondTerm = (Plus<?>) bcf.secondTerm();						
					if(bcf.firstTerm() instanceof Greatest<?> | bcf.firstTerm() instanceof Least<?>){
						if(bcf.operator == NOT_EQUAL){
							return not(this.adapt(new BinaryComparissonFormula(EQUAL, bcf.firstTerm(), bcf.secondTerm())));
						}
						else if(bcf.operator == EQUAL){
							return and(this.adapt(new BinaryComparissonFormula(GREATER_EQUAL, bcf.firstTerm(), bcf.secondTerm())), this.adapt(new BinaryComparissonFormula(LESS_EQUAL, bcf.firstTerm(), bcf.secondTerm())));
						}
						else{
							TupleFunction<?> leastOrGreatest = (TupleFunction<?>) bcf.firstTerm();
							ConditionFormula[] replacementSubFomulas = new ConditionFormula[leastOrGreatest.arity()];
							
							for(int arg = 0; arg < replacementSubFomulas.length; ++arg){
								assert leastOrGreatest.arg(arg) instanceof Plus<?>;
								Plus<?> greatestArg = (Plus<?>) leastOrGreatest.arg(arg);
								replacementSubFomulas[arg] = new BinaryComparissonFormula(bcf.operator, greatestArg.arg(0), apply(PLUS, periodType, secondTerm.arg(0), ArithmeticRelationFormulaExtractor.this.compute(apply(MINUS, ArithmeticRelationFormulaExtractor.this.durationType, secondTerm.arg(1), greatestArg.arg(1)))));
							}
							ConditionFormula replacementFormula;
						if((leastOrGreatest.ID == GREATEST & (bcf.operator == LESS | bcf.operator == LESS_EQUAL)) | (leastOrGreatest.ID == LEAST & (bcf.operator == GREATER_EQUAL | bcf.operator == GREATER))){
							replacementFormula = and(replacementSubFomulas);
						}
						else if((leastOrGreatest.ID == GREATEST & (bcf.operator == GREATER_EQUAL | bcf.operator == GREATER)) | (leastOrGreatest.ID == LEAST & (bcf.operator == LESS | bcf.operator == LESS_EQUAL))){
							replacementFormula = or(replacementSubFomulas);
						}
						else{
							throw new Error();
						}
						return replacementFormula.replaceAtomicFormula(this);
						
						}
					}
					if(secondTerm.arg(0) instanceof Greatest<?> | secondTerm.arg(0) instanceof Least<?>){
						if(bcf.firstTerm() instanceof Attribute<?> | bcf.secondTerm() instanceof Variable<?>){
							
						}
						return this.adapt(ArithmeticRelationFormulaExtractor.this.mirror(bcf));
					}
					
				}
				return source;
			}
			
		};
		this.leftToRightInequations = new Adapter<AtomicConditionFormula, ConditionFormula>() {
			
			@Override
			public ConditionFormula adapt(final AtomicConditionFormula source) {
				if(source instanceof BinaryComparissonFormula){
					BinaryComparissonFormula bcf = (BinaryComparissonFormula) source;
					switch (bcf.operator) {
					case LESS: {
						return bcf;
					}
					case LESS_EQUAL: {
						return bcf;
					}
					case EQUAL: {
						return bcf;
					}
					case GREATER: {
						return ArithmeticRelationFormulaExtractor.this.mirror(bcf);
					}
					case GREATER_EQUAL: {
						return ArithmeticRelationFormulaExtractor.this.mirror(bcf);
					}
					case NOT_EQUAL:{
						return not(new BinaryComparissonFormula(EQUAL, bcf.firstTerm(), bcf.secondTerm()));
					}
					default: throw new Error("Unexpected case: " + bcf.operator);				
					}
				}
				return source;
			}
			
		};
		this.zeroAdditionRemover = new Adapter<Term<?>, Term<?>>(){
			
			@Override
			public Term<?> adapt(final Term<?> source) {
				if(source instanceof TupleFunction<?>){
					if(source instanceof Plus<?>){
						Plus<?> plus = (Plus<?>) source;
						if(plus.arg(1) instanceof Value<?> && ArithmeticRelationFormulaExtractor.this.isZeroDuration((Value<?>)plus.arg(1))){
							return this.adapt(plus.arg(0));
						}
					}
					//if not returned before
					TupleFunction<?> function = (TupleFunction<?>) source;
					Term<?>[] normalizedArguments = new Term<?>[function.arity()];
					for(int arg = 0; arg < normalizedArguments.length; ++arg){
						normalizedArguments[arg] = this.adapt(function.arg(arg));
					}
					return apply(function.ID, function.type, normalizedArguments);
				}
				return source;
			}
			
		};
		this.localAdapter = new Adapter<AtomicConditionFormula, TemporalRelationFormula>(){
			
			@Override
			public TemporalRelationFormula adapt(final AtomicConditionFormula atom) {
				if(atom instanceof TemporalRelationFormula){
					return (TemporalRelationFormula) atom;
				}
				else if(atom instanceof BinaryComparissonFormula){
					BinaryComparissonFormula bcf = (BinaryComparissonFormula) atom;
					BinaryComparisonOperator op;
					switch (bcf.operator) {
					case EQUAL: op = BinaryComparisonOperator.EQUAL;
						break;
					case GREATER: op = BinaryComparisonOperator.GREATER;
						break;
					case GREATER_EQUAL: op = BinaryComparisonOperator.GREATER_EQUAL;
						break;
					case LESS: op = BinaryComparisonOperator.LESS;
						break;
					case LESS_EQUAL: op = BinaryComparisonOperator.LESS_EQUAL;
						break;
					case NOT_EQUAL: op = BinaryComparisonOperator.NOT_EQUAL;
						break;
					default:
						throw new Error();
						//TODO better exception
					}
					final Term<?> term1 = bcf.term(0);
		
					final String name1;
					final boolean isAttribute1;
					
					if(term1 instanceof BasicTypeAttribute){
						assert term1.type.equals(ArithmeticRelationFormulaExtractor.this.periodType);
						name1 = ((BasicTypeAttribute)term1).name;
						isAttribute1 = true;
					}
					else if( term1 instanceof BasicTypeVariable){
						assert term1.type.equals(ArithmeticRelationFormulaExtractor.this.periodType);
						BasicTypeVariable var1 = (BasicTypeVariable) term1;
						name1 = var1.name;
						isAttribute1 = false;			
					}
					else{
						throw new IllegalArgumentException();
					}
					
					
					Term<?> term2 = bcf.term(1);
					if(term2 instanceof BasicTypeAttribute){
						assert term2.type.equals(ArithmeticRelationFormulaExtractor.this.periodType);
						String name2 = ((BasicTypeAttribute)term2).name;
						boolean isAttribute2 = true;
						long constant = 0;
						return new SimpleTemporalFormula(name1, isAttribute1, op, name2, isAttribute2, constant);
					}
					else if(term2 instanceof BasicTypeVariable){
						assert term2.type.equals(ArithmeticRelationFormulaExtractor.this.periodType);
						BasicTypeVariable var2 = (BasicTypeVariable) term2;
						String name2 = var2.name;
						boolean isAttribute2 = false;
						long constant = 0;
						return new SimpleTemporalFormula(name1, isAttribute1, op, name2, isAttribute2, constant);
					}
					else if(term2 instanceof TupleFunction){
						TupleFunction<?> function2 = (TupleFunction<?>) bcf.secondTerm();
						if(function2.ID == TupleFunctionID.PLUS | function2.ID == TupleFunctionID.MINUS){
							Term<?> arg1 = function2.arg(0);
							Term<?> arg2 = function2.arg(1);
							final String name2;
							final boolean isAttribute2;
							final long constant;
							if(arg1 instanceof BasicTypeAttribute){
								arg1.type.equals(periodType);
								name2 = ((TimestampAttribute)arg1).name;
								isAttribute2 = true;
							}
							else if(arg1 instanceof BasicTypeVariable){
								arg1.type.equals(periodType);
								BasicTypeVariable var2 = (BasicTypeVariable) arg1;
								name2 = var2.name;
								isAttribute2 = false;
							}
							else{
								throw new IllegalArgumentException();
							}
							if(arg2 instanceof BasicTypeValue){
								assert arg2.type.equals(ArithmeticRelationFormulaExtractor.this.durationType);
								long value = ArithmeticRelationFormulaExtractor.this.longvalue((BasicTypeValue)arg2);
								if(function2.ID == TupleFunctionID.MINUS){
									constant = -value;
								}
								else{
									constant = value;
								}
							}
							else{
								throw new IllegalArgumentException();
							}
							return new SimpleTemporalFormula(name1, isAttribute1, op, name2, isAttribute2, constant);
						}
						else if(bcf.operator == BinaryComparissonOperatorID.EQUAL){
							String[] arguments = new String[function2.arity()];
							boolean[] isAttribute = new boolean[arguments.length];
							
							final MinMax m;
							if(function2.ID == TupleFunctionID.LEAST){
								m = MinMax.MIN;
							}
							else if(function2.ID == TupleFunctionID.GREATEST){
								m = MinMax.MAX;
							}
							else{
								throw new IllegalArgumentException();
							}
							
							for(int argument = 0; argument < arguments.length; ++argument){
								assert function2.arg(argument).type.equals(periodType);
								if(function2.arg(argument) instanceof BasicTypeAttribute){	
									arguments[argument] = ((BasicTypeAttribute)function2.arg(argument)).name;
									isAttribute[argument] = true;
								}
								else if(function2.arg(argument) instanceof BasicTypeVariable){
									
									BasicTypeVariable var2 = (BasicTypeVariable) function2.arg(argument);
									arguments[argument] = var2.name;
									isAttribute[argument] = false;
								}
								else{
									throw new IllegalArgumentException();
								}
							}							
							return new TemporalMinMax(name1, isAttribute1, m, arguments, isAttribute);
						}
						else{
							throw new IllegalArgumentException();
						}
					}	
					else{
						throw new IllegalArgumentException();
					}
				}
				else{
					throw new IllegalArgumentException();
				}
			}
		};
	}
	
	protected BinaryComparissonFormula mirror(final BinaryComparissonFormula bcf){
		assert bcf.secondTerm() instanceof Plus<?>;
		Plus<?> secondTerm = (Plus<?>) bcf.secondTerm();
		assert secondTerm.type.equals(ArithmeticRelationFormulaExtractor.this.periodType);
		assert secondTerm.arg(1).type.equals(ArithmeticRelationFormulaExtractor.this.durationType);
		assert secondTerm.arg(1) instanceof Value<?>;
		return new BinaryComparissonFormula(bcf.operator.mirrored(), secondTerm.arg(0), apply(PLUS, ArithmeticRelationFormulaExtractor.this.periodType, bcf.firstTerm(), ArithmeticRelationFormulaExtractor.this.compute(apply(MINUS, ArithmeticRelationFormulaExtractor.this.durationType, ArithmeticRelationFormulaExtractor.this.durationZero, secondTerm.arg(1)))));
	}

	//TODO liniarize plus and minus convert to ( + + + ) - ( + + + )
	
	public Type durationType(final Type periodType){
		if(periodType instanceof BasicType){
			BasicType basicType = (BasicType) periodType;
			switch (basicType.ID) {
			case IDENTIFIER: {
				return DURATION.type;
			}
			case TIMESTAMP: {
				return DURATION.type;
			}
			case LONG: {
				return LONG.type;
			}
			default:
				break;
			}
		}
		throw new IllegalArgumentException("Type \"" + periodType.toString() + "\" has no duration type.");
	}
	
	public Value<?> zeroPeriod(final Type periodType){
		if(periodType instanceof BasicType){
			BasicType basicType = (BasicType) periodType;
			switch (basicType.ID) {
			case IDENTIFIER: {
				return zeroIdentifierPeriod;
			}
			case TIMESTAMP: {
				return zeroTimestampPeriod;
			}
			case LONG: {
				return zeroLong;
			}
			default:
				break;
			}
		}
		throw new IllegalArgumentException();
	}
	
	public Value<?> zeroDuration(final Type durationType){
		if(durationType instanceof BasicType){
			BasicType basicType = (BasicType) durationType;
			switch (basicType.ID) {
			case DURATION: {
				return zeroDuration;
			}
			case LONG: {
				return zeroLong;
			}
			default:
				try{
					return this.zeroDuration(this.durationType(durationType));
				}
				catch(IllegalArgumentException e){
					break;
				}
			}
		}
		throw new IllegalArgumentException("Type \"" + durationType.toString() + "\" is not a duration type.");
	}
	
	public boolean isZeroDuration(final Value<?> value){
		return value.equals(this.zeroDuration(value.type));
	}
	
	public boolean isZeroPeriod(final Value<?> value){
		return value.equals(this.zeroPeriod(value.type));
	}
	
	public long longvalue(final Value<?> value){
		if(value.type.ID == TypeID.BASIC){
			BasicTypeValue basicTypeValue = (BasicTypeValue) value;
			switch (basicTypeValue.type.ID) {
			case LONG: {
				return ((LongValue)value).value;
			}
			case DURATION: {
				return ((DurationValue)value).value;
			}
			default:
				throw new Error("Unsupported case: " + basicTypeValue.type.ID);
			}
		}
		else{
			throw new Error("Unsupported case: " + value.type);
		}
	}
	
	//eliminate dublicates in min/max , constante aus min, max ausklammern, bai gleichem term aber unterschiedlicher konstante nur kleinere/größere konstante behalten
/*	public Adapter<Term<?>, Term<?>> sumAndDifferenceNormalizer = new Adapter<Term<?>, Term<?>>() {
		
		@Override
		public Term<?> adapt(final Term<?> source) {
			if(source instanceof TupleFunction<?>){
				TupleFunction<?> function = (TupleFunction<?>) source;
				Term<?>[] normalizedArguments = new Term<?>[function.arity()];
				boolean allValues = true;
				for(int arg = 0; arg < normalizedArguments.length; ++arg){
					normalizedArguments[arg] = this.adapt(function.arg(arg));
					if(!(normalizedArguments[arg] instanceof BasicTypeValue)){
						allValues = false;
					}
				}
				if(allValues){
					return ArithmeticRelationFormulaExtractor.this.compute(function);
				}
				else{//at least one of the arguments is not a value
					//all arguments are normalized
					switch (function.ID) {
					case PLUS: {
						if(normalizedArguments[0] instanceof BasicTypeValue){
							return this.adapt(TupleFunction.apply(function.ID, function.type, normalizedArguments[1], normalizedArguments[0]));
						}
						else if(normalizedArguments[1] instanceof BasicTypeValue){
							if(normalizedArguments[0] instanceof TupleFunction<?>){
								TupleFunction<?> innerFunction = (TupleFunction<?>) normalizedArguments[0];
								switch (innerFunction.ID) {
								case PLUS: {
									if(innerFunction.arg(1) instanceof BasicTypeValue){
										return TupleFunction.apply(function.ID, function.type, innerFunction.arg(0), ArithmeticRelationFormulaExtractor.this.compute(apply(PLUS, innerFunction.arg(1), normalizedArguments[1])));
									}
									else{
										return TupleFunction.apply(function.ID, function.type, normalizedArguments);
									}
								}
								case MINUS: {
									return this.adapt(apply(MINUS, apply(PLUS, innerFunction.arg(2), normalizedArguments[1]), innerFunction.arg(1)));
								}
								default:
									return TupleFunction.apply(function.ID, function.type, normalizedArguments);
								}
								
							}
							else{
								return TupleFunction.apply(function.ID, function.type, normalizedArguments);
							}
						}
						else if(normalizedArguments[0] instanceof Plus){
							
						}
						else if(normalizedArguments[0] instanceof Minus){
							
						}
					}
					case MINUS: {
				
						break;
					}
					case GREATEST: {
			
						break;
					}
					case LEAST: {
	
						break;
					}
					default:
						return TupleFunction.apply(function.ID, function.type, normalizedArguments);
					}
				}
			}
			return source;
		}
		
	};
*/	

	public Adapter<Term<?>, Term<?>> compute = new Adapter<Term<?>, Term<?>>() {
		
		@Override
		public de.lmu.ifi.pms.cep.tsa.conditions.Term<?> adapt(final de.lmu.ifi.pms.cep.tsa.conditions.Term<?> source) {
			if(source instanceof TupleFunction<?>){
				TupleFunction<?> function = (TupleFunction<?>) source;
				if(function.ID != USER){
					if(function.Variables().size() == 0 & function.Attributes().size() == 0){
						try{
							return ArithmeticRelationFormulaExtractor.this.compute(function);
						}
						catch(IllegalArgumentException e){
							throw new IllegalArgumentException("Can not combine constants.", e);
						}
					}
					else{
						boolean changed = false;
						Term<?>[] arguments = function.arguments().toArray(new Term<?>[0]);
						for(int arg = 0; arg < arguments.length; ++arg){
							Term<?> newArgument = this.adapt(arguments[arg]);
							if(newArgument != arguments[arg]){
								arguments[arg] = newArgument;
								changed = true;
							}
						}
						if(changed){
							if(function.isTyped){
								return Term.apply(function.ID, (Type)function.type, arguments);
							}
							else{
								return TupleFunction.apply(function.ID, arguments);
							}
						}
					}
				}
			}
			return source;
		}
		
	};
	
	//TODO componentWise addition for composite types/values
	public BasicTypeValue compute(final TupleFunction<?> term){
		final TupleFunctionID function = term.ID;
		final Term<?>[] arguments = term.arguments().toArray(new Term<?>[0]);
		if(function.minArity <= arguments.length & arguments.length <= function.maxArity){
			BasicTypeValue[] values = new BasicTypeValue[arguments.length];
			for(int arg = 0; arg < arguments.length; ++arg){
				Term<?> argument = arguments[arg];
				BasicTypeValue value;
				if(!argument.isTyped){
					throw new IllegalArgumentException("Only typed arguments are allowed;");
				}
				if(argument instanceof TupleFunction<?>){
					TupleFunction<?> innerFunction = (TupleFunction<?>) argument;
					value = this.compute(innerFunction);
				}
				else if(argument instanceof BasicTypeValue){
					value = (BasicTypeValue) argument;
				}
				else{
					throw new IllegalArgumentException("Only tuple functions and basic type values are alled.");
				}
				switch (value.type.ID) {
				case INT: {
					value = new LongValue(((IntValue)value).value);
					break;
				}
				case LONG: {				
					break;
				}
				case DURATION: {					
					break;
				}
				default:
					throw new IllegalArgumentException("Unsupported type: " + value.type.ID);
				}
				
				values[arg] = value;
			}
			BasicType commonType = values[0].type;
			long[] longValues = new long[arguments.length];
			for(int arg = 0; arg < values.length; ++arg){
				if(values[arg].type.ID != commonType.ID){
					throw new IllegalArgumentException("Incompatible types");
				}
				longValues[arg] = this.longvalue(values[arg]);
			}
			
			long result;
			switch (function) {
			case PLUS: {
				result = Math.plus(longValues[0], longValues[1]);
				break;
			}
			case DIVISION: {
				result = Math.div(longValues[0] , longValues[1]);
				break;
			}
			case MINUS: {
				result = Math.sub(longValues[0], longValues[1]);
				break;
			}
			case POWER: {
				result = Math.potenz(longValues[0], longValues[1]);
				break;
			}
			case TIMES: {
				result = Math.mult(longValues[0] , longValues[1]);
				break;
			}
			case GREATEST: {
				result = Math.max(longValues);				
				break;
			}
			case LEAST: {
				result = Math.min(longValues);
				break;
			}
			default:
				throw new IllegalArgumentException("Unsupported function: " + function);
			}
			
			switch (commonType.ID) {
			case LONG: {
				return new LongValue(result);
			}
			case DURATION: {
				return new DurationValue(result);
			}
			default:
				throw new Error("Unexpected case: " + commonType.ID);
			}
		}
		else{
			throw new IllegalArgumentException("Invalid number of arguments for function " + function);
		}	
	}
	
	public ConditionFormula arithmeticRelationFormula(final FixedArray<ConditionFormula> constraints){
		return this.arithmeticRelationFormula(new ConjunctiveConditionFormula(constraints));
	}
	
	public ConditionFormula arithmeticRelationFormula(final ConditionFormula formula){
		return formula.flaten().focus(this.focus).flaten().replaceAtomicFormula(this.allLeftInequations).replaceTerm(this.pushConstantUp).replaceAtomicFormula(this.periodSpliter).replaceAtomicFormula(this.minMaxEliminator).replaceAtomicFormula(this.leftToRightInequations).replaceTerm(this.zeroAdditionRemover).flaten();
	}

	public TemporalRelationFormula extract(final ConditionFormula formula){
		return this.convert(this.arithmeticRelationFormula(formula));
	}
	
	protected TemporalRelationFormula convert(final ConditionFormula temporalRelationFormula){
		if(temporalRelationFormula instanceof TemporalRelationFormula){
			return (TemporalRelationFormula) temporalRelationFormula;
		}
		else{
			switch (temporalRelationFormula.ID) {
			case TOP: return TemporalRelationFormula.Top;
			case BOTTOM: return TemporalRelationFormula.Bottom;
			case ATOMIC:
				TemporalRelationFormula adoption = this.localAdapter.adapt((AtomicConditionFormula)temporalRelationFormula);
				assert adoption != null;
				return adoption;
			case CONJUNCTIVE:{
				TemporalRelationFormula[] subformulas;
				ConjunctiveConditionFormula ccf = (ConjunctiveConditionFormula) temporalRelationFormula;
				subformulas = new TemporalRelationFormula[ccf.subformulas().size()];
				for(int index = 0; index < subformulas.length; ++index){
					subformulas[index] = this.convert(ccf.subformulas().get(index));
				}
				return new ConjunctiveTemporalRelationFormula(subformulas);
			}
			case DISJUNCTIVE:{
				TemporalRelationFormula[] subformulas;
				DisjunctiveConditionFormula dcf = (DisjunctiveConditionFormula) temporalRelationFormula;
				subformulas = new TemporalRelationFormula[dcf.subformulas().size()];
				for(int index = 0; index < subformulas.length; ++index){
					subformulas[index] = this.convert(dcf.subformulas().get(index));
				}
				return new DisjunctiveTemporalRelationFormula(subformulas);
			}
			case NEGATED:
				return new NegatedTemporalRelationFormula(this.convert(((NegatedConditionFormula)temporalRelationFormula).subformula()));
			default:
				throw new IllegalArgumentException();
				//TODO better exception
			}
		}
	}
	
}
