package de.lmu.ifi.pms.cep.tsa.time.analysis;

import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.time.ArithmeticRelationFormulaExtractor;
import de.lmu.ifi.pms.cep.tsa.time.constraints.TemporalRelationFormula;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;

public abstract class ArithmeticAnalyserFactory {
	
	final ArithmeticRelationFormulaExtractor formulaExtractor;
	
	public ArithmeticAnalyserFactory(final BasicTypeID type) {
		this(type.type);
	}
	
	public ArithmeticAnalyserFactory(final BasicType type) {
		this(new ArithmeticRelationFormulaExtractor(type));
	}
	
	public ArithmeticAnalyserFactory(final ArithmeticRelationFormulaExtractor formulaExtractor) {
		super();
		this.formulaExtractor = formulaExtractor;
	}

	public ArithmeticAnalyser make(final ConditionFormula focusedFormula){
		return this.make(this.formulaExtractor.extract(focusedFormula));
	}
	
	public ArithmeticAnalyser make(final ConditionFormula focusedFormula, final FixedArrayNavigableSet<String> relevantAttributes, final FixedArrayNavigableSet<String> relevantVariables){
		return this.make(this.formulaExtractor.extract(focusedFormula), relevantAttributes, relevantVariables);
	}
	
	public abstract ArithmeticAnalyser make(final TemporalRelationFormula temporalRelationFormula);
	
	public abstract ArithmeticAnalyser make(final TemporalRelationFormula temporalRelationFormula, final FixedArrayNavigableSet<String> relevantAttributes, final FixedArrayNavigableSet<String> relevantVariables);
	
}
