package de.lmu.ifi.pms.cep.tsa.time.analysis;

import de.lmu.ifi.pms.cep.tsa.conditions.AtomicConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID;
import de.lmu.ifi.pms.cep.tsa.conditions.DurationValue;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TimestampAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.TupleFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID;
import de.lmu.ifi.pms.cep.tsa.conditions.analysis.SimpleFocusedFormulaAnalyser;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.time.constraints.ConjunctiveTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.DisjunctiveTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.NegatedTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.SimpleTemporalFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.TemporalMinMax;
import de.lmu.ifi.pms.cep.tsa.time.constraints.TemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.BinaryComparisonOperator;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.MinMax;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.filter.AbstractFilter;
import de.lmu.ifi.pms.java.util.filter.Filter;

public class TemporalSimpleFocusedFormulaAnalyser extends SimpleFocusedFormulaAnalyser {
	
	public static final Adapter<AtomicConditionFormula, TemporalRelationFormula> localType =
		new Adapter<AtomicConditionFormula, TemporalRelationFormula>(){
		
			@Override
			public TemporalRelationFormula adapt(final AtomicConditionFormula atom) {
				if(atom instanceof TemporalRelationFormula){
					return (TemporalRelationFormula) atom;
				}
				else if(atom instanceof BinaryComparissonFormula){
					BinaryComparissonFormula bcf = (BinaryComparissonFormula) atom;
					BinaryComparisonOperator op;
					switch (bcf.operator) {
					case EQUAL: op = BinaryComparisonOperator.EQUAL;
						break;
					case GREATER: op = BinaryComparisonOperator.GREATER;
						break;
					case GREATER_EQUAL: op = BinaryComparisonOperator.GREATER_EQUAL;
						break;
					case LESS: op = BinaryComparisonOperator.LESS;
						break;
					case LESS_EQUAL: op = BinaryComparisonOperator.LESS_EQUAL;
						break;
					case NOT_EQUAL: op = BinaryComparisonOperator.NOT_EQUAL;
						break;
					default:
						throw new Error();
						//TODO better exception
					}
					final Term<?> term1 = bcf.term(0);
		
					final String name1;
					final boolean isAttribute1;
					
					if(term1 instanceof TimestampAttribute){
						name1 = ((TimestampAttribute)term1).name;
						isAttribute1 = true;
					}
					else if(term1 instanceof BasicTypeVariable){
						BasicTypeVariable var1 = (BasicTypeVariable) term1;
						if(var1.ID == BasicTypeID.TIMESTAMP){
							name1 = var1.name;
							isAttribute1 = false;			
						}
						else{
							return null;
						}
					}
					else{
						return null;
					}
					
					
					Term<?> term2 = bcf.term(1);
					if(term2 instanceof TimestampAttribute){
						String name2 = ((TimestampAttribute)term2).name;
						boolean isAttribute2 = true;
						long constant = 0;
						return new SimpleTemporalFormula(name1, isAttribute1, op, name2, isAttribute2, constant);
					}
					else if(term2 instanceof BasicTypeVariable){
						BasicTypeVariable var2 = (BasicTypeVariable) term2;
						if(var2.ID == BasicTypeID.TIMESTAMP){
							String name2 = var2.name;
							boolean isAttribute2 = false;
							long constant = 0;
							return new SimpleTemporalFormula(name1, isAttribute1, op, name2, isAttribute2, constant);
						}
						else{
							return null;
						}
					}
					else if(term2 instanceof TupleFunction){
						TupleFunction<?> function2 = (TupleFunction<?>) bcf.secondTerm();
						if(function2.ID == TupleFunctionID.PLUS | function2.ID == TupleFunctionID.MINUS){
							Term<?> arg1 = function2.arg(0);
							Term<?> arg2 = function2.arg(1);
							final String name2;
							final boolean isAttribute2;
							final long constant;
							if(arg1 instanceof TimestampAttribute){
								name2 = ((TimestampAttribute)arg1).name;
								isAttribute2 = true;
							}
							else if(arg1 instanceof BasicTypeVariable){
								BasicTypeVariable var2 = (BasicTypeVariable) arg1;
								if(var2.ID == BasicTypeID.TIMESTAMP){
									name2 = var2.name;
									isAttribute2 = false;
								}
								else{
									return null;
								}
							}
							else{
								return null;
							}
							if(arg2 instanceof DurationValue){
								long value = ((DurationValue) arg2).value;
								if(function2.ID == TupleFunctionID.MINUS){
									constant = -value;
								}
								else{
									constant = value;
								}
							}
							else{
								return null;
							}
							return new SimpleTemporalFormula(name1, isAttribute1, op, name2, isAttribute2, constant);
						}
						else if(bcf.operator == BinaryComparissonOperatorID.EQUAL){
							String[] arguments = new String[function2.arity()];
							boolean[] isAttribute = new boolean[arguments.length];
							
							final MinMax m;
							if(function2.ID == TupleFunctionID.LEAST){
								m = MinMax.MIN;
							}
							else if(function2.ID == TupleFunctionID.GREATEST){
								m = MinMax.MAX;
							}
							else{
								return null;
							}
							
							for(int argument = 0; argument < arguments.length; ++argument){
								if(function2.arg(argument) instanceof TimestampAttribute){
									arguments[argument] = ((TimestampAttribute)function2.arg(argument)).name;
									isAttribute[argument] = true;
								}
								else if(function2.arg(argument) instanceof BasicTypeVariable){
									BasicTypeVariable var2 = (BasicTypeVariable) function2.arg(argument);
									if(var2.ID == BasicTypeID.TIMESTAMP){
										arguments[argument] = var2.name;
										isAttribute[argument] = false;
									}
									else{
										return null;
									}
								}
								else{
									return null;
								}
							}							
							return new TemporalMinMax(name1, isAttribute1, m, arguments, isAttribute);
						}
						else{
							return null;
						}
					}	
					else{
						return null;
					}
				}
				else{
					return null;
				}
			}
		};
		
	public static final Filter<AtomicConditionFormula> preAdaptfocus = preAdaptFocus(localType);

	@Override
	public Adapter<AtomicConditionFormula, TemporalRelationFormula> localType() {
		return localType;
	}
	
	public static final Filter<AtomicConditionFormula> postAdaptfocus = new AbstractFilter<AtomicConditionFormula>() {
		
		@Override
		public boolean accepts(final AtomicConditionFormula element) {
			return element instanceof TemporalRelationFormula;
		}
		
	};

	@Override
	public Filter<AtomicConditionFormula> postAdaptFocus() {
		return postAdaptfocus;
	}

	@Override
	public boolean isConsistent(final AbstractCollection<AtomicConditionFormula> positiveLiterals, final AbstractCollection<AtomicConditionFormula> negativeLiterals) {
		@SuppressWarnings("unchecked")
		TemporalRelationFormula temporalRelationFormula = new ConjunctiveTemporalRelationFormula(new ConjunctiveTemporalRelationFormula((AbstractCollection<TemporalRelationFormula>)(AbstractCollection<?>)positiveLiterals),new NegatedTemporalRelationFormula(new DisjunctiveTemporalRelationFormula((AbstractCollection<TemporalRelationFormula>)(AbstractCollection<?>)negativeLiterals))).simpleForm().flaten();
		//TODO .simpleForm() could be omited if adapter aready guarantees simpleForm (same for flaten())
		
		
		//TODO reuse matrix, do not need to collect attribute names / variable names -> init function telling about attributes and variables
		
		final FixedArrayNavigableSet<String> attributeNames = temporalRelationFormula.AttributeNames();
		int numberOfAttributes = attributeNames.size();
		final FixedArrayNavigableSet<String> variableNames = temporalRelationFormula.VariableNames();
		int size = numberOfAttributes + variableNames.size();
		long[][] matrix = new long[size][size];
		
		for(ConjunctiveTemporalRelationFormula conjunction :temporalRelationFormula.tempDNFiteration()){
			@SuppressWarnings("unchecked")
			FixedArray<SimpleTemporalFormula> flatConjunction = (FixedArray<SimpleTemporalFormula>)(FixedArray<?>) conjunction.subFormulas();
			
			SimpleArithmeticAnalyzer.setDistanceMatrix(matrix, flatConjunction, 0, 0, numberOfAttributes, new FixedArrayNavigableSet<String>(false, false, new String[0]), attributeNames, new FixedArrayNavigableSet<String>(false, false, new String[0]), variableNames);
			SimpleFloydWarshall floydWarshall = new SimpleFloydWarshall(matrix, 0);
			floydWarshall.calculate();
			if(floydWarshall.isConsistent()){
				return true;
			}
		}
		return false;
	}

}
