package de.lmu.ifi.pms.cep.tsa.time.constraints;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TimestampAttribute;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.BinaryComparisonOperator;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.MinMax;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;

public class TemporalMinMax extends AtomicTemporalRelationFormula{
	public final String v;
	boolean isAttributeV;
	private final String[] arguments;
	private final boolean[] isAttribute;
	public final MinMax operator;
	
	/**
	 * This class represents  v=min(variables) and v=max(variables)
	 * @param m this decides whether the type is minimum oder maximum
	 * @param v 
	 * @param identifier
	 */
	public TemporalMinMax(final MinMax m, final String v, final AbstractCollection<String> arguments){
		this(v, m, arguments.toArray(new String[arguments.size()]));
	}

	public TemporalMinMax(final String v1, final boolean isAttribute1, final MinMax m, final String[] arguments, final boolean[] isAttribute){
		this.operator = m;
		this.v = v1;
		this.isAttributeV = isAttribute1;
		this.arguments = arguments;
		this.isAttribute = isAttribute;
		//TODO copy arguments?
		
		if(this.arguments.length != this.isAttribute.length){
			throw new IllegalArgumentException();
			//TODO better exception
		}

	}
	
	public TemporalMinMax(final String v, final MinMax m, final String... arguments){
		this(v, false, m, arguments, new boolean[arguments.length]);
	}

	public FixedArray<String> arguments() {
		return new FixedArray<String>(this.arguments);
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append("v = ");
		if(this.operator==MinMax.MIN){
			sb.append("min(");
		}else{
			sb.append("max(");
		}
		for(String s : this.arguments){
			sb.append(s);
			sb.append(",");
		}
		sb.setLength(sb.length()-",".length());
		sb.append(")");
	}
	
	@Override
	protected void VariableNames(final ArrayList<String> variables) {
		if(!this.isAttributeV){
			variables.add(this.v);
		}
		for(int arg = 0; arg < this.arguments.length; ++arg){
			if(!this.isAttribute[arg]){
				variables.add(this.arguments[arg]);
			}
		}
	}
	
	@Override
	protected void AttributeNames(final ArrayList<String> attributes) {
		if(this.isAttributeV){
			attributes.add(this.v);
		}
		for(int arg = 0; arg < this.arguments.length; ++arg){
			if(this.isAttribute[arg]){
				attributes.add(this.arguments[arg]);
			}
		}
	}
	
	@Override
	protected TemporalRelationFormula simpleForm(final boolean positivePolarity){
		if(positivePolarity){
			ArrayList<TemporalRelationFormula> leftSide = new ArrayList<TemporalRelationFormula>();
			ArrayList<TemporalRelationFormula> rightSide = new ArrayList<TemporalRelationFormula>();
			for(int arg = 0; arg < this.arguments.length; ++arg){
				String argument = this.arguments[arg];
				boolean isAttribute = this.isAttribute[arg];
				if(this.operator==MinMax.MIN){
					leftSide.add(new SimpleTemporalFormula(this.v, this.isAttributeV, BinaryComparisonOperator.LESS_EQUAL, argument, isAttribute, false));
					rightSide.add(new SimpleTemporalFormula(argument, isAttribute, BinaryComparisonOperator.LESS_EQUAL,this.v, this.isAttributeV, false));
				}else{assert this.operator==MinMax.MAX;
					leftSide.add(new SimpleTemporalFormula(argument, isAttribute, BinaryComparisonOperator.LESS_EQUAL, this.v, this.isAttributeV, false));
					rightSide.add(new SimpleTemporalFormula(this.v, this.isAttributeV, BinaryComparisonOperator.LESS_EQUAL,argument, isAttribute, false));
				}
			}
			return new ConjunctiveTemporalRelationFormula(new ConjunctiveTemporalRelationFormula(leftSide), new DisjunctiveTemporalRelationFormula(rightSide));
		}
		else{
			return this.simpleForm(true).simpleForm(false);
			//Does not produce loop, as this.simpleForm(true) does not contain a min or max anymore
		}
	}
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		for(int arg = 0; arg < this.arguments.length; ++arg){
			if(this.isAttribute[arg]){
				TimestampAttribute a = TimestampAttribute.create(this.arguments[arg]);
				if(filter.accepts(a)){
					terms.add(a);
				}
			}
			else{
				BasicTypeVariable v = BasicTypeVariable.create(this.arguments[arg], BasicTypeID.TIMESTAMP);
				if(filter.accepts(v)){
					terms.add(v);
				}
			}
		}
	}
	
	protected AtomicTerm<BasicType> arg1(){
		if(this.isAttributeV){
			return TimestampAttribute.create(this.v);
		}
		else{
			return BasicTypeVariable.create(this.v, BasicTypeID.TIMESTAMP);
		}
	}
	
	protected AtomicTerm<BasicType> arg(final int index){
		if(this.isAttribute[index]){
			return TimestampAttribute.create(this.arguments[index]);
		}
		else{
			return BasicTypeVariable.create(this.arguments[index], BasicTypeID.TIMESTAMP);
		}
	}
	
	@Override
	public TemporalRelationFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		Term<?> t1 = replacement.adapt(this.arg1());
		String v1;
		boolean isAttribute1;
		if(t1 instanceof TimestampAttribute){
			v1 = ((TimestampAttribute) t1).name;
			isAttribute1 = true;
		}
		else if(t1 instanceof BasicTypeVariable){
			v1 = ((BasicTypeVariable) t1).name;
			isAttribute1 = false;
		}
		else{
			throw new IllegalArgumentException();
			//TODO better exception
		}
		
		String[] arguments = new String[this.arguments.length];
		boolean[] isAttribute = new boolean[arguments.length];
		boolean changed = false;
		
		for(int argIndex = 0; argIndex < this.arguments.length; ++argIndex){
			Term<?> t2 = replacement.adapt(this.arg(argIndex));
			if(t2 instanceof TimestampAttribute){
				arguments[argIndex] = ((TimestampAttribute) t2).name;
				isAttribute[argIndex] = true;
			}
			else if(t2 instanceof BasicTypeVariable){
				arguments[argIndex] = ((BasicTypeVariable) t2).name;
				isAttribute[argIndex] = false;
			}
			else{
				throw new IllegalArgumentException();
				//TODO better exception
			}
			if(!this.arguments[argIndex].equals(arguments[argIndex]) | this.isAttribute[argIndex] != isAttribute[argIndex]){
				changed = true;
			}
		}
		
		
		
		if(v1.equals(this.v) & isAttribute1 == this.isAttributeV & !changed){
			return this;
		}
		else{
			return new TemporalMinMax(v1, isAttribute1, this.operator, arguments, isAttribute); 
		}
	}
	
	
	@Override
	public TemporalRelationFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		return this.replaceAtomicTerm(replacement);
	}
	
}