package de.lmu.ifi.pms.cep.tsa.time.constraints;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.fixed.empty.Empty;
import de.lmu.ifi.pms.java.util.filter.Filter;
import de.lmu.ifi.pms.java.util.tupel.Single;

public abstract class TemporalRelationFormula extends AtomicConditionFormula{
	
	@SuppressWarnings("hiding")
	public final ConditionFormulaID ID;
	@SuppressWarnings("hiding")
	public static final TemporalRelationFormula Bottom = new AtomicTemporalRelationFormula(ConditionFormulaID.BOTTOM, ConditionFormulaID.BOTTOM, true, true) {
		
		@Override
		public int tempDNFsize(){
			return 1;
		}
		
		@Override
		protected AbstractCollection<TemporalRelationFormula> tempDNFiteration(final boolean positivePolarity) {
			if(positivePolarity){
				return new Single<TemporalRelationFormula, TemporalRelationFormula>(this);
			}
			else{
				return Bottom.tempDNFiteration(true);
			}
		}
		
		@Override
		public AbstractCollection<ConjunctiveTemporalRelationFormula> tempDNFiteration() {
			return new Single<ConjunctiveTemporalRelationFormula,ConjunctiveTemporalRelationFormula>(new ConjunctiveTemporalRelationFormula());
		}
		
		@Override
		public int tempCNFsize() {
			return 0;
		}
		
		@Override
		protected AbstractCollection<TemporalRelationFormula> tempCNFiteration(final boolean positivePolarity) {
			if(positivePolarity){
				return Empty.instance();
			}
			else{
				return Bottom.tempCNFiteration(true);
			}
		}
		
		@Override
		public AbstractCollection<DisjunctiveTemporalRelationFormula> tempCNFiteration() {
			return new Single<DisjunctiveTemporalRelationFormula,DisjunctiveTemporalRelationFormula>(new DisjunctiveTemporalRelationFormula(this));
		}
		
		@Override
		protected void VariableNames(final ArrayList<String> variables) {
			
		}
		
		@Override
		protected void AttributeNames(final ArrayList<String> attributes) {
			
		}
		
		@Override
		protected TemporalRelationFormula simpleForm(final boolean positivePolarity) {
			return positivePolarity ? Bottom : Top;
		}
				
		@Override
		protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {}
		
		@Override
		public TemporalRelationFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
			return this;
		}
		
		
		@Override
		public TemporalRelationFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
			return this;
		}
		
		@Override
		public void writeTo(final StringBuilder sb, final int indentDepth) {
			indentTab(sb, indentDepth);
			sb.append(this.ID);
		}
		
	};
	@SuppressWarnings("hiding")
	public static final AtomicTemporalRelationFormula Top = new AtomicTemporalRelationFormula(ConditionFormulaID.TOP, ConditionFormulaID.TOP, true, true) {
		
		@Override
		public int tempDNFsize(){
			return 0;
		}
		
		@Override
		protected AbstractCollection<TemporalRelationFormula> tempDNFiteration(final boolean positivePolarity) {
			if(positivePolarity){
				return Empty.instance();
			}
			else{
				return Top.tempDNFiteration(true);
			}
		}
		
		@Override
		public AbstractCollection<ConjunctiveTemporalRelationFormula> tempDNFiteration() {
			return Empty.instance();
		}
		
		@Override
		public int tempCNFsize() {
			return 1;
		}
		
		@Override
		protected AbstractCollection<TemporalRelationFormula> tempCNFiteration(final boolean positivePolarity) {
			if(positivePolarity){
				return new Single<TemporalRelationFormula, TemporalRelationFormula>(this);
			}
			else{
				return Top.tempCNFiteration(true);
			}
		}
		
		@Override
		public AbstractCollection<DisjunctiveTemporalRelationFormula> tempCNFiteration() {
			return new Single<DisjunctiveTemporalRelationFormula,DisjunctiveTemporalRelationFormula>(new DisjunctiveTemporalRelationFormula());
		}
		
		@Override
		protected void VariableNames(final ArrayList<String> variables) {
			
		}
		
		@Override
		protected void AttributeNames(final ArrayList<String> attributes) {
			
		}
		
		@Override
		protected TemporalRelationFormula simpleForm(final boolean positivePolarity) {
			return positivePolarity ? Top : TemporalRelationFormula.Bottom;
		}
		
		@Override
		protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {}
		
		@Override
		public TemporalRelationFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
			return this;
		}
		
		@Override
		public TemporalRelationFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
			return this;
		}
		
		@Override
		public void writeTo(final StringBuilder sb, final int indentDepth) {
			indentTab(sb, indentDepth);
			sb.append(this.ID);
		}
		
	};
	
	TemporalRelationFormula(final ConditionFormulaID ID, final boolean isNormalized, final boolean isFlatened) {
		super(isNormalized, isFlatened);
		this.ID = ID;
	}
	
	TemporalRelationFormula(final ConditionFormulaID superID, final ConditionFormulaID ID, final boolean isNormalized, final boolean isFlatened) {
		super(superID, isNormalized, isFlatened);
		this.ID = ID;
	}
	
	@Override
	public FixedArrayNavigableSet<String> VariableNames(){
		ArrayList<String> variableNames = new ArrayList<String>();
		this.VariableNames(variableNames);		
		return new FixedArrayNavigableSet<String>(false, variableNames.toArray(new String[variableNames.size()]));
	}
	
	protected abstract void VariableNames(ArrayList<String> variables);
	
	@Override
	public FixedArrayNavigableSet<String> AttributeNames(){
		ArrayList<String> attributeNames = new ArrayList<String>();
		this.AttributeNames(attributeNames);		
		return new FixedArrayNavigableSet<String>(false, attributeNames.toArray(new String[attributeNames.size()]));
	}
	
	protected abstract void AttributeNames(final ArrayList<String> attributes);
	
	@Override
	public TemporalRelationFormula normalize() {
		if(this.isNormalized){
			return this;
		}
		else{
			return this.normalizeImpl();
		}
	}
	
	@Override
	protected abstract TemporalRelationFormula normalizeImpl() ;
	
	@Override
	public TemporalRelationFormula flaten() {
		if(this.isFlatened){
			return this;
		}
		else if(this.isNormalized){
			return this.flatenImpl();
		}
		else{
			return this.normalize().flaten();
		}
	}
	
	@Override
	protected abstract TemporalRelationFormula flatenImpl();
	
	public TemporalRelationFormula simpleForm(){
		return this.simpleForm(true).flaten();
	}
	
	protected abstract TemporalRelationFormula simpleForm(boolean positivePolarity);
	
	public AbstractCollection<ConjunctiveTemporalRelationFormula> tempDNFiteration(){
		return new CollectionAdapter<TemporalRelationFormula, ConjunctiveTemporalRelationFormula>(this.tempDNFiteration(true)){

			@Override
			public ConjunctiveTemporalRelationFormula adapt(final TemporalRelationFormula source) {
				TemporalRelationFormula flatenedSource = source.flaten();
				assert flatenedSource.isFlatened;
				switch (flatenedSource.ID) {
				case ATOMIC: case NEGATED:
					return new ConjunctiveTemporalRelationFormula(source);
				default:
					return (ConjunctiveTemporalRelationFormula) flatenedSource;
				}
			}
			
		};
	}
	
	public abstract int tempDNFsize();

	protected abstract AbstractCollection<TemporalRelationFormula> tempDNFiteration(boolean positivePolarity);
	
	public AbstractCollection<DisjunctiveTemporalRelationFormula> tempCNFiteration(){
		return new CollectionAdapter<TemporalRelationFormula, DisjunctiveTemporalRelationFormula>(this.tempCNFiteration(true)){

			@Override
			public DisjunctiveTemporalRelationFormula adapt(final TemporalRelationFormula source) {
				TemporalRelationFormula flatenedSource = source.flaten();
				switch (flatenedSource.ID) {
				case ATOMIC: case NEGATED:
					return new DisjunctiveTemporalRelationFormula(source);
				default:
					return (DisjunctiveTemporalRelationFormula) flatenedSource;				
				}
			}
			
		};
	}
	
	public abstract int tempCNFsize();

	protected abstract AbstractCollection<TemporalRelationFormula> tempCNFiteration(boolean positivePolarity);
	
	@Override
	public abstract TemporalRelationFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement);
	
	@Override
	public abstract TemporalRelationFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement);
	
	@Override
	protected abstract void collect(Filter<Term<?>> filter, ArrayList<Term<?>> terms);
	
	
}
