package de.lmu.ifi.pms.java.lang.iteration;

import java.util.Iterator;

import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.Collection;



public class ComposedIterator<Type> extends AbstractIterator<Type> {
	
	private final Iterator<Type>[] iterators;
	
	int index = 0;
//	int returnedValues = 0;
	
	public ComposedIterator(final Iterator<Type>... iterators){
		this.iterators = iterators;
	}
	
	@SuppressWarnings("unchecked")
	public ComposedIterator(final Collection<? extends Iterator<Type>> iterators){
		this(iterators.toArray(new Iterator[iterators.size()]));
	}
	
	public ComposedIterator(final Collection<? extends Iterable<? extends Type>> iterations, final boolean javaIsStupid){
		this(new CollectionAdapter<Iterable<? extends Type>, Iterator<Type>>(iterations){
			@SuppressWarnings("unchecked")
			@Override
			public Iterator<Type> adapt(final Iterable<? extends Type> source) {
				return (Iterator<Type>) source.iterator();
			}
		});
	}
	
	public ComposedIterator(final Iterable<Type>... iterables){
		this(convert(iterables));
	}
	
	private static <Type> Iterator<Type>[] convert(final Iterable<Type>[] iterables){
		@SuppressWarnings("unchecked")
		final
		Iterator<Type>[] iterators = new Iterator[iterables.length];
		for(int i = 0; i < iterables.length; ++i){
			iterators[i] = iterables[i].iterator();
		}
		return iterators;
	}

	@Override
	public boolean hasNext() {
		while(this.index < this.iterators.length && !this.iterators[this.index].hasNext()){
			++this.index;
//			this.returnedValues = 0;
		}
		return this.index < this.iterators.length;
	}

	@Override
	public Type next() {
		this.checkNext();
		return this.iterators[this.index].next();
	}

}
