package de.lmu.ifi.pms.java.util.adaption;

import de.lmu.ifi.pms.java.lang.iteration.FixedIterator;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.fixed.AbstractFixedCollection;
import de.lmu.ifi.pms.java.util.collection.fixed.wrapper.FiniteIteratorWrapper;

public abstract class CollectionAdapter<SourceType, DestinationType> extends
		AbstractFixedCollection<DestinationType> implements Adapter<SourceType, DestinationType>{
	
	protected final int elementCount;
	protected final Collection<? extends SourceType> source;

	public CollectionAdapter(final Iterable<? extends SourceType> source, final int elementCount) {
		super();
		this.elementCount = elementCount;
		this.source = new AbstractFixedCollection<SourceType>() {

			@Override
			public FixedIterator<SourceType> iterator() {
				return new FiniteIteratorWrapper<SourceType>(source.iterator(), elementCount);
			}

			@Override
			public int size() {
				return elementCount;
				//FIXME iterator maight stop after less than element steps
			}
			
		};
	}
	
	public CollectionAdapter(final Collection<? extends SourceType> c){
		this.elementCount = c.size();
		this.source = c;
	}
	
	@Override
	public FixedIterator<DestinationType> iterator() {
		return new FixedIteratorAdapter<SourceType, DestinationType>(this.source.iterator()) {
			@Override
			public DestinationType adapt(final SourceType source) {
				return CollectionAdapter.this.adapt(source);
			};
		};
	}

	@Override
	public int size() {
		return this.elementCount;
	}
	
	@Override
	public abstract DestinationType adapt(SourceType source);

}
