/**
 * 
 */
package de.lmu.ifi.pms.java.util.collection.array.extendible;

import java.lang.ref.PhantomReference;
import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.SoftReference;
import java.lang.ref.WeakReference;
import java.util.WeakHashMap;

import de.lmu.ifi.pms.java.util.collection.array.AbstractExtendibleArray;


/**
 * @author Simon Matthias Brodt
 *
 */
public abstract class WeakIArrayQueue<Type, ReferenceType extends Reference<Type>> extends AbstractExtendibleArray<Type>{

	private final AbstractExtendibleArray<ReferenceType> references;
	
	private final ReferenceQueue<Type> referenceQueue;
	
	protected final Class<? super Type> ensuredType;
	
	/**
	 * @param baseCapacity
	 * @param size
	 * @param type
	 */
	public WeakIArrayQueue(Class<? super Type> ensuredType, ReferenceQueue<Type> referenceQueue, AbstractExtendibleArray<ReferenceType> references) {
		super();
		this.references = references;
		this.ensuredType = ensuredType;
		this.referenceQueue = referenceQueue;
	}
	
	protected abstract ReferenceType newReferenceTo(Type value, ReferenceQueue<Type> referenceQueue);
	
	protected abstract WeakIArrayQueue<Type, ReferenceType> newWeakExtendibleArray(Class<? super Type> ensuredType, ReferenceQueue<Type> referenceQueue, AbstractExtendibleArray<ReferenceType> references);
	/**Should not affect size
	 * 
	 * @param reference
	 */
	protected abstract void enqueueAction(ReferenceType reference);
	
	protected final void clearQueue(){
		ReferenceType reference = (ReferenceType) referenceQueue.poll();
		while(reference != null){
			enqueueAction(reference);
			reference = (ReferenceType) referenceQueue.poll();
		}
	}

	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractExtendibleArray#addLast(java.lang.Object)
	 */
	@Override
	public void addLast(Type value) {
		 ensuredType.cast(value);
		 clearQueue();
		 if(value == null){
			 references.addLast(null);
		 }
		 else{
			 references. addLast(newReferenceTo(value, referenceQueue));
		 }
	}

	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractExtendibleArray#insert(int, java.lang.Object)
	 */
	@Override
	public void insert(int index, Type value) {
		 ensuredType.cast(value);
		 if(value == null){
			 references.insert(index, null);
		 }
		 else{
			 references.insert(index, newReferenceTo(value, referenceQueue));
		 }
	}

	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractExtendibleArray#remove(int)
	 */
	@Override
	public void remove(int index) {
		 clearQueue();
		 references.remove(index);
	}

	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractExtendibleArray#removeLast()
	 */
	@Override
	public void removeLast() {
		clearQueue();
		references.removeLast();
	}

	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractExtendibleArray#subArray(int, int)
	 */
	@Override
	public WeakIArrayQueue<Type, ReferenceType> subArray(int offset, int length) {
		return newWeakExtendibleArray(ensuredType, referenceQueue, references.subArray(offset, length));
	}

	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractModifyableArray#insertIndexDropLast(int, java.lang.Object)
	 */
	@Override
	public void insertIndexDropLast(int index, Type value) {
		 ensuredType.cast(value);
		 if(value == null){
			 references.insertIndexDropLast(index, null);
		 }
		 else{
			 references.insertIndexDropLast(index, newReferenceTo(value, referenceQueue));
		 }
	}

	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractModifyableArray#set(int, java.lang.Object)
	 */
	@Override
	public void set(int index, Type value) {
		ensuredType.cast(value);
		clearQueue();
		if(value == null){
			references.set(index, null);
		}
		else{
			references.set(index, newReferenceTo(value, referenceQueue));
		}
	}

	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractArray#ensuredType()
	 */
	@Override
	public Class<? super Type> ensuredType() {
		return ensuredType;
	}

	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractArray#get(int)
	 */
	@Override
	public Type get(int index) {
		clearQueue();
		ReferenceType reference = references.get(index);
		if(reference == null){
			return null;
		}
		else{
			return (Type) reference.get();
		}
	}

	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractArray#minIndexInc()
	 */
	@Override
	public int minIndexInc() {
		return 0;
	}

	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractArray#size()
	 */
	@Override
	public int size() {
		clearQueue();
		return references.size();
	}

}
