package de.lmu.ifi.pms.java.util.collection.array.fixed;

import static de.lmu.ifi.pms.java.util.collection.array.Arrays.*;

import java.io.Serializable;
import java.util.AbstractCollection;
import java.util.Iterator;

import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.Root;
import org.simpleframework.xml.core.Persister;

import de.lmu.ifi.pms.development.UnimplementedMethodError;
import de.lmu.ifi.pms.java.lang.iteration.FixedIterator;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.AbstractFixedArray;
import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedDeque;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedList;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.FromJavaCollectionWrapper;
import de.lmu.ifi.pms.java.util.string.StringBuilder;


@Root(name="list")
public class FixedArray<Type> extends AbstractFixedArray<Type> implements FixedList<Type>, FixedDeque<Type>, Serializable{	
	
	protected Type[] vector;
	
	@ElementList(entry="element",inline=true, required=false, empty=false)
	protected java.util.Collection<Type> getVector(){
		return new java.util.ArrayList<Type>(java.util.Arrays.asList(this.vector));
	}
	
	@SuppressWarnings("unchecked")
	public FixedArray(final boolean bodyOnly){
		if(!bodyOnly){
			this.vector = (Type[]) new Object[0];
		}
	}
	
	public FixedArray(final FixedArray<Type> vector){
		this.fillBody(vector);
	}
	
	public FixedArray(final boolean copyArray, final Type... elements){
		this.fillBody(copyArray, elements);
	}
	
	@SuppressWarnings("unchecked")
	public FixedArray(final Type element){
		this(false, element);
	}
	
	public FixedArray(final Type... elements){
		this(true,elements);
	}
	
	public FixedArray(final Collection<? extends Type> collection){
		this.fillBody(collection);
	}
	
	public FixedArray(@ElementList(entry="element",inline=true, required=false, empty=false) final java.util.Collection<Type> collection){
		this(new FromJavaCollectionWrapper<Type>(collection));
	}
	
	public FixedArray(final Iterable<Type> collection, final int elementCount){
		this.fillBody(collection, elementCount);
	}
	
	public final boolean hasEmptyBody(){
		return this.vector == null;
	}
	
	protected final void checkFillBody(){
		if(!this.hasEmptyBody()){
			throw new IllegalStateException("Already filled!");
		}
	}
	
	public void fillBody(final FixedArray<Type> vector){
		this.checkFillBody();
		this.vector = vector.vector;
	}
	
	@SuppressWarnings("unchecked")
	public void fillBody(final boolean copyArray, final Type... elements){
		this.checkFillBody();
		if(copyArray){
			this.vector = (Type[])getArray(elements.getClass().getComponentType(), elements.length);
		}
		else{
			this.vector = elements;
		}
		for(int i=0; i < elements.length; ++i){
			this.vector[i] = elements[i];
		}
	}
	
	@SuppressWarnings("unchecked")
	public void fillBody(final Collection<? extends Type> collection){
		this.checkFillBody();
		this.vector = (Type[])collection.toArray();
	}
	
	@SuppressWarnings("unchecked")
	public void fillBody(final Iterable<Type> collection, final int elementCount){
		this.checkFillBody();
		this.vector = (Type[]) new Object[elementCount];
		Iterator<Type> iterator = collection.iterator();
		for(int i = 0; i < elementCount; ++i){
			this.vector[i] = iterator.next();
		}
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractArray#ensuredType()
	 */
	@Override
	public Class<?> ensuredType() { 
		return this.vector.getClass().getComponentType();
	}
	
	@Override
	public Type get(final int index){
		this.checkIndex(index);
		return this.vector[index];
	}
	
	@Override
	public final Type first(){
		return this.get(0);
	}
	
	@Override
	public final Type getFirst() {
		return this.first();
	}
	
	@Override
	public final Type peekFirst() {
		if(this.isEmpty()){
			return null;
		}
		else{
			return this.getFirst();
		}
	}

	@Override
	public final Type peek() {
		return this.peekFirst();
	}

	@Override
	public final Type element() {
		return this.getFirst();
	}

	@Override
	public final Type last(){
		return this.get(this.size() - 1);
	}
	
	@Override
	public final Type getLast() {
		return this.last();
	}
	
	@Override
	public Type peekLast() {
		if(this.isEmpty()){
			return null;
		}
		else{
			return this.getLast();
		}
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractArray#minIndexInc()
	 */
	@Override
	public int minIndexInc() {
		return 0;
	}

	@Override
	public int size(){
		return this.vector.length;
	}	
	
	protected int getOffset(){
		return 0;
	}	
	
	@Override
	public FixedArray<Type> subArray(final int offset, final int length){
		return new FixedSubArray<Type>(this, offset, length);
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractFinalArray#iterator()
	 */
	@Override
	public FixedArrayIterator<Type, FixedArray<Type>> iterator() {
		return this.listIterator();
	}

	@Override
	public FixedArrayIterator<Type, FixedArray<Type>> listIterator() {
		return this.listIterator(0);
	}		

	@Override
	public FixedArrayIterator<Type, FixedArray<Type>> listIterator(final int index) {
		return new FixedArrayIterator<Type, FixedArray<Type>>(this, index);
	}	
	
	@Override
	public FixedIterator<Type> descendingIterator() {
		throw new UnimplementedMethodError();
		 //FIXME implement
	}

	@Override
	public int indexOf(final Object o) {
		FixedArrayIterator<Type, FixedArray<Type>> iterator = this.listIterator();
		Type t;
		while(iterator.hasNext()){
			t = iterator.next();
			if(o == null ? t == null : o.equals(t)){
				return iterator.nextIndex()-1;
			}
		}
		return -1;
	}

	@Override
	public int lastIndexOf(final Object o) {
		FixedArrayIterator<Type, FixedArray<Type>> iterator = this.listIterator(this.size());
		Type t;
		while(iterator.hasPrevious()){
			t = iterator.previous();
			if(o == null ? t == null : o.equals(t)){
				return iterator.nextIndex()-1;
			}
		}
		return -1;
	}
	
	@Override
	public <T> void copyTo(final int sourceStart, final T[] destination, final int destinationStart, final int count) {
		System.arraycopy(this.vector, sourceStart, destination, destinationStart, count);
	}
	
	@Override
	public void copyTo(final int sourceStart, final int[] destination,final int destinationStart, final int count) {
		System.arraycopy(this.vector, sourceStart, destination, destinationStart, count);
	}
	
	@Override
	public AbstractCollection<Type> javaCollection() {
		return super.javaCollection();
	}

	@Override
	public FixedArray<Type> subList(final int fromIndex, final int toIndex) {
		return this.subArray(fromIndex, toIndex - fromIndex);
	}

	private static class FixedSubArray<Type> extends FixedArray<Type> {
		private final int offset;
		private final int length;

		public FixedSubArray(final FixedArray<Type> vector, final int offset, final int length) {
			super(vector);
			if(offset + length < vector.maxIndexExc()){
				this.offset = offset + vector.getOffset();
				this.length = length;
			}
			else{
				throw new IndexOutOfBoundsException();
			}
		}

		@Override
		public Type get(final int index) {
			this.checkIndex(index);
			return this.vector[index + this.offset];
		}

		@Override
		public int size() {
			return this.length;
		}

		@Override
		protected int getOffset() {
			return this.offset;
		}
		
		@Override
		public <T> void copyTo(final int sourceStart, final T[] destination, final int destinationStart, final int count) {
			copy(this.vector, this.offset + sourceStart, destination, destinationStart, count);
		}
				
	}
	
	public static class FixedArrayIterator<Type, ArrayType extends FixedArray<Type>> extends GenericFixedArrayIterator<Type, ArrayType>{				
		
		public FixedArrayIterator(final ArrayType vector) {
			this(vector, 0);
		}

		public FixedArrayIterator(final ArrayType vector, final int index) {
			this(vector, vector.minIndexInc(), vector.maxIndexExc(), index);
		}

		public FixedArrayIterator(final ArrayType array, final int firstIndexInclusive, final int lastIndexExclusive, final int position) {
			super(array, firstIndexInclusive, lastIndexExclusive, position);
		}

	}
	
	public static <Type> FixedArray<Type> concat(final FixedArray<Type>... arrays){
		return concat(new FixedArray<FixedArray<Type>>(false, arrays));
	}
	
	public static <Type> FixedArray<Type> concat(final Collection<FixedArray<Type>> arrays){
		if(arrays.size() > 0){
			int newSize = 0;
			@SuppressWarnings("unchecked")
			FixedArray<Type>[] aarrays = new FixedArray[arrays.size()];
			Class<?> ensuredType = arrays.iterator().next().ensuredType();//FIXME use most generalType
			int index = 0;
			for(FixedArray<Type> array: arrays){
				newSize += array.size();
				aarrays[index] = array;
				if(!ensuredType.isAssignableFrom(array.ensuredType())){
					ensuredType = array.ensuredType();
				}
				++index;
			}
			@SuppressWarnings("unchecked")
			Type[] newArray = (Type[])Arrays.getArray(ensuredType, newSize);
			newSize = 0;
			for(FixedArray<Type> array: aarrays){
				array.copyTo(0, newArray, newSize, array.size());
				newSize += array.size();				
			}
			return new FixedArray<Type>(false, newArray);
		}
		else{
			return new FixedArray<Type>(false);
		}
	}
	
	public static void main(final String[] args) throws Exception{
		FixedArray<String> array = new FixedArray<String>("a");
		Persister p = new Persister();
		StringBuilder sb = new StringBuilder();
		p.write(array, sb);
		String serialized_array = sb.toString();
		System.out.println(serialized_array);
		
		@SuppressWarnings("unchecked")
		FixedArray<String> deserialized_array = p.read(array.getClass(), serialized_array);
		
		System.out.println(deserialized_array);
	}
		
}
