package de.lmu.ifi.pms.java.util.collection.extendible;

import de.lmu.ifi.pms.java.lang.iteration.ExtendibleIterator;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.FromJavaCollectionWrapper;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.ToJavaCollectionExtendibleWrapper;


public abstract class AbstractExtendibleCollection<Type> extends AbstractCollection<Type> implements ExtendibleCollection<Type> {

	@Override
	public abstract boolean add(Type value);
	
	@Override
	public boolean addAll(final java.util.Collection<? extends Type> c) {
		return this.addAll(new FromJavaCollectionWrapper<Type>(c));
	}

	@Override
	public boolean addAll(final Collection<? extends Type> c) {
		boolean modified = false;
		for(Type t: c){
			if(this.add(t)){
				modified = true;
			}
		}
		return modified;
	}

	@Override
	public void clear() {
		ExtendibleIterator<Type> e = this.iterator();
		while (e.hasNext()) {
		    e.next();
		    e.remove();
		}
	}

	@Override
	public boolean remove(final Object o) {
		ExtendibleIterator<Type> e = this.iterator();
		if (o==null) {
		    while (e.hasNext()) {
			if (e.next()==null) {
			    e.remove();
			    return true;
			}
		    }
		} else {
		    while (e.hasNext()) {
			if (o.equals(e.next())) {
			    e.remove();
			    return true;
			}
		    }
		}
		return false;
	}
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public boolean removeAll(final java.util.Collection<?> c) {
		return this.removeAll(new FromJavaCollectionWrapper(c));
	}

	@Override
	public boolean removeAll(final Collection<?> c) {
		boolean modified = false;
		ExtendibleIterator<Type> e = this.iterator();
		while (e.hasNext()) {
		    if (c.contains(e.next())) {
			e.remove();
			modified = true;
		    }
		}
		return modified;
	}
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public boolean retainAll(final java.util.Collection<?> c) {
		return this.retainAll(new FromJavaCollectionWrapper(c));
	}

	@Override
	public boolean retainAll(final Collection<?> c) {
		boolean modified = false;
		ExtendibleIterator<Type> e = this.iterator();
		while (e.hasNext()) {
		    if (!c.contains(e.next())) {
			e.remove();
			modified = true;
		    }
		}
		return modified;
	}
	
	

	@Override
	public abstract ExtendibleIterator<Type> iterator();
	
	@Override
	public java.util.AbstractCollection<Type> javaCollection() {
		return new ToJavaCollectionExtendibleWrapper<Type>(this);
	}

}
