/**
 * 
 */
package de.lmu.ifi.pms.java.util.collection.extendible.list;

import de.lmu.ifi.pms.java.lang.iteration.AbstractExtendibleIterator;
import de.lmu.ifi.pms.java.util.collection.extendible.AbstractExtendibleCollection;

/**
 * @author Simon Matthias Brodt
 *
 */
public abstract class AbstractSingleLinkedList<Type, ListEntryType extends SingleLinkedListEntry<Type, ListEntryType>> extends AbstractExtendibleCollection<Type>{

	final ListEntryType head;
	
	protected int size;
	
	public AbstractSingleLinkedList() {
		super();
		this.head = newListHead(this);
		this.size = 0;
	}

	public abstract ListEntryType newListHead(AbstractSingleLinkedList<Type, ListEntryType> linkingList);
	
	public abstract ListEntryType newUnlinkedEntry(Type value);
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.AbstractCollection#size()
	 */
	@Override
	public int size() {
		return this.size;
	}
	
	@Override
	public boolean add(Type value) {
		return addLast(value);
	}
	
	public boolean addLast(Type value) {
		insertBefore(this.head, newUnlinkedEntry(value));
		return true;
	}
	
	public Type last(){
		return value(previous(this.head));
	}
	
	public void removeLast(){
		removePrevious(this.head);
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.extendible.AbstractExtendibleCollection#clear()
	 */
	@Override
	public void clear() {
		removePreceeding(this.head);
	}

	/* (non-Javadoc)
	 * @see extension.java.util.collection.extendible.AbstractExtendibleCollection#iterator()
	 */
	@Override
	public AbstractExtendibleIterator<Type> iterator() {
		return new ListIterator<Type, ListEntryType, AbstractSingleLinkedList<Type, ListEntryType>>(this);
	}
	
	protected ListEntryType previous(ListEntryType entry){
		return entry.previous();
	}
	
	protected void removePrevious(ListEntryType entry){
		entry.removePrevious();
	}
	
	protected void insertBefore(ListEntryType entry, ListEntryType newEntry){
		newEntry.insertBefore(entry);
	}
	
	protected void removePreceeding(ListEntryType entry){
		entry.removePreceeding();
	}
	
	protected Type value(ListEntryType entry){
		return entry.value();
	}
	
	protected void setValue(ListEntryType entry, Type newValue){
		entry.setValue(newValue);
	}
	
	public static class ListIterator<Type, ListEntryType extends SingleLinkedListEntry<Type, ListEntryType>, ListType extends AbstractSingleLinkedList<Type, ListEntryType>> extends AbstractExtendibleIterator<Type>{
		
		protected final ListType list;
		protected int remainingSize;
		protected ListEntryType pastEntry;
		protected ListEntryType presentEntry;

		public ListIterator(ListType list) {
			super();
			this.list = list;
			this.remainingSize = list.size;
			this.presentEntry = list.head;
		}

		/* (non-Javadoc)
		 * @see extension.java.lang.iteration.AbstractModifyableIterator#hasNext()
		 */
		@Override
		public boolean hasNext() {
			return this.remainingSize > 0;
		}

		/* (non-Javadoc)
		 * @see extension.java.lang.iteration.AbstractModifyableIterator#next()
		 */
		@Override
		public Type next() {
			checkNext();
			this.pastEntry = this.presentEntry;
			this.presentEntry = this.list.previous(this.presentEntry);
			--this.remainingSize;
			return this.presentEntry.value();
		}

		/* (non-Javadoc)
		 * @see extension.java.lang.iteration.AbstractExtendibleIterator#insert(java.lang.Object)
		 */
		@Override
		public void insert(Type value) {
			this.list.insertBefore(this.presentEntry, this.list.newUnlinkedEntry(value));
		}

		/* (non-Javadoc)
		 * @see extension.java.lang.iteration.AbstractExtendibleIterator#remove()
		 */
		@Override
		public void remove() {
			if(this.pastEntry == null){
				throw new IllegalStateException();//TODO better exception
			}
			else{
			 try{
				 this.list.removePrevious(this.pastEntry);
				 this.presentEntry = this.pastEntry;
				 this.pastEntry = null;
			 }
			 catch (NullPointerException e) {
				throw new UnsupportedOperationException();
			 }
			}
		}

		/* (non-Javadoc)
		 * @see extension.java.lang.iteration.AbstractModifyableIterator#replace(java.lang.Object)
		 */
		@Override
		public void replace(Type newValue) {
			 this.presentEntry.setValue(newValue);
		}
		
		
		
		
	}
	
}
