package de.lmu.ifi.pms.java.util.filter;

import de.lmu.ifi.pms.java.util.collection.Collection;

public abstract class AbstractFilter<Type> implements Filter<Type> {

	@Override
	public abstract boolean accepts(Type element);

	@Override
	@SuppressWarnings({ "unchecked", "cast" })
	public boolean accepts(final int element) {
		return this.accepts((Type)(Object)element);
	}

	@Override
	@SuppressWarnings({ "unchecked", "cast" })
	public boolean accepts(final long element) {
		return this.accepts((Type)(Object)element);
	}

	@Override
	@SuppressWarnings({ "unchecked", "cast" })
	public boolean accepts(final float element) {
		return this.accepts((Type)(Object)element);
	}

	@Override
	@SuppressWarnings({ "unchecked", "cast" })
	public boolean accepts(final double element) {
		return this.accepts((Type)(Object)element);
	}

	@Override
	@SuppressWarnings({ "unchecked", "cast" })
	public boolean accepts(final byte element) {
		return this.accepts((Type)(Object)element);
	}

	@Override
	@SuppressWarnings({ "unchecked", "cast" })
	public boolean accepts(final short element) {
		return this.accepts((Type)(Object)element);
	}

	@Override
	@SuppressWarnings({ "unchecked", "cast" })
	public boolean accepts(final char element) {
		return this.accepts((Type)(Object)element);
	}

	@Override
	@SuppressWarnings({ "unchecked", "cast" })
	public boolean accepts(final boolean element) {
		return this.accepts((Type)(Object)element);
	}
	
	@SuppressWarnings("rawtypes")
	private static final AbstractFilter allFilter = new AbstractFilter() {
		
		@Override
		public boolean accepts(final Object element) {
			return true;
		}
		
	};
	
	@SuppressWarnings("unchecked")
	public static <Type> AbstractFilter<Type> allFilter(){
		return allFilter;
	}
	
	@SuppressWarnings("rawtypes")
	private static final AbstractFilter noneFilter = new AbstractFilter() {
		
		@Override
		public boolean accepts(final Object element) {
			return false;
		}
		
	};
	
	@SuppressWarnings("unchecked")
	public static <Type> AbstractFilter<Type> noneFilter(){
		return noneFilter;
	}
	
	public static <Type> AbstractFilter<Type> andFilter(final Collection<? extends AbstractFilter<Type>> filters){
		return new AbstractFilter<Type>() {
			
			@Override
			public boolean accepts(final Type element) {
				for(AbstractFilter<Type> f: filters){
					if(!f.accepts(element)){
						return false;
					}
				}
				return true;
			}
			
		};
	}
	
	public static <Type> AbstractFilter<Type> andFilter(final AbstractFilter<Type>... filters){
		return new AbstractFilter<Type>() {
			
			@Override
			public boolean accepts(final Type element) {
				for(AbstractFilter<Type> f: filters){
					if(!f.accepts(element)){
						return false;
					}
				}
				return true;
			}
			
		};
	}
	
	public static <Type> AbstractFilter<Type> orFilter(final Collection<? extends AbstractFilter<Type>> filters){
		return new AbstractFilter<Type>() {
			
			@Override
			public boolean accepts(final Type element) {
				for(AbstractFilter<Type> f: filters){
					if(f.accepts(element)){
						return true;
					}
				}
				return false;
			}
			
		};
	}
	
	public static <Type> AbstractFilter<Type> orFilter(final AbstractFilter<Type>... filters){
		return new AbstractFilter<Type>() {
			
			@Override
			public boolean accepts(final Type element) {
				for(AbstractFilter<Type> f: filters){
					if(f.accepts(element)){
						return true;
					}
				}
				return false;
			}
			
		};
	}
	
	public static <Type> AbstractFilter<Type> notFilter(final AbstractFilter<Type> f){
		return new AbstractFilter<Type>() {
			
			@Override
			public boolean accepts(final Type element) {
				return !f.accepts(element);
			}
			
		};
	}
	
	public static <Type> AbstractFilter<Type> intersectionFilter(final Collection<Type> c){
		return new AbstractFilter<Type>() {
			
			@Override
			public boolean accepts(final Type element) {
				return c.contains(element);
			}
			
		};
	}
	
	public static <Type> AbstractFilter<Type> differenceFilter(final Collection<Type> c){
		return new AbstractFilter<Type>() {
			
			@Override
			public boolean accepts(final Type element) {
				return !c.contains(element);
			}
			
		};
	}


}
