/**
 * 
 */
package de.lmu.ifi.pms.java.util.map;

import de.lmu.ifi.pms.java.lang.iteration.Iterator;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.set.NavigableSet;
import de.lmu.ifi.pms.java.util.map.javaWrapper.ToJavaNavigableMapWrapper;

/**
 * @author Simon Matthias Brodt
 *
 */
public abstract class AbstractNavigableMap<KeyType, ValueType, EntryType extends Entry<KeyType, ValueType>>
	extends AbstractSortedMap<KeyType, ValueType, EntryType>
	implements NavigableMap<KeyType, ValueType, EntryType> {

	@Override
	public abstract NavigableSet<KeyType> keys();

	@Override
	public abstract EntryType entry(KeyType key);

	@Override
	public abstract ValueType get(KeyType key);

	@Override
	public abstract boolean has(KeyType key);

	@Override
	public abstract Collection<ValueType> values();

	@Override
	public abstract Iterator<EntryType> iterator();

	@Override
	public abstract int size();

	@Override
	public abstract Entry<KeyType, ValueType> ceilingEntry(KeyType key);

	@Deprecated
	@Override
	public final NavigableSet<KeyType> descendingKeySet(){
		return this.keys().descendingSet();
	}

	@Override
	public abstract NavigableMap<KeyType, ValueType, ?> descendingMap();

	@Override
	public abstract Entry<KeyType, ValueType> firstEntry();

	@Override
	public abstract Entry<KeyType, ValueType> floorEntry(KeyType key);

	@Override
	public abstract KeyType floorKey(KeyType key);

	@Override
	public abstract NavigableMap<KeyType, ValueType, ?> headMap(KeyType toKey,
			boolean inclusive);

	@Override
	public NavigableMap<KeyType, ValueType, ?> headMap(final KeyType toKey){
		return this.headMap(toKey, false);
	}

	@Override
	public abstract Entry<KeyType, ValueType> higherEntry(KeyType key);

	@Override
	public abstract Entry<KeyType, ValueType> lastEntry();

	@Override
	public abstract Entry<KeyType, ValueType> lowerEntry(KeyType key);

	@Deprecated
	@Override
	public final NavigableSet<KeyType> navigableKeySet(){
		return this.keys();
	}

	@Override
	public abstract NavigableMap<KeyType, ValueType, ?> subMap(KeyType fromKey,
			boolean fromInclusive, KeyType toKey, boolean toInclusive);

	@Override
	public NavigableMap<KeyType, ValueType, ?> subMap(final KeyType fromKey,final KeyType toKey) {
		return this.subMap(fromKey, true, toKey, false);
	}

	@Override
	public abstract NavigableMap<KeyType, ValueType, ?> tailMap(KeyType fromKey, boolean inclusive);

	@Override
	public NavigableMap<KeyType, ValueType, ?> tailMap(final KeyType fromKey){
		return this.tailMap(fromKey, true);
	}

	@Override
	public AbstractJavaNavigableMap<KeyType, ValueType> javaMap() {
		return new ToJavaNavigableMapWrapper<KeyType, ValueType>(this);
	}
	
	public static abstract class AbstractJavaNavigableMap<KeyType, ValueType> extends AbstractJavaSortedMap<KeyType, ValueType> implements java.util.NavigableMap<KeyType, ValueType>{
		
	}

}
