package de.lmu.ifi.pms.java.util.map.fixed;


import java.io.Serializable;
import java.util.ListIterator;

import de.lmu.ifi.pms.java.lang.iteration.AbstractFixedIterator;
import de.lmu.ifi.pms.java.lang.iteration.FixedIterator;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArraySet;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray.FixedArrayIterator;
import de.lmu.ifi.pms.java.util.collection.fixed.empty.Empty;
import de.lmu.ifi.pms.java.util.collection.set.Set;
import de.lmu.ifi.pms.java.util.map.Map;


public abstract class AbstractFixedArrayMap<KeyType, ValueType, KeySetType extends FixedArraySet<KeyType>> extends AbstractFixedMap<KeyType, ValueType, AbstractFixedEntry<KeyType, ValueType>> implements Serializable{
	
	KeySetType keys;
	FixedArray<ValueType> values;
	
	public AbstractFixedArrayMap(){
		this(false);
	}
	
	@SuppressWarnings("unchecked")
	public AbstractFixedArrayMap(final boolean emptyBody){
		if(!emptyBody){
			this.keys = this.newKeySet(Empty.<KeyType>instance());
			this.values = new FixedArray<ValueType>();
		}
	}
	
	public AbstractFixedArrayMap(final Map<KeyType, ? extends ValueType, ?> map){
		super();
		this.fillBody(map);
	}
	
	public AbstractFixedArrayMap(final KeySetType keys, final FixedArray<? extends ValueType> values) {
		super();
		this.fillBody(keys, values);
	}
	
	protected abstract KeySetType newKeySet(Set<KeyType> keys);
	
	public final boolean hasEmptyBody(){
		return this.keys == null;
	}
	
	protected final void checkFillBody(){
		if(!this.hasEmptyBody()){
			throw new IllegalStateException("Already filled!");
		}
	}
	
	public void fillBody(final Map<KeyType, ? extends ValueType, ?> map){
		this.checkFillBody();
		this.keys = this.newKeySet(map.keys());
		this.values = new FixedArray<ValueType>(map.values());
	}
	
	@SuppressWarnings("unchecked")
	public void fillBody(final KeySetType keys, final FixedArray<? extends ValueType> values){
		this.checkFillBody();
		if(keys.size() == values.size()){
			this.keys = keys;
			this.values = (FixedArray<ValueType>)values;
		}
		else{
			throw new IllegalArgumentException();//TODO better exception
		}
	}
	
	@Override
	public boolean isEmpty() {
		return this.keys.isEmpty();
	}
	
	@Override
	public int size() {
		return this.keys.size();
	}

	public KeyType firstKey() {
		return this.keys.first();
	}

	public KeyType lastKey() {
		return this.keys.last();
	}

	public FixedEntry<KeyType, ValueType> firstEntry() {
		return new FixedEntryImpl<KeyType, ValueType>(this.keys.first(), this.values.first());
	}

	public FixedEntry<KeyType, ValueType> lastEntry() {
		return new FixedEntryImpl<KeyType, ValueType>(this.keys.last(), this.values.last());
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.map.Map#has(java.lang.Object)
	 */
	@Override
	public boolean has(final KeyType key) {
		return this.keys.contains(key);
	}

	public boolean containsValue(final Object value) {
		return this.values.contains(value);
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.map.Map#entry(java.lang.Object)
	 */
	@Override
	public AbstractFixedEntry<KeyType, ValueType> entry(final KeyType key) {
		return new FixedEntryImpl<KeyType, ValueType>(key, this.get(key));
	}

	@Override
	public ValueType get(final Object key) {
		int index = this.keys.indexOf(key);
		if(index < 0){
			return null;
		}
		else{
			return this.values.get(index);
		}
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.fixed.AbstractFixedCollection#iterator()
	 */
	@Override
	public FixedIterator<AbstractFixedEntry<KeyType, ValueType>> iterator() {
		return new ArrayMapIterator<KeyType, ValueType, AbstractFixedArrayMap<KeyType,ValueType,KeySetType>>(this);
	}

	@Override
	public KeySetType keys() {
		return this.keys;
	}

	@Override
	public FixedArray<ValueType> values() {
		return this.values;
	}
	
	public static class ArrayMapIterator<KeyType, ValueType, ArrayMapType extends AbstractFixedArrayMap<KeyType, ValueType, ?>> extends AbstractFixedIterator<AbstractFixedEntry<KeyType, ValueType>> implements ListIterator<AbstractFixedEntry<KeyType, ValueType>>{
		
		final ArrayMapType arrayMap;
		final FixedArrayIterator<KeyType,?> keyIterator;
		final FixedArrayIterator<ValueType,?> valueIterator;

		public ArrayMapIterator(final ArrayMapType arrayMap, final int firstIndexInclusive, final int lastIndexExclusive, final int position) {
			super();
			this.arrayMap = arrayMap;
			this.keyIterator = this.arrayMap.keys().iterator();
			this.valueIterator = this.arrayMap.values().iterator();
		}

		public ArrayMapIterator(final ArrayMapType arrayMap) {
			this(arrayMap, arrayMap.keys().minIndexInc(), arrayMap.keys().maxIndexExc(), 0);
		}

		@Override
		public boolean hasNext() {
			return this.keyIterator.hasNext();
		}

		@Override
		public AbstractFixedEntry<KeyType, ValueType> next() {
			return new FixedEntryImpl<KeyType, ValueType>(this.keyIterator.next(), this.valueIterator.next());
		}

		@Override
		public boolean hasPrevious() {
			return this.keyIterator.hasPrevious();
		}

		@Override
		public final int nextIndex() {
			return this.keyIterator.nextIndex();
		}

		@Override
		public final int previousIndex() {
			return this.keyIterator.previousIndex();
		}

		@Override
		public AbstractFixedEntry<KeyType, ValueType> previous() {
			return new FixedEntryImpl<KeyType, ValueType>(this.keyIterator.previous(), this.valueIterator.previous());
		}

		@Override
		@Deprecated
		public void add(final AbstractFixedEntry<KeyType, ValueType> e) {
			 throw new UnsupportedOperationException();
		}

		@Override
		@Deprecated
		public void set(final AbstractFixedEntry<KeyType, ValueType> e) {
			throw new UnsupportedOperationException();
		}
		
	}
	
}
