/**
 * 
 */
package de.lmu.ifi.pms.java.util.map.javaWrapper;

import java.util.NoSuchElementException;

import de.lmu.ifi.pms.java.lang.iteration.FixedIterator;
import de.lmu.ifi.pms.java.util.adaption.FixedIteratorAdapter;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.fixed.wrapper.FixedCollectionWrapper;
import de.lmu.ifi.pms.java.util.collection.fixed.wrapper.FixedSetWrapper;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.FromJavaCollectionWrapper;
import de.lmu.ifi.pms.java.util.collection.set.Set;
import de.lmu.ifi.pms.java.util.map.fixed.AbstractFixedMap;




/**
 * @author Simon Matthias Brodt
 *
 */
public class FromJavaMapFixedWrapper<KeyType, ValueType> extends AbstractFixedMap<KeyType, ValueType, FromJavaEntryFixedWrapper<KeyType, ValueType>> {

	final java.util.Map<KeyType, ValueType> javaMap;

	@SuppressWarnings("unchecked")
	public FromJavaMapFixedWrapper(final java.util.Map<KeyType, ? extends ValueType> javaMap) {
		super();
		this.javaMap = (java.util.Map<KeyType, ValueType>)javaMap;
	}

	@Override
	public FromJavaEntryFixedWrapper<KeyType, ValueType> entry(final KeyType key) {
		for(java.util.Map.Entry<KeyType, ValueType> javaEntry: this.javaMap.entrySet()){
			if(javaEntry.getKey().equals(key)){
				return new FromJavaEntryFixedWrapper<KeyType, ValueType>(javaEntry);
			}
		}
		throw new NoSuchElementException();//TODO better exception
	}

	@Override
	public ValueType get(final KeyType key) {
		return this.javaMap.get(key);
	}

	@Override
	public boolean has(final KeyType key) {
		return this.javaMap.containsKey(key);
	}

	@Override
	public FixedIterator<FromJavaEntryFixedWrapper<KeyType, ValueType>> iterator() {
		return new FixedIteratorAdapter<java.util.Map.Entry<KeyType, ValueType>, FromJavaEntryFixedWrapper<KeyType,ValueType>>(this.javaMap.entrySet().iterator()){

			@Override
			public FromJavaEntryFixedWrapper<KeyType, ValueType> adapt(final java.util.Map.Entry<KeyType, ValueType> source) {
				return new FromJavaEntryFixedWrapper<KeyType, ValueType>(source);
			}
			
		};
	}

	@Override
	public Set<KeyType> keys() {
		return new FixedSetWrapper<KeyType>(this.javaMap.keySet());
	}


	@Override
	public int size() {
		return this.javaMap.size();
	}

	@Override
	public Collection<ValueType> values() {
		return new FixedCollectionWrapper<ValueType>(new FromJavaCollectionWrapper<ValueType>(this.javaMap.values()));
	}
	
}
