/**
 * 
 */
package de.lmu.ifi.pms.java.util.map.javaWrapper;

import java.util.Comparator;
import java.util.NoSuchElementException;

import de.lmu.ifi.pms.development.UnimplementedMethodError;
import de.lmu.ifi.pms.java.lang.iteration.ExtendibleIterator;
import de.lmu.ifi.pms.java.util.adaption.ExtendibleIteratorAdapter;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.fixed.wrapper.FixedCollectionWrapper;
import de.lmu.ifi.pms.java.util.collection.fixed.wrapper.FixedSortedSetWrapper;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.FromJavaCollectionWrapper;
import de.lmu.ifi.pms.java.util.collection.set.SortedSet;
import de.lmu.ifi.pms.java.util.map.AbstractSortedMap;
import de.lmu.ifi.pms.java.util.map.Entry;
import de.lmu.ifi.pms.java.util.map.ExtendibleSortedMap;
import de.lmu.ifi.pms.java.util.map.Map;
import de.lmu.ifi.pms.java.util.map.SortedMap;



/**
 * @author Simon Matthias Brodt
 *
 */
public class FromJavaSortedMapExtendibleWrapper<KeyType, ValueType>
		extends AbstractSortedMap<KeyType, ValueType, FromJavaEntryExtendibleWrapper<KeyType, ValueType>> implements ExtendibleSortedMap<KeyType, ValueType, FromJavaEntryExtendibleWrapper<KeyType, ValueType>>{

	final java.util.SortedMap<KeyType, ValueType> javaMap;

	public FromJavaSortedMapExtendibleWrapper(final java.util.SortedMap<KeyType, ValueType> javaMap) {
		super();
		this.javaMap = javaMap;
	}

	@Override
	public FromJavaEntryExtendibleWrapper<KeyType, ValueType> entry(final KeyType key) {
		if(this.has(key)){
			return new FromJavaEntryExtendibleWrapper<KeyType, ValueType>(this.javaMap.tailMap(key).entrySet().iterator().next(), this.javaMap);
		}
		else{
			throw new NoSuchElementException();//TODO better exception
		}
	}

	@Override
	public ValueType get(final KeyType key) {
		return this.javaMap.get(key);
	}

	@Override
	public boolean has(final KeyType key) {
		return this.javaMap.containsKey(key);
	}

	@Override
	public SortedMap<KeyType, ValueType, ?> headMap(final KeyType toKey) {
		return new FromJavaSortedMapExtendibleWrapper<KeyType, ValueType>(this.javaMap.headMap(toKey));
	}

	@Override
	public SortedMap<KeyType,ValueType,?> tailMap(final KeyType fromKey) {
		return new FromJavaSortedMapExtendibleWrapper<KeyType, ValueType>(this.javaMap.tailMap(fromKey));
	}
	
	@Override
	public SortedMap<KeyType,ValueType,?> subMap(final KeyType fromKey, final KeyType toKey) {
		return new FromJavaSortedMapExtendibleWrapper<KeyType, ValueType>(this.javaMap.subMap(fromKey, toKey));
	}

	/* (non-Javadoc)
	 * @see extension.java.util.map.AbstractNavigableMap#iterator()
	 */
	@Override
	public ExtendibleIterator<FromJavaEntryExtendibleWrapper<KeyType, ValueType>> iterator() {
		return new ExtendibleIteratorAdapter<java.util.Map.Entry<KeyType, ValueType>, FromJavaEntryExtendibleWrapper<KeyType,ValueType>>(this.javaMap.entrySet()){
			
			/* (non-Javadoc)
			 * @see extension.java.lang.iteration.IteratorAdapter#adapt(java.lang.Object)
			 */
			@Override
			public FromJavaEntryExtendibleWrapper<KeyType, ValueType> adapt(final java.util.Map.Entry<KeyType, ValueType> source) {
				return new FromJavaEntryExtendibleWrapper<KeyType, ValueType>(source, FromJavaSortedMapExtendibleWrapper.this.javaMap);
			}
			
			@Override
			public java.util.Map.Entry<KeyType, ValueType> inverseAdapt(final FromJavaEntryExtendibleWrapper<KeyType, ValueType> source) {
				return source.javaEntry;
			}
			
		};
	}

	@Override
	public SortedSet<KeyType> keys() {
		final java.util.Set<KeyType> keys = FromJavaSortedMapExtendibleWrapper.this.javaMap.keySet();
		if(keys instanceof SortedSet<?>){
			return new FixedSortedSetWrapper<KeyType>((java.util.SortedSet<KeyType>)keys);
		}
		else{
			//FIXME implement
			throw new UnimplementedMethodError();
		}
	}
	
	@Override
	@Deprecated
	public final boolean put(final KeyType key, final ValueType value) {
		return this.add(key, value);
	}
	
	@Override
	public boolean add(final KeyType key, final ValueType value) {
		ValueType oldValue = this.javaMap.put(key, value);
		return oldValue != null;
	}
	
	@Override
	public boolean add(final Entry<KeyType, ValueType> entry){
		return this.add(entry.key(), entry.value());
	}
	
	@Override
	public boolean add(final FromJavaEntryExtendibleWrapper<KeyType,ValueType> entry){
		return this.add((Entry<KeyType, ValueType>)entry);
	}
	
	@Override
	public boolean addAll(final java.util.Collection<? extends FromJavaEntryExtendibleWrapper<KeyType,ValueType>> c) {
		return this.addAll(new FromJavaCollectionWrapper<FromJavaEntryExtendibleWrapper<KeyType,ValueType>>(c));
	}
	
	@Override
	public boolean addAll(final Collection<? extends FromJavaEntryExtendibleWrapper<KeyType,ValueType>> c){
		boolean modified = false;
		for(FromJavaEntryExtendibleWrapper<KeyType,ValueType> entry: c){
			if(this.add(entry)){
				modified = true;
			}
		}
		return modified;
	}
	
	@Override
	public boolean add(final java.util.Map<KeyType, ValueType> map) {
		return this.add(new FromJavaMapFixedWrapper<KeyType, ValueType>(map));
	}
	
	@Deprecated
	@Override
	public final boolean putAll(final java.util.Map<KeyType, ValueType> map) {
		return this.add(map);
	}
	
	@Override
	public boolean add(final Map<KeyType, ValueType, ?> map){
		boolean modified = false;
		for(Entry<KeyType, ValueType> entry: map){
			if(this.add(entry)){
				modified = true;
			}
		}
		return modified;
	}
	
	@Override
	public void set(final KeyType key, final ValueType value) {
		this.javaMap.put(key, value);
	}
	
	@Deprecated
	@SuppressWarnings("unchecked")
	@Override
	public final boolean remove(final Object o){
		return this.removeKey((KeyType)o);
	}
	
	@Override
	public boolean removeKey(final KeyType key) {
		ValueType value = this.javaMap.remove(key);
		return value != null;
	}
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public boolean removeAll(final java.util.Collection<?> c) {
		return this.removeAll(new FromJavaCollectionWrapper(c));
	}
	
	@Override
	public boolean removeAll(final Collection<?> c) {
		boolean modified = false;
		ExtendibleIterator<FromJavaEntryExtendibleWrapper<KeyType, ValueType>> e = this.iterator();
		while (e.hasNext()) {
		    if (c.contains(e.next())) {
			e.remove();
			modified = true;
		    }
		}
		return modified;
	}
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public boolean retainAll(final java.util.Collection<?> c) {
		return this.retainAll(new FromJavaCollectionWrapper(c));
	}

	@Override
	public boolean retainAll(final Collection<?> c) {
		boolean modified = false;
		ExtendibleIterator<FromJavaEntryExtendibleWrapper<KeyType, ValueType>> e = this.iterator();
		while (e.hasNext()) {
		    if (!c.contains(e.next())) {
			e.remove();
			modified = true;
		    }
		}
		return modified;
	}
	
	@Override
	public void clear() {
		this.javaMap.clear();
	}
	
	@Override
	public boolean remove(final Entry<KeyType, ValueType> o) {
		KeyType key = o.key();
		if(this.has(key)){
			FromJavaEntryExtendibleWrapper<KeyType, ValueType> entry = this.entry(key);
			if(o.value().equals(entry.value())){
				entry.remove();
				return true;
			}
			else{
				return false;
			}
		}
		else{
			return false;
		}
	}

	@Override
	public int size() {
		return this.javaMap.size();
	}

	@Override
	public Collection<ValueType> values() {
		return new FixedCollectionWrapper<ValueType>(new FromJavaCollectionWrapper<ValueType>(this.javaMap.values()));
	}
	
	@Override
	public Comparator<? super KeyType> comparator() {
		return this.javaMap.comparator();
	}
		
}
