package de.lmu.ifi.pms.java.util.stream.io;

public abstract class AbstractCloseableImpl extends AbstractCloseable implements Openable{
	
	public static enum OpenState{
		
		CONSTRUCTED{
			
			/* (non-Javadoc)
			 * @see extension.java.util.stream.io.AbstractCloseable.OpenState#open(extension.java.util.stream.io.AbstractCloseable)
			 */
			@Override
			public void open(final AbstractCloseableImpl closeable) {
				closeable.open = OPEN;
				closeable.openAction();
			}
			
		},
		
		OPEN{
			
			/* (non-Javadoc)
			 * @see extension.java.util.stream.io.AbstractCloseable.OpenState#isOpen()
			 */
			@Override
			public boolean isOpen() {
				return true;
			}
			
		}
		,
		
		CLOSED{
			/* (non-Javadoc)
			 * @see extension.java.util.stream.io.AbstractCloseable.OpenState#close(extension.java.util.stream.io.AbstractCloseable)
			 */
			@Override
			public void close(final AbstractCloseableImpl closeable) {
				throw new IllegalStateException();
			}
		}
		;
		
		public void open(final AbstractCloseableImpl closeable){
			throw new IllegalStateException();
		}
		
		public void close(final AbstractCloseableImpl closeable){
			closeable.open = CLOSED;
			closeable.closeAction();
		}
		
		public boolean isOpen(){
			return false;
		}
		
	}
	
	public static enum Synchronisation{
		
		None{
			
		},
		
		Synchronized{
			@Override
			public void open(final AbstractCloseableImpl closeable){
				synchronized (closeable.closeLock()) {
					closeable.open.open(closeable);
				}
			}
			
			@Override
			public void close(final AbstractCloseableImpl closeable){
				synchronized (closeable.closeLock()) {
					closeable.open.close(closeable);
				}
			}
			
			@Override
			public boolean isOpen(final AbstractCloseableImpl closeable){
				synchronized (closeable.closeLock()) {
					return closeable.open.isOpen();	
				}
			}
		}
		
		;
		
		public void open(final AbstractCloseableImpl closeable){
			closeable.open.open(closeable);
		}
		
		public void close(final AbstractCloseableImpl closeable){
			closeable.open.open(closeable);
		}
		
		public boolean isOpen(final AbstractCloseableImpl closeable){
			return closeable.open.isOpen();
		}
		
	}
	
	
	OpenState open;
	
	protected final Synchronisation synchronisation;
	
	public AbstractCloseableImpl() {
		this(OpenState.OPEN, Synchronisation.None);
	}
	
	public AbstractCloseableImpl(final Synchronisation synchronisation) {
		this(OpenState.OPEN ,synchronisation);
	}
	
	public AbstractCloseableImpl(final OpenState state) {
		this(state, Synchronisation.None);
	}

	public AbstractCloseableImpl(final OpenState state, final Synchronisation synchronisation) {
		super();
		this.open = state;
		this.synchronisation = synchronisation;
	}
	
	public abstract Object closeLock();
	
	@Override
	public final void open(){
		this.synchronisation.open(this);
	}
	
	protected void openAction(){
		
	}
	
	protected void closeAction(){
		
	}
	
	@Override
	public final void close(){
		this.synchronisation.close(this);
	}
	
	@Override
	public final boolean isOpen(){
		return this.synchronisation.isOpen(this);
	}
	
	public final OpenState openState(){
		return this.open;
	}
	
}
