package de.lmu.ifi.pms.java.util.string;

import java.util.regex.Pattern;

import org.simpleframework.xml.ElementList;

import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.FromJavaCollectionWrapper;
import de.lmu.ifi.pms.java.util.filter.AbstractFilter;


public class RegExFilter extends AbstractFilter<String> {
	
	public final FixedArray<Pattern> patterns;
	
	@ElementList(inline=true, entry="pattern", type=String.class, required=false)
	public java.util.ArrayList<String> getPatterns(){
		return new java.util.ArrayList<String>(new CollectionAdapter<Pattern, String>(this.patterns){
			
			@Override
			public String adapt(final Pattern source) {
				return source.pattern();
			}
			
		}.javaCollection());
	}
	
	public RegExFilter(final String... patterns) {
		this(new FixedArray<String>(false, patterns));
	}
	
	RegExFilter(@ElementList(inline=true, entry="pattern", type=String.class, required=false) final java.util.ArrayList<String> patterns) {
		this(new FromJavaCollectionWrapper<String>(patterns));
	}
	
	RegExFilter(final java.util.Collection<String> patterns) {
		this(new FromJavaCollectionWrapper<String>(patterns));
	}
	
	public RegExFilter(final Collection<String> patterns) {
		this(toPatterns(patterns));
	}

	public RegExFilter(final FixedArray<Pattern> patterns) {
		super();
		this.patterns = patterns;
	}
	
	public static FixedArray<Pattern> toPatterns(final Collection<String> regExs){
		Pattern[] patterns = new Pattern[regExs.size()];
		int index = 0;
		for(String regEx: regExs){
			patterns[index] = Pattern.compile(regEx);
			++index;
		}
		return new FixedArray<Pattern>(false, patterns);
	}

	@Override
	public boolean accepts(final String element) {
		for(Pattern pattern: this.patterns){
			if(pattern.matcher(element).matches()){
				return true;
			}
		}
		return false;
	}

}
