package nl.cwi.monetdb.event_mill;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Properties;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.configuration.SQL_DatabaseDriver;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_EXCEPTIONS;


public class MonetDriver extends SQL_DatabaseDriver{
	
	public MonetDriver() throws ClassNotFoundException {
		super("nl.cwi.monetdb.jdbc.MonetDriver", "monetdb");
	}

	@Override
	protected String urlDatabaseExpansion(final String database) {
		return "/" + database;
	}
	
	@Override
	protected boolean isSetLogging(final Properties properties) {
		return properties.contains("debug") || properties.contains("logfile");
	}
	
	@Override
	protected void setLoggingProperty(final Properties properties, final LogLevel logLevel, final String logDestination) {
		if(logLevel.ordinal() <= LogLevel.DEBUG.ordinal()){
			properties.put("debug", "true");
			properties.put("logfile", logDestination);
		}
	}

	
	@Override
	public Connection connect(final String jdbcUrl, final java.util.Properties properties) throws SQLException{
		this.info("Connecting to database...");
		
		Connection connection;
		try{
			connection = super.connect(jdbcUrl, properties);
		} catch(SQLException e1){
			this.debug("Failed to connect", e1);
			/*if(this.db_server != null){
				this.stopDB();
			}
			if(this.db_configuration.start_command != null){
				this.startDB();
			}
			else{
				this.error("Error on connecting to database!", e1);
				throw new RuntimeException(e1);
			}*/
			throw e1;
		}
		this.info("Connected to database.");

		return connection;
	}
	
	@Override
	public void initGreatestFunctions(final Statement statement) throws SQLException{
		//Create greater functions
		for(int arity = 1; arity <= 20; ++arity){
			while(true){
				try {
					statement.execute(sqlCreateGreatestFunction(arity));
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						throw e;
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else{
						this.error("", e);
						throw e;
					}
				}
			}
		}
	}
	
	@Override
	public void clearGreatestFunctions(final Statement statement) throws SQLException{
		for(int arity = 1; arity <= 20; ++arity){
			while(true){
				try {
					statement.execute(sqlDropGreatestFunction(arity));
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						throw e;
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else{
						this.error("", e);
						break;
					}
				}
			}
		}
	}
	
	public static String sqlCreateGreatestFunction(final int arity){
		if(arity >= 1){
			java.lang.StringBuilder functionStatement = new java.lang.StringBuilder();
			functionStatement.append("CREATE FUNCTION GREATEST ");
			functionStatement.append("( ");
			for(int arg = 0; arg < arity; ++arg){
				functionStatement.append("\"n" + arg + "\" BIGINT");
				functionStatement.append(", ");
			}		
			functionStatement.setLength(functionStatement.length() - ", ".length());		
			functionStatement.append(" ) ");
			functionStatement.append("RETURNS BIGINT \n");
			
			functionStatement.append("BEGIN \n");
			for(int arg = 1; arg < arity; ++arg){
				functionStatement.append("\t");
				functionStatement.append("IF ( \"n");
				functionStatement.append(arg);
				functionStatement.append("\" > \"n0\" ) THEN SET \"n0\" = \"n");
				functionStatement.append(arg);
				functionStatement.append("\"; END IF;"); 
				functionStatement.append("\n");
			}
			functionStatement.append("\t");
			functionStatement.append("RETURN \"n0\";\n"); 
			functionStatement.append("END;");
			
			return functionStatement.toString();
		}
		else{
			throw new IllegalArgumentException("Greater Function has a minimum arity of 1.");
		}
	}
	
	public static String sqlDropGreatestFunction(final int arity){
		if(arity >= 1){
			java.lang.StringBuilder functionStatement = new java.lang.StringBuilder();
			functionStatement.append("DROP FUNCTION GREATEST ");
			functionStatement.append("( ");
			for(int arg = 0; arg < arity; ++arg){
				functionStatement.append("BIGINT");
				functionStatement.append(", ");
			}		
			functionStatement.setLength(functionStatement.length() - ", ".length());		
			functionStatement.append(" );");
			
			return functionStatement.toString();
		}
		else{
			throw new IllegalArgumentException("Greater Function has a minimum arity of 1.");
		}
	}
	
	@Override
	public void initLeastFunctions(final Statement statement) throws SQLException{
		//Create least functions
		for(int arity = 1; arity <= 20; ++arity){
			while(true){
				try {
					statement.execute(sqlCreateLeastFunction(arity));
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						throw e;
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else{
						this.error("", e);
						throw e;
					}
				}
			}
		}
	}
	
	@Override
	public void clearLeastFunctions(final Statement statement) throws SQLException{
		for(int arity = 1; arity <= 20; ++arity){
			while(true){
				try {
					statement.execute(sqlDropLeastFunction(arity));
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						throw e;
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else{
						this.error("", e);
						break;
					}
				}
			}
		}
	}
	
	public static String sqlCreateLeastFunction(final int arity){
		//FIXME Move to Driver
		if(arity >= 1){
			java.lang.StringBuilder functionStatement = new java.lang.StringBuilder();
			functionStatement.append("CREATE FUNCTION LEAST ");
			functionStatement.append("( ");
			for(int arg = 0; arg < arity; ++arg){
				functionStatement.append("\"n" + arg + "\" BIGINT");
				functionStatement.append(", ");
			}		
			functionStatement.setLength(functionStatement.length() - ", ".length());		
			functionStatement.append(" ) ");
			functionStatement.append("RETURNS BIGINT \n");
			
			functionStatement.append("BEGIN \n");
			for(int arg = 1; arg < arity; ++arg){
				functionStatement.append("\t");
				functionStatement.append("IF ( \"n");
				functionStatement.append(arg);
				functionStatement.append("\" < \"n0\" ) THEN SET \"n0\" = \"n");
				functionStatement.append(arg);
				functionStatement.append("\"; END IF;"); 
				functionStatement.append("\n");
			}
			functionStatement.append("\t");
			functionStatement.append("RETURN \"n0\";\n"); 
			functionStatement.append("END;");
			
			return functionStatement.toString();
		}
		else{
			throw new IllegalArgumentException("Least Function has a minimum arity of 1.");
		}
	}
	
	public static String sqlDropLeastFunction(final int arity){
		//FIXME Move to Driver
		if(arity >= 1){
			java.lang.StringBuilder functionStatement = new java.lang.StringBuilder();
			functionStatement.append("DROP FUNCTION LEAST ");
			functionStatement.append("( ");
			for(int arg = 0; arg < arity; ++arg){
				functionStatement.append("BIGINT");
				functionStatement.append(", ");
			}		
			functionStatement.setLength(functionStatement.length() - ", ".length());		
			functionStatement.append(" );");
			
			return functionStatement.toString();
		}
		else{
			throw new IllegalArgumentException("Least Function has a minimum arity of 1.");
		}
	}
	
	/*
	protected void startDB(){
		assert this.db_configuration.start_command != null;
		try {
			this.info("Immediate connection failed. Try starting database server.");
			this.db_server = Runtime.getRuntime().exec(this.db_configuration.start_command);
			this.db_input = new PrintStream(this.db_server.getOutputStream(), true);
		} catch (IOException e2) {
			this.error("Error on connecting to database!", e2);
			throw new RuntimeException(e2);
		}
		int tryCount = 0;
		while(this.db_server != null & this.connection == null){
			try {
				this.connection = DriverManager.getConnection(this.url, this.driverProperties);
			} catch (SQLException e2) {
			
				try{
					int exitValue = this.db_server.exitValue();
					this.debug("DB-Server terminated with exit code" + exitValue, e2);
					this.db_server = null;
					this.db_input.close();
					this.db_input = null;
				}
				catch(IllegalThreadStateException e){
					
				}
				
				this.debug("Failed to connect", e2);
				if(tryCount < 11){
					try {
						Thread.sleep(10000);
					} catch (InterruptedException e) {

					}
					++tryCount;
				}
				else{
					this.error("Error on connecting to database!", e2);
					throw new RuntimeException(e2);
				}
			}
		}
	}
	
	protected void stopDB(){
		if(this.db_server != null){ 
			if(this.db_configuration.stop_command != null){
				this.db_input.print(this.db_configuration.stop_command);
				try {
					Thread.sleep(1000);
				} catch (InterruptedException e) {

				}
			}
			
			try{
				this.db_server.exitValue();
			}
			catch (IllegalThreadStateException e) {
				this.db_server.destroy();
			}
			
			boolean terminated = false;
			while(!terminated){
				try {
					this.db_server.waitFor();
					terminated = true;
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			this.db_server = null;
			this.db_input.close();
			this.db_input = null;
		}
	}
	*/
	
	//logging
	@SuppressWarnings("hiding")
	protected static final Logger logger = LoggerFactory.getLogger(MonetDriver.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}
	
}
