package de.lmu.ifi.pms.cep.unitTests.time.constraints.tests;

import static org.junit.Assert.assertEquals;

import java.util.Arrays;

import org.junit.Test;

import de.lmu.ifi.pms.cep.tsa.time.constraints.ConjunctiveTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.DisjunctiveTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.SimpleTemporalFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.TemporalMinMax;
import de.lmu.ifi.pms.cep.tsa.time.constraints.TemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.BinaryComparisonOperator;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.MinMax;



public class TemporalRelationFormulaTest {
	
	@Test
	public void testFormulaComposition(){
		SimpleTemporalFormula stf = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS, "w",0);
		SimpleTemporalFormula stf2 = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS, "w",10000);
		TemporalRelationFormula trf = new ConjunctiveTemporalRelationFormula(
				stf,stf2);
		assert ((ConjunctiveTemporalRelationFormula)trf).subFormulas().size()!=0; 
		assertEquals(stf,((ConjunctiveTemporalRelationFormula)trf).subFormulas().get(0));
		assertEquals(stf2,((ConjunctiveTemporalRelationFormula)trf).subFormulas().get(1));
	}
	
	@Test
	public void testTemporalMinMax(){
		TemporalMinMax tmm = new TemporalMinMax("v", MinMax.MIN, "a","b");
		assertEquals("v = min(a,b)",tmm.toString());
		tmm = new TemporalMinMax("v", MinMax.MAX, "a","b");
		assertEquals("v = max(a,b)",tmm.toString());
	}
	
	@Test
	public void testPrintout(){
		SimpleTemporalFormula stf = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS, "w",0);
		SimpleTemporalFormula stf2 = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS, "w",10000);
		TemporalRelationFormula trf = new ConjunctiveTemporalRelationFormula(stf,stf2);
		TemporalMinMax tmm = new TemporalMinMax("v", MinMax.MIN, "a","b");
		TemporalRelationFormula trf2 = new DisjunctiveTemporalRelationFormula(trf,tmm);
		assertEquals("(v < w AND v < w+10000)",trf.toString());
		assertEquals("((v < w AND v < w+10000) OR v = min(a,b))",trf2.toString());
		TemporalRelationFormula trf3 = new DisjunctiveTemporalRelationFormula(trf2,trf2,trf2);
		System.out.println(trf3);
		System.out.println("b".compareTo("a"));
	}
	
	@Test
	public void testVariables(){
		SimpleTemporalFormula stf = new SimpleTemporalFormula("w", BinaryComparisonOperator.LESS, "v",0);
		assertEquals(Arrays.asList("v","w"),stf.VariableNames());
		TemporalMinMax tmm = new TemporalMinMax("z", MinMax.MIN, "x","z","y","m");
		assertEquals(Arrays.asList("m","x","y","z"),tmm.VariableNames());
		TemporalRelationFormula trf = new ConjunctiveTemporalRelationFormula(stf,tmm);
		assertEquals(Arrays.asList("m","v","w","x","y","z"),trf.VariableNames());
		TemporalRelationFormula trf2 = new DisjunctiveTemporalRelationFormula(trf,tmm);
		System.out.println(trf2.VariableNames());
	}
}
