package de.lmu.ifi.pms.cep.unitTests.time.constraints.tests;

import static org.junit.Assert.assertTrue;

import org.junit.Test;

import de.lmu.ifi.pms.cep.tsa.time.analysis.SimpleFloydWarshall;
import de.lmu.ifi.pms.cep.tsa.time.exceptions.MatrixNotQuadradicException;



public class TestFloydWarshall {
	long[][] test;
	
	public void initThreeLettersMatrix(){
		this.test = new long[3][3];
		this.test[0][0] = 0;
		this.test[0][1] = 1;
		this.test[0][2] = Long.MAX_VALUE;
		this.test[1][0] = Long.MAX_VALUE;
		this.test[1][1] = 0;
		this.test[1][2] = 3;
		this.test[2][0] = 3;
		this.test[2][1] = 1;
		this.test[2][2] = 0;
	}
	public void initFiveLettersMatrix(){
		this.test = new long[5][5];
		this.test[0][0] = 0;
		this.test[0][1] = 1;
		this.test[0][2] = Long.MAX_VALUE;
		this.test[0][3] = Long.MAX_VALUE;
		this.test[0][4] = Long.MAX_VALUE;
		this.test[1][0] = Long.MAX_VALUE;
		this.test[1][1] = 0;
		this.test[1][2] = Long.MAX_VALUE;
		this.test[1][3] = 2;
		this.test[1][4] = 4;
		this.test[2][0] = Long.MAX_VALUE;
		this.test[2][1] = Long.MAX_VALUE;
		this.test[2][2] = 0;
		this.test[2][3] = Long.MAX_VALUE;
		this.test[2][4] = Long.MAX_VALUE;
		this.test[3][3] = 0;
		this.test[3][2] = 6;
		this.test[3][1] = Long.MAX_VALUE;
		this.test[3][0] = 2;
		this.test[3][4] = Long.MAX_VALUE;
		this.test[4][4] = 0;
		this.test[4][0] = Long.MAX_VALUE;
		this.test[4][1] = Long.MAX_VALUE;
		this.test[4][2] = 2;
		this.test[4][3] = Long.MAX_VALUE;
	}
	@Test
	public void testCalc(){
		this.initThreeLettersMatrix();
		SimpleFloydWarshall fw = new SimpleFloydWarshall(this.test, this.test.length);
		
		try{
			fw.calculate();
		}catch(Exception e){
			e.printStackTrace();
		}
		System.out.println(fw.getMatrix()[2][0]);
		assertTrue(fw.getMatrix()[2][0] == 3);
		assertTrue(fw.getMatrix()[1][0] == 6);
		assertTrue(fw.getMatrix()[0][2] == 4);
		
		this.initFiveLettersMatrix();
		fw = new SimpleFloydWarshall(this.test, this.test.length);
		try{
			fw.calculate();
		}catch(Exception e){
			e.printStackTrace();
		}
		assertTrue(fw.getMatrix()[0][2]==7);
		assertTrue(fw.getMatrix()[1][2]==6);
		assertTrue(fw.getMatrix()[1][0]==4);
		assertTrue(fw.getMatrix()[0][3]==3);
		
	}
	
	@Test(expected=MatrixNotQuadradicException.class)
	public void testException() throws Exception{
		this.test= new long[4][3];
		SimpleFloydWarshall fw = new SimpleFloydWarshall(this.test, this.test.length);
		fw.calculate();
	}
	
}
